;;; -*- Mode: LISP; PACKAGE: AGS; SYNTAX: COMMON-LISP; BASE: 10 -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;


(in-package :ags)

(mod~defmod doc
	    :uses (mod pp sys)
	    :exports (doc~latex
		      doc~latex-p
		      doc~replace-char-by-string
		      doc~replace-string-by-string
		      doc~extract
		      doc+doc
		      doc+latex-interface
		      doc+latex-complete
		      doc~latex-lispdef
		      doc~parse-def
		      doc+interface))

#{ \section{Documentation}
The DOC module provides pretty-print styles and functions for extracting
documentation from Lisp source files.

#}

(eval-when (load compile eval)
(defvar doc*export-list nil
  "A variable containing the export list of a module.")
)

(eval-when (load compile eval)
(defvar doc*defined-generic-functions nil
  "A list of names of defgeneric functions that have been documented so far"))


(defmacro doc~latex (&rest doc)
  (declare 
   (ignore doc)
   (value "none")
   (authors nesmith)
   (effect "Ignores its arguments, returns no value")
   (input "You can input anything, it'll be ignored.  Actually, this
macro is defined only because the form \#\\{ blah blah blah \#\\} is
defined in our readtable to expand to (doc~latex \"blah blah blah\"). So
doc~latex has to evaluate to something when we're loading a lisp file.
When we're extracting documentation, we'll just write the string to
the .tex file."))
  (values))


(defun doc~latex-p (obj)
  (declare (edited  "28-MAR-1994 12:06")
	   (authors AFIEDLER)
	   (input   "An object.")
	   (effect  )
	   (value   #{T, iff OBJ is a {\vb doc~latex} list.#}))
  (and (listp obj) (eql (car obj) 'doc~latex)))

(defun doc=interface-p (obj)
  (declare (edited  "05-APR-1994 08:59")
	   (authors AFIEDLER)
	   (input   #{A lisp definition without leading {\vb def\em thing\/}.#})
	   (effect  )
	   (value   #{T, iff the first element of OBJ is either a symbol in the
export list of the module now being read, doc*export-list, or OBJ is of the
form (setf NAME) and NAME is in the export list.#}))
  (when (listp obj)
    (let ((obj-name (car obj)))
      (when (and (listp obj-name)
		 (eq 'setf (car obj-name)))
	(setq obj-name (cadr obj-name)))
      (and obj-name (member obj-name doc*export-list) t))))

(deftype doc+interface ()
  "The type of a definition list of an interface function or macro."
  '(satisfies doc=interface-p))


(defun doc~replace-char-by-string (str alist)
  (declare (edited  "11-APR-1994 08:23")
	   (authors AFIEDLER nesmith)
	   (input   "A string and an association list of characters and strings.")
	   (effect  "Replaces in STR each occurrence of the characters in ALIST"
		    "by the associated strings.")
	   (value   "The new string."))
  (with-output-to-string
   (stream)
   (let ((char nil) (match nil))
     (dotimes (i (length str))
       (setq char (char str i)
	     match (cdr (assoc char alist :test #'char=)))
       (if match
	   (write-string match stream)
	   (write-char char stream))))))


(defun doc~replace-string-by-string (str alist)
  (declare (edited  "11-APR-1994 08:24")
	   (authors AFIEDLER nesmith)
	   (input   "A string and an association list of string-string-pairs.")
	   (effect  "Replaces each occurrence of the first string in a pair in STR by the"
		    "associated string.")
	   (value   "The new string."))
  (let ((new-string str))
    (dolist (pair alist new-string)
      (setq new-string 
	    (with-output-to-string (out)
              (do* ((old (car pair))
		    (new (cdr pair))
		    (old-len (length old))
		    (start 0 (+ pos old-len))
		    (pos (search old new-string 
				 :test #'char= 
				 :start2 start)
			 (search old new-string 
				 :test #'char= 
				 :start2 start)))
		  ((null pos)
		   (write-string (subseq new-string start) out))
		(write-string (subseq new-string start pos) out)
		(write-string new out)))))))

(eval-when (load compile eval)
(defmacro doc=write-each (l stream)
  (declare (edited  "29-MAR-1994 14:47")
	   (authors AFIEDLER nesmith)
	   (input   "A list and a stream.")
	   (effect  "Writes each member of L to STREAM.")
	   (value   "Undefined."))
  `(dolist (e ,l)
    (write e :stream ,stream)))
)

;;; latex

(eval-when (load compile eval)
(defvar doc*replace-char-latex
  (list '(#\& . "\\&") '(#\$ . "\\$") '(#\_ . "\\_") '(#\# . "\\#")
	'(#\~ . "\\protect\\(\\sim\\protect\\)")
	'(#\^ . "\\^{\\ }"))
  "An association list with characters associated to strings that should replace them in order to obtain sound \\LaTeX\\ output of a string.")

(defvar doc*replace-char-in-symbol-latex
  (list* '(#\< . "$<$") '(#\> . "$>$")
	 doc*replace-char-latex)
  "An association list with characters associated to strings that should replace them in order to obtain sound \\LaTeX\\ output of a lisp symbol.")

(defvar doc*replace-string-latex
  (list '("--> " . "$\\rightarrow$\\ ")
	'("<-- " . "$\\leftarrow$\\ ")
	'("<--> " . "$\\leftrightarrow$\\ ")
	'("==> " . "$\\Rightarrow$\\ ")
	'("<== " . "$\\Leftarrow$\\ ")
	'("<==> " . "$\\Leftrightarrow$\\ "))
  "An association list with strings associated to other strings that should replace the first in order to obtain sound \\LaTeX\\ output of a string.")
)

(defun doc=write-latex-symbol (stream sym)
  (declare (edited  "28-MAR-1994 09:37")
	   (authors AFIEDLER)
	   (input   "A symbol, a stream.")
	   (effect  "Writes the \\LaTeX\\ output of SYM to STREAM.")
	   (value   "Undefined."))
  (write-string (doc~replace-char-by-string 
		 (write-to-string sym :pretty nil)
		 doc*replace-char-in-symbol-latex)
		stream))

(defun doc=write-latex-string (stream str)
  (declare (edited  "28-MAR-1994 09:38")
	   (authors AFIEDLER)
	   (input   "A string and a stream.")
	   (effect  "Writes the \\LaTeX\\ output of STRING to STREAM.")
	   (value   "Undefined."))
  (write-string 
   (doc~replace-string-by-string
    (doc~replace-char-by-string str doc*replace-char-latex)
    doc*replace-string-latex)
   stream))

(defun doc=write-latex-list (stream l)
  (pprint-logical-block (stream l :prefix "(" :suffix ")")
    (pprint-exit-if-list-exhausted)
    (loop
     (write (pprint-pop))
     (pprint-exit-if-list-exhausted)
     (write-char #\space))))
       

(defun doc=write-latex-head (stream name params type doc-items)
  (declare (edited  "07-MAR-1994 16:06")
	   (authors AFIEDLER)
	   (input   "A name, a list of parameters, a string for the type of NAME,"
		    "and a stream.")
	   (effect  "Writes the head of a lispdefinition-environment in \\LaTeX\\ format"
		    "to STREAM.")
	   (value   "Undefined."))
  ;(break ">>0")
  (let ((*standard-output* stream)) ;(break ">>1")
    (fresh-line) ;(break ">>2")
  (pprint-logical-block (nil nil :prefix "\\begin{lispdefinition}"
				:suffix "\\end{lispdefinition}")
    (pprint-logical-block (nil nil :prefix "{" :suffix "}")
      (write type))
    (pprint-logical-block (nil nil :prefix "{" :suffix "}")
      (write name))
    (terpri)
    (pprint-logical-block (nil params :prefix "{" 
			       :suffix (if params "}" "\\hfill}"))
      (let ((elt nil))
	(loop
	 (pprint-exit-if-list-exhausted)
	 (setq elt (pprint-pop))
	 (pprint-logical-block (nil nil
				:prefix (if (symbolp elt)
					    "\\mbox{" "")
				:suffix (if (symbolp elt)
					    "}" ""))
	   (write elt :case :downcase) (terpri)))))
    (doc=write-latex-doc-items nil doc-items))
  (terpri)))

(defgeneric doc~parse-def (type def)
  (declare
   (authors nesmith)
   (input "A symbol TYPE and a Lisp definition DEF.  TYPE is used to select
which method is used.")
   (value #{Four values are returned: a symbol NAME; a string DESC; a list ARGS
of arguments (or when more relevant, parents); a list of DOC-ITEMS (see below).

NAME is the name of the defined object (function or macro name, class name,
etc.).  DESC is a short string description of the defined object, for example
``Function'' for defun forms, and ``Macro'' for defmacro forms.
ARGS is for functional definitions the lambda-list of arguments for
objects such as classes this should be the superclasses.  DOC-ITEMS should
a list of which each member is of the
form `(symbol string1 string2 ...)'.  This list should normally be gathered
from the `(declare )' forms at the beginning of DEF.  Miscellaneous commentary
such as doc-strings allowed in many definitions can be returned in the form
`(misc string*)'. #})
   (example "\\begin{code}
   (doc~parse-def 'DEFUN '(DEFUN FOO (BAR) 
                         (DECLARE (INPUT \"A number BAR\")
                                  (VALUE \"NIL\"))
                        \"This is a useful function.\"
		       nil))
\\end{code}
might return the four values
\\begin{code}
FOO \"Function\" (BAR) ((INPUT \"A number BAR\") (VALUE \"NIL\") 
		        (MISC \"This is a useful function.\"))
\\end{code}"))
  (:method ((type t) (def t))
    (values nil nil nil nil))
  (:method ((type (eql 'defun)) def)
    (let ((name (second def))
	  (desc "Function")
	  (args (third def))
	  (doc-items nil)
	  (comment-string nil)
	  (last-form (car (last def))))
      (dolist (x (cdddr def))
	(cond ((stringp x)
	       ;; only one comment-string is allowed, and it must not be
	       ;; the last form...
	       (if (or comment-string
		       (eq x last-form))
		   (return)
		   (setq comment-string x)))
	      ((and (listp x) (eq 'declare (car x)))
	       (setq doc-items (append doc-items (cdr x))))
	      (t (return))))
      (values name desc args 
	      (if comment-string
		  (cons (list 'misc comment-string) doc-items)
		  doc-items))))
  (:method ((type (eql 'defmacro)) def)
    (multiple-value-bind (name desc args doc-items)
	(doc~parse-def 'defun def)
      (declare (ignore desc))
      (values name "Macro" args doc-items)))
  (:method ((type (eql 'defgeneric)) def)
    (multiple-value-bind (name desc args doc-items)
	(doc~parse-def 'defun def)
      (declare (ignore desc))
      (pushnew name doc*defined-generic-functions)
      (values name "Generic Function" 
	      args
	      (let ((doc-string 
		     (cadr (assoc :documentation (cdddr def)))))
		(if doc-string
		    (cons (list 'misc doc-string) doc-items)
		    doc-items)))))
  (:method ((type (eql 'deftype)) def)
    (multiple-value-bind (name desc args doc-items)
	(doc~parse-def 'defun def)
      (declare (ignore desc))
      (values name "Type" args doc-items)))  
  (:method ((type (eql 'defsetf)) def)
    (let ((name (list 'setf (second def)))
	  (args (if (> (length def) 4)
		    (third def)))
	  (doc-items nil)
	  (comment-string ""))
      (when (> (length def) 3)
	(if (stringp (fourth def))
	    (setq comment-string (fourth def))
	    (dolist (x (cddddr def))
	      (cond ((stringp x) (concatenate 'string comment-string x))
		    ((and (listp x) (eq 'declare (car x)))
		     (setq doc-items (append doc-items (cdr x))))
		    (t (return))))))
      (values name "Setf Function" args
	      (if (string= comment-string "")
		  doc-items
		  (cons (list 'misc comment-string) doc-items)))))
  (:method ((type (eql 'defvar)) def)
    (let ((name (second def))
	  (desc "Variable")
	  (comment-string (fourth def)))
      (values name desc nil 
	      (if comment-string (list (list 'misc comment-string))))))
  (:method ((type (eql 'defparameter)) def)
    (let ((name (second def))
	  (desc "Parameter")
	  (comment-string (fourth def)))
      (values name desc nil 
	      (if comment-string (list (list 'misc comment-string))))))
  (:method ((type (eql 'defconstant)) def)
    (let ((name (second def))
	  (desc "Constant")
	  (comment-string (fourth def)))
      (values name desc nil 
	      (if comment-string (list (list 'misc comment-string))))))
  (:method ((type (eql 'defstruct)) obj)
    (let* ((sec (second obj))
	   (thi (if (> (length obj) 2) (third obj)))
	   (name (if (listp sec) (car sec) sec))
	   (options (if (listp sec) (cdr sec)))
	   (desc "Structure"))
      (values name desc options (if (stringp thi)
				    (list (list 'misc thi))))))
  (:method ((type (eql 'defclass)) def)
    (let ((name (second def))
	  (parents (third def))
	  (doc-string
	   (second (assoc :documentation (cdddr def)))))
      (values name "Class" parents
	      (if doc-string (list (list 'misc doc-string))))))
  (:method ((type (eql 'pp~defstyle)) def)
    (let ((name (second def))
	  (desc "Pretty-print Style")
	  (parent (cadr (member :parent def)))
	  (help (cadr (member :help def))))
      (values name desc (if parent (list parent))
	      (if help (list (list 'misc help))))))
  (:method ((type (eql 'define-condition)) def)
    (let ((name (second def))
	  (parents (third def))
	  (doc-string
	   (second (assoc :documentation (cdddr def)))))
      (values name "Condition Type" parents
	      (if doc-string (list (list 'misc doc-string))))))
  )


(defun doc=write-latex-doc-items (stream doc-items)
  (declare 
   (authors nesmith)
   (input "An list DOC-ITEMS, of which each member is of the
form `(symbol string1 string2 ...)', and a stream.  The symbol used in a
DOC-ITEM must be one of the following: INPUT, VALUE, VALUES, EFFECT, REMARK,
EXAMPLE, EXAMPLES, and COMMENTS. If a symbol appears twice in the list, only
the first occurrence will be used.")
   (value "undefined")
   (effect #{For each of the DOC-ITEMS, prints a LaTeX \item command to the
STREAM, in the following way:
\\item[Symbol\\hfill] string1string2\ldots

If none of the DOC-ITEMS appears, an item will be printed to that effect,
so that at least one item will always be produced.#}))
  (let (;(authors (cdr (assoc 'authors doc-items)))
	(input (or (assoc 'input doc-items :test #'string=)))
	(value  (or (assoc 'value doc-items :test #'string=)))
	(values (assoc 'values doc-items :test #'string=))
	(effect (assoc 'effect doc-items :test #'string=))
	(remark (assoc 'remark doc-items :test #'string=))
	(example (assoc 'example doc-items :test #'string=))
	(examples (assoc 'examples doc-items :test #'string=))
	(comments
	 (let ((coms (assoc 'comments doc-items :test #'string=))
	       (misc (assoc 'misc doc-items :test #'string=)))
	   (cond ((and coms misc)
		  (append coms (cdr misc)))
		 (coms)
		 (misc (cons 'comments (cdr misc)))
		 (t nil))))
	;; we need to write at least one item
	(wrote-something nil))
    (dolist (item (list input value values effect remark 
			example examples comments))
      (when (cdr item)
	(setq wrote-something t)
	(let ((head (car item))
	      (strings (cdr item)))
	  (terpri stream)
	  (write-string "\\item[" stream)
	  (write-string
	   (string-capitalize (string head)) stream)
	  (write-string "\\hfill] " stream)
	  (terpri stream)
	  (when strings
	    (dolist (str strings)
	      (write str :stream stream)
	      (terpri stream))))))
    (unless wrote-something
      (write-string "\\item[Comment\\hfill] No documentation available."
		    stream))))

(defun doc~latex-lispdef (str obj)
  (declare
   (authors nesmith)
   (input "A stream STR and a list OBJ")
   (value "not defined")
   (effect "Calls DOC~PARSE-DEF (see the description of that function)
with the car of OBJ and OBJ itself, then
prints the resulting values NAME DESC ARGS-OR-PARENTS DOC-ITEMS in the form:"
#{
\begin{verbatim}
\begin{lispdefinition}{DESC}{NAME}
{ARGS-OR-PARENTS
}
\item[SYMBOL1\hfill] 
    STRINGS1
\item[SYMBOL2\hfill]
    STRINGS2
...
\end{lispdefinition}
\end{verbatim} #})
   )
  (multiple-value-bind (name description args-or-parents doc-items)
      (doc~parse-def (car obj) obj)
    (when name
      (pp~with-style 'doc+latex-direct
	(doc=write-latex-head str 
			      name args-or-parents description doc-items)))))

(defun doc=latex-defclass (str obj)
  (declare
   (authors nesmith)
   (input "A stream STR and a list representing a defclass definition, i.e.,
a list of the form `(defclass ...)'.")
   (value "not defined")
   (effect "First writes a description of the class in LaTeX style to
STR (using the lispdefinition environment).  Then parses the slot definitions
of the class looking for accessors, readers and writers.  Each of the symbols
is looked up in the list doc*defined-generic-functions.  If the symbol is
in this list, then the corresponding generic function has already been
documented and need not be documented again.  If not, a dummy defgeneric
form is created and printed to STR (using write, so all pretty-printing
variables are in effect)."))
  (doc~latex-lispdef str obj)
  (let* ((slot-defs (fourth obj))
	(generic-defs nil)
	(class-name (string (second obj)))
	(inst-name (intern (concatenate 'string class-name "-OBJ"))))
    (dolist (slot-def slot-defs)
      (let ((reader-names nil)
	    (writer-names nil)
	    (accessor-names nil)
	    (doc nil)
	    (slot-name (car slot-def)))
	(setq slot-def (cdr slot-def))
	(loop
	 (unless slot-def (return))
	 (case (car slot-def)
	   (:reader (pushnew (cadr slot-def) reader-names))
	   (:writer (pushnew (cadr slot-def) writer-names))
	   (:accessor (pushnew (cadr slot-def) accessor-names))
	   (:documentation
	    (setq doc (cadr slot-def))))
	 (setq slot-def (cddr slot-def)))
	(when reader-names
	  (dolist (reader-name reader-names)
	    (unless (member reader-name doc*defined-generic-functions)
	      (push `(defgeneric ,reader-name (,inst-name)
		   (declare (input "An object of class "
				   ,class-name)
			    (effect "None")
			    (value "The value of the " 
				   ,(string slot-name) 
				   " slot of "
				   ,(string inst-name))))
		generic-defs))
	  ))
	(when accessor-names
	  (dolist (a-reader-name accessor-names)
	    (unless (member a-reader-name doc*defined-generic-functions)
	      (push `(defgeneric ,a-reader-name (,inst-name)
		       (declare (input "An object of class "
				       ,class-name)
				(effect "None")
				(value "The value of the " 
				       ,(string slot-name) 
				       " slot of "
				       ,(string inst-name)))
		       (:documentation "Can also be used with setf to set the slot's value"))
		    generic-defs))
	    (unless (member (list 'setf a-reader-name)
			    doc*defined-generic-functions
			    :test #'equal)
	      (push `(defgeneric (setf ,a-reader-name) (,inst-name new-value)
		       (declare (input "An object of class "
				       ,class-name
				       " and a new value ")
				(effect "The "
					,(string slot-name) 
					" slot of "
					,(string inst-name)
					" is set to NEW-VALUE")
				(value "NEW-VALUE is returned.")))
		    generic-defs))))
	(when writer-names
	  (dolist (writer-name writer-names)
	    (unless (member writer-name doc*defined-generic-functions)
	    (push `(defgeneric ,writer-name (,inst-name)
		     (declare (input "An object of class "
				   ,class-name)
			    (effect "None")
			    (value "The value of the " 
				   ,(string slot-name) 
				   " slot of "
				   ,(string inst-name))))
		  generic-defs))))))
    (dolist (generic-def generic-defs)
      (write generic-def :stream str)))
  )

(defun doc=style-symbol-p (sym)
  "True if SYM is the name of a defined PP+STYLE"
  (typep (pp~find-style sym) 'pp+style))


(deftype doc=style-symbol ()
  "A symbol naming a defined pprint style."
  `(and symbol (satisfies doc=style-symbol-p)))

(defun doc=modify-style-p (obj)
  (declare
   (authors nesmith)
   (input "Any lisp object")
   (value "T if OBJ is a list of the form (pp~modify-style FOO ...) and
FOO is of type DOC=STYLE-SYMBOL, else nil"))
  (and (listp obj)
       (eq (car obj) 'pp~modify-style)
       (typep (cadr obj) 'doc=style-symbol)))

(deftype doc=modify-style ()
  `(satisfies doc=modify-style-p))

(deftype doc=parse-def-method ()
  `(satisfies doc=parse-def-method-p))

(defun doc=parse-def-method-p (obj)
  (declare
   (authors nesmith)
   (input "Any lisp object")
   (value "T if OBJ is a list of the form (defmethod doc~parse-def ...), else
nil"))
  (and (listp obj)
       (eq (car obj) 'defmethod)
       (eq (cadr obj) 'doc~parse-def)))


(defun doc=extract-doc-pathname (stream inpath)
  (declare
   (authors nesmith)
   (input "A stream and a pathname INPATH")
   (value "undefined")
   (effect "reads each lisp form from INPATH and writes it to stream.
If pretty-printing is in effect, the current pretty-print dispatch table
will be used."))
  (let ((*package* (find-package "AGS"))) ; always read in the AGS package
    (terpri *error-output*)
    (write-string "; Extracting documentation from file " *error-output*)
    (write-string (let ((*print-pretty* nil))
		    (prin1-to-string (truename inpath)))
		  *error-output*)
    (let ((*standard-output* stream))
      (with-open-file
	(in-file inpath :direction :input)
	(let ((eof (gensym))
	      (obj nil)
	      (doc*defined-generic-functions nil))
	  (loop
	   (setq obj (read in-file nil eof))
	   (when (eq obj eof) (return))
	   (write obj :stream stream)))))))


(defun doc=latex-defmod (str obj)
  (declare
   (authors nesmith)
   (input "A stream and a MOD~DEFMOD form.")
   (effect "sets global value of DOC*EXPORT-LIST to the export list of
the MOD~DEFMOD form")
   (ignore str))
  (let ((pos (position :exports obj)))
    (setq doc*export-list (if pos (nth (1+ pos) obj))))
  )


(pp~defstyle doc+doc :parent pp+top
	     :help "The top style for documentation extraction.  Some
special methods are defined for this style.  For example, a MOD~DEFMOD
definition will set the value of the variable DOC*EXPORT-LIST, so that
the interface functions for the module can be determined.
An EXPORT declaration will add the exported symbols to DOC*EXPORT-LIST.

A PP~MODIFY-STYLE form will be evaluated during printing, if it refers to
an existing pretty-print style.  In this way modules can define
extensions to the documentation styles during the documentation process.

A list of the form {\\vb (DEFMETHOD DOC~PARSE-DEF \\ldots)} will be evaluated,
so that this generic function can be extended."
	     :pprint-methods
	     ((pathname doc=extract-doc-pathname 10)
	      ((cons (member mod~defmod))
	       doc=latex-defmod
	       100)
	      ((cons (member export))
	       (lambda (str obj)
		 (declare (ignore str))
		 (setq doc*export-list 
		       (append doc*export-list
			       (if (listp (second (second obj)))
				   (second (second obj))
				   (list (second (second obj)))))))
	       100)
	      (doc=modify-style
	       (lambda (str obj)
		 (declare (ignore str))
		 (eval obj))
	       100)
	      ((cons (member sys~define-condition))
	       (lambda (str obj)
		 (write (macroexpand-1 obj) :stream str))
	       100)
	      (doc=parse-def-method
	       (lambda (str obj)
		 (declare (ignore str))
		 (eval obj))
	       100)
	      ((cons (member eval-when let let*))
	       (lambda (str obj)
		 (dolist (e (cddr obj))
		   (write e :stream str)))
	       5)
	      ((cons (member prog1 prog2 progn))
	       (lambda (str obj)
		 (dolist (e (cdr obj))
		   (write e :stream str)))
	       5)
	      ((cons symbol)
	       (lambda (str obj)
		 (declare (ignore str))
		 (let ((*print-pretty* nil))
		   (format *error-output* 
			   "~%; Ignoring ~S ~S" 
			   (first obj) (second obj)))
	       0))
	      ((cons (member doc~latex))
	       (lambda (str obj)
		 (write-string (second obj) str))
	       5)))

(pp~defstyle doc+latex-direct :parent pp+top
	     :pprint-methods
	     ((symbol
	       doc=write-latex-symbol
	       10)
	      (string
	       doc=write-latex-string
	       10)
	      ((cons (member doc~latex))
	       (lambda (str obj)
		 (write-string (second obj) str))
	       15)
	      (null 
	       doc=write-latex-symbol
	       11)
	      (list
	       doc=write-latex-list
	       10)
	      ((cons (member quote))
	       (lambda (str obj)
		 (write-char #\' str)
		 (write (cadr obj) :stream str))
	       15)))


(pp~defstyle doc+latex-interface :parent doc+doc
	     :help "The style for \\LaTeX\\ output of interface functions."
	     :pprint-methods
	     (((cons (member defun defgeneric defmacro deftype 
			     defvar defconstant defparameter defsetf
			     pp~defstyle define-condition)
		     doc+interface)
	       doc~latex-lispdef 10)
	      ((cons (member defclass) doc+interface)
	       doc=latex-defclass
	       10)
	      ((cons (member defstruct)
		     (or doc+interface (cons doc+interface)))
	       doc~latex-lispdef
	       10)
	      ((cons (member doc~latex))
	       (lambda (str obj)
		 (write-string (second obj) str))
	       5)
	      ))

(pp~defstyle doc+latex-complete :parent doc+latex-interface
	     :help "The style for complete \\LaTeX\\ output."
	     :pprint-methods
	     (((cons (member defun defgeneric defmacro deftype 
			     defvar defconstant defparameter defsetf
			     pp~defstyle define-condition))
	       doc~latex-lispdef 15)
	      ((cons (member defclass))
	       doc=latex-defclass
	       15)
	      ((cons (member defstruct))
	       doc~latex-lispdef
	       15)))

(sys~define-condition
 doc=file-error (sys+error)
 ((pathname))
 (lambda (cond stream)
   (format stream "Can't find file ~A." (doc=file-error-pathname cond)))
 (:documentation "Signalled when PATHNAME can't be found."))



  

(defun doc~extract (infile &optional style (outfile nil) (type :unspecific))
  (declare (edited  "30-MAR-1994 09:34")
	   (authors AFIEDLER nesmith)
	   (input   "An input file, an optional style, an optional output file, and optional the"
		    "type of the output file, which defaults to :unspecific.")
	   (effect  "Writes the documentation for INFILE in style STYLE to OUTFILE."
		    "If OUTFILE is NIL, the default output file will be in the same"
		    "directory as INFILE, but with type TYPE.")
	   (value   "Undefined."))
  (flet ((wild-pathname-same-dir (path type)
	  (make-pathname :host (pathname-host path)
			 :device (pathname-device path)
			 :directory (pathname-directory path)
			 :name :wild
			 :type type)))
  (let* ((inpath (probe-file infile))
	 (outpath 
	  (if outfile
	      (merge-pathnames outfile
			       *default-pathname-defaults*)
	    (merge-pathnames (make-pathname :type type) inpath)
	      #+weg(translate-pathname 
	       inpath
	       (wild-pathname-same-dir inpath :wild)
	       (wild-pathname-same-dir inpath type)))))
      (if inpath
	  (with-open-file (out-stream outpath :direction :output
				      :if-does-not-exist :create
				      :if-exists :supersede)
	    (pp~pprint inpath style out-stream))
	  (sys~signal
	   (sys~make-condition 'doc=file-error 
			       :pathname (merge-pathnames infile)))))))

