;;; repair.el
;;;
;;; Created 13-Aug-1994 by Joern Richts (richts@cs.uni-sb.de).
;;;

;; vc must be required here because vc-checkin is redefined in this file.  If vc would be loaded after this
;; file, the redefintion would be lost.

(require 'vc)

(defun rp-get-repair-file-name (file)
  "Ask the inferior lisp for the repair file of FILE or if FILE is a buffer of the file for this buffer.
   If the file is not registered by defsystem in the inferior lisp, return NIL;
   if it has no repair file, return T. Else return the pathname of the repair file."
  (let* ((filename (if (bufferp file) (buffer-file-name file) file))
	 (repair-filename (ilisp-send (format "(and (fboundp 'mk::file-repair-file) (mk::file-repair-file %S))"
					      filename)
				      (format "Looking for repair file of %s" filename)
				      'repair-file)))
    (cond ((equal repair-filename "NIL") nil)
	  ((equal repair-filename "T") t)
	  (t (substring repair-filename 1 (1- (length repair-filename)))))))

(defun rp-get-repair-file-buffer (buffer)
  "Return the buffer with the repair file for BUFFER.
   If there is no repair file or there is no buffer with it, signal an error."
  (let* ((repair-filename (rp-get-repair-file-name buffer))
	 (repair-buffer (get-file-buffer repair-filename)))
    (cond ((null repair-filename) (error "This file is not registered by defsystem in the inferior lisp."))
	  ((equal repair-filename t) (error "This file has no repair file."))
	  (t (if (get-file-buffer repair-filename)
		 (get-file-buffer repair-filename)
	       (error "Repair file %S is not loaded." repair-filename))))))

(defun rp-start-patch (source-buffer)
  (interactive "bStart patch for buffer: ")
  (rp-find-repair-file-buffer source-buffer))

(defun rp-find-repair-file-buffer (&optional source-buffer repair-file)
  "Load the repair file of an lisp file (registered by defsystem in the inferior lisp) in a buffer.
   If the file is allready visited in a buffer, use that buffer.  Return this buffer.  If the file is not
   registered with defsystem in the inferior lisp or if it has no repair file, return NIL and signal an
   error.  If the repair file is read-only and checked in with some version control, try to check it out."
  (let* ((buffer (or source-buffer (current-buffer)))
	 (repair (or repair-file (rp-get-repair-file-name buffer))))
    (cond ((null repair) (error "This file is not registered by defsystem in the inferior lisp."))
	  ((equal repair t) (error "This file has no repair file."))
	  ((not (file-exists-p repair)) (error "Repair file %S does not exist." repair))
	  ((get-file-buffer repair)
	   (save-excursion
	       (let ((repair-buffer (get-file-buffer repair)))
		 (set-buffer repair-buffer)
		 (if (and buffer-read-only (vc-backend-deduce (buffer-file-name repair-buffer)))
		     ;; if the repair file is checked in try to check it out (but catch possible errors)
		     (condition-case nil (vc-next-action nil) (error nil))) 
		 (if buffer-read-only
		     (message "Repair file %S allready exists  in buffer %s but it is read-only." repair repair-buffer)
		   (message "Repair file %S allready exists  in buffer %s." repair repair-buffer))
		 repair-buffer)))
	  (t (save-excursion
	       (let ((repair-buffer (find-file-noselect repair)))
		 (if (null repair-buffer)
		     (message "Repair file %S could not be loaded." repair repair-buffer)
		   (set-buffer repair-buffer)
		   (if (and buffer-read-only (vc-backend-deduce (buffer-file-name repair-buffer)))
		     ;; if the repair file is checked in try to check it out (but catch possible errors)
		     (condition-case nil (vc-next-action nil) (error nil)))
		   (if buffer-read-only
		       (message "Repair file %S loaded in buffer %s but it is read-only." repair repair-buffer)
		     (message "Repair file %S loaded in buffer %s." repair repair-buffer)))
		 repair-buffer))))))


(defun rp-write-preamble (source-buffer repair-buffer)
  "Insert the preamble (appropriate for SOURCE-BUFFER) for a new entry into REPAIR-BUFFER.
   In a Lisp comment the date, the author, and the file of SOURCE-BUFFER are added followed by
   an in-package statement with the package of SOURCE-BUFFER."
  (require 'exdocu)
  (save-excursion
    (set-buffer source-buffer)
    (let ((pack (lisp-buffer-package)))
      (set-buffer repair-buffer)
      (goto-char (point-max))
      (insert "\n;;; Inserted at " (exdoc-time) " by " (exdoc-author) )
      (insert "\n;;; from " (buffer-file-name source-buffer) "\n")
      (if (and pack (not (equal pack "Error:")))
	  ;; "Error:" seems to be the value if the package does not exist in the inferior lisp
	  (insert "\n(in-package \"" pack "\")\n")))))


(defun rp-insert-region (start end source-buffer repair-buffer)
  "The region from START to END in SOURCE-BUFFER is added to REPAIR-BUFFER."
  (save-excursion 
    (set-buffer source-buffer)
    (let ((defun-string (buffer-substring start end)))
      (set-buffer repair-buffer)
      (goto-char (point-max))
      (insert "\n" defun-string "\n"))))
      

(defun rp-add-definition ()
  "Add the definition point is in to the repair file of the file of the current buffer."
  (interactive)
  (let* ((buffer (current-buffer))
	 (repair-buffer (rp-get-repair-file-buffer buffer))
	 (form (lisp-defun-region-and-name))
	 (start (car form))
	 (end (car (cdr form))))
    (save-excursion
      (rp-write-preamble buffer repair-buffer)
      (rp-insert-region start end buffer repair-buffer)
      (message "%s added to %s." (caddr form) repair-buffer))))


(defun rp-finish-patch (&optional repair-buffer)
  "Finish the patch for REPAIR-BUFFER (default is current buffer).
   Save the file and check it in if it is under version control."
  (interactive "bFinish patch in buffer: ")
  (let ((repair-buffer (or repair-buffer (current-buffer))))
    (save-excursion 
      (set-buffer repair-buffer)
      (save-buffer)
      (if (and (not buffer-read-only)
	       (vc-backend-deduce (buffer-file-name repair-buffer)))
	  (vc-next-action-on-file (buffer-file-name repair-buffer) nil (format "Repairing"))))))

(defun rp-patch-file (source-file &optional repair-file)
  "Add the changed defininitions of SOURCE-FILE to REPAIR-FILE after asking the user.
   For more information see rp-patch-buffer."
  (let* ((repair-file (or repair-file (rp-get-repair-file-name source-file)))
	 (source-buffer (get-file-buffer source-file)))
    (if (and (buffer-name source-buffer)
	     (y-or-n-p (format "Add changes of %s to %S "
			       (buffer-name source-buffer) repair-file)))
	(rp-patch-buffer source-buffer repair-file))))


(defun rp-patch-buffer (&optional source-buffer repair-file)
  "Add all changed definitions of buffer SOURCE-BUFFER (to the last version) to REPAIR-FILE.
   If REPAIR-FILE is omitted the inferior Lisp is asked for the repair file to
   SOURCE-BUFFER's file. Then SOURCE-BUFFER is compared with the last version that was
   checked in. For every definition that was changed, the user is asked if he wants to add
   this definintion to the repair file.  Finally the repair file is saved and compiled."
  (interactive "bPatch buffer: ")
  (let* ((buffer (or source-buffer (current-buffer)))
	 (repair-buffer (rp-find-repair-file-buffer buffer repair-file))
	 (old-diff-switches diff-switches))
    (unwind-protect
	(progn (setq diff-switches "-c0")
	       (if (> (save-excursion (vc-backend-diff (buffer-file-name buffer))) 1)
		   (error "Running 'diff' results in error.")))
      (setq diff-switches old-diff-switches))
    (unwind-protect
	(save-excursion
	  (let ((vc-buffer (get-buffer "*vc*")))
	    (if (null vc-buffer)
		(error "Running 'diff' produced no output"))
	    (set-buffer vc-buffer)
	    (goto-char (point-min))
	    (if (search-forward-regexp "^--- [0-9]+" nil t)
		(rp-add-diffs buffer repair-buffer vc-buffer)
	      (message "Running 'diff' produced no differences")
	      (sleep-for 1))))
      (rp-finish-patch repair-buffer))))

(defun rp-add-diffs (source-buffer repair-buffer vc-buffer)
  "Add the changed definitions of SOURCE-BUFFER to REPAIR-BUFFER according to the diff protocoll in VC-BUFFER."
  (let ((end-of-def -1)
	(def-changed nil))
    (set-buffer vc-buffer)
    (goto-char (point-min))
    (while (progn (set-buffer vc-buffer) (search-forward-regexp "^--- [0-9]+" nil t))
      (let ((start-line (rp-diff-start-line))
	    (end-line (rp-diff-end-line)))
	(do ((line start-line (1+ line)))
	    ((> line end-line))
	  (set-buffer source-buffer)
	  (goto-line line) (beginning-of-line)
	  (if (> (point) end-of-def)
	      (let* ((pos (point))
		     (form (lisp-defun-region-and-name))
		     (start (first form))
		     (end (second form)))
		(message "Line %d, Pos %d, Start %d, End %d" line pos start end)
		(when (and (<= start pos) (<= pos end))	; if we are inside the definition, i.e. not in a comment
		  (switch-to-buffer source-buffer)
		  (goto-char start)
		  (recenter 2)
		  (setq end-of-def end)
		  (when (save-excursion
			  (y-or-n-p (format "Add %s to repair file %S "
					    (third form) (buffer-name repair-buffer))))
		    (when (null def-changed)
		      (rp-write-preamble source-buffer repair-buffer)
		      (setq def-changed t))
		    (rp-insert-region start end source-buffer repair-buffer))))))))
    (if def-changed
	(message "Finished patch for %S" (buffer-file-name source-buffer))
      (message "No definitions addded to %s" (buffer-name repair-buffer)))
    (sleep-for 1)))

(defun rp-diff-start-line ()
  "Return the number that is preceding point."
  (save-excursion
    (backward-char)
    (car (read-from-string (thing-at-point 'word)))))

(defun rp-diff-end-line ()
  "If point is on a ',', return the number behind point; else the number preceding point."
  (save-excursion
    (forward-char)
    (if (= 44 (preceding-char))
	(car (read-from-string (thing-at-point 'word)))
      (backward-char)
      (rp-diff-start-line))))


(defun vc-checkin (file &optional rev comment)
  "Check in the file specified by FILE.  This function is changed in '~omega/emacs/keim/repair.el'.
   The optional argument REV may be a string specifying the new version level \(if nil increment
   the current level).  The file is either retained with write permissions zeroed, or deleted
   (according to the value of `vc-keep-workfiles').  COMMENT is a comment string; if omitted, a
   buffer is popped up to accept a comment.

   This function is changed: First it looks up in the inferior Lisp if there is a repair file for
   this file (defined with defsystem).  If this is the case, rp-patch-file is called, i.e. the user
   is asked if he wants to add his changes to the repair file."
  (setq vc-log-after-operation-hook 'vc-checkin-hook)
  (if (and (boundp 'ilisp-status)
	   (stringp ilisp-status)
	   (eq ':ready (car (read-from-string ilisp-status))))
      (let ((repair-file (rp-get-repair-file-name file)))
	(if (stringp repair-file)
	    (rp-patch-file file repair-file))))
  (vc-start-entry file rev comment
		  "Enter a change comment." 'vc-backend-checkin))


(provide 'repair)
