;;; vc-query.el  ---  checking in files when killing Emacs
;;;
;;; Created 15-Aug-1994 by Joern Richts (richts@informatik.rwth-aachen.de).
;;;
;;; Adapted for emacs 19.29 on 4-Jul-1995 with some useful suggestions by
;;;  Martin Steppler <steppler@dfv.rwth-aachen.de>
;;;
;;; This file installs a function in the kill-emacs-query-functions and
;;; kill-emacs-query-functions.  When killing a buffer that is still checked
;;; out, this function queries the user if he really wants to kill this buffer.
;;; When killing Emacs, for each file that is still checked out, the possibility
;;; is offered to check in this file; if some of these files are not checked in,
;;; at the end the user is asked if he really wants to kill Emacs.
;;; 
;;; The command checkin-some-buffer queries the user for each checked-out buffer
;;; if he wants to check it in.

(require 'vc)

(defun vcq-file-checked-out-p (file)
  "Determine if FILE is checked out by the user so it must be checked in."
  (and file 
       (file-exists-p file)
       (if (fboundp 'vc-backend)
	   (vc-backend file)
	 (vc-backend-deduce file))
       (string-equal (user-login-name) (vc-locking-user file))))

(defun vcq-checked-out-buffers-p (buffers)
  "Return T if there are buffers that have not been checked in yet."
  (cond ((null buffers) nil)
	((and (buffer-name (car buffers))
	      (buffer-file-name (car buffers))
	      (vcq-file-checked-out-p (buffer-file-name (car buffers))))
	 t)
	(t (vcq-checked-out-buffers-p (cdr buffers)))))

(defun vcq-file-unchanged-p (buffer)
  "Determine if FILE is unchanged, i.e. it must not be checked in but reverted."
  (and (not (buffer-modified-p buffer))
       (vc-workfile-unchanged-p (buffer-file-name buffer))))


(defun checkin-some-buffers (&optional arg)
  "Check in some visited files which are checked out by the user.  Asks user about each one.
   Optional argument (the prefix) non-nil means check in all without questions."
  (interactive "P")
  (if arg
      (mapcar (lambda (buffer)
		(if (vcq-file-checked-out-p (buffer-file-name buffer))
		    (if (vcq-file-unchanged-p buffer)
			;; if a checked out buffer is unmodified and the working file is unchanged,
			;; the user is asked in a very confusing way in vc-next-action-on-file.
			;; So we catch this case here and revert the buffer
			;; from the last checked in version ourselves.
			(save-excursion
			  (set-buffer buffer)
			  (vc-revert-buffer))
		      (vc-next-action-on-file (buffer-file-name buffer) nil (format "Emacs")))))
	      (buffer-list))
    (map-y-or-n-p '(lambda (buffer)
		     (if (vcq-file-checked-out-p (buffer-file-name buffer))
			 (format "Check in (uncommented) file %s? " (buffer-file-name buffer))
		       nil))
		  '(lambda (buffer)
		     (if (vcq-file-unchanged-p buffer)
			 ;; if a checked out buffer is unmodified and the working file is unchanged,
			 ;; the user is asked in a very confusing way in vc-next-action-on-file.
			 ;; So we catch this case here and revert the buffer
			 ;; from the last checked in version ourselves.
			 (save-excursion
			   (set-buffer buffer)
			   (vc-revert-buffer))
		       (vc-next-action-on-file (buffer-file-name buffer) nil (format "Emacs"))))
		  (buffer-list)
		  '("file" "files" "checkin"))))


(defun vcq-chekin-before-kill-buffer ()
  "(for kill-buffer-query-functions:) if the buffer is checked out, ask the user if he wants to kill it anyway."
  (if (vcq-file-checked-out-p buffer-file-name)
      (prog1 (y-or-n-p (format "File %s checked out; kill anyway? " buffer-file-name))
	(message nil))
    t))


(defun vcq-chekin-before-kill-emacs ()  
  "(for kill-emacs-query-functions:) Ask the user if he wants to check in the checked out buffers.
   For each buffer that is checked out the user is asked if he wants to check it in.
   If this is the case the buffer's file is checked in (uncommented, i.e. the comment 
   will be 'Exit emacs').  If some buffers have not been checked in in the end,
   the user is asked if he wants to kill Emacs anyway."
  (checkin-some-buffers)
  (if (vcq-checked-out-buffers-p (buffer-list))
      (prog1 (y-or-n-p "Checked out buffers exist; exit anyway? ")
	(message nil))
    t))





(add-hook 'kill-buffer-query-functions 'vcq-chekin-before-kill-buffer t)

(add-hook 'kill-emacs-query-functions 'vcq-chekin-before-kill-emacs)
