;;;; **************************************************************
;;;; * ********************************************************** *
;;;; * *                                                        * *
;;;; * *  MODULE:            convert.lisp                       * *
;;;; * *                                                        * *
;;;; * *  VERSION:           1.0                                * *
;;;; * *  LAST MODIFICATION: April 1993                         * *
;;;; * *                                                        * *
;;;; * *  AUTHOR:            C. Meyer (meyer@mpi-sb.mpg.de)     * *
;;;; * *                     MPI fuer Informatik                * *
;;;; * *                     Im Stadtwald                       * *
;;;; * *                     6600 Saarbruecken                  * *
;;;; * *                     Germany                            * *
;;;; * *                                                        * *
;;;; * ********************************************************** *
;;;; **************************************************************

#| This module gives a function 'convert~convert' to convert a term
   file into a file where terms are printed in lisp notation.
   See the documentation of 'convert~convert'.
|#

;;;**************************************************************
;;;* ********************************************************** *
;;;* *                                                        * *
;;;* *  SOURCE HANDLER                                        * *
;;;* *                                                        * *
;;;* ********************************************************** *
;;;**************************************************************

(defvar *stream* nil)
(defvar *current-char* nil)

(defun convert=init-source (name)
  (setq *stream* (open name :direction :input :element-type :default :if-does-not-exist nil))
  (convert=next-char)
  t
  )

(defun convert=finish-source ()
  (close *stream*)
  t
  )

(defun convert=current-char ()
  *current-char*
  )

(defun convert=next-char ()
  (setq *current-char* (read-char *stream* nil #\^Z))
  nil
  )

(defun convert=look-char ()
  (peek-char nil *stream* nil #\^Z)
  )

;;;**************************************************************
;;;* ********************************************************** *
;;;* *                                                        * *
;;;* *  SCANNER                                               * *
;;;* *                                                        * *
;;;* ********************************************************** *
;;;**************************************************************

(defvar *token* nil)

(defun convert=is-token-p (token)
  (eq (convert=token-id) token)
  )

(defun convert=token-id ()
  (first *token*)
  )

(defun convert=token-value ()
  (second *token*)
  )

(defun convert=accept (token)
  (if (convert=is-token-p token)
      (convert=get-token)
      (progn (convert=finish-source)
	     (error "Scanner (accept): unexpected token: ~A" token))
      )
  nil
  )

(defun convert=get-token ()
  (let ((current (convert=current-char)))
    (cond ((char= current #\space ) (convert=scan-white-space) (convert=get-token))
	  ((char= current #\newline ) (convert=scan-white-space) (convert=get-token))
	  ((char= current #\tab ) (convert=scan-white-space) (convert=get-token))
	  ((char= current #\( ) (convert=scan-left-bracket))
	  ((char= current #\) ) (convert=scan-right-bracket))
	  ((char= current #\, ) (convert=scan-comma))
	  ((char= current #\. ) (convert=scan-point))
	  ((convert=digit-p current)  (convert=scan-digit-name))
	  ((char= current #\^Z) (convert=scan-eof))
	  ((convert=letter-p current) (convert=scan-name))
	  (t (convert=finish-source) (error "Scanner (get-token): Unknown character: ~S" current))
	  )
    )
  )

(defun convert=letter-p (value)
  (or (and (char>= value #\A) (char<= value #\Z))
      (and (char>= value #\a) (char<= value #\z)))
  )

(defun convert=digit-p (value)
  (and (char>= value #\0) (char<= value #\9))
  )

(defun convert=valid-char-p (value)
  (or (char= value #\space)
      (char= value #\newline)
      (char= value #\tab)
      (char= value #\()
      (char= value #\))
      (char= value #\,)
      (char= value #\.)
      )
  )

(defun convert=scan-white-space ()
  (convert=next-char)
  )

(defun convert=scan-left-bracket ()
  (convert=next-char)
  (setq *token* '(:left))
  )

(defun convert=scan-right-bracket ()
  (convert=next-char)
  (setq *token* '(:right))
  )

(defun convert=scan-comma ()
  (convert=next-char)
  (setq *token* '(:comma))
  )

(defun convert=scan-point ()
  (convert=next-char)
  (setq *token* '(:point))
  )

(defun convert=scan-eof ()
  (convert=next-char)
  (setq *token* '(:eof))
  )

(defun convert=scan-name ()
  (do ((name (prog1 (string (convert=current-char)) (convert=next-char))
	     (prog1 (concatenate 'string name (string (convert=current-char))) (convert=next-char)))) 
      ((not (or (convert=letter-p (convert=current-char))
		(convert=digit-p (convert=current-char))))
       (setq *token* (list :name name)))
    )
  )

(defun convert=scan-digit-name ()
  (do ((name (prog1 (string (convert=current-char)) (convert=next-char))
	     (prog1 (concatenate 'string name (string (convert=current-char))) (convert=next-char))))
      ((not (or (convert=letter-p (convert=current-char))
		(convert=digit-p (convert=current-char))))
       (setq *token* (list :name (concatenate 'string name "c"))))
    )
  )

;;;**************************************************************
;;;* ********************************************************** *
;;;* *                                                        * *
;;;* *  PARSER                                                * *
;;;* *                                                        * *
;;;* ********************************************************** *
;;;**************************************************************

(defun convert=parse-decl (filename)
  (convert=init-source filename)
  (convert=get-token)
  (loop
   (when (convert=is-token-p :eof)
     (return t))
   (convert=parse-term-decl)
   (convert=accept :point)
   )
  (convert=finish-source)
  *decl-list*
  )

(defun convert=parse-term-decl ()
  (let* ((name        (convert=token-value))
	 (symbol-name (read-from-string name))
	 (arity       0))
    (convert=accept :name)
    (cond ((convert=is-token-p :left)
	   (convert=get-token)
	   (setq arity (convert=parse-decl-arg-list))
	   (convert=accept :right)

	   (unless (find symbol-name *declarated* :test #'eq)
	     (push symbol-name *declarated*)
	     (push (list 'constants (list symbol-name (make-list (1+ arity) :initial-element 'i))) *decl-list*)))
	  (t (unless (find symbol-name *declarated* :test #'eq)
	       (push symbol-name *declarated*)
	       (push (if (and (string>= name "u") (string<= name "zzzz"))
			 (list 'variables (list symbol-name 'i))
			 (list 'constants (list symbol-name 'i)))
		     *decl-list*))))
    )
  )

(defun convert=parse-decl-arg-list ()
  (convert=parse-term-decl)
  (do ((i 1 (1+ i)))
      ((convert=is-token-p :right) i)
    (convert=accept :comma)
    (convert=parse-term-decl)
    )
  )

(defun convert=parse-terms (filename stream)
  (convert=init-source filename)
  (convert=get-token)
  (loop
   (when (convert=is-token-p :eof)
     (return t))
   (print (convert=parse-term) stream)
   (convert=accept :point)
   )
  (convert=finish-source)
  )

(defun convert=parse-term ()
  (let ((symbol-name (read-from-string (convert=token-value)))
	(arg-list    nil))
    (convert=accept :name)
    (cond ((convert=is-token-p :left)
	   (convert=get-token)
	   (setq arg-list (convert=parse-arg-list))
	   (convert=accept :right)
	   
	   (list* symbol-name arg-list))
	  (t symbol-name))
    )
  )
	   
(defun convert=parse-arg-list ()
  (do ((term (list (convert=parse-term)) (nconc term (list (convert=parse-term)))))
      ((convert=is-token-p :right) term)
    (convert=accept :comma)
    )
  )

(defun convert~convert (filename)
  " **************************************************************
    INPUT:   A file name.
    RETURNS: None.
    SUMMARY: First the file is scanned for symbols to make a
             declaration whether a symbol is a constant,
             application or variable, also the arity of the
             application symbols is saved. Then all terms are
             scanned and converted into lisp notation.
             Output is written to a file called: 'filename'.con.
    ************************************************************** "

  (with-open-file (converted (concatenate 'string filename ".con") :direction :output :if-exists :rename)
    (declare (special *declarated* *decl-list*))
    (setq *declarated* nil *decl-list* nil)
    (format t "Symbol declaration.~%")
    (pprint (convert=parse-decl filename) converted)
    (format t "Convert terms.~%")
    (convert=parse-terms filename converted)
    )
  )
