;;;; **************************************************************
;;;; * ********************************************************** *
;;;; * *                                                        * *
;;;; * *  MODULE:            fpa.lisp                           * *
;;;; * *                                                        * *
;;;; * *  VERSION:           1.0                                * *
;;;; * *  LAST MODIFICATION: April 1993                         * *
;;;; * *                                                        * *
;;;; * *  AUTHOR:            C. Meyer (meyer@mpi-sb.mpg.de)     * *
;;;; * *                     MPI fuer Informatik                * *
;;;; * *                     Im Stadtwald                       * *
;;;; * *                     6600 Saarbruecken                  * *
;;;; * *                     Germany                            * *
;;;; * *                                                        * *
;;;; * ********************************************************** *
;;;; **************************************************************

;; evaluate the form (push :fpa-debug *features*)
;; for a debugging version of "fpa.lisp".
;; Stores more information in leaf nodes for
;; an output function for query trees.

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  LOCAL DEFINITIONS                                     * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

(eval-when (load compile eval)
  (defconstant *hash-max* 501        "Number of entries for the hash table")

  ;; The depth of the index balances good retrieval (deep index) against fast retrieval (flat index).
  (defconstant *real-index-depth* 60 "Number of entries for a key")

  (defvar *my-fpa-index* nil         "Actual index")

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  KEYS                                                  * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

  ;; A key contains path information. The position of b in the
  ;; User_TERM f(f(a,b),c) is referred to by the key [f1f2b], for example.

  (defvar *my-key* nil "All key functions operate on this global array")
  (setq *my-key* (make-array *real-index-depth* :initial-element 0))

  (defconstant *key-equal*   1 "Return value for fpa=key-compare")
  (defconstant *key-smaller* 2 "Return value for fpa=key-compare")
  (defconstant *key-bigger*  3 "Return value for fpa=key-compare")
    
  (defmacro fpa=key-element (key i)
    `(svref ,key ,i)
    )

  (defmacro fpa=key-set-all (key)
    `(setf (fpa=key-element ,key 0) 0)
    )

  (defmacro fpa=key-size (key)
    `(fpa=key-element ,key 0)
    )

  (defmacro fpa=set-key-size (size key)
    `(setf (fpa=key-size ,key) ,size)
    )
  
  (defmacro fpa=key-is-full-p (key)
    `(= (fpa=key-size ,key) (1- *real-index-depth*))
    )

  (defmacro fpa=push-key (element key)
    (let ((newvar (gensym)))
      `(let ((,newvar ,key))
	 (setf (fpa=key-element ,newvar (incf (fpa=key-size ,newvar))) 
	       ,element)
    )))

  (defmacro fpa=pop-and-push-key (element key)
    (let ((newvar (gensym)))
      `(let ((,newvar ,key))
	 (setf (fpa=key-element ,newvar (fpa=key-size ,newvar)) ,element)
	 )))

  (defmacro fpa=pop-key (key)
    `(decf (fpa=key-size ,key))
    )

  (defmacro fpa=pop-two-times (key)
    `(decf (fpa=key-size ,key) 2)
    )

  (defmacro fpa=copy-key (key)
    (let ((newvar (gensym)))
      `(let ((,newvar ,key))
	 (subseq ,newvar 0 (1+ (fpa=key-size ,newvar)))
	 )))
  
  (defun fpa=key-compare (key1 key2)
    " **************************************************************
      INPUT:   Two Keys.
      RETURNS: *key-equal*,   if Key1 is the same key as Key2.
               *key-bigger*,  if Key1 is bigger than Key2.
               *key-smaller*, if Key1 is smaller than Key2.
      SUMMARY: Compares the keys.
      CAUTION: ALL-Keys are the smallest keys.
      ************************************************************** "

    (cond ((> (fpa=key-size key1) (fpa=key-size key2)) *key-bigger*)
	  ((< (fpa=key-size key1) (fpa=key-size key2)) *key-smaller*)
	  (t (do ((i 1 (1+ i)))
		 ((> i (fpa=key-size key1)) *key-equal*)
	       (cond ((> (fpa=key-element key1 i) (fpa=key-element key2 i)) (return *key-bigger*))
		     ((< (fpa=key-element key1 i) (fpa=key-element key2 i)) (return *key-smaller*))
		     (t nil)))))
    )

  (defmacro fpa=key-equal-p (compared)
    `(= ,compared *key-equal*)
    )

  (defmacro fpa=key-smaller-or-equal-p (compared)
    `(/= ,compared *key-bigger*)
    )

  (defmacro fpa=key-equal-or-bigger-p (compared)
    `(/= ,compared *key-smaller*)
    )
  )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  HEADERS                                               * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

;; Headers solve hash collisions in the hash table by chaining.

(defstruct (fpa=header (:constructor fpa=make-header))
  "Slots: a key to find the header after collision; a connector to hold the terms"
  (key nil :read-only t)
  (connector nil))

(eval-when (load compile eval)
  (defmacro fpa=create-header (lookup key)
    " **************************************************************
      INPUT:   A lookup to the entry of the index which has to be
               inserted. A key that represents a path to the entry
               in the fpa-index.
      RETURNS: A new created header.
      ************************************************************** "

    `(fpa=make-header :key (fpa=copy-key ,key) :connector (list ,lookup))
    )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  HASHING                                               * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

  (defmacro fpa=hash-function (key)
    " **************************************************************
      INPUT:   A key.
      RETURNS: An index of the hash table.
      SUMMARY: Performs the hashing.
      ************************************************************** "
    (let ((newvar (gensym)))
      `(let ((,newvar ,key))
	 (do ((i 0 (1+ i))
	      (val 0 (+ (* val 7) (fpa=key-element ,newvar i))))
	     ((= i (fpa=key-size ,newvar)) (mod val *hash-max*)))
	 )))

  (defmacro fpa=get-hash (hash)
    `(svref (index=index *my-fpa-index*) ,hash)
    )

  (defmacro fpa=key-to-header (key)
    " **************************************************************
      INPUT:   A Key.
      RETURNS: A header which is the head of a list of connectors
               pointing to declarations having the path properties
               of key.
      ************************************************************** "
    (let ((newvar (gensym)))
      `(let ((,newvar ,key))
	 (dolist (header (fpa=get-hash (fpa=hash-function ,newvar)) nil)
	   (case (fpa=key-compare ,newvar (fpa=header-key header))
	     (*key-equal* (return header))
	     (*key-smaller* (return nil))))
	 )))

  (defmacro fpa=get-index-entry (key)
    `(let ((header (fpa=key-to-header ,key)))
       (when header
	     (fpa=header-connector header)))
    )

  (defmacro fpa=create-index-entry (key hash lookup)
    `(push (fpa=create-header ,lookup ,key) (fpa=get-hash ,hash))
    )

  (defmacro fpa=delete-index-entry (hash)
    `(pop (fpa=get-hash ,hash))
    )
  
  (defmacro fpa=term-id (lookup)
    `(car ,lookup)
    )

  (defmacro fpa=term-term (lookup)
    `(cdr ,lookup)
    )
  )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  INDEX CREATION AND DELETION                           * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

(defclass fpa+index (index+index)
  ((term-counter :initform 0 :accessor fpa=term-counter :documentation "id for the last inserted term"))
  (:documentation "fpa index class"))

(defmethod index~make-index ((index-type (eql :fpa)) path-type)
  " **************************************************************
    INPUT:   An index type (:fpa), a path type (:path).
    RETURNS: An instance of class fpa+index.
    SUMMARY: Initializes the index.
    CAUTION: MUST BE CALLED BEFORE ANY OTHER INDEX FUNCTION.
    ************************************************************** "

  (let ((table (make-array *hash-max* :initial-element nil)))
	(make-instance 'fpa+index :index table :path-type path-type))
  )

(defmethod index~create-entry ((fpa-index fpa+index) term)
  " **************************************************************
    INPUT:   An index to which an entry will be added. A term that
             has the path properties that will be searched later.
    RETURNS: An id for term.
    SUMMARY: Integrates a new entry in the fpa-index.
    ************************************************************** "

  (setq *my-fpa-index* fpa-index)
  (case (index=path-type fpa-index)
	(:path (let ((lookup (cons (incf (fpa=term-counter fpa-index)) term))) ; create a cons cell with id and 'term'.
		 ;; create an entry for the all-key, ...
		 (fpa=key-set-all *my-key*)
		 (fpa=create-single-entry lookup)

		 ;; for the 'term' and maybe its arguments.
		 (fpa=push-key (if (user~term-is-variable-p term)
			       (user~symbol-special-integer)
			       (user~symbol-to-integer (user~term-top term)))
			       *my-key*)
		 (fpa=apply-to-all-keys 'fpa=create-single-entry lookup term)
		 (fpa=term-counter fpa-index)))) ; return the id of the inserted 'term'.
  )

(defmethod index~delete-entry ((fpa-index fpa+index) id term)
  " **************************************************************
    INPUT:   An index from which an entry will be extracted,
             the id of term in the fpa-index, a term that
             has the path properties that will be deleted.
    RETURNS: The id of term.
    SUMMARY: Deletes all structures which are needed in order to
             keep a specific entry in the index.
    ************************************************************** "

  (setq *my-fpa-index* fpa-index)
  (case (index=path-type fpa-index)
	(:path
	 ;; delete the entry for 'term' referenced by its 'id' and the all-key, ...
	 (fpa=key-set-all *my-key*)
	 (fpa=delete-single-entry id)

	 ;; referenced the key for 'term' and maybe the key for all its arguments.
	 (fpa=push-key (if (user~term-is-variable-p term)
			   (user~symbol-special-integer)
			   (user~symbol-to-integer (user~term-top term)))
		       *my-key*)
	 (fpa=apply-to-all-keys 'fpa=delete-single-entry id term)
	 id)) ; return the 'id' of the deleted 'term'.
  )

(defmethod index~retrieve ((fpa-index fpa+index) mode term)
  " **************************************************************
    INPUT:   An index, a search modus and a term.
    RETURNS: A list of terms.
    SUMMARY: Finds all entries which are candidates for being
             unifiable/more general/instances (depending on mode)
             of term.
    CAUTION: As during retrieval no new entries may be added to
             the index, this procedure first completes the search.
    ************************************************************** "

  (case (index=path-type fpa-index)
	(:path (let ((tree (fpa=tree-build fpa-index mode term))) ; build the query tree for 'term'.
		 (do ((lookup 
		       ;; first evaluate the tree with the highest possible term id.
		       (fpa=tree-next tree (fpa=term-counter fpa-index))
		       ;; then evaluate the tree with the last retrieved term id.
		       (fpa=tree-next tree (fpa=term-id lookup)))
		      (terms nil (cons (fpa=term-term lookup) terms))) ; build the list of retrieved terms.
		     ((null lookup) terms))))) ; no more answers.
  )

(defun fpa=apply-to-all-keys (create-or-delete lookup term)
  " **************************************************************
    INPUT:   Create-or-delete is either the insertion or deletion
             function on connectors. Lookup is either a cons cell
             consisting of an id and a 'term' which is to be
             inserted in the fpa-index or an id of a 'term' which
             is to be deleted. Term is a subterm of this 'term'.
    RETURNS: None.
    SUMMARY: For all keys of term either call
             fpa=create-single-entry or fpa=delete-single-entry.
    ************************************************************** "

  (funcall create-or-delete lookup) ; create or delete the entry for 'term' depending on *my-key*.
  (let ((position 0)) ; current argument position of the application 'term', to build the key.
    (dolist (arg (user~term-list-of-args term))
	    ;; create or delete the entry for 'term' in the index for every key of 'term'.
	    (fpa=push-key (incf position) *my-key*)
	    (fpa=push-key (if (user~term-is-variable-p arg)
			      (user~symbol-special-integer)
			      (user~symbol-to-integer (user~term-top arg)))
			  *my-key*)
	    (fpa=apply-to-all-keys create-or-delete lookup arg)
	    (fpa=pop-two-times *my-key*))) ; pop both 'fpa=push-key' calls for the next argument of 'term'.
  )

(defun fpa=create-single-entry (lookup)
  " **************************************************************
    INPUT:   A cons cell consisting of an id and a 'term' which
             has to be integrated in the fpa-index.
    RETURNS: None.
    ************************************************************** "

  (let* ((hash (fpa=hash-function *my-key*)) ; calculate the index for the hash array.
	 (header-list (fpa=get-hash hash)))  ; get the list of headers from the hash array.
    (if (and header-list
	     (fpa=key-equal-or-bigger-p (fpa=key-compare *my-key* (fpa=header-key (first header-list)))))
	;; there is a list of headers and the *my-key* we are looking for is equal or bigger then
	;; the key of the first header (the list of headers is sorted by their keys).
	(let* ((last-head nil)   ; 'last-head' points to the header before 'header'.
	       (compared-keys 0) ; remember the comparison to avoid calling 'fpa=key-compare' twice.
	       (header
		;; the 'header-list' is scanned for a header with a key equal or bigger than *my-key*.
		(do ((header-scan header-list (rest header-scan)))
		    ((or (null header-scan)
			 (fpa=key-smaller-or-equal-p (setq compared-keys
							   (fpa=key-compare *my-key* (fpa=header-key (car header-scan))))))
		     (car header-scan))
		    (setq last-head header-scan))))
	  (if (and header
		   (fpa=key-equal-p compared-keys))
	      ;; a 'header' is found and its key is equal to *my-key*.
	      (push lookup (fpa=header-connector header))

	      ;; either there is no 'header' with a key equal to *my-key*, so 'last-head' points
	      ;; to the last header in the list and we have to concatenate it with a new header;
	      ;; or a 'header' is found but with a bigger key than *my-key* but 'last-head' points
	      ;; to a header with a smaller key than *my-key*, so we have to insert a new header
	      ;; between 'last-head' and 'header'.
	      (rplacd last-head (cons (fpa=create-header lookup *my-key*) (cdr last-head)))))

        ;; either there is no 'header-list' in the index array, so a new header must be stored
        ;; in the array or the key of the first header in 'header-list' is bigger than *my-key*,
        ;; so a new header must be pushed on the list of headers.
        (fpa=create-index-entry *my-key* hash lookup)))
  )

(defun fpa=delete-single-entry (id)
  " **************************************************************
    INPUT:   An id of a 'term' which has to be deleted.
    RETURNS: None.
    SUMMARY: Deletes the Connector and (if necessary) the Header
             leading to 'term' with the id.
    ************************************************************** "
  
  (let* ((hash (fpa=hash-function *my-key*)) ; calculate the index for the hash array.
	 (header-list (fpa=get-hash hash)))  ; get the list of headers from the hash array.
    (when header-list
	  ;; there is a 'header-list' for 'hash', otherwise there is nothing to delete.
	  (let* ((last-head nil)   ; 'last-head' points to the header before 'header'.
		 (compared-keys 0) ; remember the comparison to avoid calling 'fpa=key-compare' twice.
		 (connector nil)   ; for faster access on the list of terms.
		 (header 
		  ;; the 'header-list' is scanned for a header with a key equal or bigger than *my-key*.
		  (do ((header-scan header-list (rest header-scan)))
		      ((or (null header-scan)
			   (fpa=key-smaller-or-equal-p (setq compared-keys
							     (fpa=key-compare *my-key* (fpa=header-key (car header-scan))))))
		       (car header-scan))
		      (setq last-head header-scan))))		 
	    (when (and header
		       (fpa=key-equal-p compared-keys))
	      ;; there is a 'header' for *my-key*, otherwise there is nothing to delete.
	      (setq connector (fpa=header-connector header))
	      (if (rest connector)
		  ;; 'connector' consists of more than one term.
		  (if (= id (fpa=term-id (first connector)))
		      ;; the first element of 'connector' is the 'term' we are looking for.
		      (pop (fpa=header-connector header))
		      ;; 'term' should be one of the terms in the rest of 'connector'.
		      (delete id connector :count 1 :start 1 :test #'= :key #'car))
		  ;; 'connector' is a list of one term.
		  (when (= id (fpa=term-id (first connector)))
			;; this term is the desired one, so the hold 'header' must be deleted.
			(if last-head
			    ;; 'header' is not the first header in the 'header-list'.
			    (rplacd last-head (cddr last-head))
			    ;; 'header' is the first header, so pop it.
			    (fpa=delete-index-entry hash))))))))
  )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  QUERY TREE NODE CONSTRUCTION                          * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

#| The query trees consist of three different kinds of nodes
   (AND-, OR-, or LEAF-nodes). For a query the query tree
   is constructed and ids to entries of the index are
   bubbled up to the root of the query tree. |#

(eval-when (load compile eval)
  (defconstant *or-node-type*   1 "Return value for fpa=node-type")
  (defconstant *leaf-node-type* 2 "Return value for fpa=node-type")
  (defconstant *and-node-type*  3 "Return value for fpa=node-type")
  
  (defvar *free-leaf-list* nil "Free list for leaf nodes")
  (defvar *free-and-list* nil  "Free list for and nodes")
  (defvar *free-or-list* nil   "Free list for or nodes")
  )

(defstruct fpa=node
  "Slots: type contains one of the *..-node-type* constants for fast access on nodes"
  (type 0 :read-only t)
  )

(defstruct (fpa=leaf-node (:constructor fpa=make-leaf-node-struct)
			  (:include fpa=node (type *leaf-node-type*)))
  "Slots: terms contains a list of terms"
  #+fpa-debug (key nil)
  (terms nil)
  )

(defstruct (fpa=and-node (:constructor fpa=make-and-node-struct)
			 (:include fpa=node (type *and-node-type*)))
  "Slots: subs contains a list of sons; sons is the number of sons"
  (subs nil)
  (sons 2)
  )

(eval-when (load compile eval)
  (defmacro fpa=and-node-struct-p (node)
    `(= (fpa=node-type ,node) *and-node-type*)
    )
  )

(defstruct (fpa=or-node (:constructor fpa=make-or-node-struct)
			(:include fpa=node (type *or-node-type*)))
  "Slots: left, right contain the two sons;
          lastleft, lastright contain cons cells with id and a term not used on the last retrieval"
  (left nil)
  (right nil)
  (lastleft nil)
  (lastright nil)
  )

(eval-when (load compile eval)
  (defmacro fpa=free-leaf-node (node)
    (let ((newvar (gensym)))
      `(let ((,newvar ,node))
	 (progn (setf (fpa=leaf-node-terms ,newvar) nil
		       #+fpa-debug (fpa=leaf-node-key ,newvar) #+fpa-debug nil)
		 (push ,newvar *free-leaf-list*))
	 )))

  (defmacro fpa=free-or-node (node)
    (let ((newvar (gensym)))
      `(let ((,newvar ,node))
	 (progn (setf (fpa=or-node-lastleft ,newvar) nil
		       (fpa=or-node-lastright ,newvar) nil
		       (fpa=or-node-left ,newvar) nil
		       (fpa=or-node-right ,newvar) nil)
		 (push ,newvar *free-or-list*))
	 )))

  (defmacro fpa=free-and-node (node)
    (let ((newvar (gensym)))
      `(let ((,newvar ,node))
	 (progn (setf (fpa=and-node-sons ,newvar) 2
		      (fpa=and-node-subs ,newvar) nil)
		(push ,newvar *free-and-list*))
	 )))
  
  (defmacro fpa=free-and-without-nil-node (node nil-node)
    "Frees node and all subnodes of node except the subnode nil-node"
    (let ((newvar (gensym)))
      `(let ((,newvar ,node))
	 (progn (dolist (sub-node (fpa=and-node-subs ,newvar))
		    (unless (eq ,nil-node sub-node)
			    (fpa=tree-free sub-node))
		    )
	    (fpa=free-and-node ,newvar))
	 )
      ))
)
(defun fpa=tree-free (tree)
  " **************************************************************
    INPUT:   A tree which has to be deleted.
    RETURNS: None.
    SUMMARY: Frees tree with all subtrees.
    ************************************************************** "

  (when tree
    (let ((type (fpa=node-type tree)))
      (cond ((= type #.*or-node-type*)
	     (fpa=tree-free (fpa=or-node-left tree))
	     (fpa=tree-free (fpa=or-node-right tree))
	     (fpa=free-or-node tree))
	    ((= type #.*leaf-node-type*)
	     (fpa=free-leaf-node tree))
	    (t (dolist (sub-node (fpa=and-node-subs tree))
		 (fpa=tree-free sub-node))
	       (fpa=free-and-node tree)))))
  )

;; the following functions should be inline compiled

(eval-when (load compile eval)
	   (proclaim '(inline fpa=make-leaf-node))
	   (proclaim '(inline fpa=make-and-node))
	   (proclaim '(inline fpa=make-and-node-inst))
	   (proclaim '(inline fpa=make-and-node-unif))
	   (proclaim '(inline fpa=make-and-node-gen))
	   (proclaim '(inline fpa=make-or-node))
	   )

(defun fpa=make-leaf-node (key)
  " **************************************************************
    INPUT:   A key to get the list of terms from the index.
    RETURNS: A leaf-node containing the connector.
    SUMMARY: If there are any old leaf-nodes left in the free list
             such a leaf-node is returned filled with the
             connector recieved from the index,
             otherwise a new one is created and returned.
    ************************************************************** "

  (let ((connector (fpa=get-index-entry key)))
    (if connector
	;; there are terms in the index referenced by 'key'.
	(if *free-leaf-list*
	    ;; there is at least one leaf-node left in the free-list, take it.
	    (let ((leaf-node (pop *free-leaf-list*)))
	      (setf (fpa=leaf-node-terms leaf-node) connector
		    #+fpa-debug (fpa=leaf-node-key leaf-node) #+fpa-debug (fpa=copy-key key))
	      leaf-node)
	    ;; no more old leaf-nodes, create a new one. 
	    (fpa=make-leaf-node-struct :terms connector
				       #+fpa-debug :key #+fpa-debug (fpa=copy-key key)))
        ;; no terms are saved in the index, so just return nil.
	nil))
  )
	     
(defun fpa=make-and-node (subs)
  " **************************************************************
    INPUT:   Subs contains a list of two subnodes.
    RETURNS: An and-node containing subs.
    SUMMARY: If there are any old and-nodes left in the free list
             such an and-node is returned filled with subs,
             otherwise a new one is created and returned.
    CAUTION: The slot 'sons' is initialized with 2,
             so this function works correctly only if subs
             consists of two subnodes.
    ************************************************************** "

  (if *free-and-list*
      ;; there is at least one and-node left in the free-list, take it.
      (let ((and-node (pop *free-and-list*)))
	(setf (fpa=and-node-subs and-node) subs) ; the slot 'sons' is initialized to 2.
	and-node)
      ;; no more old and-nodes, create a new one. 
      (fpa=make-and-node-struct :subs subs))
  )

(defun fpa=make-and-node-inst (term key)
  " **************************************************************
    INPUT:   A term which has the path properties of key.
    RETURNS: An and-node with all subnodes described
             by the arguments of term. If only one successor
             is found the root-node of this tree is returned.
             On retrieval this and-node or this root-node
             returns instances of term.
    SUMMARY: Scans the arguments of term, builds for all
             non-variable arguments the subtrees, collects
             them in a list and stores this list in 'subs'.
    CAUTION: The function terminates only if the argument list of
             term consists of at least one non-variable argument.
    ************************************************************** "

  (let ((position 1)                             ; current argument position of the application 'term', to build the key.
	(counter  (fpa=key-size key))            ; to save the key size at function entry.
	(arglist  (user~term-list-of-args term)) ; for faster access on the list of arguments.
	(new      nil)                           ; the current new created subnode.
	(old      nil))                          ; the latest created node, old will be returned.
    (loop
     ;; until the first non-variable argument is found.
     (cond ((user~term-is-variable-p (first arglist))
	    ;; argument is a variable, so try the next argument.
	    (setq arglist (rest arglist))
	    (incf position))
	   (t 
	    ;; a non-variable argument is found, compute the instance-tree for this argument.
	    (fpa=push-key position key)
	    (setq old (fpa=instance-tree-build (first arglist)))
	    (fpa=set-key-size counter key) ; reset the old key size.

	    ;; all returned subtrees must be non-nil, otherwise return nil.
	    (if old
		(return (dolist (arg (rest arglist) old) ; consider the rest of arguments, old will be returned.
				(cond ((user~term-is-variable-p arg)
				       ;; do not consider variables.
				       (incf position))

				      (t 
				       ;; a non-variable term.
				       (fpa=push-key (incf position) key)
				       (cond ((setq new (fpa=instance-tree-build arg))
					      ;; the subtree of the current argument is non-nil.
					      (cond ((fpa=and-node-struct-p old)
						     (if (fpa=and-node-struct-p new)
							 ;; 'old' and 'new' are both and-nodes.
							 (progn (incf (fpa=and-node-sons old) (fpa=and-node-sons new))
								(nconc (fpa=and-node-subs old) (fpa=and-node-subs new)))
						         ;; only 'old' is an and-node.
						         (progn (incf (fpa=and-node-sons old))
								(push new (fpa=and-node-subs old)))))

						    ;; 'old' can be a leaf- or or-node only computed by
						    ;; the first call to fpa=instance-tree-build outside the 'dolist'.
						    ((fpa=and-node-struct-p new)
						     ;; only 'new' is an and-node
						     (incf (fpa=and-node-sons new))
						     (push old (fpa=and-node-subs new))
						     (setq old new))

						    (t
						     ;; both 'old' and 'new' are no and-nodes.
						     (setq old (fpa=make-and-node (list old new))))))

					     (t 
					      ;; we have an empty subtree for the current argument.
					      (fpa=set-key-size counter key) ; reset the old key size.
					      (fpa=tree-free old)            ; free the already created nodes.
					      (return nil)))
				       (fpa=set-key-size counter key)))))    ; reset the old key size.
	      ;; the first non-variable argument has a nil tree, there is nothing to free.
	      (return nil))))))
  )

(defun fpa=make-and-node-unif (term key)
  " **************************************************************
    INPUT:   A term which has the path properties of key.
    RETURNS: An and-node with all subnodes described
             by the arguments of term. If only one successor
             is found the root-node of this tree is returned.
             On retrieval this and-node or this root-node
             returns unification partners of term.
    SUMMARY: Scans the arguments of term, builds for all
             non-variable arguments the subtrees, collects
             them in a list and stores this list in 'subs'.
    CAUTION: The function terminates only if the argument list of
             term consists of at least one non-variable argument.
    ************************************************************** "

  (let ((position 1)                             ; current argument position of the application 'term', to build the key.
	(counter  (fpa=key-size key))            ; to save the key size at function entry.
	(arglist  (user~term-list-of-args term)) ; for faster access on the list of arguments.
	(new      nil)                           ; the current new created subnode.
	(old      nil))                          ; the latest created node, old will be returned.
    (loop
     ;; until the first non-variable argument is found.
     (cond ((user~term-is-variable-p (first arglist))
	    ;; argument is a variable, so try the next argument.
	    (setq arglist (rest arglist))
	    (incf position))

	   (t 
	    ;; a non-variable argument is found, compute the unification-tree for this argument.
	    (fpa=push-key position key)
	    (setq old (fpa=unification-tree-build (first arglist)))
	    (fpa=set-key-size counter key) ; reset the old key size.

	    ;; all returned subtrees must be non-nil, otherwise return nil.
	    (if old
		(return (dolist (arg (rest arglist) old) ; consider the rest of arguments, old will be returned.
				(cond ((user~term-is-variable-p arg)
				       ;; do not consider variables.
				       (incf position))

				      (t 
				       ;; a non-variable term.
				       (fpa=push-key (incf position) key)
				       (cond ((setq new (fpa=unification-tree-build arg))
					      ;; the subtree of the current argument is non-nil.
					      (cond ((fpa=and-node-struct-p old)			     
						     (if (fpa=and-node-struct-p new)
							 ;; 'old' and 'new' are both and-nodes.
							 (progn (incf (fpa=and-node-sons old) (fpa=and-node-sons new))
								(nconc (fpa=and-node-subs old) (fpa=and-node-subs new)))
						         ;; only 'old' is an and-node.
						         (progn (incf (fpa=and-node-sons old))
								(push new (fpa=and-node-subs old)))))

						    ;; 'old' can be a leaf- or or-node only computed by
						    ;; the first call to fpa=unification-tree-build outside the 'dolist'.
						    ((fpa=and-node-struct-p new)
						     ;; only 'new' is an and-node
						     (incf (fpa=and-node-sons new))
						     (push old (fpa=and-node-subs new))
						     (setq old new))

						    (t 
						     ;; both 'old' and 'new' are no and-nodes.
						     (setq old (fpa=make-and-node (list old new))))))

					     (t
					      ;; we have an empty subtree for the current argument.
					      (fpa=set-key-size counter key) ; reset the old key size.
					      (fpa=tree-free old)            ; free the already created nodes.
					      (return nil)))
				       (fpa=set-key-size counter key))))) ; reset the old key size.
	      ;; the first non-variable argument has a nil tree, there is nothing to free.
	      (return nil))))))
  )

(defun fpa=make-and-node-gen (term key)
  " **************************************************************
    INPUT:   A term which has the path properties of key.
    RETURNS: An and-node with all subnodes described
             by the arguments of term. If only one successor
             is found the root-node of this tree is returned.
             On retrieval this and-node or this root-node
             returns generalizations of term.
    SUMMARY: Scans the arguments of term, builds for all
             arguments the subtrees, collects them in a list
             and stores this list in 'subs'.
    ************************************************************** "

  (let ((position 1)                             ; current argument position of the application 'term', to build the key.
	(counter  (fpa=key-size key))            ; to save the key size at function entry.
	(arglist  (user~term-list-of-args term)) ; for faster access on the list of arguments.
	(new      nil)                           ; the current new created subnode.
	(old      nil))                          ; the latest created node, old will be returned.

    ;; we have to consider every kind of argument.
    (fpa=push-key 1 key)
    (setq old (fpa=generalization-tree-build (first arglist)))
    (fpa=set-key-size counter key) ; reset the old key size.

    ;; all returned subtrees must be non-nil, otherwise return nil.
    (if old 
	(dolist (arg (rest arglist) old)
		(fpa=push-key (incf position) key)
		(cond ((setq new (fpa=generalization-tree-build arg))
		       ;; the subtree of the current argument is non-nil.
		       (cond ((fpa=and-node-struct-p old)
			      (if (fpa=and-node-struct-p new)
				  ;; 'old' and 'new' are both and-nodes.
				  (progn (incf (fpa=and-node-sons old) (fpa=and-node-sons new))
					 (nconc (fpa=and-node-subs old) (fpa=and-node-subs new)))
				  ;; only 'old' is an and-node.
				  (progn (incf (fpa=and-node-sons old))
					 (push new (fpa=and-node-subs old)))))

			     ;; 'old' can be a leaf- or or-node only computed by
			     ;; the first call to fpa=generalization-tree-build outside the 'dolist'.
			     ((fpa=and-node-struct-p new)
			      ;; only 'new' is an and-node
			      (incf (fpa=and-node-sons new))
			      (push old (fpa=and-node-subs new))
			      (setq old new))

			     (t
			      ;; both 'old' and 'new' are no and-nodes.
			      (setq old (fpa=make-and-node (list old new))))))

		      (t
		       ;; we have an empty subtree for the current argument.
		       (fpa=set-key-size counter key) ; reset the old key size.
		       (fpa=tree-free old)            ; free the already created nodes.
		       (return nil)))
		(fpa=set-key-size counter key)) ; reset the old key size.
      ;; the first argument has a nil tree, there is nothing to free.
      nil))
  )

(defun fpa=make-or-node (left right)
  " **************************************************************
    INPUT:   Two nodes which become the subnodes of the or-node.
    RETURNS: An or-node containing the two nodes left and right.
    SUMMARY: If one of the two nodes is nil there is no reason
             to build an or-node, so the other node is returned,
             If there are any old or-nodes left in the free list
             such an or-node is returned filled with the two
             nodes, otherwise a new one is created and returned.
    ************************************************************** "

  (cond ((null left) right)
	((null right) left)
	(t 
	 ;; both subnodes have non-nil values, so we have to build an or-node.
	 (if *free-or-list*
	     ;; there is at least one or-node left in the free-list, take it.
	     (let ((or-node (pop *free-or-list*)))
	       (setf (fpa=or-node-left or-node) left
		     (fpa=or-node-right or-node) right) ; the other slots are initialized to nil.
	       or-node)
	     ;; no more old or-nodes, create a new one. 
	     (fpa=make-or-node-struct :left left :right right))))
  )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  QUERY TREE CONSTRUCTION                               * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

(eval-when (load compile eval)
	   (defconstant *instance*       1 "Mode constant for term retrieval")
	   (defconstant *unification*    2 "Mode constant for term retrieval")
	   (defconstant *generalization* 3 "Mode constant for term retrieval")
	   )

(defun fpa=instance-tree-build (term)
  " **************************************************************
    INPUT:   A term.
    RETURNS: A query tree which returns instances of term
             on retrieval.
    ************************************************************** "

  (cond ((user~term-is-variable-p term)
	 ;; all terms are instances of a variable.
	 (fpa=key-set-all *my-key*)
	 (fpa=make-leaf-node *my-key*))

	((user~term-is-application-p term)
	 ;; an instance of an application is the application itself.
	 (fpa=push-key (user~symbol-to-integer (user~term-top term)) *my-key*)
	 (if (or (fpa=key-is-full-p *my-key*) (user~term-all-args-are-var-p term))
	     ;; there is no room left to consider all arguments of the application or
	     ;; there are no more constraints because all arguments of 'term' are variables.
	     (fpa=make-leaf-node *my-key*)
	     ;; we have enough space and we have to consider the argument list of 'term'.
	     (fpa=make-and-node-inst term *my-key*)))

	(t 
	 ;; an instance of a constant is the constant itself.
	 (fpa=push-key (user~symbol-to-integer (user~term-top term)) *my-key*)
	 (fpa=make-leaf-node *my-key*))
	)
  )

(defun fpa=unification-tree-build (term)
  " **************************************************************
    INPUT:   A term.
    RETURNS: A query tree which returns unification partners of
             term on retrieval.
    ************************************************************** "

  (let ((treenode nil)) ; holds a subtree for a successor of an or-node
    (cond ((user~term-is-variable-p term)
	   ;; all terms are unification partner of a variable.
	   (fpa=key-set-all *my-key*)
	   (fpa=make-leaf-node *my-key*))

	  ((user~term-is-application-p term)
	   ;; an unification partner of an application can be a variable or ...
	   (fpa=push-key (user~symbol-special-integer) *my-key*)
	   (setq treenode (fpa=make-leaf-node *my-key*))
	   ;; or the application itself.
	   (fpa=pop-and-push-key (user~symbol-to-integer (user~term-top term)) *my-key*)
	   (if (or (fpa=key-is-full-p *my-key*) (user~term-all-args-are-var-p term))
	       ;; there is no room left to consider all arguments of the application or
	       ;; there are no more constraints because all arguments of 'term' are variables.
	       (fpa=make-or-node treenode (fpa=make-leaf-node *my-key*))
	       ;; we have enough space and we have to consider the argument list of 'term'.
	       (fpa=make-or-node treenode (fpa=make-and-node-unif term *my-key*))))

	  (t 
	   ;; an unification partner of a constant can be a variable or ...
	   (fpa=push-key (user~symbol-special-integer) *my-key*)
	   (setq treenode (fpa=make-leaf-node *my-key*))
	   ;; or the constant itself.
	   (fpa=pop-and-push-key (user~symbol-to-integer (user~term-top term)) *my-key*)
	   (fpa=make-or-node treenode (fpa=make-leaf-node *my-key*)))
	  )				;cond
    )					;let
  )

(defun fpa=generalization-tree-build (term)
  " **************************************************************
    INPUT:   A term.
    RETURNS: A query tree which returns generalizations of term
             on retrieval.
    ************************************************************** "

  (let ((treenode nil)) ; holds a subtree for a successor of an or-node.
    (cond ((user~term-is-variable-p term)
	   ;; a generalization of a variable is a variable.
	   (fpa=push-key (user~symbol-special-integer) *my-key*)
	   (fpa=make-leaf-node *my-key*))

	  ((user~term-is-application-p term)
	   ;; a generalization of an application can be a variable or ...
	   (fpa=push-key (user~symbol-special-integer) *my-key*)
	   (setq treenode (fpa=make-leaf-node *my-key*))
	   ;; or the application itself.
	   (fpa=pop-and-push-key (user~symbol-to-integer (user~term-top term)) *my-key*)
	   (if (fpa=key-is-full-p *my-key*)
	       ;; there is no room left to consider all arguments of the application.
	       (fpa=make-or-node treenode (fpa=make-leaf-node *my-key*))
	       ;; we have enough space to compute the subtrees of the argument list.
	       (fpa=make-or-node treenode (fpa=make-and-node-gen term *my-key*))))

	  (t 
	   ;; a generalization of a constant can be a variable or ...
	   (fpa=push-key (user~symbol-special-integer) *my-key*)
	   (setq treenode (fpa=make-leaf-node *my-key*))
	   ;; or the constant itself.
	   (fpa=pop-and-push-key (user~symbol-to-integer (user~term-top term)) *my-key*)
	   (fpa=make-or-node treenode (fpa=make-leaf-node *my-key*)))
	  )                             ;cond
    )                                   ;let
  )

(defun fpa=tree-build (fpa-index mode term)
  " **************************************************************
    INPUT:   An index.
             Mode is one of the constants:
             *unification*, *generalization*, or *instance*.
             Term is the query term.
    RETURNS: A query tree for term.
    ************************************************************** "

  (setq *my-fpa-index* fpa-index)
  ;; here it is very important to use a case construct on integers
  ;; because only this version will be fully optimized by the compiler.
  (case mode
    (#.*instance*       (fpa=key-set-all *my-key*)
			(fpa=instance-tree-build term))
    (#.*unification*    (fpa=key-set-all *my-key*)
			(fpa=unification-tree-build term))
    (#.*generalization* (fpa=key-set-all *my-key*)
			(fpa=generalization-tree-build term)))
  )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  QUERY TREE NODE EVALUATION                            * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

;; the following functions should be inline compiled

(eval-when (load compile eval)
	   (proclaim '(inline fpa=leaf-node-next))
	   (proclaim '(inline fpa=and-node-next))
	   (proclaim '(inline fpa=or-node-next))
	   )

(defun fpa=leaf-node-next (node maximum)
  " **************************************************************
    INPUT:   A node and a maximum id.
    RETURNS: A cons cell which consists of an id equal or smaller
             than the maximum and the associated term.
    SUMMARY: The connector of the leaf-node is shorten to the
             next cell right after the returned cell.
    ************************************************************** "

  (do ((connector (fpa=leaf-node-terms node) (rest connector))) ; holds the list of terms.
      ((null connector)
       ;; all term ids are bigger than the maximum.
       (fpa=free-leaf-node node)
       nil)
      (when (>= maximum (fpa=term-id (first connector)))
	    ;; the actual term id is equal or smaller to maximum.
	    (setf (fpa=leaf-node-terms node) ; the next call will have a smaller maximum,
                  (rest connector))          ; so remember only the rest of the list of terms.
	    (return (first connector))))
  )

(defun fpa=and-node-next (node maximum)
  " **************************************************************
    INPUT:   A node and a maximum id.
    RETURNS: A cons cell which consists of an id equal or smaller
             than the maximum and the associated term.
    SUMMARY: The list of subnodes 'subs' is scanned until all
             subnodes return equal ids which are equal or
             smaller than the maximum.
    ************************************************************** "

  (let* ((number-of-equal-sons 0)                        ;must be equal to 'sons' to return a non-nil value.
	(sons                  (fpa=and-node-sons node)) ;number of subnodes of 'node'.
	(orig-subs             (fpa=and-node-subs node)) ;list of subnodes of 'node' for faster access.
	(subs orig-subs)                                 ;list of subnodes of 'node'.
	(lookup                nil))                     ;the actual cons cell with id and term.
    (loop
     (if (setq lookup (fpa=tree-next (first subs) maximum))
	 ;; lookup consists of a cons cell with id and term.
	 (let ((term-id (fpa=term-id lookup))) ;for faster access on the id.
	   (if (< term-id maximum)
	       ;; the id is unequal to all ids computed before.
	       (setq number-of-equal-sons 1 ; so we just have only one equal son
		     maximum term-id)	    ; and a new maximum...
	       ;; the id is equal to all other ids computed before
	       (when (= sons (incf number-of-equal-sons))
		 ;; all sons returned equal ids.
		 (return lookup)))
	   ;; simulate a circular list.
	   (setq subs (or (rest subs) orig-subs))
	   )
       ;; the actual subtree did not return a term.
       (progn (fpa=free-and-without-nil-node node (first subs))
	      (return nil)))))
  )

(defun fpa=or-node-next (node maximum)
  " **************************************************************
    INPUT:   A node and a maximum id.
    RETURNS: A cons cell which consists of an id equal or smaller
             than the maximum and the associated term.
    SUMMARY: If there are old ids stored in the 'last' slots,
             the bigger one - equal or smaller than the maximum -
             is returned, otherwise missing ids are computed
             and again the bigger one - equal or smaller than the
             maximum - is returned and the other value is saved in
             the appropriate 'last' slot for the next retrieval.
    ************************************************************** "

  (let ((lookup1 (fpa=or-node-lastleft node))     ;old id recieved from the left subtree.
	(lookup2 (fpa=or-node-lastright node)))   ;old id recieved from the right subtree.

    ;; are there any old ids equal or smaller than maximum ?
    (when (or (null lookup1) (> (fpa=term-id lookup1) maximum))
      (setq lookup1 (fpa=tree-next (fpa=or-node-left node) maximum)))
    (when (or (null lookup2) (> (fpa=term-id lookup2) maximum))
      (setq lookup2 (fpa=tree-next (fpa=or-node-right node) maximum)))

    (if lookup1
	(if lookup2
	    (if (>= (fpa=term-id lookup1) (fpa=term-id lookup2))
		;; id of lookup1 >= id of lookup2.
		(progn (setf (fpa=or-node-lastleft node) nil
			     (fpa=or-node-lastright node) lookup2)
		       lookup1)
	        ;; id of lookup1 < id of lookup2.
		(progn (setf (fpa=or-node-lastleft node) lookup1
			     (fpa=or-node-lastright node) nil)
		       lookup2))
	    ;; the right subtree did not return an id.
	    (progn (setf (fpa=or-node-right node) nil
			 (fpa=or-node-lastleft node) nil)
		   lookup1)
	    )
        ;; the left subtree did not return an id.
	(if lookup2
	    ;; only the right subtree returned an id.
	    (progn (setf (fpa=or-node-left node) nil
			 (fpa=or-node-lastright node) nil)
		   lookup2)
	    ;; both subtrees did not return values.
	    (progn (fpa=free-or-node node)
		   nil))))
  )

(defun fpa=tree-next (node maximum)
  " **************************************************************
    INPUT:   A node and a maximum id.
    RETURNS: A cons cell consisting of an id and the associated
             'term'.
    SUMMARY: Each time fpa=tree-next is called, ids are bubbled up
             in the query tree until one id reaches the root of
             the query tree (node) which is smaller than maximum.
             This id refers to an entry of the index which
             fulfills all restrictions expressed by the
             query tree.
    ************************************************************** "

  (when node
    ;; here it is very important to use a very fast choice mechanism
    ;; because this function will be called very often.
    (let ((type (fpa=node-type node)))
      (if (= type #.*or-node-type*) 
	  (fpa=or-node-next node maximum)
	(if (= type #.*leaf-node-type*) 
	    (fpa=leaf-node-next node maximum)
	  (fpa=and-node-next node maximum)))))
  )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  OUTPUT FUNCTIONS                                      * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

#+fpa-debug
(defmacro fpa=node-out-tab (tabulator)
  `(dotimes (i ,tabulator)
    (princ " "))
  )

#+fpa-debug
(defun fpa=key-out (key)
  (format t "~D:" (fpa=key-size key))
  (do ((i 1 (+ i 2)))
      ((> i (fpa=key-size key)) nil)
    (if (= 0 (fpa=key-element key i))
	(princ '*)
	(princ (fpa=key-element key i)))
    (when (< i (fpa=key-size key)) (princ (fpa=key-element key (1+ i))))
    )
  )

#+fpa-debug
(defun fpa=nil-node-out (tabulator)
  (fpa=node-out-tab tabulator)
  (format t "NIL~%")
  )

#+fpa-debug
(defun fpa=leaf-node-out (node tabulator)
  (fpa=node-out-tab tabulator)
  (format t "Key: ")
  (fpa=key-out (fpa=leaf-node-key node))
  (format t " : First term (ID): ~S~%" (fpa=term-id (first (fpa=leaf-node-terms node))))
  )

#+fpa-debug
(defun fpa=and-node-out (node tabulator)
  (fpa=node-out-tab tabulator)
  (format t "AND~%")
  (dolist (sub-node (fpa=and-node-subs node))
    (fpa=node-out-tab tabulator)
    (format t "And-sub:~%")
    (fpa=node-out sub-node (+ 2 tabulator)))
  (fpa=node-out-tab tabulator)
  (format t "and-------------------------~%")
  )

#+fpa-debug
(defun fpa=or-node-out (node tabulator)
  (fpa=node-out-tab tabulator)
  (format t "OR~%")
  (fpa=node-out-tab tabulator)
  (format t
	  "Last left term (ID): ~S ; Last right term (ID): ~S~%"
	  (fpa=term-id (fpa=or-node-lastleft node))
	  (fpa=term-id (fpa=or-node-lastright node)))
  (fpa=node-out-tab tabulator)
  (format t "Left:~%")
  (fpa=node-out (fpa=or-node-left node) (+ 2 tabulator))
  (fpa=node-out-tab tabulator)
  (format t "Right:~%")
  (fpa=node-out (fpa=or-node-right node) (+ 2 tabulator))
  (fpa=node-out-tab tabulator)
  (format t "or--------------------------~%")
  )

#+fpa-debug
(defmacro fpa=node-out-macro (node tabulator)
  `(if ,node
    (case (fpa=node-type ,node)
      (,*or-node-type*   (fpa=or-node-out ,node ,tabulator))
      (,*leaf-node-type* (fpa=leaf-node-out ,node ,tabulator))
      (,*and-node-type*  (fpa=and-node-out ,node ,tabulator)))
    (fpa=nil-node-out ,tabulator))
  )

#+fpa-debug
(defun fpa=node-out (node tabulator)
  ""
  (fpa=node-out-macro node tabulator)
  )
