;;; -*- syntax: common-lisp; package: keim; base: 10; mode: lisp -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1994 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;


(in-package :keim)

(mod~defmod fac :uses (cl lit mod node res subst term top uni)
	    :exports (fac~get-factors)
	    :documentation "Procedures for computing the factors of a clause.")

#{\section{Factoring}
 These routines are used in factoring clauses.  
 FAC~GET-FACTORS, given a clause, returns the first-level
factors of the clause, as clauses.
#}


(defun fac~get-factors (clause)
  (declare 
   (authors nesmith)
   (input "A CLAUSE")
   (value "A list of clauses, each of which is a factor of CLAUSE")
   (effect "None"))
  ;; Inside this module we work with the clauses as lists of literals.
  ;; fac=get-factors-aux recursively gets all factors, and we construct
  ;; the clauses from them.
  (let* ((lits (cl~literals clause))
	 (factors (fac=get-factors-aux lits lits)))
      (mapcar #'(lambda (factor) 
		  (fac=construct-clause-factor clause (car factor)
					       (cadr factor)
					       (caddr factor)))
	factors)))


(defun fac=get-factors-aux (part-of-clause whole-clause)
  (declare 
   (authors nesmith)
   (input "Two lists of literals, with the first being a subset of the second")
   (value "All factors of WHOLE-CLAUSE which are generated by unifying pairs
of literals in PART-OF-CLAUSE")
   (effect "None"))
  ;; First the first literal of the clause is checked against all following
  ;; literals, and FAC=SINGLE-FACTOR returns all factors which be made by
  ;; unifying that literal with some other.  FAC=GET-FACTORS-AUX is called 
  ;; recursively for each literal in the clause, and it appends any factors 
  ;; created into a single list of factors.  Null factors are removed, and
  ;; also duplicate factors.  
  (flet ((delete-set-equals (list)
	  (delete-duplicates list :test #'fac=literals-set-equal
			     :key #'car)))
    ;; if we're down to one literal, stop
  (if (<= (length part-of-clause) 1) 
      nil
      (delete-set-equals
       (delete nil
	       (nconc
		(mapcar #'(lambda (x)
			    (fac=single-factor 
			     (car part-of-clause) x whole-clause))
			(cdr part-of-clause))
		(fac=get-factors-aux (cdr part-of-clause) whole-clause))
	       :key #'car)))))


(defun fac=single-factor (literal1 literal2 literal-list)
  (declare 
   (authors nesmith)
   (input "Two literals and a list of literals")
   (value "If LITERAL1 and LITERAL2 have the same polarity and are unifiable,
then their unifier is applied to the LITERAL-LIST and the duplicate
literals are removed; the resulting list is returned as the first element
of a list. The second element of the returned 
value is the unifying substitution,
and the third element is a dotted pair of the two literals.
If the literals have different polarities or are not unifiable, 
then an empty list will be returned.")
   (effect "None"))
  (flet ((del-sub (subst list)
	   (delete-duplicates (subst~apply subst list)
			      :test #'res~literal-equal-p)))
    (let ((pos1 (lit~positive-p literal1))
	  (pos2 (lit~positive-p literal2)))
      (when (eq pos1 pos2)
	(let ((sub (uni~unify (lit~atom literal1)
			      (lit~atom literal2))))
	  (when sub (list (del-sub sub literal-list)
			  sub
			  (cons literal1 literal2))))))))



;; FAC=CONSTRUCT-CLAUSE-FACTOR makes a clause, renames its variables,
;; then makes its justification a factoring justification.  Finally,
;; the class of the new clause is updated to type new-clause.
(defun fac=construct-clause-factor (parent literal-list sub pair)
  (declare 
   (authors nesmith)
   (input "A PARENT clause and a list of literals")
   (value "A new clause whose literals are LITERAL-LIST and whose justification
is by factoring from the PARENT.  The variables in the new clause will be 
renamed.")
   (effect "None"))
  (multiple-value-bind (new-clause renaming)
      (top~rename (cl~create literal-list :name "tmp") 
       :return-renaming t)
    (node~set-justification! 
     new-clause
     ;; don't save positions, unifier ?
     (res~factoring-create parent
			   (list (pos~list-position 
				  (list 
				   (position (car pair)
					     (cl~literals parent))))
				 (pos~list-position 
				  (list 
				   (position (cdr pair)
					     (cl~literals parent)))))
			   sub renaming (gensym "STEP")))
    new-clause))


(defun fac=literals-set-equal (list1 list2)
  (declare 
   (authors nesmith)
   (input "Two lists of literals.")
   (value "T if the lists are equal as sets, using res~literal-equal-p
as a predicate, else nil. We assume the lists contain no duplicates.")
   (effect "None"))
  ;; if lists have no dupes, and they are set-equal, then they must be
  ;; the same size
  (and (= (length list1) (length list2))
       ; copy list2 so we can use it destructively
       (let ((temp nil)
	     (list2 (copy-list list2)))
	 (dolist (l1 list1 t)
	   ;; loop over list1, removing each element from list2,
	   ;; if an element is removed, then temp will be set to T
	   (setq list2
		 (delete l1 list2 
			 :test #'(lambda (x y)
				   (setq temp (res~literal-equal-p x y)))
			 :count 1))
	   ;; if temp is not T, then there is an element of list1 not
	   ;; in list2, so can immediately return nil
	   (unless temp
	     (return-from fac=literals-set-equal nil))))))



