


(IN-PACKAGE "KEIM")

(mod~defmod sort :uses (mod type )
	    :documentation ""
	    :exports ()
	    )

;;; This files provides a simple sort system for \keim. It supplies a
;;; feature for storing sort declarations at symbols and introduces
;;; sort symbols. In this sort system sorts are symbols and term
;;; declarations have the form f: S_1 x ... x S_n -> S for an n-ary
;;; function symbol f. Multiple declarations for one function symbol
;;; are allowed.
;;;
;;; This module was designed for first-order terms in regular
;;; signatures where the sort hierarchy is a semi-lattice.  Therefore
;;; the higher level functions like SORT~SORT, SORT~SUBSORT etc. only
;;; are correct if this condition holds. See the documentation of
;;; these functions for details.


;;;; Sorted Terms

;;; In this sort module only the sort declarations of symbols can be
;;; stored; the sorts of each term are computed every time the
;;; function SORT~SORT is called. For the user this should make no
;;; difference. The declarations are stored in the sharedsymbol of
;;; each symbol. While functions can have multiple sirt declarations,
;;; variables can have only one declaration. The sorts of a symbol are
;;; used to compute its type. Only the types of terms are checked;
;;; when creating an application, no test is made if the resulting
;;; term is well-sorted.

(eval-when (load compile eval)
  (defclass sort+sharedsymbol (shsym+sharedsymbol)
    ((sort-decls :initarg :sort-decls :reader sort=sort-decls :writer sort=write-sort-decls!))
    (:documentation "This is the class of all shared symbols of sorted terms. All instances of the classes
                     SYM+VAR and SYM+CONST contain instances of this class if sorts are present.")))




(defgeneric sort~shsym-create (symbol name type &key sort-decls)
  (declare (edited  "08-AUG-1991 13:30")
	   (authors RICHTS)
	   (input   "A reference symbol, a name, a TYPE and a list of sorts (all corresponding to TYPE).")
	   (effect  "None.")
	   (value   "A shared-symbol of the same class as SYMBOL with"
		    "name NAME and type TYPE"
		    "as specified by the keyword pairs.")
	   (example "(make-instance 'term+term)  'sh1  I    --> SH1"
		    "SH1                         'SH2   O   --> SH2" ))
  (:method ((symbol (eql nil)) name type &key sort-decls)
	   (make-instance 'sort+sharedsymbol :name (string name) :type type :sort-decls sort-decls))
  (:method ((symbol keim+term) name type &key sort-decls)
	   (make-instance 'sort+sharedsymbol :name (string name) :type type :sort-decls sort-decls)))


(defmethod shsym~create ((symbol keim+term) name type &key (sort-decls nil supplied))
  (if supplied
      (make-instance 'sort+sharedsymbol :name (string name) :type type :sort-decls sort-decls)
      (make-instance 'sort+sharedsymbol :name (string name) :type type :sort-decls (sort~sort-decls symbol))))


(defgeneric sort~declarations (symbol)
  (declare (edited  "03-JAN-1994 12:35")
	   (authors RICHTS)
	   (input   "A term symbol.")
	   (effect  "None.")
	   (value   "A list of sort declarations for SYMBOL."))
  (:method ((symbol sym+sym))
     (sort=sort-decls (sym~shared-symbol symbol))))


(defgeneric sort~sort (term)
  (declare (edited  "03-JAN-1994 12:43")
	   (authors RICHTS)
	   (input   "A first order term in a regular signature.")
	   (effect  "None.")
	   (value   "The minimal sort of TERM. This is unique since the signatyre is regular."))
  (:method ((term appl+appl))
     (let* ((sorts (mapcar #'sort~sort (appl~arguments term)))
	    (decls (remove-if-not #'(lambda (decl)
				      (every #'(lambda (term-sort decl-sort)
						 (sort~subsort-p term-sort decl-sort))
					     sorts (type~n-domain decl)))
				      (sort~declarations (appl~function term))))
	    (ranges (mapcar #'type~n-range decls))
	    (result (delete-duplicates
		     (remove-if-not #'(lambda (sort)
					(every #'(lambda (range)
						   (sort~subsort-p sort range))
					       ranges))
				    ranges)
		     :test #'sort~equal)))
       (cond ((null result)  ;length < 1
	      (error "KEIM (sort): The term ~A is not well sorted. ~
                      SORT~~SORT would not compute correctly." term))
	     ((rest result)   ;length > 1
	      (error "KEIM (sort): The term ~A has not a unique sort. ~
                      The signature is not regular and SORT~~SORT would not compute correctly." term))
	     (t (first result)))))
  (:method ((sym sym+sym))
     (let ((decls (sort~declarations sym)))
       (if (rest decls)
	   (error "KEIM (sort): The symbol ~A has more than one term declaration.~
                   The signature is not regular and SORT~~SORT would not compute correctly." sym)
	   (first decls)))))



;;;; Sort Symbols


;;; Sort symbols in this module are extended predicate symbols or
;;; types. In particular they are unary predicate symbols (i.e. of
;;; type \alpha -> o) which have three additional slots in their
;;; sharedsymbol. The sort-label slot is used for searching in the
;;; sort hierarchy. The subsort slot contains the subsort relations
;;; that are explicitly declared. In order to make the subsort test
;;; more efficient, the subsort-rel slot contains also subsort
;;; relations that are not declared explicitly but are introduced
;;; implicitly because of transitivity.
;;;
;;; Sort symbols can not be created directly. They must be created as
;;; predicate symbols and then be declared as a sort symbol. Note
;;; that a sort symbol can still be used as a predicate.
;;;
;;; Types are also sorts. Primitive types (like \iota and o) serve as the
;;; top sorts {\it any}, i.e. each sort S of type \alpha -> o is an
;;; subsort of \alpha. Each primitive type has a sorts slot that
;;; contains all sorts for this type. Complex types are used to
;;; represent complex sorts of the form S_1 x ... x S_n -> S.


(eval-when (load compile eval)
  (defclass sort+sharedsort (sort+sharedsymbol)
    ((sort-label :initform nil :reader sort=sort-label :writer sort=write-sort-label!)
     (subsorts :initform nil :reader sort=subsorts :writer sort=write-subsorts!)
     (sort-rel :initform nil :reader sort=sort-rel :writer sort=write-sort-rel!))
    (:documentation "This is the class of all shared symbols of sorts. If an instance of the class 
                     SYM+SYM contains an instance of this class, it is an sort.")))


(eval-when (load compile eval)
  (defclass sort+type-constant (type+constant)
    ((sorts :initform nil :reader sort=type-sorts :writer sort=write-type-sorts!))
    (:documentation "This is the class of all type constants that have subsorts.")))

(eval-when (load compile eval)
  (defclass sort+type-variable (type+variables)
    ((sorts :initform nil :reader sort=type-sorts :writer sort=write-type-sorts!))
    (:documentation "This is the class of all type variables that have subsorts.")))


(defgeneric sort~p (object)
  (declare (edited  "04-JAN-1994 16:21")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True iff OBJECT is a sort, i.e. a primitive type, a complex type"
		    "or a predicate symbol that has been declared as a sort."))
  (:method ((type type+type))
     t)
  (:method ((const sym+const))
     (typep (sym~shared-symbol const) 'sort+sharedsort))
  (:method ((object t))
     nil))


(defgeneric sort~symbol-p (object)
  (declare (edited  "04-JAN-1994 16:21")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True iff OBJECT is a sort, i.e. a primitive type or a predicate symbol"
		    "that has been declared as a sort."))
  (:method ((type type+primitive))
     t)
  (:method ((const sym+const))
     (typep (sym~shared-symbol const) 'sort+sharedsort))
  (:method ((object t))
     nil))


(defgeneric sort~declare (SYMBOL)
  (declare (edited  "08-AUG-1991 13:30")
	   (authors RICHTS)
	   (input   "Something that can be declared as a sort;"
		    "i.e. a constant symbol with type \\alpha -> o, i.e. a unary predicate symbol,"
		    "or a primitive type.")
	   (effect  "The class of  SYMBOL is changed so that it can be used as a sort."
		    "If SYMBOL is a predicate, the class of its shared-symbol is changed"
		    "from SORT+SHAREDSYMBOL to SORT+SHAREDSORT;"
		    "if SYMBOL is a primitive type it is changed from TYPE+CONSTANT to SORT+TYPE-CONSTANT"
		    "or from TYPE+VARIABLE to SORT+TYPE-VARIABLE.")
	   (value   "Undefined."))
  (:method ((predicate sym+const))
     (change-class (sym~shared-symbol predicate) 'sort+sharedsort)
     (let* ((type (type~c-domain (term~type predicate)))
	    (subsorts (sort~subsorts type)))
       (unless (sort~subsorts type)  ;if type has no subsorts, it hs not been declared as a sort yet.
	 (sort~declare type))
       (sort=write-type-sorts! (cons predicate subsorts) type)))
  (:method ((type type+constant))
     (change-class type 'sort+type-constant))
  (:method ((type type+variable))
     (change-class type 'sort+type-variable)))

(defgeneric sort~equal (sort1 sort2)
  (declare (edited  "05-JAN-1994 14:41")
	   (authors RICHTS)
	   (input   "Two sorts, i.e. extended predicates, primitive or complex types.")
	   (effect  "None.")
	   (value   "True iff SORT1 and SORT2 are the same sort."))
  (:method ((sort1 sym+sym) (sort2 sym+sym))
     (term~equal sort1 sort2))
  (:method ((sort1 type+primitive) (sort2 type+primitive))
     (keim~equal sort1 sort2))
  (:method ((sort1 type+complex) (sort2 type+complex))
     (keim~equal sort1 sort2)))

  
(defgeneric sort~subsorts (sort)
  (declare (edited  "03-JAN-1994 17:27")
	   (authors RICHTS)
	   (input   "A sort symbol.")
	   (effect  "None.")
	   (value   "A list of 'direct' subsorts, i.e. sorts declared explicitly as subsorts of SORT."
		    "If sort is a type, all subsorts of this type."))
  (:method ((shsort sort+sharedsort))
     (sort=subsorts shsort))
  (:method ((sort sym+sym))
     (sort=subsorts (sym~shared-symbol sort)))
  (:method ((sort sort+type-constant))
     (sort=type-sorts sort))
  (:method ((sort sort+type-variable))
     (sort=type-sorts sort))
  (:method ((sort type+primitive))
     nil))


(defgeneric sort=add-sort-rel (sort1 sort2 gcs)
  (declare (edited  "03-JAN-1994 17:27")
	   (authors RICHTS)
	   (input   "Two sorts and their greatest common subsort or NIL.")
	   (effect  "The sort-rel slot of each sort is updated."
		    "(SORT1 . GCS) is added to SORT2 and (SORT2 . GCS) is added to SORT1.")
	   (value   "Undefined."))
  (:method ((sort1 sym+sym) (sort2 sym+sym) gcs)
     (let ((shsort1 (sym~shared-symbol sort1))
	   (shsort2 (sym~shared-symbol sort2)))
       (sort=write-sort-rel! (acons sort1 gcs (sort=sort-rel shsort2)) shsort2)
       (sort=write-sort-rel! (acons sort2 gcs (sort=sort-rel shsort1)) shsort1))))


(defun sort~subsort-p (sort1 sort2)
  (declare (edited  "03-JAN-1994 14:22")
	   (authors RICHTS)
	   (input   "Two sorts.")
	   (effect  "If it has not been computed yet, the sort relation between SORT1 and SORT2"
		    "is stored in SORT1 and SORT2.")
	   (value   "True iff SORT1 is a subsort of SORT2. This is reflexive,"
		    "i.e. (sort~subsort-p s s) is always true."))
  (let ((gcs (sort~gc-subsort sort1 sort2)))
    (and gcs (sort~equal sort1 gcs))))

(defun sort~gc-subsort (sort1 sort2)
  (declare (edited  "03-JAN-1994 14:22")
	   (authors RICHTS)
	   (input   "Two sorts in a sort hierarchy that is an semi-lattice (i.e. two sorts"
		    "have at most one greatest common subsort).")
	   (effect  "If it has not been computed yet, the sort relation between SORT1 and SORT2"
		    "(i.e. its greatest common subsort) is stored in SORT1 and SORT2.")
	   (value   "The greatest common subsort of SORT1 and SORT2 or NIL if they have none."))
  (cond ((and (type~p sort1) (type~p sort2))
	 (if (keim~equal sort1 sort2) sort1 nil))
	((type~p sort1)
	 (if (keim~equal sort1 (type~c-domain (term~type sort2)))
	     sort2
	     nil))
	((type~p sort2)
	 (if (keim~equal sort2 (type~c-domain (term~type sort1)))
	     sort1
	     nil))
	((sort~equal sort1 sort2) sort1)
	(t
	 (let ((sort-gcs-pair (assoc sort2 (sort=sort-rel (sym~shared-symbol sort1)) :test #'sort~equal)))
	   (if sort-gcs-pair
	       (cdr sort-gcs-pair)
	       (sort=compute-gc-subsort sort1 sort2))))))


(defun sort=compute-gc-subsort (sort1 sort2)
  (declare (edited  "03-JAN-1994 14:44")
	   (authors RICHTS)
	   (input   "Two sort symbols.")
	   (effect  "The greatest common subsort is inserted into the subsort-rel slot of both sorts.")
	   (value   "The greatest common subsort of SORT1 and SORT2."))
  (sort=mark-subsorts sort1 t)
  (let ((common-subsorts (sort=collect-marked-subsorts sort2)))
    (sort=mark-subsorts sort1 nil t)
    (let* ((result-list (delete-duplicates
			 (remove-if #'(lambda (sort)
					(some #'(lambda (common-subsort)
						  (and (not (sort~equal sort common-subsort))
						       (sort~subsort-p sort common-subsort)))
					      common-subsorts))
				    common-subsorts)
			 :test #'sort~equal))
	   (result (if (rest result-list) ;length > 1
		       (error "KEIM (sort): The sorts ~A and ~A have more than one greatest common subsort.~
                             The sort hierarchy is not a semi-lattice and SORT~~GC-SUBSORT ~
                             would not compute correctly." sort1 sort2)
		     (first result-list))))
      (sort=add-sort-rel sort1 sort2 result)
      result)))


(defun sort=mark-subsorts (sort marker &optional rec)
  (let ((shsort (sym~shared-symbol sort)))
    (unless (and (eql (sort=sort-label shsort) marker) (null rec))
      (sort=write-sort-label! marker shsort)
      (mapc #'(lambda (subsort)
		(sort=mark-subsorts subsort marker rec))
	    (sort=subsorts shsort)))))

(defun sort=collect-marked-subsorts (sort)
  (let ((shsort (sym~shared-symbol sort)))
    (cond ((sort=sort-label shsort)
	   (sort=mark-subsorts sort nil)
	   (list sort))
	  (t (mapcan #'sort=collect-marked-subsorts (sort=subsorts shsort))))))


(defun sort~type (sort)
  (declare (edited  "04-JAN-1994 15:50")
	   (authors RICHTS)
	   (input   "A sort.")
	   (effect  "None.")
	   (value   "The corresponding type to SORT.")
	   (example "If Nat is a predicate symbol of type \\iota -> o that has been declared as sort:"
		    "(sort~type (Nat x \\iota -> o))  -->  (\\iota x \\iota -> o)."))
  (cond ((type~complex-p sort)
	 (type~abstract (sort~type (type~n-range sort))
			(mapcar #'sort~type (type~n-domain sort))))
	((type~p sort) sort)
	((sym~constant-p sort) (type~c-domain (term~type sort)))
	(t (error "KEIM (sort): ~A is not a sort. It is not a complex type, a basic type or a predicate symbol."
		  sort))))

(defmethod type~n-range ((sort sym+const))
  sort)

(defmethod type~n-domain ((sort sym+const))
  nil)


(defun sort~all-sorts (env)
  (declare (edited  "05-JAN-1994 15:52")
	   (authors RICHTS)
	   (input   "An environment.")
	   (effect  "None.")
	   (value   "A list of all sort symbols (predictes only)."
		    "This is the concatenation of the subsorts of each type symbol in ENV."))
  (let ((type-keys (env~class-keys env 'type+primitive)))
    (mapcan #'sort~subsorts
	    (mapcar #'(lambda (key)
			(env~lookup-object key env))
		    type-keys))))


;;;; {\post}

;;; The {\post} syntax for sorts is an extension of the {\post} syntax for unsorted terms.


;;;;; Subsorts Declarations

;;; In subsorts declarations the subsort relations between sorts can be
;;; declared. Furthermore this is the only possibility to define a
;;; predicate symbol as a sort; all predicates that are used as sort
;;; symbols, must have been occured in a subsorts declaration before.

;;; begin{lisp}
;;; <decl>         ::=  (subsorts ( <sort>+ )+ ).
;;; end{lisp}

;;; begin{latex}
;;; \begin{postsyntax}
;;; \syntax{
;;; \nt{decl} ::= (subsorts ( \nt{sort}+ )+ ).}
;;; \end{postsyntax}
;;; end{latex}

;;; The semantic of a {\post} declaration {\tt (subsorts (S_1 ... S_n))}
;;; is as follows: All S_i must be predicate symbols of the same type
;;; \alpha -> o. Those predicates that have not been defined as sorts
;;; yet, i.e. that have not occured in a {\tt subsorts} declaration,
;;; are now defined as sorts with SORT~DECLARE. For these symbols the
;;; subsort relation S_i < \alpha is defined. Then the subsort
;;; relations S_1 < ... < S_n are defined. These relations are exactly
;;; those that can be obtained with SORT~SUBSORTS, i.e. {\tt
;;; (SORT~SUBSORTS \alpha)} will result in a list containing S_1, ...,
;;; S_n and the resulting list of {\tt (SORT~SUBSORTS S_n)} will only
;;; contain S_{n-1} and not S_1, ..., S_{n-2}.

(defmethod post~read-object ((subsort-decls list) (env env+environment) 
			    (indicator (eql :subsorts)))
  (dolist (subsort-decl subsort-decls)
    (let ((sort-list (mapcar #'(lambda (post-sort)
				 (let ((symbol (env~lookup-object post-sort env)))
				   (cond ((sort~p symbol) symbol)
					 ((sym~p symbol)
					  (sort~declare symbol)
					  symbol)
					 (t (error "KEIM (sort): ~A is not a term symbol ~
                                                    and can not be declared as a sort."
						   symbol)))))
			     subsort-decl)))
      (do ((subsort (first sort-list) (first sorts))
	   (sorts (rest sort-list) (rest sorts)))
	  ((null sorts))
	(sort=write-subsorts! (cons subsort (sort~subsorts (first sorts)))
			      (sym~shared-symbol (first sorts)))))))



;;;;; Sorts

;;; Sorts in {\post} are similar to types. A sort symbol can be a
;;; declared type symbol or a declared predicate symbol that has been
;;; defined as a sort in a subsort declaration.

;;; begin{lisp}
;;; <sort>         ::=  <sort-symbol> | (<sort> <sort>+).
;;; <sort-symbol>  ::=  <name>.
;;; end{lisp}

;;; begin{latex}
;;; \begin{postsyntax}
;;; \syntax{
;;; \nt{sort} ::= \nt{sort-symbol} | (\nt{sort} \nt{sort}+).
;;; \nt{sort-symbol} ::= \nt{name}.}
;;; \end{postsyntax}
;;; end{latex}

(defun sort~read-existing-sort (sort env)
  (declare (edited  "04-JAN-1994 11:35")
	   (authors RICHTS huang nesmith)
	   (input  "A {\\post} representation of an existing sort and an environment.")
	   (effect "None")
	   (value  "The sort is looked up in the environment and a sort is returned."
		   "If the sort is not in the environment an error will be signaled.")
	   (example "(sort~read-existing-sort '(o i Nat) env):   (Nat x i) -> o"))
  (sort~env-lookup sort env))


(defmethod post~read-object ((sort symbol) (env env+environment) 
			    (indicator (eql :existing-sort)))
  (sort~env-lookup sort env))


(defmethod post~read-object ((sort cons) (env env+environment) 
			    (indicator (eql :existing-sort)))
  (sort~env-lookup sort env))


(defgeneric sort~env-lookup (sort env)
  (declare (edited  "04-JAN-1994 11:35")
	   (authors RICHTS NESMITH)
	   (input   "A {\\post} representation of an existing sort and an environment.")
	   (effect  "None.")
	   (value   "The sort is looked up in the environment and a sort is returned."
		    "If the sort is not in the environment an error will be signaled.")
	   (example "(sort~read-existing-sort  '(o i Nat) env):   (Nat x i) -> o"))
  (:method ((sort list) (env env+environment))
	   (if (= (length sort) 1)
	       (sort~env-lookup (car sort) env)
	       (sort=env-lookup-complex sort env)))
  (:method ((sort symbol) (env env+environment))
	   (sort=env-lookup-symbol sort env)))


(defun sort=env-lookup-symbol (sort env)
  (declare (edited  "04-JAN-1994 11:35")
	   (authors RICHTS nesmith)
	   (input  "A {\\post} sort symbol and an environment.")
	   (effect "None.")
	   (value  "The sort symbol that is associated with SORT in ENV."
		   "An error is signaled if SORT is not associated with a sort symbol."))
  (let ((obj (env~lookup-object sort env)))
    (cond ((not obj) 
	   (error "~A is not an existing sort." sort))
	  ((not (or (sort~p obj) (type~p obj)))
	   (error "~A is declared in the environment as a ~A, not as a sort."
		  sort (class-of obj)))
	  (t obj))))


(defun sort=env-lookup-complex (sort env)
  (declare (edited  "04-JAN-1994 11:35")
	   (authors RICHTS nesmith)
	   (input  "A complex {\\post} sort and an environment.")
	   (effect "None.")
	   (value  "The sort that is associated with SORT in ENV."
		   "This is a complex type where the basic types are sort symbols."
		   "An error is signaled if SORT is not a correct {\\post} sort."))
  (if (consp sort)
      (type~abstract (sort~env-lookup (car sort) env)
		     (mapcar #'(lambda (x) (sort~env-lookup x env))
			     (reverse (cdr sort))))
      (error "complex sort ~S should be a cons" sort)))



;;;;; Term Declarations

;;; This module only provides term declarations of the form f: S_1 x
;;; ... x S_n -> S. This declaration in {\post} would be {\tt
;;; (constant (f (S S_n ... S_1)))}.

;;; begin{lisp}
;;; <decl>         ::=  (variables (<name> <sort>)+ ).
;;; <decl>         ::=  (constants (<name> <sort>)+ ).
;;; end{lisp}

;;; begin{latex}
;;; \begin{postsyntax} \syntax{
;;; \nt{decl} ::= (variables (\nt{name} \nt{sort})+ ).
;;; \nt{decl} ::= (constants (\nt{name} \nt{sort})+ ).}
;;; \end{postsyntax}
;;; end{latex}

;;; The sort symbols in the sorts must have been defined: They must be
;;; declared type symbols or predicate symbols that have occured in a
;;; subsorts declaration. The semantic of these declarations is that
;;; of SYM~ENV-ENTER-VARIABLE and SYM~ENV-ENTER-CONSTANT as defined in
;;; this module. These two functions and the methods of POST~READ-OBJECT for
;;; the indicators :VARIABLE and :CONSTANT must be redefined in this
;;; module.

(defmethod post~read-object ((var cons) (env env+environment) 
			    (indicator (eql :variable)))
  (let ((sort (sort~env-lookup (cdr var) env)))
    (sym~env-enter-variable (car var) sort env nil)))


(defmethod post~read-object ((var cons) (env env+environment) 
			    (indicator (eql :variable-multiple)))
  (let ((sort (sort~env-lookup (cdr var) env)))
    (sym~env-enter-variable (car var) sort env t)))


(defmethod post~read-object ((const cons) (env env+environment) 
			    (indicator (eql :constant)))
  (let ((sort (sort~env-lookup (cdr const) env)))
    (sym~env-enter-constant (car const) sort env)))

(defmethod post~print-declaration ((variable sym+var) stream)
  (format stream "(~A " (keim~name variable))
  (post~print (sort~sort variable) stream)
  (format stream ")"))

(setf sort*var-create 'sym~variable-create)
(unintern 'sym~variable-create (find-package "KEIM"))

(defgeneric sym~variable-create (object sort &optional refsym)
  (declare (edited  "04-JAN-1994 11:35")
	   (authors RICHTS nesmith)
	   (input   "A term symbol, a shared symbol or a symbol, and a sort."
		    "Optionally a reference symbol (currently ignored).")
	   (effect  "None")
	   (value   "Creates a new variable using the name of the given symbol and the"
		    "sort, and possibly using the properties of the optional reference symbol,"
		    "and returns it."))
  (:method ((symbol symbol) sort &optional refsym)
     (let* ((type (sort~type sort))
	    (shsym (sort~shsym-create refsym (string symbol) type :sort-decls (list sort))))
       (make-instance 'sym+var :symbol shsym :type type)))
  (:method ((symbol shsym+sharedsymbol) sort &optional refsym)
     (let* ((type (sort~type sort))
	    (shsym (sort~shsym-create refsym (keim~name symbol) type :sort-decls (list sort))))
       (make-instance 'sym+var :symbol shsym :type type)))
  (:method ((symbol sym+sym) sort &optional refsym)
     (let* ((type (sort~type sort))
	    (shsym (sort~shsym-create refsym (format nil "x-~A" (incf term*variable-counter)) type
				      :sort-decls (list sort))))
       (make-instance 'sym+var :symbol shsym :type type))))

;(export 'sym~variable-create)
(setf sort*const-create 'sym~constant-create)
(unintern 'sym~constant-create (find-package "KEIM"))

(defgeneric sym~constant-create (object sort &optional refsym)
  (declare (edited  "04-JAN-1994 13:27")
	   (authors RICHTS nesmith)
	   (input   "A shared symbol or a symbol, and a sort. Optionally a reference"
		    "symbol (currently ignored).")
	   (effect  "None")
	   (value   "Creates a new constant using the name of the given symbol and"
		    "the sort, possibly using the properties of the optional reference symbol,"
		    "and returns it."))
  (:method ((symbol symbol) sort &optional refsym)
     (let* ((type (sort~type sort))
	    (shsym (sort~shsym-create refsym (string symbol) type :sort-decls (list sort))))
       (make-instance 'sym+const :symbol shsym :type type)))
  (:method ((symbol shsym+sharedsymbol) sort &optional refsym)
     (let* ((type (sort~type sort))
	    (shsym (sort~shsym-create refsym (keim~name symbol) type :sort-decls (list sort))))
       (make-instance 'sym+const :symbol shsym :type type)))
  (:method ((symbol sym+sym) sort &optional refsym)
     (let* ((type (sort~type sort))
	    (shsym (sort~shsym-create refsym (keim~name symbol) type :sort-decls (list sort))))
       (make-instance 'sym+const :symbol shsym :type type))))

;(export sym~constant-create)


(defun sym~env-enter-variable (symbol sort env &optional allow-multiple)
  (declare
   (authors nesmith)
   (input  "A lisp symbol, a sort and an environment.  Optional argument"
	   "is boolean ALLOW-MULTIPLE.")
   (effect "If the symbol has no association in environment, then a new term variable"
	   "with the symbol as name and given sort is created and entered in the environment."
	   "If ALLOW-MULTIPLE is NIL and the symbol is associated with a non-variable"
	   "or with a variable of a differing sort, an error is signaled;"
	   "if the symbol is already associated with a variable of the proper sort,"
	   "the existing variable is returned."
	   "If ALLOW-MULTIPLE is non-NIL, in every case a new variable is made"
	   "and entered in the environment.")
   (value  "The variable, if one is successfully made or found."))
  (when (not (symbolp symbol))
    (error "~S is not a symbol." symbol))
  (when (not (sort~p sort))
    (error "~S is not a sort." sort))
  (when (not (env~p env))
    (error "~S is not an environment." env))
  (let ((object (env~lookup-object symbol env)))
    (cond ((or allow-multiple (null object))
	   (env~enter symbol (sym~variable-create symbol sort) env))
	  ((not (sym~variable-p object))
	   (post~error "Can't declare ~A as a  variable, because it already exists in the environment as a ~A." 
		       symbol (class-name (class-of object))))
	  ((not (keim~equal (sort~sort object) sort))
	   (post~error "Can't declare ~A as a variable of sort ~A, because it already exists~
                        in the environment with sort ~S." 
		       symbol sort (sort~sort object)))
	  (t object))))

(defun sym~env-enter-constant (symbol sort env)
  (declare
   (authors nesmith)
   (input  "A lisp symbol, a sort and an environment.")
   (effect "If the symbol has no association in environment, then a new"
	   "term constant with the symbol as name and given sort is created and "
	   "entered in the environment.  If the symbol is associated with a "
	   "non-constant or with a constant of a differing type, an error is signaled."
	   "If the symbol is already associated with a constant of the proper type, "
	   "then SORT is added to the sort declarations of that constant and that constant is returned.")
   (value  "The constant, if one is successfully made or found."))
  (when (not (symbolp symbol))
    (error "~S is not a symbol." symbol))
  (when (not (sort~p sort))
    (error "~S is not a sort." sort))
  (when (not (env~p env))
    (error "~S is not an environment." env))
  (let ((object (env~lookup-object symbol env)))
    (cond ((null object)
	   (let ((constant (sym~constant-create symbol sort)))
	     (env~enter symbol constant env)
	     constant))
	  ((not (sym~constant-p object))
	   (post~error "Can't declare ~A as a  constant, because it already exists in the environment as a ~A. " 
		       symbol (class-name (class-of object))))
	  ((not (keim~equal (term~type object) (sort~type sort)))
	   (post~error "Can't declare ~A as a constant of sort ~A, because it already exists ~
                        in the environment with type ~S." 
		       symbol sort (term~type object)))
	  (t (sort=write-sort-decls! (cons sort (sort~declarations object))
				     (sym~shared-symbol object))
	     object))))

(defmethod top=replace-free-variables-and-rename ((variable sym+var) variables)
   (let ((label (sym~label variable)))
     (if label
	 (term~copy label)
	 (if (or (eq variables t) (member variable variables :test #'keim~equal))
	     (top=label-symbol variable (sym~variable-create variable (sort~sort variable)))
	     (term~copy variable)))))

(defmethod top=replace-free-variables-and-rename!! ((variable sym+var) variables)
   (let ((label (sym~label variable)))
     (if label 
	 (unless (eq label t)
	   (term~set-term variable label))
	 (when (or (eq variables t) (member variable variables :test #'keim~equal))
	   (let ((new-variable (sym~variable-create variable (sort~sort variable))))
	     (top=label-symbol (term~copy variable) new-variable)
	     (term~set-term variable new-variable))))
     variable))

#|
(defmethod print-object ((var sym+var) stream)
  (format stream "~A:~A" (keim~name var) (sort~sort var)))
|#
