;;; -*- Mode: LISP; Base: 10; Syntax: Common-lisp; Package: KEIM -*-

(in-package "KEIM")

(mod~defmod houni :uses (appl keim mod subst sym term top )
	    :documentation "Syntactic Higher-Order Unification."
	    :exports (houni~with-permanent-bindings
		      houni~unify)
	    )

;; Unification Algorithms

(defvar houni*permanent-bindings nil)

(defmacro houni~with-permanent-bindings (&body forms)
  `(unwind-protect
       (progn (setf houni*permanent-bindings t)
	      ,@forms)
     (setf houni*permanent-bindings nil)
     (top~unbind-binding-list!)))

(defun houni~unify (term1 term2)
  (houni=unify-free-terms term1 term2))

(defun houni=unify-free-terms (term1 term2  &key (eq-test #'keim~equal))
  (unwind-protect
       (progn
	 (unless houni*permanent-bindings (top~new-binding-list!))
	 (if (if (listp term1)
		 (houni=unify-free-terms-rec (car term1) (car term2) (cdr term1) (cdr term2) :eq-test eq-test)
		 (houni=unify-free-terms-rec term1 term2 nil nil :eq-test eq-test))
	     (if houni*permanent-bindings
		 t
	         (subst~create-binding-substitution))))
    (unless houni*permanent-bindings (top~unbind-binding-list!))))

(defgeneric houni=unify-free-terms-rec (term1 term2 termlist1 termlist2  &key eq-test )
  (:method :around ((t1 term+term) (t2 term+term)  termlist1 termlist2 &key (eq-test #'keim~equal))
   (if (keim~equal (term~type t1) (term~type t2))
       (call-next-method)
       nil))
  (:method ((t1 (eql nil)) (t2 (eql nil)) termlist1 termlist2 &key (eq-test #'keim~equal))
   (when (or termlist1 termlist2)
     (error "Programmierfehler: Es sind noch Terme da."))
   t)
  (:method ((variable1 sym+var) (variable2 sym+var) termlist1 termlist2  &key (eq-test #'keim~equal))
   (let ((binding1 (term~binding variable1))
	 (binding2 (term~binding variable2)))
     (cond ((funcall eq-test variable1 variable2)
	    (houni=unify-free-terms-rec (car termlist1) (car termlist2) (cdr termlist1) (cdr termlist2) :eq-test eq-test))
	   ((and binding1 binding2)
	    (houni=unify-free-terms-rec binding1 binding2 termlist1 termlist2  :eq-test eq-test))
	   (binding1
	    (houni=unify-free-terms-rec binding1 variable2 termlist1 termlist2  :eq-test eq-test))
	   (binding2
	    (houni=unify-free-terms-rec variable1 binding2 termlist1 termlist2  :eq-test eq-test))
	   ((houni=occurs-p variable1 variable2) nil)
	   (t (top~bind-term! variable1 variable2)
	      (houni=unify-free-terms-rec (car termlist1) (car termlist2) (cdr termlist1) (cdr termlist2)  :eq-test eq-test)))))
  (:method ((variable sym+var) (term term+term) termlist1 termlist2  &key (eq-test #'keim~equal))
   (cond ((term~binding variable)
	  (houni=unify-free-terms-rec (term~binding variable) term termlist1 termlist2 :eq-test eq-test))
	 ((houni=occurs-p variable term :eq-test eq-test) nil)
	 (t
	  (top~bind-term! variable term)
	  (houni=unify-free-terms-rec (car termlist1) (car termlist2) (cdr termlist1) (cdr termlist2)  :eq-test eq-test))))
  (:method ((term term+term) (variable sym+var) termlist1 termlist2  &key (eq-test #'keim~equal))
   (houni=unify-free-terms-rec variable term termlist1 termlist2 :eq-test eq-test))
  (:method ((constant1 sym+const) (constant2 sym+const) termlist1 termlist2  &key (eq-test #'keim~equal))
   (if (funcall eq-test constant1 constant2)
       (houni=unify-free-terms-rec (car termlist1) (car termlist2) (cdr termlist1) (cdr termlist2) :eq-test eq-test)
       nil))
  (:method ((application1 appl+appl) (application2 appl+appl) termlist1 termlist2  &key (eq-test #'keim~equal))
   (let* ((subterms1 (term~subterms application1))
	  (subterms2 (term~subterms application2))
	  (difference (- (length subterms1) (length subterms2))))
     (cond ((zerop difference)
	    (houni=unify-free-terms-rec (car subterms1) (car subterms2)
					(append (cdr subterms1) termlist1) (append (cdr subterms2) termlist2)
					:eq-test eq-test))
	   ((minusp difference)
	    (houni=unify-free-terms-rec (car subterms1)
					(appl~create (car subterms2) (subseq subterms2 1 (1+ (- difference))))
					(append (cdr subterms1) termlist1)
					(append (subseq subterms2 (1+ (- difference))) termlist2)
					:eq-test eq-test))
	   ((plusp difference)
	    (houni=unify-free-terms-rec (appl~create (car subterms1) (subseq subterms1 1 (1+ difference)))
					(car subterms2)
					(append (subseq subterms1 (1+ difference)) termlist1)
					(append (cdr subterms2) termlist2)
					:eq-test eq-test)))))
  (:method ((abstraction1 abstr+abstr) (abstraction2 abstr+abstr) termlist1 termlist2 &key (eq-test #'keim~equal))
   (unwind-protect (progn (top~bind-term! (abstr~bound-variable abstraction1) (abstr~bound-variable abstraction2))
			  (sym~set-label! (abstr~bound-variable abstraction1) :bound)
			  (sym~set-label! (abstr~bound-variable abstraction2) :bound)
			  (houni=unify-free-terms-rec (abstr~scope abstraction1) (abstr~scope abstraction2)
						      termlist1 termlist2
						      :eq-test eq-test))
     (top~unbind-terms! (list (abstr~bound-variable abstraction1)))
     (sym~set-label! (abstr~bound-variable abstraction1) nil)
     (sym~set-label! (abstr~bound-variable abstraction2) nil)))
  (:method ((term1 term+term) (term2 term+term) termlist1 termlist2  &key (eq-test #'keim~equal))
   (declare (ignore termlist1 termlist2))
   nil)
  (:method ((lit1 lit+literal) (lit2 lit+literal) litlist1 litlist2  &key (eq-test #'keim~equal))
   (if (eq (lit~positive-p lit1) (lit~positive-p lit2))
       (houni=unify-free-terms-rec (lit~atom lit1) (lit~atom lit2) litlist1 litlist2 :eq-test eq-test)
       nil)))


(defgeneric houni=occurs-p (variable obj   &key eq-test)
  (:method (variable1 (variable2 sym+var)   &key (eq-test #'keim~equal))
	   (cond ((funcall eq-test variable1 variable2) t)
		 ((term~binding variable2) (houni=occurs-p variable1 (term~binding variable2) :eq-test eq-test))
		 ((eq (sym~label variable1) :bound) t)
		 ((eq (sym~label variable2) :bound) t)
		 (t nil)))
  (:method (variable (constant sym+const)  &key (eq-test #'keim~equal))
	   (declare (ignore variable))
	   nil)
  (:method (variable (application appl+appl) &key (eq-test #'keim~equal))
	   (some #'(lambda (argument)
		     (houni=occurs-p variable argument :eq-test eq-test))
		 (term~subterms application)))
  (:method (variable (abstraction abstr+abstr) &key (eq-test #'keim~equal))
	   (if (funcall eq-test variable (abstr~bound-variable abstraction))
	       nil
	       (houni=occurs-p variable (abstr~scope abstraction)))))

#|
(defun houni=match-free-terms (term1 term2  &key (eq-test #'keim~equal))
  (unwind-protect
      (progn
	(top~new-binding-list!)
	(if (if (listp term1)
		(houni=match-free-terms-rec (car term1) (car term2) (cdr term1) (cdr term2) :eq-test eq-test)
		(houni=match-free-terms-rec term1 term2 nil nil :eq-test eq-test))
	    (subst~create-binding-substitution nil)))
    (top~unbind-binding-list!)))

(defgeneric houni=match-free-terms-rec (term1 term2 termlist1 termlist2  &key eq-test)
  (:method ((t1 (eql nil)) (t2 (eql nil)) termlist1 termlist2  &key (eq-test #'keim~equal))
   (when (or termlist1 termlist2)
     (error "Programmierfehler: Es sind noch Terme da."))
   t)
  (:method ((variable sym+var) term termlist1 termlist2  &key (eq-test #'keim~equal))
   (let ((binding (term~binding variable)))
     (cond ((term~binding variable)
	    (if (funcall eq-test binding term)
		(houni=match-free-terms-rec (car termlist1) (car termlist2) (cdr termlist1) (cdr termlist2) :eq-test eq-test)
		nil))
	   (t (top~bind-term! variable term)
	      (houni=match-free-terms-rec (car termlist1) (car termlist2) (cdr termlist1) (cdr termlist2) :eq-test eq-test)))))
  (:method ((constant1 sym+const) (constant2 sym+const) termlist1 termlist2  &key (eq-test #'keim~equal))
   (if (funcall eq-test constant1 constant2)
       (houni=match-free-terms-rec (car termlist1) (car termlist2) (cdr termlist1) (cdr termlist2) :eq-test eq-test)
       nil))
  (:method ((application1 appl+appl) (application2 appl+appl) termlist1 termlist2  &key (eq-test #'keim~equal))
   (if (funcall eq-test (appl~function application1) (appl~function application2))
       (let ((arguments1 (appl~arguments application1))
	     (arguments2 (appl~arguments application2)))
	 (when (/= (length arguments1) (length arguments2))
	   (error "~A and ~A have different numbers of arguments are not first order terms"))
	 (houni=match-free-terms-rec (car arguments1) (car arguments2) (append (cdr arguments1) termlist1) (append (cdr arguments2) termlist2) :eq-test eq-test))
       nil))
  (:method ((term1 term+term) (term2 term+term) termlist1 termlist2  &key (eq-test #'keim~equal))
   (declare (ignore termlist1 termlist2))
   (if (funcall eq-test (term~top term1) (term~top term2))
       (error "Programmierfehler: Methoden falsch definiert; Topsymbole sind gleich."))
   nil)) 

|#



#+ho-unify(defun uni~unify (t1 t2 &key eq-test)
             (houni~unify t1 t2))
