;;; -*- syntax: common-lisp; package: keim; base: 10; mode: lisp -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;

(in-package :keim)

(mod~defmod ad2post :uses ( mod pp prob scan )
	    :documentation "Reads an object in ADLER syntax and constructs a KEIM object"
	    :exports
	    (ad2post~transform
	     ad2post~read-file
	     ad2post~write-object-to-file
	     ad2post~read-problem
	     ))
#{\section{Adler to Post Transformation}
This module implements a transformer from the Adler syntax adopted as
standard by the German Schwerpunkt Deduktion members. 
 An object in Adler syntax is read in and
a KEIM object is constructed. 

This cannot yet properly handle formulas with sorts. 
#}


(defun ad2post~transform (source-file target-file)
  (declare (edited  "07-MAR-1994 09:37")
	   (authors ACSEHN)
	   (input "A soure file in ADLER syntax and a target file."  )
	   (effect "The source file is read and transformed into POST syntax and written to target file."  )
	   (value "none."  ))
  (let ((symbollist (ad2post~read-file source-file)))
    (ad2post~write-object-to-file target-file symbollist)))
  
(defun ad2post~read-file (filename)
  (declare (edited  "15-OCT-1993 14:39")
	   (authors RICHTS)
	   (input   "The name of a file with a problem.")
	   (effect  "An error occurs if the file does not contain a syntactical correct adler problem.")
	   (value   "The POST representation of the problem (A list of lists and symbols)."))
  (let ((*package* (find-package :keim)))
    (with-open-file (stream filename)
      (scan~init stream))
    (progn (ad2post=initialize-predicates)
	   (ad2post=initialize-functions)
	   (ad2post=initialize-type-relation)
	   (ad2post=parse-problem-))))

(defun ad2post~read-problem (directory file)
  (declare (edited  "16-SEP-1993 15:33")
	   (authors ACSEHN)
	   (input "a directory and a file. "  )
	   (effect "none." )
	   (value  "the POST problem read form file in that directory."  ))
  (let ((new-path (concatenate 'string directory file)))
    (with-open-file (stream new-path :direction :input)
		    (prob~read (read stream)))))



;; (ad2post=parse-problem-)

;---------------------------- Names ----------------------------
(defun ad2post=is-name- ()
  (scan~name-p (scan~current-token)))

(defun ad2post=parse-name- ()
  (if (scan~name-p (scan~current-token))
      (prog1 (scan~token-value (scan~current-token))
	(scan~next-token))
    (ad2post=syn-err- "Name expected.")))

(defun ad2post=is-variable-name- ()
  (ad2post=is-name-))

(defun ad2post=parse-variable-name- ()
  (let ((token (scan~token-value (scan~current-token))))
    (ad2post=insert-variable token)
    (ad2post=parse-name-)))

;----------------------------- Variables ------------------------

(defvar ad2post*formula-variables NIL "The list where the variables of the actually ~
 scanned formula are saved.")

(defun ad2post=formula-variables ()
  (declare (edited  "02-MAR-1994 11:02")
	   (authors ACSEHN)
	   (input "None."  )
	   (effect "None."  )
	   (value "The list of variables for the actual formula."  ))
  ad2post*formula-variables)

(defun ad2post=in-formula-variables-p (object)
  (if (member object (ad2post=formula-variables))
      T
    NIL))
  
(defun ad2post=initialize-formula-variables ()
  (declare (edited  "02-MAR-1994 11:00")
	   (authors ACSEHN)
	   (input "None."  )
	   (effect  "The list of formula variables, i.e. ad2post*formula-variables, is set to the empty list." )
	   (value "NIL"  ))
  (setq ad2post*formula-variables NIL))

(defun ad2post=insert-variable (object)
  (declare (edited  "02-MAR-1994 10:54")
	   (authors ACSEHN)
	   (input "An object."  )
	   (effect "The object is inserted into the list of formula variables, i.e. ad2post*formula-variables" )
	   (value "The altereed list of variables."  ))
  (progn (setq ad2post*formula-variables (adjoin object ad2post*formula-variables))
	 ad2post*formula-variables))

;----------------------------- Predicates ------------------------

(defvar ad2post*predicates NIL "The list where the predicates of the actually ~
 scanned problem.")

(defun ad2post=predicates ()
  (declare (edited  "02-MAR-1994 11:02")
	   (authors ACSEHN)
	   (input "None."  )
	   (effect "None."  )
	   (value "The list of predicates for the actual problem."  ))
  ad2post*predicates)
  
(defun ad2post=initialize-predicates ()
  (declare (edited  "02-MAR-1994 11:00")
	   (authors ACSEHN)
	   (input "None."  )
	   (effect  "The list of predicates, i.e. ad2post*predicates, is set to the empty list." )
	   (value "NIL"  ))
  (setq ad2post*predicates NIL))

(defun ad2post=insert-predicate (object)
  (declare (edited  "02-MAR-1994 10:54")
	   (authors ACSEHN)
	   (input "An object."  )
	   (effect "The object is inserted into the list of predicates, i.e. ad2post*predicates" )
	   (value "The altered list of predicates."  ))
  (progn (when (not (ad2post=in-formula-variables-p object))
	   (setq ad2post*predicates (adjoin object ad2post*predicates)))
	 ad2post*predicates))

(defun ad2post=in-predicates-p (object)
  (if (member object (ad2post=predicates))
      T
    NIL))

;----------------------------- Functions ------------------------

(defvar ad2post*functions NIL "The list where the functions of the actually ~
 scanned problem.")

(defun ad2post=functions ()
  (declare (edited  "02-MAR-1994 11:02")
	   (authors ACSEHN)
	   (input "None."  )
	   (effect "None."  )
	   (value "The list of functions for the actual problem."  ))
  ad2post*functions)
  
(defun ad2post=initialize-functions ()
  (declare (edited  "02-MAR-1994 11:00")
	   (authors ACSEHN)
	   (input "None."  )
	   (effect  "The list of functions, i.e. ad2post*functions, is set to the empty list." )
	   (value "NIL"  ))
  (setq ad2post*functions NIL))

(defun ad2post=insert-function (object)
  (declare (edited  "02-MAR-1994 10:54")
	   (authors ACSEHN)
	   (input "An object."  )
	   (effect "The object is inserted into the list of functions, i.e. ad2post*functions" )
	   (value "The altered list of functions."  ))
  (progn (when (not (ad2post=in-formula-variables-p object))
	   (setq ad2post*functions (adjoin object ad2post*functions)))
	 ad2post*functions))

(defun ad2post=in-functions-p (object)
  (if (member object (ad2post=functions))
      T
    NIL))

;---------------------------- Tokens ----------------------------
(defun ad2post=is-token- (string)
  (and (scan~terminal-p (scan~current-token))
       (string-equal string (scan~token-value (scan~current-token)))))

(defun ad2post=accept- (string)
  (if (and (scan~terminal-p (scan~current-token))
	   (string-equal string (scan~token-value (scan~current-token))))
      (scan~next-token)
    (ad2post=syn-err- (format nil "Expected ~S." string))))


(defun ad2post=res- (arg)
  arg)

(defun ad2post=res1- (arg)
  arg)

;----------------------------- Error messages ----------------------------------------------------
(defun ad2post=syn-err- (string)
  (scan~error-msg string))

;---------------------------------------- The syntax rules ---------------------------------------


#|

#!syn ad2post problem is "begin_problem" <res {|'problem|}>
                         "("
                         name <res {..%res %name}>
                         ")"
                         "."
                         logical-part <res {..%res %logical-part}>
                         "end_problem"
                         "."
end-syn

|#



(PROGN (DEFUN AD2POST=IS-PROBLEM- () (LET () (AD2POST=IS-TOKEN- "begin_problem")))
       (DEFUN AD2POST=PARSE-PROBLEM- ()
         (LET (%LOGICAL-PART- %NAME- %RES-)
           (AD2POST=ACCEPT- "begin_problem")
           (SETF %RES- (AD2POST=RES- (APPEND (LIST 'PROBLEM))))
           (AD2POST=ACCEPT- "(")
           (SETF %NAME- (AD2POST=PARSE-NAME-))
           (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %NAME-))))
           (AD2POST=ACCEPT- ")")
           (AD2POST=ACCEPT- ".")
           (SETF %LOGICAL-PART- (AD2POST=PARSE-LOGICAL-PART-))
           (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %LOGICAL-PART-))))
           (AD2POST=ACCEPT- "end_problem")
           (AD2POST=ACCEPT- ".")
           %RES-)))

#|

#!syn ad2post logical-part is ( formula-list <res {..%res %formula-list}> )*
end-syn

|#

(PROGN (DEFUN AD2POST=IS-LOGICAL-PART- () (LET () (OR (OR (AD2POST=IS-FORMULA-LIST-)) NIL)))
       (DEFUN AD2POST=PARSE-LOGICAL-PART- ()
         (LET (%FORMULA-LIST- %RES-)
           (LOOP (COND ((AD2POST=IS-FORMULA-LIST-)
                        (SETF %FORMULA-LIST- (AD2POST=PARSE-FORMULA-LIST-))
                        (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-LIST-)))))
                       (T (RETURN))))
           %RES-)))

#|

#!syn ad2post formula-list is ( "list_of_axioms" <res {|'axiom-list|}>
			     / "list_of_theorems" <res {|'conclusion-list|}>
			   )
                          "."
                          ( "formula"
			    "("
			    formula <res1 %formula>
			    <#ad2post=find-types %formula >
			    [ "," name <res1 {|'named-formula| %name %res1}> ]
			    <res {..%res %res1}>			    
			    ")"
			    "." 
			  )* 
                         "end_of_list"
                         "."
end-syn
|#

(PROGN (DEFUN AD2POST=IS-FORMULA-LIST- () (LET () (OR (AD2POST=IS-TOKEN- "list_of_axioms") (AD2POST=IS-TOKEN- "list_of_theorems"))))
       (DEFUN AD2POST=PARSE-FORMULA-LIST- ()
         (LET (%NAME- %AD2POST=FIND-TYPES %RES1- %FORMULA- %RES-)
           (COND ((AD2POST=IS-TOKEN- "list_of_axioms")
                  (AD2POST=ACCEPT- "list_of_axioms")
                  (SETF %RES- (AD2POST=RES- (APPEND (LIST 'AXIOM-LIST)))))
                 ((AD2POST=IS-TOKEN- "list_of_theorems")
                  (AD2POST=ACCEPT- "list_of_theorems")
                  (SETF %RES- (AD2POST=RES- (APPEND (LIST 'CONCLUSION-LIST)))))
                 (T (AD2POST=SYN-ERR- "Alternative failed")))
           (AD2POST=ACCEPT- ".")
           (LOOP (COND ((AD2POST=IS-TOKEN- "formula")
                        (AD2POST=ACCEPT- "formula")
                        (AD2POST=ACCEPT- "(")
                        (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                        (SETF %RES1- (AD2POST=RES1- %FORMULA-))
                        (SETF %AD2POST=FIND-TYPES (AD2POST=FIND-TYPES %FORMULA-))
                        (COND ((AD2POST=IS-TOKEN- ",")
                               (AD2POST=ACCEPT- ",")
                               (SETF %NAME- (AD2POST=PARSE-NAME-))
                               (SETF %RES1- (AD2POST=RES1- (APPEND (LIST 'NAMED-FORMULA) (LIST %NAME-) (LIST %RES1-))))))
                        (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %RES1-))))
                        (AD2POST=ACCEPT- ")")
                        (AD2POST=ACCEPT- "."))
                       (T (RETURN))))
           (AD2POST=ACCEPT- "end_of_list")
           (AD2POST=ACCEPT- ".")
           %RES-)))


#|

#!syn ad2post formula is  ( ( "forall" <res {|'forall|}> <#ad2post=insert-predicate |'forall| >
			      / "exists" <res {|'exists|}> <#ad2post=insert-predicate |'exists| >
			      / <!#ad2post=quant-p > quant <res {%quant} > <#ad2post=insert-predicate %quant >
			    )
			    "("
			    var-list <res {..%res %var-list}>
			    ","
			    formula <res {..%res %formula}>
			    ")"
			  / log-op-formula <res %log-op-formula>
			  )
end-syn

|#


(PROGN (DEFUN AD2POST=IS-FORMULA- ()
         (LET (%AD2POST=QUANT-P)
           (OR (OR (AD2POST=IS-TOKEN- "forall") (AD2POST=IS-TOKEN- "exists") (AD2POST=QUANT-P)) (AD2POST=IS-LOG-OP-FORMULA-))))
       (DEFUN AD2POST=PARSE-FORMULA- ()
         (LET (%LOG-OP-FORMULA- %FORMULA- %VAR-LIST- %QUANT- %AD2POST=INSERT-PREDICATE %AD2POST=QUANT-P %RES-)
           (COND ((OR (AD2POST=IS-TOKEN- "forall") (AD2POST=IS-TOKEN- "exists") (AD2POST=QUANT-P))
                  (COND ((AD2POST=IS-TOKEN- "forall")
                         (AD2POST=ACCEPT- "forall")
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST 'FORALL))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'FORALL)))
                        ((AD2POST=IS-TOKEN- "exists")
                         (AD2POST=ACCEPT- "exists")
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST 'EXISTS))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'EXISTS)))
                        ((AD2POST=QUANT-P)
                         (SETF %AD2POST=QUANT-P (AD2POST=QUANT-P))
                         (SETF %QUANT- (AD2POST=PARSE-QUANT-))
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST %QUANT-))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE %QUANT-)))
                        (T (AD2POST=SYN-ERR- "Alternative failed")))
                  (AD2POST=ACCEPT- "(")
                  (SETF %VAR-LIST- (AD2POST=PARSE-VAR-LIST-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %VAR-LIST-))))
                  (AD2POST=ACCEPT- ",")
                  (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-))))
                  (AD2POST=ACCEPT- ")"))
                 ((AD2POST=IS-LOG-OP-FORMULA-)
                  (SETF %LOG-OP-FORMULA- (AD2POST=PARSE-LOG-OP-FORMULA-))
                  (SETF %RES- (AD2POST=RES- %LOG-OP-FORMULA-)))
                 (T (AD2POST=SYN-ERR- "Alternative failed")))
           %RES-)))

;; in ``log-op-formula'' the nonterminal ``atom'' is included.

#|

#!syn ad2post var-list is  <#ad2post=initialize-formula-variables > "["  variable-name <res {%variable-name}>
                               ( "," variable-name <res {..%res %variable-name}> )* 
			   "]"
end-syn

|#

(PROGN (DEFUN AD2POST=IS-VAR-LIST- () (LET () (AD2POST=IS-TOKEN- "[")))
       (DEFUN AD2POST=PARSE-VAR-LIST- ()
         (LET (%VARIABLE-NAME- %AD2POST=INITIALIZE-FORMULA-VARIABLES %RES-)
           (SETF %AD2POST=INITIALIZE-FORMULA-VARIABLES (AD2POST=INITIALIZE-FORMULA-VARIABLES))
           (AD2POST=ACCEPT- "[")
           (SETF %VARIABLE-NAME- (AD2POST=PARSE-VARIABLE-NAME-))
           (SETF %RES- (AD2POST=RES- (APPEND (LIST %VARIABLE-NAME-))))
           (LOOP (COND ((AD2POST=IS-TOKEN- ",")
                        (AD2POST=ACCEPT- ",")
                        (SETF %VARIABLE-NAME- (AD2POST=PARSE-VARIABLE-NAME-))
                        (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %VARIABLE-NAME-)))))
                       (T (RETURN))))
           (AD2POST=ACCEPT- "]")
           %RES-)))


(defun ad2post=is-quant- ()
  (ad2post=is-name-))

(defun ad2post=parse-quant- ()
  (let ((token (scan~token-value (scan~current-token))))
    (if (ad2post=quant-p)
	(let ((number (ad2post=suffix 'quant token)))
	  (if (ad2post=number-p number)
	      (prog1 token
		(scan~next-token))
	    (ad2post=syn-err- "Number after ``quant'' expected.")))
      (ad2post=parse-name-))))

(defun ad2post=quant-p ()
  (let* ((operator 'quant)
	 (token (scan~token-value (scan~current-token)))
	 (number (ad2post=suffix operator  token))
	 (subpos (string<= (string operator) token)))
    (when subpos
      (> subpos 0))))

#|

#!syn ad2post log-op-formula is ( "not" <res {|'not|}>
				  <#ad2post=insert-predicate |'not| >
				  "("
				  formula <res {..%res %formula}>
				  ")"
				  / <!#ad2post=un-op-p > un-op <res {%un-op}>
				          <#ad2post=insert-predicate %un-op >
 				     "("
				     formula <res {..%res %formula}>
				     ")"
				  / ( "implies" <res {|'implies|}> <#ad2post=insert-predicate |'implies| >
				      / "implied" <res {|'implies|}> <#ad2post=insert-predicate |'implies| >
				      / "equiv" <res {|'equiv|}> <#ad2post=insert-predicate |'equiv| >
				      / <!#ad2post=bin-op-p > bin-op <res {%bin-op}> <#ad2post=insert-predicate %bin-op >
				     ) 
				     "("
				     formula <res {..%res %formula}>
				     ","
				     formula <res {..%res %formula}>				     
				     ")"
				  / ( "and" <res {|'and|}> <#ad2post=insert-predicate |'and| >
				      / "or" <res {|'or|}> <#ad2post=insert-predicate |'or| > )
				    "("
				    formula <res {..%res %formula}>
				    [ ( "," formula <res {..%res %formula}> )* ]
				    ")"
				  / atom <res %atom>
				)
end-syn

|#

(PROGN (DEFUN AD2POST=IS-LOG-OP-FORMULA- ()
         (LET (%AD2POST=BIN-OP-P %AD2POST=UN-OP-P)
           (OR (AD2POST=IS-TOKEN- "not") (AD2POST=UN-OP-P)
               (OR (AD2POST=IS-TOKEN- "implies") (AD2POST=IS-TOKEN- "implied") (AD2POST=IS-TOKEN- "equiv") (AD2POST=BIN-OP-P))
               (OR (AD2POST=IS-TOKEN- "and") (AD2POST=IS-TOKEN- "or")) (AD2POST=IS-ATOM-))))
       (DEFUN AD2POST=PARSE-LOG-OP-FORMULA- ()
         (LET (%ATOM- %BIN-OP- %AD2POST=BIN-OP-P %UN-OP- %AD2POST=UN-OP-P %FORMULA- %AD2POST=INSERT-PREDICATE %RES-)
           (COND ((AD2POST=IS-TOKEN- "not")
                  (AD2POST=ACCEPT- "not")
                  (SETF %RES- (AD2POST=RES- (APPEND (LIST 'NOT))))
                  (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'NOT))
                  (AD2POST=ACCEPT- "(")
                  (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-))))
                  (AD2POST=ACCEPT- ")"))
                 ((AD2POST=UN-OP-P)
                  (SETF %AD2POST=UN-OP-P (AD2POST=UN-OP-P))
                  (SETF %UN-OP- (AD2POST=PARSE-UN-OP-))
                  (SETF %RES- (AD2POST=RES- (APPEND (LIST %UN-OP-))))
                  (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE %UN-OP-))
                  (AD2POST=ACCEPT- "(")
                  (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-))))
                  (AD2POST=ACCEPT- ")"))
                 ((OR (AD2POST=IS-TOKEN- "implies") (AD2POST=IS-TOKEN- "implied") (AD2POST=IS-TOKEN- "equiv") (AD2POST=BIN-OP-P))
                  (COND ((AD2POST=IS-TOKEN- "implies")
                         (AD2POST=ACCEPT- "implies")
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST 'IMPLIES))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'IMPLIES)))
                        ((AD2POST=IS-TOKEN- "implied")
                         (AD2POST=ACCEPT- "implied")
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST 'IMPLIES))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'IMPLIES)))
                        ((AD2POST=IS-TOKEN- "equiv")
                         (AD2POST=ACCEPT- "equiv")
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST 'EQUIV))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'EQUIV)))
                        ((AD2POST=BIN-OP-P)
                         (SETF %AD2POST=BIN-OP-P (AD2POST=BIN-OP-P))
                         (SETF %BIN-OP- (AD2POST=PARSE-BIN-OP-))
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST %BIN-OP-))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE %BIN-OP-)))
                        (T (AD2POST=SYN-ERR- "Alternative failed")))
                  (AD2POST=ACCEPT- "(")
                  (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-))))
                  (AD2POST=ACCEPT- ",")
                  (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-))))
                  (AD2POST=ACCEPT- ")"))
                 ((OR (AD2POST=IS-TOKEN- "and") (AD2POST=IS-TOKEN- "or"))
                  (COND ((AD2POST=IS-TOKEN- "and")
                         (AD2POST=ACCEPT- "and")
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST 'AND))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'AND)))
                        ((AD2POST=IS-TOKEN- "or")
                         (AD2POST=ACCEPT- "or")
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST 'OR))))
                         (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE 'OR)))
                        (T (AD2POST=SYN-ERR- "Alternative failed")))
                  (AD2POST=ACCEPT- "(")
                  (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-))))
                  (COND ((OR (OR (AD2POST=IS-TOKEN- ",")) NIL)
                         (LOOP (COND ((AD2POST=IS-TOKEN- ",")
                                      (AD2POST=ACCEPT- ",")
                                      (SETF %FORMULA- (AD2POST=PARSE-FORMULA-))
                                      (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %FORMULA-)))))
                                     (T (RETURN))))))
                  (AD2POST=ACCEPT- ")"))
                 ((AD2POST=IS-ATOM-) (SETF %ATOM- (AD2POST=PARSE-ATOM-)) (SETF %RES- (AD2POST=RES- %ATOM-)))
                 (T (AD2POST=SYN-ERR- "Alternative failed")))
           %RES-)))

(defun ad2post=is-un-op- ()
  (ad2post=is-name-))

(defun ad2post=parse-un-op- ()
  (let ((token (scan~token-value (scan~current-token))))
    (if (ad2post=un-op-p)
	(let ((number (ad2post=suffix 'un_op token)))
	  (if (ad2post=number-p number)
	      (prog1 token
		(scan~next-token))
	    (ad2post=syn-err- "Number after ``un_op'' expected.")))
      (ad2post=parse-name-))))

(defun ad2post=un-op-p ()
  (let* ((operator 'un_op)
	 (token (scan~token-value (scan~current-token)))
	 (number (ad2post=suffix operator  token))
	 (subpos (string<= (string operator) token)))
    (when subpos
      (> subpos 0))))


(defun ad2post=is-bin-op- ()
  (ad2post=is-name-))

(defun ad2post=parse-bin-op- ()
  (let ((token (scan~token-value (scan~current-token))))
    (if (ad2post=bin-op-p)
	(let ((number (ad2post=suffix 'bin_op token)))
	  (if (ad2post=number-p number)
	      (prog1 token
		(scan~next-token))
	    (ad2post=syn-err- "Number after ``bin_op'' expected.")))
      (ad2post=parse-name-))))

(defun ad2post=bin-op-p ()
  (let* ((operator 'bin_op)
	 (token (scan~token-value (scan~current-token)))
	 (number (ad2post=suffix operator  token))
	 (subpos (string<= (string operator) token)))
    (when subpos
      (> subpos 0))))

#|

#!syn ad2post atom  is  ( "true" <res |'true| > <#ad2post=insert-predicate %res >
			   / "false" <res |'false| > <#ad2post=insert-predicate %res >
			   / "equal" <res {|'=|}> <#ad2post=insert-predicate |'=| >
			           "("
				   term <res {..%res %term}>
				   ","
				   term <res {..%res %term}>
				   ")"
			   /  <!#ad2post=zero-op-p > zero-op <res %zero-op> <#ad2post=insert-predicate %zero-op >
			   /  pred-expr <res %pred-expr>
			   )
end-syn

|#

(PROGN (DEFUN AD2POST=IS-ATOM- ()
         (LET (%AD2POST=ZERO-OP-P)
           (OR (AD2POST=IS-TOKEN- "true") (AD2POST=IS-TOKEN- "false") (AD2POST=IS-TOKEN- "equal") (AD2POST=ZERO-OP-P)
               (AD2POST=IS-PRED-EXPR-))))
       (DEFUN AD2POST=PARSE-ATOM- ()
         (LET (%PRED-EXPR- %ZERO-OP- %AD2POST=ZERO-OP-P %TERM- %AD2POST=INSERT-PREDICATE %RES-)
           (COND ((AD2POST=IS-TOKEN- "true")
                  (AD2POST=ACCEPT- "true")
                  (SETF %RES- (AD2POST=RES- 'TRUE))
                  (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE %RES-)))
                 ((AD2POST=IS-TOKEN- "false")
                  (AD2POST=ACCEPT- "false")
                  (SETF %RES- (AD2POST=RES- 'FALSE))
                  (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE %RES-)))
                 ((AD2POST=IS-TOKEN- "equal")
                  (AD2POST=ACCEPT- "equal")
                  (SETF %RES- (AD2POST=RES- (APPEND (LIST '=))))
                  (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE '=))
                  (AD2POST=ACCEPT- "(")
                  (SETF %TERM- (AD2POST=PARSE-TERM-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %TERM-))))
                  (AD2POST=ACCEPT- ",")
                  (SETF %TERM- (AD2POST=PARSE-TERM-))
                  (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %TERM-))))
                  (AD2POST=ACCEPT- ")"))
                 ((AD2POST=ZERO-OP-P)
                  (SETF %AD2POST=ZERO-OP-P (AD2POST=ZERO-OP-P))
                  (SETF %ZERO-OP- (AD2POST=PARSE-ZERO-OP-))
                  (SETF %RES- (AD2POST=RES- %ZERO-OP-))
                  (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE %ZERO-OP-)))
                 ((AD2POST=IS-PRED-EXPR-) (SETF %PRED-EXPR- (AD2POST=PARSE-PRED-EXPR-)) (SETF %RES- (AD2POST=RES- %PRED-EXPR-)))
                 (T (AD2POST=SYN-ERR- "Alternative failed")))
           %RES-)))



(defun ad2post=is-zero-op- ()
  (ad2post=is-name-))

(defun ad2post=parse-zero-op- ()
  (let ((token (scan~token-value (scan~current-token))))
    (if (ad2post=zero-op-p)
	(let ((number (ad2post=suffix 'zero_op token)))
	  (if (ad2post=number-p number)
	      (prog1 token
		(scan~next-token))
	    (ad2post=syn-err- "Number after ``zero_op'' expected.")))
      (ad2post=parse-name-))))

(defun ad2post=zero-op-p ()
  (let* ((operator 'zero_op)
	 (token (scan~token-value (scan~current-token)))
	 (number (ad2post=suffix operator  token))
	 (subpos (string<= (string operator) token)))
    (when subpos
      (> subpos 0))))


(defun ad2post=number-p (string)
  (declare (edited  "07-FEB-1994 18:29")
	   (authors ACSEHN)
	   (input "A string."  )
	   (effect "None." )
	   (value  "T, iff the string is a number" ))
  (eql (length (string-trim '(#\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\0) string)) 0))
      



(defun ad2post=suffix (prefix symbol)
  (declare (edited  "07-FEB-1994 17:22")
	   (authors ACSEHN)
	   (input "A SYMBOL starting with symbol PREFIX"  )
	   (effect "None." )
	   (value "The rest of the string where prefix is removed from the front."  ))
  (let ((position (string<= (string prefix) (string symbol))))
    (if position
	(subseq (string symbol) position)
      symbol)))

#|

#!syn ad2post pred-expr is pred-sym <res %pred-sym>
                         [  "(" term  <res {%res %term}>
			       ( "," term <res {..%res %term}> )*
			    ")"	 ]
end-syn

|#

(PROGN (DEFUN AD2POST=IS-PRED-EXPR- () (LET () (AD2POST=IS-PRED-SYM-)))
       (DEFUN AD2POST=PARSE-PRED-EXPR- ()
         (LET (%TERM- %PRED-SYM- %RES-)
           (SETF %PRED-SYM- (AD2POST=PARSE-PRED-SYM-))
           (SETF %RES- (AD2POST=RES- %PRED-SYM-))
           (COND ((AD2POST=IS-TOKEN- "(")
                  (AD2POST=ACCEPT- "(")
                  (SETF %TERM- (AD2POST=PARSE-TERM-))
                  (SETF %RES- (AD2POST=RES- (APPEND (LIST %RES-) (LIST %TERM-))))
                  (LOOP (COND ((AD2POST=IS-TOKEN- ",")
                               (AD2POST=ACCEPT- ",")
                               (SETF %TERM- (AD2POST=PARSE-TERM-))
                               (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %TERM-)))))
                              (T (RETURN))))
                  (AD2POST=ACCEPT- ")")))
           %RES-)))

#|

#!syn ad2post pred-sym is name <#ad2post=insert-predicate %name > <res %name>
end-syn

|#

(PROGN (DEFUN AD2POST=IS-PRED-SYM- () (LET () (AD2POST=IS-NAME-)))
       (DEFUN AD2POST=PARSE-PRED-SYM- ()
         (LET (%AD2POST=INSERT-PREDICATE %NAME- %RES-)
           (SETF %NAME- (AD2POST=PARSE-NAME-))
           (SETF %AD2POST=INSERT-PREDICATE (AD2POST=INSERT-PREDICATE %NAME-))
           (SETF %RES- (AD2POST=RES- %NAME-))
           %RES-)))


#|

#!syn ad2post term is  ( symbol  <res %symbol> <#ad2post=insert-function %symbol >
			    [  "(" term  <res {%res %term}>
			          ( "," term <res {..%res %term}> )*
			      ")" ]
		       )
end-syn

|#

(PROGN (DEFUN AD2POST=IS-TERM- () (LET () (OR (AD2POST=IS-SYMBOL-))))
       (DEFUN AD2POST=PARSE-TERM- ()
         (LET (%TERM- %AD2POST=INSERT-FUNCTION %SYMBOL- %RES-)
           (COND ((AD2POST=IS-SYMBOL-)
                  (SETF %SYMBOL- (AD2POST=PARSE-SYMBOL-))
                  (SETF %RES- (AD2POST=RES- %SYMBOL-))
                  (SETF %AD2POST=INSERT-FUNCTION (AD2POST=INSERT-FUNCTION %SYMBOL-))
                  (COND ((AD2POST=IS-TOKEN- "(")
                         (AD2POST=ACCEPT- "(")
                         (SETF %TERM- (AD2POST=PARSE-TERM-))
                         (SETF %RES- (AD2POST=RES- (APPEND (LIST %RES-) (LIST %TERM-))))
                         (LOOP (COND ((AD2POST=IS-TOKEN- ",")
                                      (AD2POST=ACCEPT- ",")
                                      (SETF %TERM- (AD2POST=PARSE-TERM-))
                                      (SETF %RES- (AD2POST=RES- (APPEND %RES- (LIST %TERM-)))))
                                     (T (RETURN))))
                         (AD2POST=ACCEPT- ")"))))
                 (T (AD2POST=SYN-ERR- "Alternative failed")))
           %RES-)))

#|

#!syn ad2post symbol is name <res %name>
end-syn

|#

(PROGN (DEFUN AD2POST=IS-SYMBOL- () (LET () (AD2POST=IS-NAME-)))
       (DEFUN AD2POST=PARSE-SYMBOL- () (LET (%NAME- %RES-) (SETF %NAME- (AD2POST=PARSE-NAME-)) (SETF %RES- (AD2POST=RES- %NAME-)) %RES-)))

;--------------------- Predefined Symbols -------------------------

(defun ad2post=predefined-symbols ()
  '(
    (defn (o aa aa))
    (image (o aa (o bb) (aa bb)))
    (powerset (o (o aa) (o aa)))
    (ifthen (aa o aa))
    (intersection (o aa (o aa) (o aa)))
    (union (o aa (o aa) (o aa)))
    (superset (o (o aa) (o aa)))
    (subset (o (o aa) (o aa)))
    (exists (o (o aa)))
    (forall (o (o aa)))
    (not (o o))
    (equiv (o o o))
    (implies (o o o))
    (or (o o o))
    (and (o o o))
    (true o)
    (false o)
    (= (o aa aa))
    ))



;(type-variables bb)
;    (type-variables aa)
;    (type-constants i)
;    (type-constants o) 

;------------------- Type Reconstruction ---------------------

(defvar ad2post*type-relation NIL "This is a list of the relation between symols and their types.")

(defun ad2post=initialize-type-relation ()
  (declare (edited  "02-MAR-1994 19:03")
	   (authors ACSEHN)
	   (input  "None." )
	   (effect "The type relation is set to the empty set." )
	   (value "NIL"  ))
  (setq ad2post*type-relation NIL))

(defun ad2post=insert-type-pair (object type)
  (declare (edited  "02-MAR-1994 19:02")
	   (authors ACSEHN)
	   (input "An object and a type."  )
	   (effect "The object with its type is added to the type relation." )
	   (value "The altered type relation."  ))
  (progn (setq ad2post*type-relation (adjoin (cons object type) ad2post*type-relation :test #'equal))
	 ad2post*type-relation))

(defun ad2post=get-type (object)
  (declare (edited  "02-MAR-1994 19:02")
	   (authors ACSEHN)
	   (input "An object."  )
	   (effect "None." )
	   (value "The corresponding type of OBJECT."  ))
  (let ((type-list (remove-if-not #'(lambda (x) (eql (first x) object)) ad2post*type-relation)))
    (if (eql (length type-list) 1)
	(rest (first type-list))
      (mapcar #'cdr type-list))))
	

(defun ad2post=quantor-p (object)
  (declare (edited  "02-MAR-1994 19:01")
	   (authors ACSEHN)
	   (input "An object."  )
	   (effect "None." )
	   (value "T, iff the object has the suffix ``quant''."  ))
  (if (not (listp object))
      (let* ((operator 'quant)
	     (number (ad2post=suffix operator  object))
	     (subpos (string<= (string operator) object)))
	(when subpos
	  (> subpos 0)))
    NIL))

(defun ad2post=find-types (list)
  (declare (edited  "02-MAR-1994 15:16")
	   (authors ACSEHN)
	   (input "A list representing a term."  )
	   (effect "All symbols of the term are given its types." )
	   (value "The top symbol."  ))
  (if (and (listp list)
	   (not (null list)))
      (let* ((head (first list))
	     ;; make sure bound variables do not get kept in this list!!
	    (rest-types 
	     (if (or (ad2post=quantor-p head)
		     (equal head 'FORALL)
		     (equal head 'EXISTS))
		 (mapcar #'ad2post=find-types (cdr (rest list)))
		 (mapcar #'ad2post=find-types (rest list)))))
	(cond ((or (ad2post=quantor-p head)
		     (equal head 'FORALL)
		     (equal head 'EXISTS))
		 (ad2post=insert-type-pair head '(O (O I)))
		 'O)
		((or (equal head 'AND)
		     (equal head 'OR))
		 (ad2post=insert-type-pair head '(O O O))
		 'O)
		((ad2post=in-functions-p head)
		 (ad2post=insert-type-pair head (cons 'I rest-types)) 
		 'I)
		((ad2post=in-predicates-p head)
		 (ad2post=insert-type-pair  head (cons 'O rest-types))
		 'O)
		(t (ad2post=get-type head))))
    (cond ((ad2post=in-formula-variables-p list)
	   (ad2post=insert-type-pair list 'I)
	   'I)
	  ((ad2post=in-functions-p list)
	   (ad2post=insert-type-pair list 'I)
	   'I)
	  ((ad2post=in-predicates-p list)
	   (ad2post=insert-type-pair list 'O)
	   'O))))

(defun ad2post=find-types (term)
  (let ((new-pairs (ad2post=find-types-aux term 'O ad2post*type-relation)))
    (setq ad2post*type-relation
	  (remove-duplicates (append ad2post*type-relation new-pairs)
			     :test #'equal))))

(defun ad2post=find-types-aux (term current-type found-so-far)
  (cond ((null term) nil)
	((and (symbolp term)
	      (or (member term '(true false))
		  ;; is of form zero_opNUM
		  (and (eql 0 (search "ZERO_OP" (symbol-name term)
				      :start2 0 :test #'char-equal))
		       (multiple-value-bind (int index)
			   (parse-integer (symbol-name term)
					  :start 5
					  :junk-allowed t)
			 (and (integerp int)
			      (= index (length (symbol-name term)))))))
	      (acons term 'o new-stuff)))
	((symbolp term)
	 (if (assoc term found-so-far)
	     nil
	     (list (cons term (or current-type 'I)))))
	((listp term)
	 (let ((head (first term))
	       (args (rest term)))
	   (cond
	    ((or (ad2post=quantor-p head)
		 (member head '(forall exists)))
	     (let ((var-types
		    (mapcar #'(lambda (var) (cons var 'i))
			    (first args))))
	       (acons head '(o (o i))
		      (ad2post=find-types-aux
		       (second args)
		       'o
		       (append var-types found-so-far)))))
	    ((or (member head '(and or implies implied equiv equal))
		 ;; is of form bin_opNUM
		 (and (eql 0 (search "BIN_OP" (symbol-name head)
				     :start2 0 :test #'char-equal))
		      (multiple-value-bind (int index)
			  (parse-integer (symbol-name head)
					 :start 5
					 :junk-allowed t)
			(and (integerp int)
			     (= index (length (symbol-name head)))))))
	     (let ((new-stuff nil))
	       (dolist (arg args  (acons head '(o o o) new-stuff))
		 (setq new-stuff
		       (append 
			(ad2post=find-types-aux 
			 arg 'o
			 (append new-stuff found-so-far))
			new-stuff)))))
	    ((or (eq head 'not)
		 ;; is of form un_opNUM
		 (and (eql 0 (search "UN_OP" (symbol-name head)
				   :start2 0 :test #'char-equal))
			 (multiple-value-bind (int index)
			     (parse-integer (symbol-name head)
					    :start 5
					    :junk-allowed t)
			   (and (integerp int)
				(= index (length (symbol-name head)))))))
	     (let ((new-stuff 
		    (ad2post=find-types-aux 
		     (first args) 'o
		     found-so-far)))
	       (acons head '(o o) new-stuff)))
	    (t
	     (let ((new-stuff nil))
	       (dolist (arg args 	
			    (acons head
				   (cons current-type
					 (make-list (length args)
						    :initial-element 'i))
				   new-stuff))
		 (setq new-stuff
		       (append 
			(ad2post=find-types-aux 
			 arg nil
			 (append new-stuff found-so-far))
			new-stuff))))))))))
				   
	       

;;;  writing to file

(defun ad2post=quantor-p* (obj)
  (and (listp obj)
       (or (member (first obj) '(forall exists))
	   (ad2post=quantor-p (first obj)))))

(deftype ad2post=quantified-formula ()
  `(satisfies ad2post=quantor-p*))

(defun ad2post=equal-p (obj)
  (and (symbolp obj)
       (string-equal (symbol-name obj) "EQUAL")))

(deftype ad2post=equal ()
  `(satisfies ad2post=equal-p))

(defun ad2post=print-problem (stream object)
  (let ((name (second object))
	(env (third object))
	(*standard-output* stream)
	(*print-escape* nil))
    (pprint-logical-block 
     (nil nil :prefix "(problem " :suffix ")")
     (write name)
     (write-string " adler")
     (terpri)
     (pprint-indent :block 1)
     (write (append
	     (list '("type-constants" I O))
	     (mapcar #'(lambda (pair)
			 (list 'constant (car pair) (cdr pair)))
		     ad2post*type-relation)
	     env))
     )))

(defun ad2post=print-quantifier (stream formula)
  (let ((quantor (first formula))
	(var-list (second formula))
	(scope (third formula))
	(*standard-output* stream))
    (if var-list
	(pprint-logical-block 
	 (nil var-list :prefix "(" :suffix "))")
	 (loop
	  (pprint-exit-if-list-exhausted)
	  (write quantor)
	  (write-string "(lam (")
	  (write (pprint-pop))
	  (write-string " I)")
	  (write (list quantor (rest var-list) scope))))
	(write scope))))


(pp~defstyle ad2post-style
	     :help "Style for printing POST from Adler"
	     :parent pp+top
	     :pprint-methods
	     ((list
	       pprint-fill 
	       0)
	      (ad2post=quantified-formula
	       ad2post=print-quantifier
	       5)
	      ((cons (member problem))
	       ad2post=print-problem
	       5)
	      ((cons (member constant))
	       (lambda (str obj)
		 (write-string "(constant " str)
		 (write (second obj) :stream str)
		 (write-string " " str)
		 (write (third obj) :stream str)
		 (write-string ")" str))
	       5)
	      ((cons (member axiom-list))
	       (lambda (str obj)
		 (let ((*standard-output* str))
		 (pprint-logical-block 
		  (nil (rest obj))
		  (loop
		   (terpri)
		   (pprint-exit-if-list-exhausted)
		   (let ((object (pprint-pop)))
		     (pprint-logical-block 
		      (nil (list
			    (if (eq (first object) 'named-formula)
				(second object)
				(gensym "AX"))
			    " "
			    (if (eq (first object) 'named-formula)
				(third object)
				object))
			 :prefix "(assumption " :suffix ")")
		      (loop
		       (pprint-exit-if-list-exhausted)
		       (write (pprint-pop)))))))))
		 5)
	      ((cons (member conclusion-list))
	       (lambda (str obj)
		 (let ((*standard-output* str))
		   (pprint-logical-block 
		    (nil (rest obj))
		    (loop
		     (terpri)
		     (pprint-exit-if-list-exhausted)
		     (let ((object (pprint-pop)))
		       (pprint-logical-block 
			(nil (list
			      (if (eq (first object) 'named-formula)
				  (second object)
				  (gensym "CONC"))
			      " "
			      (if (eq (first object) 'named-formula)
				  (third object)
				  object))
			     :prefix "(conclusion " :suffix ")")
			(loop
			 (pprint-exit-if-list-exhausted)
			 (write (pprint-pop)))))))))
	       5)
	      (ad2post=equal
	       (lambda (str obj)
		 (write '= :stream str)))
	      ((cons (member AND OR))
	       (lambda (stream obj)
		 (let ((head (car obj))
		       (args (cdr obj)))
		   (pprint-logical-block 
		    (nil nil :prefix "(" :suffix ")")
		    (write head :stream stream)
		    (write-char #\space stream)
		    (write (car args) :stream stream)
		    (write-string " " stream)
		    (if (eql (length args) 2)
			(write (cadr args) :stream stream)
			(write (cons head (cdr args)) :stream stream))
		    )))
	       5)))

(defun ad2post~write-object-to-file (filename object)
  (declare (edited  "02-MAR-1994 17:01")
	   (authors ACSEHN)
	   (input "A file and an object. "  )
	   (effect "none." )
	   (value  "the object is written to the file."  ))
  (with-open-file (stream filename :direction :output :if-exists :supersede)
    (pp~pprint object 'ad2post-style stream)))

;;


