;;; -*- syntax: common-lisp; package: keim; base: 10; mode: lisp -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1994 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;


(in-package :KEIM)

(mod~defmod cont :uses (mod )
	    :documentation "Continue Objects."
	    :exports (cont~sequence cont~forms cont~start cont~continue cont~next cont~result)
	    )


#{
  This module provides some functions for a simple continue mechanism. This
  mechanism allows to return from an algorithm (for returning a solution or
  because some limit has been exceeded) and to continue this algorithm
  later. Moreover this is combined with a backtracking mechanism, i.e. the
  algorithm can set several {\em continue points} that will be executed in
  a defined order. The points can be sorted like a stack (for depth-search),
  like a queue (for breadth-first-search), or with heuristic values.

  The key point of this mechanism is the {\em continue object} that holds the
  continue points and the last result. For this purpose the class {\tt
  cont+continue} is defined. It is intended to be used as a mixin
  class. After the creation of a continue object, the continue mechanism is
  initialized and/or started with {\tt (cont~start}. To set some points that
  shall be continued later, you have to call {\tt (cont~sequence ({\it val-1}
  {\it form-1}) \ldots ({\it val-n} {\it form-n}))}. This macro is some
  syntactic sugar for the function {\tt (cont~forms (list (cons {\it val-1}
  \#'(lambda () {\it form-1})) \ldots (cons {\it val-n} \#'(lambda () {\it
  form-n}))))}. Both expressions evaluate the values and add the forms
  associated with these values to the current list of continue points. The
  values can be arbitrary objects; there meaning is determined by the call to
  {\tt cont~next} (although {\tt t} and {\tt nil} have a predefined
  meaning). When an algorithm started with {\tt cont~start} (or continued with
  {\tt cont~continue}) returns a result, this result is stored in the continue
  object where it can be accessed with {\tt (cont~result
  cont-object)}. Furthermore all continue points that have been set with {\tt
  cont~sequence} or {\tt cont~forms} are stored in the continue object. The
  algorithm can be continued with {\tt (cont~continue {\it continue-object}
  \&optional {\it test})}. Inside the algorithm the next continue point can be
  executed with {\tt (cont~next \&optional {\it test})}.

  Now we give an example that shows the possibilities of this mechanism. Suppose
  we want to find the ways through a three-dimensional, acyclic maze. If we come
  to fork, we have to go left or right; at a stair we have to go up and or down;
  and at a dead-end we have to make another decision.  If we find an exit, we
  want to stop searching to do something different but we want be able to
  continue to the search for another way in the maze.

  We order the alternatives in the following way: the stairs down are ordered
  like a stack, i.e. the last stair down is selected if there is one; at a fork
  the ways are ordered by their width, i.e. if there are no more stairs down,
  the smallest way is selected; the stairs up are ordered like a queue, i.e. if
  there are no more stays down and no more ways after a fork, the first stair
  up is selected. (This ordering is a little bit silly but it shows all
  possibilities of the mechanism.)

\begin{code}
(defun go-maze (maze)
  (cond ((fork-p maze)
	 (cont~sequence ((width (left maze)) (go-maze (left maze)))
			((width (right maze)) (go-maze (right maze))))
	 (cont~next))
	((stair-p maze)
	 (cont~sequence (t (go-maze (down maze)))
			(nil (go-maze (up maze))))
	 (cont~next))
	((dead-end-p maze)
	 (cont~next))
	((exit-p maze)
	 maze)))

(defun list-all-ways (maze)
  (do ((cont-object (cont~start (go-maze maze))
		    (cont~continue cont-object))
       (ways nil (if (cont~result cont-object)
		     (cons (cont~result cont-object) ways)
		   ways)))
      ((null cont-object) ways))
  (do-something-different))
      

\end{code}

  The function {\tt list-all-ways} finds all ways thruogh the maze, but
  everytime an exit is found and a solution is returned there can be done
  something different. After this the search in the maze is continued.

#}

(eval-when (load compile eval)
(defclass cont+continue ()
  ((result :initform nil :reader cont~result :accessor cont=result)
   (points :initform nil :accessor cont=points))
  (:documentation "This is class is intended to be a mixin for objects which are associated with a computation
that must be able to be stopped and continued later.")
))

(defun cont~create () 
  (declare (edited  "01-AUG-1994 11:35")
	   (authors RICHTS)
	   (input   "None. The continue object are initialized with cont~start.")
	   (effect  "None.")
	   (value   "A new continue object. Since cont+continue is intended to be used as a mixin class,"
		    "normally the creator of the subclass should be used instead."))
  (make-instance 'cont+continue))


(defmacro cont~start (cont-object form)
  (declare (edited  "03-JUN-1994 12:05")
	   (authors RICHTS)
	   (input   "A continue object and a list of forms.")
	   (effect  "The slot of continue points of CONT-OBJECT is initialized and set to NIL."
		    "During the evaluation of FORM, this slot is further modified. After the evaluation the"
		    "result slot of CONT-OBJECT is set to the result of FORM.")
	   (value   "The result of FORM."))
  `(let ((cont*object ,cont-object))
     (declare (special cont*object))
     (cont=end ,cont-object ,form)))


(defun cont~continue (cont-object &key (test #'<))
  (declare (edited  "03-JUN-1994 12:07")
	   (authors RICHTS)
	   (input   "A continue object and a binary predicate.")
	   (effect  "The slot of continue points of CONT-OBJECT is modified. The next continue point is"
		    "seleceted, removed, and evaled like in cont~next. During the evalution the slot is further"
		    "modified. After the evaluation the result slot of CONT-OBJECT is set to the result of this"
		    "evaluation.")
	   (value   "The result of the selected continue point, or NIL if there are no more points."))
  (let ((cont*object cont-object))
     (declare (special cont*object))
     (cont=end cont-object (cont~next :test test))))

(defun cont~next (&key (test #'<))
  (declare (edited  "03-JUN-1994 11:50")
	   (authors RICHTS)
	   (input   "A binary predicate.")
	   (effect  "Then the next continue point is seleceted: If"
		    "there is one with value T, that point that has been added last, is selected"
		    "(LIFO, stack); else if there is a non-nil value, the point with the minimal value"
		    "(according to TEST) is selected; else if there is a point with value NIL, that point"
		    "that has been added first, is selected (FIFO, queue). The selected point is removed from"
		    "the slot of continue points of the current continue object and is then evaled.")
	   (value   "The result of the selected continue point, or NIL if there are no more points.")
	   (special cont*object))
  (let ((cont-points (cont=points cont*object)))
    (cond ((null cont-points)
	   nil)
	  ((or (null (car (first cont-points)))
	       (eq t (car (first cont-points))))
	   (funcall (cdr (pop (cont=points cont*object)))))
	  (t (let ((minimum (cont=min (first cont-points) (rest cont-points) test)))
	       (setf (cont=points cont*object) (delete minimum cont-points))
	       (funcall (cdr minimum)))))))

(defun cont=min (minimum list test)
  (cond ((or (null list) (null (car (first list))))
	 minimum)
	((funcall test (car minimum) (car (first list)))
	 (cont=min minimum (rest list) test))
	(t (cont=min (first list) (rest list) test))))


(defmacro cont~sequence (&body clauses)
  (declare (edited  "09-FEB-1994 1013")
	   (authors RICHTS)
	   (input   "A list of clauses of the form (VALUE EXPRESSION).")
	   (effect  "The values are evaled and the pairs of VALUEs and EXPRESSIONs are added to the current"
		    "list of continue points.")
	   (value   "Undefined."))
  `(cont~forms (list ,@(mapcar #'(lambda (clause) 
			    `(cons ,(first clause)
				   (function (lambda () ,(second clause)))))
			clauses))))

(defun cont~forms (pairs)
  (declare (edited  "03-JUN-1994 11:05")
	   (authors RICHTS)
	   (input   "A list of pairs, consisting of a value and a (nullary) function.")
	   (effect  "The pairs are added to the current list of continue points.")
	   (value   "Undefined.")
	   (special cont*object))
  (let* ((old-points (cont=points cont*object))
	 (t-list (remove-if-not #'(lambda (x) (eq x t)) pairs :key #'car))
	 (nil-list (remove-if-not #'null pairs :key #'car))
	 (val-list  (remove-if #'(lambda (x) (or (null x) (eq x t))) pairs :key #'car))
	 (pos (or (position-if-not #'(lambda (x) (eq x t)) old-points :key #'car) 0)))
    (setf (cont=points cont*object) (nconc t-list
					  (subseq old-points 0 pos)
					  val-list
					  (subseq old-points pos)
					  nil-list))))



(defun cont=end (cont-object result)
  (declare (edited  "03-JUN-1994 11:59")
	   (authors RICHTS)
	   (input   "A continue object and the result of the computation.")
	   (effect  "The result slot of CONT-OBJECT is set to RESULT.")
	   (value   "RESULT."))
  (setf (cont=result cont-object) result)
  result)
  

(defun cont~done-p (cont-object)
  (declare (edited  "01-AUG-1994 12:01")
	   (authors RICHTS)
	   (input   "A continue object.")
	   (effect  "None.")
	   (value   "True iff there are no more continue points in CONT-OBJECT."))
  (null (cont=points cont-object)))



#|
(cont~sequence
 (0 (list 0))
 ((1+ 0) ((wuerg)))
 ((+ 7 9) (format t "hallo")))

(list-all-ways '(6 8 (stair (1 11 a b) c) (7 5 (stair d e) f)))

|#

(defun go-maze (maze)
  (cond ((exit-p maze)
	 maze)
	((fork-p maze)
	 (cont~sequence ((width-left maze) (go-maze (left maze)))
			((width-right maze) (go-maze (right maze))))
	 (cont~next :test #'>))
	((stair-p maze)
	 (cont~sequence (t (go-maze (down maze)))
			(nil (go-maze (up maze))))
	 (cont~next :test #'>))
	((dead-end-p maze)
	 (cont~next :test #'>))))

(defun list-all-ways (maze)
  (let ((cont-object (cont~create)))
    (do* ((way (cont~start cont-object (go-maze maze))
	       (cont~continue cont-object :test #'>))
	  (ways (if way (list way) nil)
		(if way (cons way ways) ways)))
	((cont~done-p cont-object) (nreverse ways))
      (do-something-different))))

(defun do-something-different ()
  nil)

(defun fork-p (maze)
  (and (numberp (first maze))
       (numberp (second maze))))

(defun stair-p (maze)
  (eq (first maze) 'stair))

(defun dead-end-p (maze)
  (null maze))

(defun exit-p (maze)
  (symbolp maze))

(defun width-left (maze)
  (first maze))

(defun width-right (maze)
  (second maze))

(defun left (maze)
  (third maze))

(defun right (maze)
  (fourth maze))

(defun down (maze)
  (second maze))

(defun up (maze)
  (third maze))
      




