;;; -*- syntax: common-lisp; package: keim; base: 10; mode: lisp -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1994 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;


(in-package "KEIM")

(mod~defmod bktrk :uses (mod )
	    :documentation "Backtracking."
	    :exports (bktrk~sequence bktrk~end bktrk~fail-p bktrk~fail)
	    )


(defvar bktrk*sequences nil)

(defvar bktrk*fail (list 'bktrk=fail))

;;; This module provides two macros for a simple backtracking
;;; mechanism. This backtracking mechanism works like that of a
;;; PROLOG-Interpreter. It bases on a list of goals, where always the
;;; first goal is processed. A goal can be replaced by several
;;; subgoals or it can be deleted if it can be solved directly. In
;;; this case the next goal from the goal list is processed. If there
;;; is a choice point, how to solve a goal, the mechanism will chose
;;; the first alternative and try to solve the rest of the goal
;;; list. If the goal list can not be solved with this alternative,
;;; the mechanism will backtrack: it will chose the next alternative
;;; and try to solve the rest of the goal list again.

;;; An example will show how this backtracking mechanism can be used
;;; in LISP.
;;; The PROLOG program
#{\begin{code}
p(X) :- q(X), r(X).
q(a).
q(b).
r(b).
\end{code}#}
;;; in LISP would be
#{\begin{code}
(defun P (var)
   (bktrk~sequence
      (Q var)
      (R var)))

(defun Q (var)
  (or (and (bind var 'a)
	   (prog1 (bktrk~end)
	     (unbind var)))
      (and (bind var 'b)
	   (prog1 (bktrk~end)
	     (unbind var)))))

(defun R (var)
   (and (bind var 'b)
	(prog1 (bktrk~end)
	  (unbind var))))
\end{code}#}
;;; where {\tt (bind var term)} is a function that returns NIL iff a
;;; binding is not possible and {\tt (unbind var)} removes the last
;;; binding made with VAR.

;;; A backtracking programm consists of three major types of
;;; operations: A goal can be replaced by subgoals, it can be deleted,
;;; and there can be alternatives to solve a goal. This operations can
;;; be programmed in LISP as follows: Replacing a goal with some
;;; subgoals is done with {\tt bktrk~sequence}. {\tt bktrk~sequence}
;;; gets some forms as arguments and append these forms in front of
;;; the list of unsolved subgoals. Then it calls the first form and
;;; returns its result. (If this first goal did not fail, all other
;;; subgoals must have been solved, too (see below)).

;;; Deleting a goal that can be solved directly and processing the
;;; next goal is done with {\tt bktrk~end}. If there are no more
;;; unsolved subgoals {\tt bktrk~end} returns T; else it evaluates the
;;; next form from the list of unsolved subgoals and returns its
;;; result. Since all subgoals that do not fail must call {\tt
;;; bktrk~end}, all unsolved subgoals have been solved if {\tt
;;; bktrk~end} returns a non-NIL value.

;;; The choice points in the LISP-program are any kind of
;;; OR-conections with {\tt or}, {\tt some}, or other similar
;;; constructions. Note that the backtracking mechanism allows to
;;; return more meaningful values than NIL and true. The only
;;; restriction is that NIL means fail and non-NIL means success.

;;; It is very important that every positive branch ends with a {\tt
;;; bktrk~end}, i.e. every expression in a {\tt bktrk~\-sequence}
;;; returns NIL or calls {\tt bktrk~end}.

(defun bktrk=forms (&rest forms)
  (mapcar #'(lambda (x) 
	      `(function (lambda () ,x)))
	  forms))

(defmacro bktrk~sequence (&body forms)
  (declare (edited  "09-FEB-1994 19:13")
	   (authors RICHTS)
	   (input   "A list of LISP expressions.")
	   (effect  "The first expression is evaluated and the rest is appended in front"
		    "of the goal list. A warning is signaled if one"
		    "form returns a non-NIL value while there are still expressions in the"
		    "goal list. (This happens if a goal that returns a non-NIL value, does"
		    "not end with BKTRK~END.)")
	   (value   "The result of the first expression."))
  (let ((bktrkforms (apply #'bktrk=forms (rest forms))))
  `(let ((old-sequence bktrk*sequences)
	 result-out)
     (unwind-protect
	 (progn
	   (setf bktrk*sequences 
		 (append (list ,@bktrkforms)
			 bktrk*sequences))
	   (let ((result ,(first forms)))
	     (setf result-out result)
	     (when (and result bktrk*sequences)
	       (warn "Backtrace: You forgot to set an ending point for backtracking. ~
                      Positive Branches have to end with (bktrk~~end t).")
	       (setf bktrk*sequences nil))
	     result))
       (unless result-out
	 ;(format t "~% Resetting bktrk*sequences to ~A." old-sequence)
	 (setf bktrk*sequences old-sequence))))))



(defmacro bktrk~end ()
  (declare (edited  "09-FEB-1994 19:16")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "If there are still expressions in the goal list, the next one is"
		    "popped from the goal list and is evaluated. If the result is NIL,"
		    "the expression is pushed onto the goal list again")
	   (value   "If the goal list is empty, T;"
		    "else the result of the evaluation of the popped goal."))
  `(if (null bktrk*sequences)  
       t
     (let* ((next-step (pop bktrk*sequences))
	    (result (funcall next-step)))
       (cond (result result)
	     (t (push next-step bktrk*sequences)
		nil)))))




#|
(defun P (var)
   (bktrk~sequence
      (Q var)
      (R var)))

(defun Q (var)
  (or (and (bind var 'a)
	   (prog1 (bktrk~end)
	     (unbind var)))
      (and (bind var 'b)
	   (prog1 (bktrk~end)
	     (unbind var)))))

(defun R (var)
   (and (bind var 'b)
	(prog1 (bktrk~end)
	  (unbind var))))

(defun bind (var const)
  (let ((binding (get var :binding)))
    (when (null binding)
      (setf (get var :binding) const))
    (if (or (eq binding const) (null binding))
	(list (list var const))
      nil)))

(defun unbind (var)
  (setf (get var :binding) nil))
|#



