;;; -*- syntax: common-lisp; package: keim; base: 10; mode: lisp -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;


(in-package "KEIM")

(mod~defmod scan :uses (mod )
	    :documentation "The scanner used for the transformation of ADLER objects into POST objects."
	    :exports (
		      scan+token
		      scan+terminal
		      scan+name
		      scan+number
		      scan~terminal-p
		      scan~name-p
		      scan~number-p
		      scan~token-value
		      scan~init
		      scan~current-token
		      scan~next-token
		      scan~error-msg
		      ))

#{\section{Scanner}
This module is used in scanning input files that are not in lisp format
for tokens.
#}

; Token:   
; terminal (reserved words, special characters)
; name     (identifiers)
; number   (cardinals)

; Classes

(defclass scan+token ()
  ((value :initarg :value :reader scan=token-value)
   (row :initarg :row :reader scan=token-row)
   (col :initarg :col :reader scan=token-col)
   (length :initarg :length :reader scan=token-length))
  (:documentation "The superclass of all token classes."))


(defclass scan+terminal (scan+token) ()
  (:documentation "This class contains all terminal tokens, i.e. reserved words and special characters."))

(defclass scan+name (scan+token) ()
  (:documentation "This class contains all identifier tokens, i.e. names of variables, constants, functions etc."))

(defclass scan+number (scan+token) ()
  (:documentation "This class contaiins all number tokens."))


; Creators

(defun scan=terminal-create (value row col length)
  (make-instance 'scan+terminal :value value :row row :col col :length length))

(defun scan=name-create (value row col length)
  (make-instance 'scan+name :value value :row row :col col :length length))

(defun scan=number-create (value row col length)
  (make-instance 'scan+number :value value :row row :col col :length length))


; Prdicates

(defun scan~terminal-p (object)
  (declare (edited  "15-OCT-1993 15:05")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "T iff OBJECT is a terminal token."))
  (typep object 'scan+terminal))

(defun scan~name-p (object)
  (declare (edited  "15-OCT-1993 15:05")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "T iff OBJECT is a name token."))
  (typep object 'scan+name))

(defun scan~number-p (object)
  (declare (edited  "15-OCT-1993 15:05")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "T iff OBJECT is a number token."))
  (typep object 'scan+number))


; Interface selectors

(defun scan~token-value (token)
  (declare (edited  "15-OCT-1993 15:07")
	   (authors RICHTS)
	   (input   "A token.")
	   (effect  "None.")
	   (value   "The string represented by TOKEN."))
  (scan=token-value token))



; Scanning
; Setup

(defvar scan*lines)
(defvar scan*current-line)
(defvar scan*row)
(defvar scan*col)
(defvar scan*current-token)

(defun scan~init (stream)
  (declare (edited  "15-OCT-1993 15:10")
	   (authors RICHTS)
	   (input   "An open stream.")
	   (effect  "The lines are read in from STREAM, the scanner is initiallized and"
		    "the first token is scanned, i.e. it becomes the current token.")
	   (value   "Undefined."))
  (do ((line (read-line stream nil :eof) (read-line stream nil :eof))
       (lines nil (cons line lines)))
      ((eq line :eof) (setf scan*lines (nreverse lines))))
  (setf scan*row -1)
  (scan=next-line)
  (scan~next-token))

;; "[" "]"
(defvar scan*reserved-words
  '("begin_problem" "end_problem" "formula" "end_of_list" "list_of_theorems"
    "list_of_axioms" "forall" "exists" "implies" "implied" "equiv" "bin_op" "not" "and"
    "or" "un_op" "zero_op" "true" "false" "equal" "eof" )
  "This the list of the reserved words. If the scanner scanns one of this words,
   it creates a terminal token instead of a name token.")

#| "problem" "is"
    "variables" "constants" "type-variables" "type-constants"
    "assumption" "conclusion"
    "sort" "lam" "lambda" "in"
    "end"
    "not" "and" "or" "equiv" "implies"
     "lam"
    "in" "subset"
    "union" "intersection"|#



; terminals
;; 
(defvar scan*special-chars
  '(#\, #\- #\. #\( #\) #\[ #\])
  "This a list of special characters. If the scanner scans one of these while scanning a word, the word ends.
   If the scanner starts scanning the next token and gets one of these as the current character, it creates a
   terminal token with a string containing this character.")

(defvar scan*blank-chars
  '(#\space #\tab)
  "This a list of all blank characters. If the scanner scans one of these while scanning a word, the word ends.
   When the scanner starts scanning the next token, these characters are zapped until a non-blank characters appears.")

; Scanning lines and characters

(defun scan=next-line ()
  (declare (edited  "15-OCT-1993 15:37")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "The current position of the input stream is set to the beginning of the next line."
		    "If there is no more line, the current line becomes 'eof'.")
	   (value   "Undefined."))
  (incf scan*row)
  (if (< scan*row (length scan*lines))
      (setf scan*current-line (elt scan*lines scan*row))
      (setf scan*current-line "eof"))
  (setf scan*col 0))

(defun scan=next-char ()
  (declare (edited  "15-OCT-1993 15:45")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "The current position of the input stream is increased by one."
		    "If the end of the line is reached a 'space' becomes the current character"
		    "and the next call of SCAN=NEXT-CHAR will set current position to the next line.")
	   (value   "Undefined."))
  (if (>= scan*col (length scan*current-line))
      (scan=next-line)      ;#\space
      (incf scan*col)))     ;(prog1 (elt scan*current-line scan*col) (incf scan*col))

(defun scan=current-char ()
  (declare (edited  "15-OCT-1993 15:59")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "None.")
	   (value   "The first character of the input stream that was not scanned yet."
		    "(At the end of a line this will be a 'space' instead of a 'newline'.)"))
  (if (>= scan*col (length scan*current-line))
      #\space
      (elt scan*current-line scan*col)))


; scanning tokens

(defun scan~current-token ()
  (declare (edited  "15-OCT-1993 16:09")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "None.")
	   (value   "The current token, i.e. the last one that was scanned."))
  scan*current-token)

(defun scan~next-token ()
  (declare (edited  "15-OCT-1993 16:03")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "The next token is scanned and the current token is set to it."
		    "The position of the input stream is set to the end of this token.")
	   (value   "The scanned token which is now the current token."))
  (let ((char (scan=current-char)))
    (setf scan*current-token
	  (cond ((digit-char-p char) (scan=number))
		((member char scan*blank-chars) (scan=spaces) (scan~next-token))
		((member char scan*special-chars) (scan=special))
		(t (scan=word))))))

(defun scan=spaces ()
  (when (member (scan=current-char) scan*blank-chars)
    (scan=next-char)
    (scan=spaces)))

(defun scan=word ()
  (let ((start scan*col))
    (do ((char (scan=current-char) (scan=current-char)))
	((or (member char scan*special-chars)
	     (member char scan*blank-chars)))
      (scan=next-char))
    (let ((word (subseq scan*current-line start scan*col)))
      (if (member word scan*reserved-words :test #'string-equal)
	  (scan=terminal-create word scan*row start (- scan*col start))
	  (scan=name-create (read-from-string word) scan*row start (- scan*col start))))))

(defun scan=number ()
  (let ((start scan*col))
    (do ((char (scan=current-char) (scan=current-char)))
	((not (digit-char-p char)))
      (scan=next-char))
    (let ((word (format nil "\|~A\|" (subseq scan*current-line start scan*col))))
      (scan=name-create (read-from-string word) scan*row start (- scan*col start)))))
	  
(defun scan=special ()
  (let ((start scan*col))
    (scan=next-char)
    (scan=terminal-create (subseq scan*current-line start scan*col) scan*row start (- scan*col start))))



; Parsing 


; Errors

(defun scan=error-msg (string col row)
  (error "~A" (format nil "~%~A~%~A^~%~%Syntax error at line ~A (~A): ~A"
		      scan*current-line (make-sequence 'string col :initial-element #\space)
		      row col string)))

(defun scan~error-msg (string &optional (token scan*current-token))
  (declare (edited  "15-OCT-1993 16:12")
	   (authors RICHTS)
	   (input   "A string and optinal a token which defaults to the current token.")
	   (effect  "An error is signaled showing the position of token in the input stream.")
	   (value   "None."))
  (scan=error-msg string (scan=token-col token) (scan=token-row token)))


