

(in-package "KEIM")

;;;;; An example for cbktrk: A simple SLD prover

;;; The polarity of the literals is NOT taken into account.  Only the position
;;; matters: The first literal of a rule is the head (corresponding to positive
;;; polarity) and the rest is the tail (corresponding to negative polarity).


(defun sld~init (goal-clause clauses)
  (declare (edited  "07-APR-1995")
           (authors Richts)
           (input   "A clause and a list of (Horn-) clauses.")
           (effect  "None.")
           (value   "An object that stores the state of the SLD-resolution"
                    "with GOAL and CLAUSES.  The next solution can be"
                    "computed with sld~next-solution.")
           (remark  "The function sld~value defines the search strategy (see there)."))
  (let ((new-clauses (top~rename clauses))
        (variables (top~free-variables goal-clause)))
    (cbktrk~create-cont-object
     (cbktrk~and-fns (mapcar #'(lambda (literal)
                                 #'(lambda ()
                                     (sld=solve-goal literal new-clauses variables)))
                             (cl~literals goal-clause)))
     :result-function #'(lambda (dummy)
                          (sld=result-function variables)))))

(defun sld~next-solution (sld-cont-object)
  (declare (edited  "07-APR-1995")
           (authors Richts)
           (input   "An object that was returned by sld~init.")
           (effect  "The next solution is computed and the state in"
                    "SLD-CONT-OBJECT is updated.")
           (value   "The next solution, i.e. a substitution for the"
                    "variables of the initial goal clause."
                    "NIL if there are no more solutions."))
  (uni~with-permanent-bindings (cbktrk~next-solution sld-cont-object)))


(defun sld=solve-goal (goal clauses variables)
  (declare (edited  "07-APR-1995")
           (authors Richts)
           (input   "A literal, a list of clauses and a list of variables.")
           (effect  "All alternatives are set up where GOAL is resolved"
                    "with the head of a clause from CLAUSES.  Then control"
                    "is returned to the central search loop."
                    ""
                    "(Some variable bindings will be set when these"
                    "alternatives will be executed.)")
           (value   "Since this function calls cbktrk~or-fns the value will"
                    "always be NIL.  Thereby control is returned to the"
                    "central search loop, which selects the next choice"
                    "point.")
           (remark  "In order to restore the actual state of variable"
                    "bindings when the alternatives will be executed, we have to"
                    "compute the bindings of VARIABLES and reset them at"
                    "the beginning of each alternative."))
  (let ((bindings (mapcar #'(lambda (var)
                              (if (sym~symbol-binding var)
                                  (top~insert-bindings! var)))
                          variables)))
    (cbktrk~or-fns (mapcar #'(lambda (clause)
                               (cons (sld~value goal clause)
                                     #'(lambda ()
                                         (top~unbind-binding-list!)
                                         (top~bind-terms! variables bindings)
                                         (sld=resolve goal clause clauses variables))))
                           clauses))))

(defun sld=resolve (goal clause clauses variables)
  (declare (edited  "07-APR-1995")
           (authors Richts)
           (input   "A literal, a clause, a list of clauses, and a list of variables.")
           (effect  "GOAL is unified with the head of CLAUSE and there by"
                    "the appropriate variable bindings are set.  If the"
                    "unification was successfull, we try to solve all"
                    "literals in the tail of CLAUSE with CLAUSES.")
           (value   "T if GOAL and the new subgoals can be solved without"
                    "encountering a choice point.  If GOAL cannot be solved"
                    "(for example if the unification fails) or if a choice point is"
                    "encountered, NIL is returned (in the latter case the"
                    "alternatives are returned via a dynamic bound"
                    "variable.")
           (remark  "Since some variables in GOAL might be bound to"
                    "variables of CLAUSE, CLAUSE is replaced in CLAUSES with"
                    "a renamed version of CLAUSE."))
  (let ((head (first (cl~literals clause))))
    (if (uni~unify (lit~atom goal) (lit~atom head))
        (let ((new-clauses (substitute (top~rename clause) clause clauses))
              (new-variables (append (top~free-variables clause) variables)))
          (cbktrk~and-fns (mapcar #'(lambda (literal)
                                      #'(lambda ()
                                          (sld=solve-goal literal new-clauses new-variables)))
                                  (rest (cl~literals clause))))))))

(defun sld=result-function (variables)
  (declare (edited  "07-APR-1995")
           (authors Richts)
           (input   "A list of variables.")
           (effect  "None.")
           (value   "A substitution mapping VARIABLES to there actual bindings.")
           (remark  "This function is called by the central search loop if"
                    "a solution is found."))
  (subst~create variables
                (top~insert-bindings! variables)))

(defun sld~value (goal clause)
  (declare (edited  "07-APR-1995")
           (authors Richts)
           (input   "A goal literal and the clause this literal will be resolved with.")
           (effect  "None.")
           (value   "This function determines the search strategy of the"
                    "SLD-resolution.  If it returns T, depth-first-search is"
                    "performed (which is incomplete like in Prolog); if it"
                    "returns NIL breadth-search is"
                    "performed.  If this function returns another"
                    "value, even heuristic search strategies are"
                    "possible (see the cbktrk module for details)."))
  t)


#|

; The following problem shows the incompleteness of a
; Prolog search strategy.  If sld~value returns T, only
; solutions (f a), (f (f a)) etc. are returned.  If it
; returns NIL all solutions (f a), (g a), (f (f a)), (g
; (g a)) etc. are returned by consecutive calls of
; sld~next solution.
;
; goal:   :- p(X).
;
; rules:  p(f(a)).
;         p(f(f(X))) :- p(f(X)).
;         p(g(b)).
;         p(g(g(X))) :- p(g(X)).
;

(setf env (env~create))

(post~read-object-list '((type-constants i o)
                         (constants (P (o i)) (Q (o i))
                                    (f (i i)) (g (i i))
                                    (a i) (b i)))
                       env)

(setf goal (cl~read '(clause goal ((x i)) (- (P x))) env))

(setf clauses (list (cl~read '(clause c1 ((x i)) (+ (P (f a)))) env)
                    (cl~read '(clause c2 ((x i)) (+ (P (f (f x)))) (- (P (f x)))) env)
                    (cl~read '(clause c1 ((x i)) (+ (P (g b)))) env)
                    (cl~read '(clause c3 ((x i)) (+ (P (g (g x)))) (- (P (g x)))) env)))

(setf cont-object (sld~init goal clauses))

(sld~next-solution cont-object)



; The following examples is a little program which
; implements the quick-sort algorithm.


(setf env (env~create))

(post~read-object-list '((type-constants i o)
                         (constants (app (o i i i)) (cons i i i) (nil i)
                                    (even (o i)) (s (i i)) (null i)
                                    (less (o i i)) (lesseq (o i i))
                                    (sort (o i i)) (split (o i i i i))
                                    (P (o i))
                                    (a i) (b i) (c i) (d i) (e i)))
                       env)

(setf clauses
      (list (cl~read '(clause c1 ((l i)) (+ (app nil l l))) env)
            (cl~read '(clause c2 ((x i) (l1 i) (l2 i) (res i))
                              (+ (app (cons x l1) l2 (cons x res)))
                              (- (app l1 l2 res))) env)
            
            (cl~read '(clause c3 () (+ (even null))) env)
            (cl~read '(clause c4 ((x i))
                              (+ (even (s (s x))))
                              (+ (even x))) env)

            (cl~read '(clause c5a ((x i)) (+ (lesseq x x))) env)
            (cl~read '(clause c6a ((x i) (y i))
                              (+ (lesseq x y))
                              (- (less x y))) env)
            
            (cl~read '(clause c5 ((x i)) (+ (less null (s x)))) env)
            (cl~read '(clause c6 ((x i) (y i))
                              (+ (less (s x) (s y)))
                              (- (less x y))) env)

            (cl~read '(clause c7 () (+ (sort nil nil))) env)
            (cl~read '(clause c8 ((x i) (l i) (l1 i) (l2 i) (ls1 i) (ls2 i) (res i))
                              (+ (sort (cons x l) res))
                              (- (split x l l1 l2))
                              (- (sort l1 ls1))
                              (- (sort l2 ls2))
                              (- (app ls1 (cons x ls2) res))) env)

            (cl~read '(clause c9 () (+ (split x nil nil nil))) env)
            (cl~read '(clause c10 ((x i) (y i) (l i) (l1 i) (l2 i))
                              (+ (split x (cons y l) (cons y l1) l2))
                              (- (less y x))
                              (- (split x l l1 l2))) env)
            (cl~read '(clause c11 ((x i) (y i) (l i) (l1 i) (l2 i))
                              (+ (split x (cons y l) l1 (cons y l2)))
                              (- (lesseq x y))
                              (- (split x l l1 l2))) env)
            ))


(setf goal (cl~read '(clause goal ((res i)) (- (app (cons a (cons b nil))
                                                    (cons c (cons d nil))
                                                    res))) env))

(setf goal (cl~read '(clause goal ((l1 i) (l2 i))
                             (- (app l1 l2 (cons a (cons b (cons c nil))))))
                    env))

(setf goal (cl~read '(clause goal ((x i) (l1 i) (l2 i))
                             (- (app l1 (cons x l2)
                                     (cons null
                                           (cons (s null)
                                                 (cons null
                                                       (cons (s (s null)) nil))))))
                             (- (even x)))
                    env))

(setf goal
      (cl~read '(clause goal ((res i))
                        (- (sort (cons   (s null)
                                   (cons (s (s (s null)))
                                   (cons (s (s null))
                                   (cons null
                                   (cons (s (s (s (s null))))
                                   nil)))))
                                 res)))
               env))


(setf cont-object (sld~init goal clauses))

(sld~next-solution cont-object)

|#
