;;; Associative Unification with constant restrictions

(in-package :KEIM)

(defvar a*maxdepth 10
  "Depth bound for the maximal length of solutions")

#|
 Some interesting test cases for A-unification

 1) axbzx =?= zczyyy
    x = cabaca, y = baca, z = a

 2) xxayby =?= cayvabd
    x = caabd, y = abd, v = caabdaabdb

 3) byxay =?= xxyaba
    no solution

 4) cxyzzbxcy =?= zbaxzcybayzba
    x = abaaba, y = aba, z = ca

 5) xxaybzyzbxz =?= zyzbxabxzabzy
    x = aab, y = aba, z = a

 6) zyxabxyz =?= ttcyx
    x = c, y = ab, z = c, t = cab

 7) xauby =?= cayvxac
    x = c, y = cac, u = cac, v = b

 8) axbyx =?= ytxt
    x = ba, y = a, t = ba

|#

(defgeneric a=term-to-word (object)
  (declare (edited  "3-August-1995")
	   (authors Kepser)
	   (input   "A Keim Term or list thereof.")
	   (effect  "None.")
	   (value   "If the term is an associative term, the word"
		    "representing the term. Else an error is"
		    "signaled."))

  (:method ((term sym+var))
	   (list term))
  (:method ((term sym+const))
	   (list term))
  (:method ((term appl+appl))
	    (mapcan #'a=term-to-word
		    (appl~arguments term)))
  (:method ((object-list list))
	   (mapcar #'a=term-to-word
		   object-list))) 



(defun a=constant-p (term vars-as-consts)
  (declare (edited  "7-August-1995")
	   (authors Kepser)
	   (input   "An element (Keim Term) and a list of KEIM"
		    "variables treated as constants.")
	   (effect  "None.")
	   (value   "Non-NIL, iff term is a constant for a-unif."))
  (or (sym~constant-p term)
      (member term vars-as-consts :test #'keim~equal)))




(defmacro a=remove (elem lst)
  (declare (edited  "11-August-1995")
	   (authors Kepser)
	   (input   "A Keim Term and an a-list thereof.")
	   (effect  "None.")
	   (value   "The list with the element removed."))
  ;; a=remove is just a shorthand for its expansion
  ;; it's needed too often to have the full expansion apear in the source
  `(remove ,elem ,lst :test #'keim~equal
	              :key  #'first))





(defun a=symmetric-difference (letters word1 word2 vars-as-consts
					&optional done vars1 consts1
					vars2 consts2)
  (declare (edited  "11-August-1995")
	   (authors Kepser)
	   (input   "letters: a list of letters from word1 and word2; "
		    "word1, word2: two words (list of Keim terms) to be "
		    "unified; vars-as-consts: a list of Keim variables "
		    "that are treated as constants.")
	   (effect  "None.")
	   (value   "Four a-lists are returned containing the symmetric "
		    "difference of vars and consts in word1 and word2. "))
  ;; value 1: for each var that occurs more often in word1 than in word2 the
  ;;          difference of occurences;
  ;; value 2: for each const that occurs more often in word1 than in word2
  ;;          the difference of occurences;
  ;; value 3: for each var that occurs more often in word2 than in word1 the
  ;;          difference of occurences;
  ;; value 4: for each const that occurs more often in word2 than in word1
  ;;          the difference of occurences
  (cond ((endp letters)
	 (values vars1 consts1 vars2 consts2))
	((member (first letters) done :test #'term~equal)
	 (a=symmetric-difference
	  (rest letters) word1 word2 vars-as-consts done vars1 consts1
	  vars2 consts2))
	(T
	 (let* ((l (first letters))
		(l-w  (- (count l word1 :test #'term~equal)
			 (count l word2 :test #'term~equal))))
	   (cond ((zerop l-w)
		  (a=symmetric-difference
		   (rest letters) word1 word2 vars-as-consts
		   (cons l done) vars1 consts1 vars2 consts2))
		 ((and (plusp l-w)
		       (a=constant-p l vars-as-consts))
		  (a=symmetric-difference
		   (rest letters) word1 word2 vars-as-consts
		   (cons l done) vars1 (acons l l-w consts1)
		   vars2 consts2))
		 ((plusp l-w)
		  (a=symmetric-difference
		   (rest letters) word1 word2 vars-as-consts
		   (cons l done) (acons l l-w vars1) consts1
		   vars2 consts2))
		 ((a=constant-p l vars-as-consts)
		  (a=symmetric-difference
		   (rest letters) word1 word2 vars-as-consts
		   (cons l done) vars1 consts1 vars2
		   (acons l (- l-w) consts2)))
		 (T
		  (a=symmetric-difference
		   (rest letters) word1 word2 vars-as-consts
		   (cons l done) vars1 consts1
		   (acons l (- l-w) vars2) consts2)))))))



(defun a=new-sym-diff-v (var sub newvar vars1 consts1 vars2
			   consts2)
  (declare (edited  "10-August-1995")
	   (authors Kepser)
	   (input   "var: a variable; "
		    "sub: a variable to substitute var; "
		    "newvar: possibly a new variable or NIL; "
		    "vars1, consts1, vars2, consts2: a-lists of symmetric "
		    "difference for vars and consts.")
	   (effect  "None.")
	   (value   "The new symmetric difference after replacing var "
		    "by sub."))
  (let ((v1 (rest (assoc var vars1 :test #'keim~equal)))
	(v2 (rest (assoc var vars2 :test #'keim~equal))))
    (cond ((eq v1 v2)  ; v1 = v2 = NIL
	   (values vars1 consts1 vars2 consts2))
	  
	  (v1  
	   (let ((c1 (rest (assoc sub vars1 :test #'keim~equal)))
		 (c2 (rest (assoc sub vars2 :test #'keim~equal)))
		 (n-vs1 (if newvar
			    (acons newvar v1 (a=remove var vars1))
			  (a=remove var vars1))))
	     (cond (c1  ; > 0
		    (values (acons sub (+ v1 c1) n-vs1)
			    consts1 vars2 consts2))
		   ((and c2 (> c2 v1))
		    (values n-vs1 consts1
			    (acons sub (- c2 v1) vars2)
			    consts2))
		   ((and c2 (> v1 c2))
		    (values (acons sub (- v1 c2) n-vs1)
			    consts1 (a=remove sub vars2)
			    consts2))
		   (c2  ; and v1 = c2
		    (values n-vs1 consts1
			    (a=remove sub vars2) consts2))
		   (T ; c1 = c2 = NIL
		    (values (acons sub v1 n-vs1)
			    consts1 vars2 consts2)))))

	  (v2  
	   (let ((c1 (rest (assoc sub vars1 :test #'keim~equal)))
		 (c2 (rest (assoc sub vars2 :test #'keim~equal)))
		 (n-vs2 (if newvar
			    (acons newvar v2 (a=remove var vars2))
			  (a=remove var vars2))))
	     (cond (c2  ; > 0
		    (values vars1 consts1
			    (acons sub (+ v2 c2) n-vs2)
			    consts2))
		   ((and c1 (> c1 v2))
		    (values (acons sub (- c1 v2) vars1)
			    consts1 n-vs2 consts2))
		   ((and c1 (> v2 c1))
		    (values (a=remove sub vars1) consts1
			    (acons sub (- v2 c1) n-vs2)
			    consts2))
		   (c1  ; and v2 = c1
		    (values (a=remove sub vars1) consts1
			    n-vs2 consts2))
		   (T ; c1 = c2 = NIL
		    (values vars1 consts1
			    (acons sub v2 n-vs2) consts2))))))))



(defun a=new-sym-diff-c (var sub newvar vars1 consts1 vars2
			     consts2)
  (declare (edited  "10-August-1995")
	   (authors Kepser)
  	   (input   "var: a variable; "
		    "sub: a constant to substitute var; "
		    "newvar: possibly a new variable or NIL; "
		    "vars1, consts1, vars2, consts2: a-lists of symmetric "
		    "difference for vars and consts.")
	   (effect  "None.")
	   (value   "The new symmetric difference after replacing var "
		    "by sub."))
  (let ((v1 (rest (assoc var vars1 :test #'keim~equal)))
	(v2 (rest (assoc var vars2 :test #'keim~equal))))
    (cond ((eq v1 v2)  ; v1 = v2 = NIL
	   (values vars1 consts1 vars2 consts2))
	  (v1
	   (let ((c1 (rest (assoc sub consts1 :test #'keim~equal)))
		 (c2 (rest (assoc sub consts2 :test #'keim~equal)))
		 (n-vs1 (if newvar
			    (acons newvar v1 (a=remove var vars1))
			  (a=remove var vars1))))
	     (cond (c1  ; > 0
		    (values n-vs1 (acons sub (+ c1 v1) consts1)
			    vars2 consts2))
		   ((and c2  (> c2 v1))
		    (values n-vs1 consts1 vars2
			    (acons sub (- c2 v1) consts2)))
		   ((and c2 (> v1 c2))
		    (values n-vs1 (acons sub (- v1 c2) consts1)
			    vars2 (a=remove sub consts2)))
		   (c2  ; and v1 = c2
		    (values n-vs1 consts1 vars2
			    (a=remove sub consts2)))
		   (T ; c1 = c2 = NIL
		    (values n-vs1 (acons sub v1 consts1)
			    vars2 consts2)))))
	  
	  (v2
	   (let ((c1 (rest (assoc sub consts1 :test #'keim~equal)))
		 (c2 (rest (assoc sub consts2 :test #'keim~equal)))
		 (n-vs2 (if newvar
			    (acons newvar v2 (a=remove var vars2))
			  (a=remove var vars2))))
	     (cond (c2  ; > 0
		    (values vars1 consts1 n-vs2
			    (acons sub (+ c2 v2) consts2)))
		   ((and c1 (> c1 v2))
		    (values vars1 (acons sub (- c1 v2) consts1)
			    n-vs2 consts2))
		   ((and c1 (> v2 c1))
		    (values vars1 (a=remove sub consts1)
			    n-vs2 (acons sub (- v2 c1) consts2)))
		   (c1  ; and v2 = c1
		    (values vars1 (a=remove sub consts1)
			    n-vs2 consts2))
		   (T ; c1 = c2 = NIL
		    (values vars1 consts1 n-vs2
			    (acons sub v2 consts2)))))))))




(defun a=solvable (vars1 consts1 vars2 consts2)
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "Symmetric difference.")
	   (effect  "None.")
	   (value   "Non-NIL, iff a-unification with this symmetric "
		    "difference still solvable."))
  (or (and vars1 vars2)
      (and (consp consts1) (endp vars1) (endp consts2) (consp vars2))
      (and (endp consts1) (consp vars1) (consp consts2) (endp vars2))
      (and (endp consts1) (endp consts2) (endp vars1) (endp vars2))))



(defun a=add-unif (var subst unifier)
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "A variable, a substitution and a unifier (a-list).")
	   (effect  "None.")
	   (value   "The new unifier."))
  (acons var (cons subst (rest (assoc var unifier :test
				      #'term~equal)))
	 (a=remove var unifier)))



(defun a=add-2unif (var subst1 subst2 unifier)
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "A variable, two segments of a substitution and "
		    "a unifier (a-list).")
	   (effect  "None.")
	   (value   "The new unifier."))
  (acons var (cons subst2
		   (cons subst1 (rest (assoc var unifier :test
					     #'term~equal))))
	 (a=remove var unifier)))



(defun a=substitute (new1 new2 old l)
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "new1, new2: two symbols to replace for old; "
		    "old: a symbol to be replaced in l; "
		    "l: a list of Keim symbols.")
	   (effect  "None.")
	   (value   "A copy of l with each occurence of old replaced by both "
		    "new1 and new2."))
  (cond ((endp l)
	 NIL)
	((term~equal old (first l))
	 (cons new1
	       (cons new2
		     (a=substitute new1 new2 old (rest l)))))
	(T
	 (cons (first l)
	       (a=substitute new1 new2 old (rest l))))))



(defun a=extend-restr (v1 v2 v3 restr)
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "v1, v2, v3: Keim vars; "
		    "restr: constant restrictions (a-list).")
	   (effect  "None.")
	   (value   "New constant restrictions, each restriction that "
		    "contains v1 now also contains v2 and v3, where v3 "
		    "mustn't be present before."))
  (mapcar #'(lambda (r)
	      (if (member v1 r :test #'term~equal)
		  (cons (first r)
			(cons v3
			      (adjoin v2 (rest r))))
		r))
	  restr))



;;;--- decide unifiability


;;; the next three functions are used to do the real work in unifying two
;;; words (decision procedure).
;;; we use a method attributed to G.D. Plotkin with a depth bound


(defun a=unif-var-const-d (var const word1 word2 vars-as-consts restr
			     vars1 consts1 vars2 consts2 unifier
			     depth words1 words2 old-unif auxvars)
  (declare (edited  "13-August-1995")
	   (authors Kepser)
	   (input   "var: a Keim variable; const: a Keim symbol; "
		    "word1, word2: rest words to be unified; "
		    "vars-as-consts: list of Keim vars to be treated as "
		    "consts; restr: linear constant restrictions; "
		    "vars1, consts1, vars2, consts2: symmetric difference; "
		    "unifier: unifier for the a-unification of the two words"
		    " constructed so far; depth: the current depth of the "
		    "computation; "
		    "words1, words2: lists of words still to be unified; "
		    "old-unif: unifier of previous words already unified; "
		    "auxvars: list of auxiliary vars introduced in the "
		    "current unification.")
	   (effect  "None.")
	   (value   "Result of the whole a-unification."))
  (or ; var --> const
      (multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	  (a=new-sym-diff-c var const NIL vars1 consts1 vars2 consts2)
	(if (a=solvable n-v1 n-c1 n-v2 n-c2)
	    (a=unif-d
	     (substitute const var word1 :test #'keim~equal)
	     (substitute const var word2 :test #'keim~equal)
	     vars-as-consts restr n-v1 n-c1 n-v2 n-c2
	     (a=add-unif var const unifier) (+ depth 1)
	     words1 words2 old-unif auxvars)))
      ; var --> const var
      (let ((newvar (fo~variable-create (gentemp "X-"))))
	(multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	    (a=new-sym-diff-c var const newvar vars1 consts1 vars2 consts2)
	  (if (a=solvable n-v1 n-c1 n-v2 n-c2)
	      (a=unif-d
	       (cons newvar (a=substitute const newvar var word1))
	       (a=substitute const newvar var word2)
	       vars-as-consts restr n-v1 n-c1 n-v2 n-c2
	       (a=add-2unif var const newvar unifier) (+ depth 1)
	       words1 words2 old-unif (cons newvar auxvars)))))))



(defun a=unif-var-var-d (v1 v2 w1 w2 vars-as-consts restr vars1 consts1
			  vars2 consts2 unifier depth words1 words2
			  old-unif auxvars)
  (declare (edited  "13-August-1995")
	   (authors Kepser)
	   (input   "v1, v2: two Keim variables to be unified; "
		    "w1, w2: rest words to be unified; "
		    "vars-as-consts: list of Keim vars to be treated as "
		    "consts; restr: linear constant restrictions; "
		    "vars1, consts1, vars2, consts2: symmetric difference; "
		    "unifier: unifier for the a-unification of the two words"
		    " constructed so far; depth: the current depth of the "
		    "computation; "
		    "words1, words2: lists of words still to be unified; "
		    "old-unif: unifier of previous words already unified; "
		    "auxvars: list of auxiliary vars introduced in the "
		    "current unification.")
	   (effect  "None.")
	   (value   "Result of the whole a-unification."))
  (or ; v1 --> v2
      (multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	  (a=new-sym-diff-v v1 v2 NIL vars1 consts1 vars2 consts2)
	(if (a=solvable n-v1 n-c1 n-v2 n-c2)
	    (a=unif-d
	     (substitute v2 v1 w1 :test #'keim~equal)
	     (substitute v2 v1 w2 :test #'keim~equal)
	     vars-as-consts (subst v2 v1 restr :test #'keim~equal)
	     n-v1 n-c1 n-v2 n-c2 (a=add-unif v1 v2 unifier)
	     (+ depth 1) words1 words2 old-unif auxvars)))
      ; v1 --> v2 v1
      (let ((newvar (fo~variable-create (gentemp "X-"))))
	(multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	    (a=new-sym-diff-v v1 v2 newvar vars1 consts1 vars2 consts2)
	  (if (a=solvable n-v1 n-c1 n-v2 n-c2)
	      (a=unif-d
	       (cons newvar (a=substitute v2 newvar v1 w1))
	       (a=substitute v2 newvar v1 w2)
	       vars-as-consts (a=extend-restr v1 v2 newvar restr)
	       n-v1 n-c1 n-v2 n-c2
	       (a=add-2unif v1 v2 newvar unifier) (+ depth 1)
	       words1 words2 old-unif (cons newvar auxvars)))))
      ; v2 --> v1 v2
      (let ((newvar (fo~variable-create (gentemp "X-"))))
	(multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	    (a=new-sym-diff-v v2 v1 newvar vars2 consts2 vars1 consts1)
	  (if (a=solvable n-v1 n-c1 n-v2 n-c2)
	      (a=unif-d
	       (cons newvar (a=substitute v1 newvar v2 w2))
	       (a=substitute v1 newvar v2 w1)
	       vars-as-consts (a=extend-restr v2 v1 newvar restr)
	       n-v1 n-c1 n-v2 n-c2
	       (a=add-2unif v2 v1 newvar unifier) (+ depth 1)
	       words1 words2 old-unif (cons newvar auxvars)))))))



(defun a=unif-d (word1 word2 vars-as-consts restr vars1 consts1 vars2
		     consts2 unifier depth words1 words2 old-unif auxvars)
  (declare (edited  "13-August-1995")
	   (authors Kepser)
	   (input   "word1, word2: rest words to be unified; "
		    "vars-as-consts: list of Keim vars to be treated as "
		    "consts; restr: linear constant restrictions; "
		    "vars1, consts1, vars2, consts2: symmetric difference; "
		    "unifier: unifier for the a-unification of the two words"
		    " constructed so far; "
		    "depth: the current depth of the computation; "
		    "words1, words2: lists of words still to be unified; "
		    "old-unif: unifier of previous words already unified; "
		    "auxvars: list of auxiliary vars introduced in the "
		    "current unification.")
	   (effect  "None.")
	   (value   "Result of the whole a-unification."))
  (let ((l1 (first word1))
	(l2 (first word2))
	(r1 (rest word1))
	(r2 (rest word2)))
    (cond ((or (> (length word1) a*maxdepth)
	       (> (length word2) a*maxdepth)
	       (> depth a*maxdepth))
	   NIL)
	  ((and (endp word1) (endp word2))
	   (let ((u (a=elim-aux unifier auxvars)))
	     (a=decide-unify
	      (a=subst-l u words1) (a=subst-l u words2)
	      vars-as-consts restr (append u old-unif))))
	  ((or (endp word1) (endp word2))
	   NIL)
	  ((keim~equal l1 l2)
	   (a=unif-d r1 r2 vars-as-consts restr vars1 consts1 vars2
		   consts2 unifier (+ depth 1) words1 words2 old-unif
		   auxvars)) 
	  ((and (not (a=constant-p l1 vars-as-consts))
		(a=constant-p l2 vars-as-consts)
		;; check linear constant restrictions
		(not (member l1 (assoc l2 restr :test #'keim~equal)
			     :test #'keim~equal)))
	   (a=unif-var-const-d l1 l2 r1 r2 vars-as-consts restr vars1
			       consts1 vars2 consts2 unifier depth
			       words1 words2 old-unif auxvars))
	  ((and (not (a=constant-p l2 vars-as-consts))
		(a=constant-p l1 vars-as-consts)
		;; check linear constant restrictions
		(not (member l2 (assoc l1 restr :test #'keim~equal)
			     :test #'keim~equal)))
	   (a=unif-var-const-d l2 l1 r2 r1 vars-as-consts restr vars2
			       consts2 vars1 consts1 unifier depth
			       words1 words2 old-unif auxvars))
	  ((and (not (a=constant-p l1 vars-as-consts))
		(not (a=constant-p l2 vars-as-consts)))
	   (a=unif-var-var-d l1 l2 r1 r2 vars-as-consts restr vars1
			     consts1 vars2 consts2 unifier depth words1
			     words2 old-unif auxvars))
	  (T NIL))))



(defun a=rev-unif (unifier)
  (declare (edited  "13-August-1995")
	   (authors Kepser)
	   (input   "A substitution as an a-list.")
	   (effect  "The codomain of each pair is reversed.")
	   (value   "The substitution with the codomain reversed."))
  (mapcar #'(lambda (x)
	      (cons (car x)
		    (nreverse (cdr x))))
	  unifier))



(defun a=subst-w (var subst lst &optional result sub)
  ; one substitution on one word
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "A variable, a word as substitute and a word to "
		    "substitute in.")
	   (effect  "None.")
	   (value   "The result of substituting subst for var in lst."))
  (cond (sub
	 (a=subst-w var subst lst (cons (first sub) result) (rest sub)))
	((endp lst)
	 (nreverse result))
	((keim~equal (first lst) var)
	 (a=subst-w var subst (rest lst) result subst))
	(T
	 (a=subst-w var subst (rest lst) (cons (first lst)
						result)))))




(defun a=subst-a (subst words)
  ; one substitution on many words
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "A substitution (a-list) and a list of words.")
	   (effect  "None.")
	   (value   "The result of applying the substitution to the words.")) 
  (mapcar #'(lambda (x)
	      (a=subst-w (car subst) (cdr subst) x))
	  words))



(defun a=subst-l (subst word-lst)
  ; many substitutions on many words
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "A list of substitutions and a list of words.")
	   (effect  "None.")
	   (value   "The result of applying the substitutions to the words."))
  (if (endp subst)
      word-lst
    (a=subst-l (rest subst)
	       (a=subst-a (first subst) word-lst))))



(defun a=elim-aux (unifier auxvars &optional (result NIL))
  (declare (edited  "12-August-1995")
	   (authors Kepser)
	   (input   "unifier: an a-list of symbols; "
		    "auxvars: a list of auxiliary vars; ")
	   (effect  "None.")
	   (value   "A new unifier with all auxiliary vars removed and the "
		    "codomains reversed, a nice unifier."))
  (cond ((endp unifier)
	 (a=rev-unif result))
	((member (caar unifier) auxvars)
	 (a=elim-aux (a=subst-a (first unifier) (rest unifier))
		     auxvars (a=subst-a (first unifier) result)))
	(T
	 (a=elim-aux (rest unifier) auxvars
		     (cons (first unifier) result)))))




(defun a~decide-unify (theory lhs rhs restr vars-as-consts)
  (declare (edited  "15-August-1995")
	   (authors Kepser)
	   (input   "An associative theory, a list of left side terms, "
		    "a list of right side terms, linear constant "
		    "restrictions and a list of vars to be treated as "
		    "constants.")
	   (effect  "None.")
	   (value   "If the two sides are a-unifiable, an a-unifier; "
		    "else NIL."))
  ;; the interface function
  (let* ((w1 (a=term-to-word lhs))
	 (w2 (a=term-to-word rhs))
	 (fsym (first (th~symbols theory)))
	 (unif (a=decide-unify w1 w2 vars-as-consts restr NIL)))
    (if unif
	(a=construct-unifier (rest unif) fsym))))




(defun a=decide-unify (words1 words2 vars-as-consts restr unifier)
  (declare (edited  "15-August-1995")
	   (authors Kepser)
	   (input   "words1, words2: two lists of words to be unified; "
		    "vars-as-consts: a list of vars to be treated as "
		    "constants; "
		    "restr: an a-list of linear constant restrictions; "
		    "unifier: a unifier (a-list) constructed so far.")
	   (effect  "None.")
	   (value   "A unifier, if the words are a-unifiable, NIL "
		    "otherwise."))
  (if (endp words1)
      (if (a=check-gr-substs (a=ground-substs unifier) restr)
	  (cons T unifier)
	NIL)
    (let ((w1 (first words1))
	  (w2 (first words2)))
      (multiple-value-bind (v1 c1 v2 c2)
	  (a=symmetric-difference (append w1 w2) w1 w2 NIL)
	(if (a=solvable v1 c1 v2 c2)
	    (a=unif-d w1 w2 vars-as-consts restr
		      v1 c1 v2 c2 NIL 0 (rest words1) (rest words2)
		      unifier NIL))))))




;----------------------------------------------------------------

;;; some auxiliary functions for cleaning up things after a unification


(defun a=ground-substs (substs &optional (result NIL))
  (declare (edited  "16-August-1995")
	   (authors Kepser)
	   (input   "A list of substitutions.")
	   (effect  "None.")
	   (value   "A new list of substitutions, where each of the "
		    "substitutions is applied to all the others. "
		    "Thus the resulting substitutions are as far grounded "
		    "as possible."))
  (if (endp substs)
      result
    (a=ground-substs (a=subst-a (first substs) (rest substs))
		     (cons (first substs)
			   (a=subst-a (first substs) result)))))



(defun a=check-gr-substs (gr-substs restr)
  (declare (edited  "16-August-1995")
	   (authors Kepser)
	   (input   "Grounded substitutions and the constant restrictions.")
	   (effect  "None.")
	   (value   "Non-NIL, iff the substitutions obey to the constant "
		    "restrictions."))
  (or (endp gr-substs)
      (and (a=check-gr-subst (caar gr-substs) (cdar gr-substs) restr)
	   (a=check-gr-substs (rest gr-substs) restr)))) 



(defun a=check-gr-subst (var image restr)
  (declare (edited  "16-August-1995")
	   (authors Kepser)
	   (input   "A grounded substitution as a var and its image and "
		    "the constant restrictions.")
	   (effect  "None.")
	   (value   "Non-NIL, iff the substitution obeys to the constant "
		    "restrictions."))
  (or (endp image)
      (and (not (member var (assoc (first image) restr
				   :test #'keim~equal)
			:test #'keim~equal))
	   (a=check-gr-subst var (rest image) restr)))) 



(defun a=construct-unifiers (unif fsym &optional (substs NIL))
  (declare (edited  "16-August-1995")
	   (authors Kepser)
	   (input   "A list unifier and the associative function symbol.")
	   (effect  "A list of Keim substitutions is created")
	   (value   "A list of Keim substitutions, the unifiers."))
  (if (endp unif)
      substs
    (a=construct-unifiers (rest unif) fsym 
			  (cons (a=construct-unifier (cdar unif)
						     fsym)
				substs))))



(defun a=construct-unifier (unif fsym &optional (subst (subst~create
							NIL NIL)))
  (declare (edited  "16-August-1995")
	   (authors Kepser)
	   (input   "A unifier as a word a-list and an associative "
		    "function symbol.")
	   (effect  "A Keim substitution is created.")
	   (value   "The unifier as a Keim substitution."))
  (if (endp unif)
      subst
    (a=construct-unifier (rest unif) fsym
			 (subst~insert-component
			  (caar unif)
			  (a=word-to-term (cdar unif) fsym)
			  subst))))



(defun a=word-to-term (word fsym)
  (declare (edited  "16-August-1995")
	   (authors Kepser)
	   (input   "A word an an associative function symbol.")
	   (effect  "None.")
	   (value   "The term representation of the word."))
  (if (null (rest word))
      (first word)
    (appl~create fsym (list (first word)
			    (a=word-to-term (rest word) fsym)))))



;;;---- unify



(defun a=unif-var-const (var const word1 word2 vars-as-consts restr
			     vars1 consts1 vars2 consts2 unifier
			     depth words1 words2 old-unif auxvars)
  (declare (edited  "17-August-1995")
	   (authors Kepser)
	   (input   "var: a Keim variable; const: a Keim symbol; "
		    "word1, word2: rest words to be unified; "
		    "vars-as-consts: list of Keim vars to be treated as "
		    "consts; restr: linear constant restrictions; "
		    "vars1, consts1, vars2, consts2: symmetric difference; "
		    "unifier: unifier for the a-unification of the two words"
		    " constructed so far; depth: the current depth of the "
		    "computation; "
		    "words1, words2: lists of words still to be unified; "
		    "old-unif: unifier of previous words already unified; "
		    "auxvars: list of auxiliary vars introduced in the "
		    "current unification.")
	   (effect  "None.")
	   (value   "All unifiers of the whole a-unification."))
  (nconc ; var --> const
      (multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	  (a=new-sym-diff-c var const NIL vars1 consts1 vars2 consts2)
	(if (a=solvable n-v1 n-c1 n-v2 n-c2)
	    (a=unif
	     (substitute const var word1 :test #'keim~equal)
	     (substitute const var word2 :test #'keim~equal)
	     vars-as-consts restr n-v1 n-c1 n-v2 n-c2
	     (a=add-unif var const unifier) (+ depth 1)
	     words1 words2 old-unif auxvars)))
      ; var --> const var
      (let ((newvar (fo~variable-create (gentemp "X-"))))
	(multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	    (a=new-sym-diff-c var const newvar vars1 consts1 vars2 consts2)
	  (if (a=solvable n-v1 n-c1 n-v2 n-c2)
	      (a=unif
	       (cons newvar (a=substitute const newvar var word1))
	       (a=substitute const newvar var word2)
	       vars-as-consts restr n-v1 n-c1 n-v2 n-c2
	       (a=add-2unif var const newvar unifier) (+ depth 1)
	       words1 words2 old-unif (cons newvar auxvars)))))))



(defun a=unif-var-var (v1 v2 w1 w2 vars-as-consts restr vars1 consts1
			  vars2 consts2 unifier depth words1 words2
			  old-unif auxvars)
  (declare (edited  "17-August-1995")
	   (authors Kepser)
	   (input   "v1, v2: two Keim variables to be unified; "
		    "w1, w2: rest words to be unified; "
		    "vars-as-consts: list of Keim vars to be treated as "
		    "consts; restr: linear constant restrictions; "
		    "vars1, consts1, vars2, consts2: symmetric difference; "
		    "unifier: unifier for the a-unification of the two words"
		    " constructed so far; depth: the current depth of the "
		    "computation; "
		    "words1, words2: lists of words still to be unified; "
		    "old-unif: unifier of previous words already unified; "
		    "auxvars: list of auxiliary vars introduced in the "
		    "current unification.")
	   (effect  "None.")
	   (value   "All unifiers of the whole a-unification."))
  (nconc ; v1 --> v2
      (multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	  (a=new-sym-diff-v v1 v2 NIL vars1 consts1 vars2 consts2)
	(if (a=solvable n-v1 n-c1 n-v2 n-c2)
	    (a=unif
	     (substitute v2 v1 w1 :test #'keim~equal)
	     (substitute v2 v1 w2 :test #'keim~equal)
	     vars-as-consts (subst v2 v1 restr :test #'keim~equal)
	     n-v1 n-c1 n-v2 n-c2 (a=add-unif v1 v2 unifier)
	     (+ depth 1) words1 words2 old-unif auxvars)))
      ; v1 --> v2 v1
      (let ((newvar (fo~variable-create (gentemp "X-"))))
	(multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	    (a=new-sym-diff-v v1 v2 newvar vars1 consts1 vars2 consts2)
	  (if (a=solvable n-v1 n-c1 n-v2 n-c2)
	      (a=unif
	       (cons newvar (a=substitute v2 newvar v1 w1))
	       (a=substitute v2 newvar v1 w2)
	       vars-as-consts (a=extend-restr v1 v2 newvar restr)
	       n-v1 n-c1 n-v2 n-c2
	       (a=add-2unif v1 v2 newvar unifier) (+ depth 1)
	       words1 words2 old-unif (cons newvar auxvars)))))
      ; v2 --> v1 v2
      (let ((newvar (fo~variable-create (gentemp "X-"))))
	(multiple-value-bind (n-v1 n-c1 n-v2 n-c2)
	    (a=new-sym-diff-v v2 v1 newvar vars2 consts2 vars1 consts1)
	  (if (a=solvable n-v1 n-c1 n-v2 n-c2)
	      (a=unif
	       (cons newvar (a=substitute v1 newvar v2 w2))
	       (a=substitute v1 newvar v2 w1)
	       vars-as-consts (a=extend-restr v2 v1 newvar restr)
	       n-v1 n-c1 n-v2 n-c2
	       (a=add-2unif v2 v1 newvar unifier) (+ depth 1)
	       words1 words2 old-unif (cons newvar auxvars)))))))



(defun a=unif (word1 word2 vars-as-consts restr vars1 consts1 vars2
		     consts2 unifier depth words1 words2 old-unif auxvars)
  (declare (edited  "17-August-1995")
	   (authors Kepser)
	   (input   "word1, word2: rest words to be unified; "
		    "vars-as-consts: list of Keim vars to be treated as "
		    "consts; restr: linear constant restrictions; "
		    "vars1, consts1, vars2, consts2: symmetric difference; "
		    "unifier: unifier for the a-unification of the two words"
		    " constructed so far; "
		    "depth: the current depth of the computation; "
		    "words1, words2: lists of words still to be unified; "
		    "old-unif: unifier of previous words already unified; "
		    "auxvars: list of auxiliary vars introduced in the "
		    "current unification.")
	   (effect  "None.")
	   (value   "All unifiers of the whole a-unification."))
  (let ((l1 (first word1))
	(l2 (first word2))
	(r1 (rest word1))
	(r2 (rest word2)))
    (cond ((or (> (length word1) a*maxdepth)
	       (> (length word2) a*maxdepth)
	       (> depth a*maxdepth))
	   NIL)
	  ((and (endp word1) (endp word2))
	   (let ((u (a=elim-aux unifier auxvars)))
	     (a=unify
	      (a=subst-l u words1) (a=subst-l u words2)
	      vars-as-consts restr (append u old-unif))))
	  ((or (endp word1) (endp word2))
	   NIL)
	  ((keim~equal l1 l2)
	   (a=unif r1 r2 vars-as-consts restr vars1 consts1 vars2
		   consts2 unifier (+ depth 1) words1 words2 old-unif
		   auxvars)) 
	  ((and (not (a=constant-p l1 vars-as-consts))
		(a=constant-p l2 vars-as-consts)
		;; check linear constant restrictions
		(not (member l1 (assoc l2 restr :test #'keim~equal)
			     :test #'keim~equal)))
	   (a=unif-var-const l1 l2 r1 r2 vars-as-consts restr vars1
			     consts1 vars2 consts2 unifier depth
			     words1 words2 old-unif auxvars))
	  ((and (not (a=constant-p l2 vars-as-consts))
		(a=constant-p l1 vars-as-consts)
		;; check linear constant restrictions
		(not (member l2 (assoc l1 restr :test #'keim~equal)
			     :test #'keim~equal)))
	   (a=unif-var-const l2 l1 r2 r1 vars-as-consts restr vars2
			     consts2 vars1 consts1 unifier depth
			     words1 words2 old-unif auxvars))
	  ((and (not (a=constant-p l1 vars-as-consts))
		(not (a=constant-p l2 vars-as-consts)))
	   (a=unif-var-var l1 l2 r1 r2 vars-as-consts restr vars1
			   consts1 vars2 consts2 unifier depth words1
			   words2 old-unif auxvars))
	  (T NIL))))



(defun a=unify (words1 words2 vars-as-consts restr unifier)
  (declare (edited  "17-August-1995")
	   (authors Kepser)
	   (input   "words1, words2: two lists of words to be unified; "
		    "vars-as-consts: a list of vars to be treated as "
		    "constants; "
		    "restr: an a-list of linear constant restrictions; "
		    "unifier: a unifier (a-list) constructed so far.")
	   (effect  "None.")
	   (value   "A (possibly empty) list of all unifiers."))
  (if (endp words1)
      (if (a=check-gr-substs (a=ground-substs unifier) restr)
	  (list (cons T unifier))
	NIL)
    (let ((w1 (first words1))
	  (w2 (first words2)))
      (multiple-value-bind (v1 c1 v2 c2)
	  (a=symmetric-difference (append w1 w2) w1 w2 NIL)
	(if (a=solvable v1 c1 v2 c2)
	    (a=unif w1 w2 vars-as-consts restr
		    v1 c1 v2 c2 NIL 0 (rest words1) (rest words2)
		    unifier NIL))))))



(defun a~unify (theory lhs rhs restr vars-as-consts)
  (declare (edited  "17-August-1995")
	   (authors Kepser)
	   (input   "An associative theory, a list of left side terms, "
		    "a list of right side terms, linear constant "
		    "restrictions and a list of vars to be treated as "
		    "constants.")
	   (effect  "None.")
	   (value   "A (possibly empty) list of all a-unifiers."))
  (let* ((w1 (a=term-to-word lhs))
	 (w2 (a=term-to-word rhs))
	 (fsym (first (th~symbols theory)))
	 (unif (a=unify w1 w2 vars-as-consts restr NIL)))
    (if unif
	(a=construct-unifiers unif fsym))))



(defmethod th~decide-unify ((theory th+a) lhs rhs restr
			    vars-as-consts)
  (a~decide-unify theory lhs rhs restr vars-as-consts))



(defmethod th~unify ((theory th+a) lhs rhs restr vars-as-consts)
  (a~unify theory lhs rhs restr vars-as-consts))

