
(in-package "KEIM")


;;; +----------------------------------------------------------------------
;;; +  queues for word equations
;;; +----------------------------------------------------------------------


(defclass cpa+queue ()
  ((size :initform 0 :accessor cpa=qsize)
   (l-todo :initform NIL :accessor cpa=l-todo)
   (r-todo :initform NIL :accessor cpa=r-todo)
   (l-done :initform NIL :accessor cpa=l-done)
   (r-done :initform NIL :accessor cpa=r-done))
  (:documentation "The class of queues in deductive component A")) 


(defmethod cpa=print-queue ((queue cpa+queue))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A queue (cpa+queue).")
	   (effect  "Prints the queue.")
	   (value   "Nil."))
  (format t "~A - ~A~%~A - ~A : ~A~%" (cpa=l-todo queue) (cpa=l-done queue)
	  (cpa=r-todo queue) (cpa=r-done queue) (cpa=qsize queue)))


(defun cpa=copy-queue (queue)
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A queue (cpa+queue).")
	   (effect  "Creates a copy of the queue.")
	   (value   "The copy of the queue."))
  (let ((new-q (make-instance 'cpa+queue)))
    (setf (cpa=qsize new-q) (cpa=qsize queue)
	  (cpa=l-todo new-q) (copy-list (cpa=l-todo queue))
	  (cpa=r-todo new-q) (copy-list (cpa=r-todo queue))
	  (cpa=l-done new-q) (copy-list (cpa=l-done queue))
	  (cpa=r-done new-q) (copy-list (cpa=r-done queue)))
    new-q))


(defmethod cpa=empty-queue-p ((queue cpa+queue))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A queue (cpa+queue).")
	   (effect  "None.")
	   (value   "T, iff the queue is empty."))
  (= 0 (cpa=qsize queue))) 


(defmethod cpa=push ((queue cpa+queue) lhs rhs)
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A queue (cpa+queue), two words (lists).")
	   (effect  "Pushes the words to the end of the queue.")
	   (value   "The altered queue."))
  (setf (cpa=qsize queue) (+ 1 (cpa=qsize queue))
	(cpa=l-done queue)
	(cons lhs (cpa=l-done queue))
	(cpa=r-done queue)
	(cons rhs (cpa=r-done queue)))
  queue)


(defmethod cpa=pop ((queue cpa+queue))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A queue (cpa+queue).")
	   (effect  "Pops the first words from the queue.")
	   (value   "The altered queue and the two words."))
  (if (not (cpa=empty-queue-p queue))
      (if (endp (cpa=l-todo queue))
	  (progn
	    (setf (cpa=l-todo queue)
		  (reverse (cpa=l-done queue))
		  (cpa=r-todo queue)
		  (reverse (cpa=r-done queue))
		  (cpa=l-done queue) NIL
		  (cpa=r-done queue) NIL)
	    (cpa=pop queue))
	(let ((lhs (first (cpa=l-todo queue)))
	      (rhs (first (cpa=r-todo queue))))
	  (setf (cpa=qsize queue) (- (cpa=qsize queue) 1)
		(cpa=l-todo queue) (rest (cpa=l-todo queue))
		(cpa=r-todo queue) (rest (cpa=r-todo queue)))
	  (values queue lhs rhs)))))


(defmethod cpa=apply-to-queue (function (queue cpa+queue))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A function and a queue (cpa+queue).")
	   (effect  "The function is applied to each word in the queue.")
	   (value   "The altered queue"))
  (if (cpa=empty-queue-p queue)
      queue
      (progn 
	(setf (cpa=l-todo queue)
	      (mapcar function (cpa=l-todo queue))
	      (cpa=r-todo queue)
	      (mapcar function (cpa=r-todo queue))
	      (cpa=l-done queue)
	      (mapcar function (cpa=l-done queue))
	      (cpa=r-done queue)
	      (mapcar function (cpa=r-done queue)))
	queue))) 



;;; +----------------------------------------------------------------------
;;; +  calculate new decisions
;;; +----------------------------------------------------------------------


(defmethod cpa=new-decs ((queue cpa+queue)
			 &optional (no-unchanged 0) (ident NIL)
			 (discr NIL) (true-vars NIL) (restr NIL))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A queue (cpa+queue).")
	   (effect  "The queue may be altered.")
	   (value   "New identification, discrimination, constants, and "
		    "restriction information.")
	   (special cpa*theory))
  
  ;; this is the main function
  ;; in principle, it does the following:
  ;; * take the first word pair from the queue
  ;; * analyse it to find some new decisions
  ;; * push it back to the queue, if needed
  ;; * go on, until the queue becomes empty or no new decisions
  ;;   can be deduced.
  ;; The function is long, but not as complicated, as it looks;
  ;; basically, it is just a long analysis of cases. 
  ;; But there are many cases, several of them being symmetrical. 
  ;; Here, we only consider those cases where we can reduce the word
  ;; equations. If the current equation cannot be reduced, we call
  ;; cpa=new-decs-front which analyses the first two letters of the current
  ;; equation now handling those cases, where "only" identification,
  ;; discrimination etc. information can be deduced.
  ;; cpa=new-decs-front calls cpa=new-decs-back which does the same as
  ;; cpa=new-decs-front, but with the last letters of the current equation.
  ;; cpa=new-decs-back then calls cpa=new-decs again to handle the next
  ;; equation. 
  
  (if (cpa=empty-queue-p queue)
      (values ident discr true-vars restr)
  (multiple-value-bind (new-queue lhs rhs)
      (cpa=pop queue)
    (cond ((> no-unchanged (cpa=qsize queue))
	   (cpa=push new-queue lhs rhs)
	   (values ident discr true-vars restr))
	  ((and (= 2 (length lhs) (length rhs))
		(or (and (unicp~constant-p (first lhs) cpa*theory)
			 (unicp~constant-p (second lhs) cpa*theory))
		    (and (unicp~constant-p (first rhs) cpa*theory)
			 (unicp~constant-p (second rhs) cpa*theory))))
	   (cpa=new-decs
	    (cpa=apply-to-queue
	     #'(lambda (l)
		 (sublis (list (cons (first lhs) (first rhs))
			       (cons (second lhs) (second rhs)))
			 l :test #'keim~equal))
	     new-queue)
	    0 (cons (unicp~id-create (first lhs) (first rhs))
		    (cons (unicp~id-create (second lhs) (second rhs)) ident))
	    discr true-vars restr))
	  ((= 1 (length lhs) (length rhs))
	   (cpa=new-decs
	    (cpa=apply-to-queue
	     #'(lambda (l)
		 (substitute (first lhs) (first rhs) l :test #'term~equal))
	     new-queue)
	    0 (cons (unicp~id-create (first lhs) (first rhs)) ident)
	    discr true-vars restr))
	  ((= 1 (length lhs))
	   (cpa=new-decs
	    (cpa=apply-to-queue
	     #'(lambda (l) (a=subst-w (first lhs) rhs l)) new-queue)
	    0 ident (cpa=make-discr (first lhs) rhs discr)
	    (cons (first lhs) true-vars)
	    (cons (unicp~cr-create (first lhs) rhs) restr)))
	  ((= 1 (length rhs))
	   (cpa=new-decs
	    (cpa=apply-to-queue
	     #'(lambda (l) (a=subst-w (first rhs) lhs l)) new-queue)
	    0 ident (cpa=make-discr (first rhs) lhs discr)
	    (cons (first rhs) true-vars)
	    (cons (unicp~cr-create (first rhs) lhs) restr)))
	  ((unicp~vars-identified-p (first lhs) (first rhs))
	   (cpa=new-decs
	    (cpa=push new-queue (rest lhs) (rest rhs))
	    0 ident discr true-vars restr))
	  ((unicp~vars-identified-p (first (last lhs)) (first (last rhs)))
	   (cpa=new-decs
	    (cpa=push new-queue (butlast lhs) (butlast rhs))
	    0 ident discr true-vars restr))
	  ((and (unicp~constant-p (first lhs) cpa*theory)
		(unicp~constant-p (first rhs) cpa*theory))
	   (cpa=new-decs
	    (cpa=apply-to-queue
	     #'(lambda (l)
		 (substitute (first lhs) (first rhs) l :test #'term~equal))
	     (cpa=push new-queue (rest lhs) (rest rhs)))
	    0 (cons (unicp~id-create (first lhs) (first rhs)) ident)
	    discr true-vars restr))
	  ((and (unicp~constant-p (first (last lhs)) cpa*theory)
		(unicp~constant-p (first (last rhs)) cpa*theory))
	   (cpa=new-decs
	    (cpa=apply-to-queue
	     #'(lambda (l)
		 (substitute (first (last lhs)) (first (last rhs)) l
			     :test #'term~equal))
	     (cpa=push new-queue (butlast lhs) (butlast rhs)))
	    0 (cons (unicp~id-create (first (last lhs)) (first (last rhs)))
		    ident)
	    discr true-vars restr))

	  ;; default
	  (T (cpa=new-decs-front
	      lhs rhs (cpa=push new-queue lhs rhs)
	      (+ 1 no-unchanged) ident discr true-vars restr))))))


(defun cpa=new-decs-front (lhs rhs queue no-unchanged ident discr true-vars
			       restr)
  (declare (edited  "05-NOV-1998")
	   (authors Kepser)
	   (input   "A lhs and rhs of a word equation, a queue (cpa*queue), "
		    "the number of unchanged word equations, "
		    "identification, discrimination, constants and "
		    "constant restriction information.")
	   (effect  "None.")
	   (value   "New identification, discrimination, constants, and "
		    "restriction information.")
	   (special cpa*theory))
  (let ((l1 (first lhs))
	(r1 (first rhs)))
    (cond ((and (unicp~constant-p l1 cpa*theory)
		(unicp~vars-discriminated-p l1 r1))
	   (cpa=new-decs-back
	    lhs rhs queue no-unchanged ident discr (cons r1 true-vars)
	    (cons (unicp~cr-create r1 l1) restr)))
	  ((and (unicp~constant-p l1 cpa*theory)
		(unicp~variable-p r1 cpa*theory))
	   (cpa=new-decs-back
	    lhs rhs queue no-unchanged
	    ident (cons (unicp~id-create l1 r1) discr)
	    true-vars (cons (unicp~cr-create r1 l1) restr)))
	  ((and (unicp~constant-p r1 cpa*theory)
		(unicp~vars-discriminated-p r1 l1))
	   (cpa=new-decs-back
	    lhs rhs queue no-unchanged ident discr (cons l1 true-vars)
	    (cons (unicp~cr-create l1 r1) restr)))
	  ((and (unicp~constant-p r1 cpa*theory)
		(unicp~variable-p l1 cpa*theory))
	   (cpa=new-decs-back
	    lhs rhs queue no-unchanged 
	    ident (cons (unicp~id-create r1 l1) discr)
	    true-vars (cons (unicp~cr-create l1 r1) restr)))
	  (T
	   (cpa=new-decs-back
	    lhs rhs queue no-unchanged ident discr true-vars restr)))))


(defun cpa=new-decs-back (lhs rhs queue no-unchanged ident discr true-vars
			      restr)
  (declare (edited  "05-NOV-1998")
	   (authors Kepser)
	   (input   "A lhs and rhs of a word equation, a queue (cpa*queue), "
		    "the number of unchanged word equations, "
		    "identification, discrimination, constants and "
		    "constant restriction information.")
	   (effect  "None.")
	   (value   "New identification, discrimination, constants, and "
		    "restriction information.")
	   (special cpa*theory))
  (let ((ln (first (last lhs)))
	(rn (first (last rhs))))
    (cond ((and (unicp~constant-p ln cpa*theory)
		(unicp~vars-discriminated-p ln rn))
	   (cpa=new-decs
	    queue no-unchanged ident discr (cons rn true-vars)
	    (cons (unicp~cr-create rn ln) restr)))
	  ((and (unicp~constant-p ln cpa*theory)
		(unicp~variable-p rn cpa*theory))
	   (cpa=new-decs
	    queue no-unchanged ident (cons (unicp~id-create ln rn) discr)
	    true-vars (cons (unicp~cr-create rn ln) restr)))
	  ((and (unicp~constant-p rn cpa*theory)
		(unicp~vars-discriminated-p rn ln))
	   (cpa=new-decs
	    queue no-unchanged ident discr (cons ln true-vars)
	    (cons (unicp~cr-create ln rn) restr)))
	  ((and (unicp~constant-p rn cpa*theory)
		(unicp~variable-p ln cpa*theory))
	   (cpa=new-decs
	    queue no-unchanged ident (cons (unicp~id-create rn ln) discr)
	    true-vars (cons (unicp~cr-create ln rn) restr)))
	  (T ;; default
	   (cpa=new-decs queue no-unchanged ident discr true-vars restr)))))



(defun cpa=make-discr (var consts &optional (result NIL))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A variable and a list of constants.")
	   (effect  "None.")
	   (value   "A list containing a var-const pair for each "
		    "const in consts."))
  (if (endp consts)
      result
    (cpa=make-discr var (rest consts)
		    (cons (unicp~id-create var (first consts))
			  result)))) 


(defun unicp~variable-p (var theory)
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A variable and a theory.")
	   (effect  "None.")
	   (value   "T, iff the variable is already assigned to that theory."))
  (eq (unicp~var-label-theory var) theory)) 


(defun cpa=apply-idents (ident queue)
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A list of identifications and a queue (cpa+queue).")
	   (effect  "Replaces each variable of ident by its representative"
		    " in the queue thereby applying the identifications "
		    "to the queue.")
	   (value   "The altered queue."))
  (if (endp ident)
      queue
      (let ((new-id (cpa=idents ident)))
	(cpa=apply-to-queue
	 #'(lambda (l) (sublis new-id l :test #'keim~equal))
	 queue))))


(defun cpa=idents (ident &optional (new-id NIL))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A list of identifications.")
	   (effect  "None.")
	   (value   "An a-list, each pair consists of a variable in "
		    "ident and its representative."))
  (if (endp ident)
      (remove-duplicates new-id :test #'keim~equal)
      (cpa=idents
       (rest ident)
       (let* ((vl (unicp~id-left (first ident)))
	      (vr (unicp~id-right (first ident)))
	      (vlr (unicp~var-label-repres vl))
	      (vrr (unicp~var-label-repres vr)))
	 (cond ((and (eq vl vlr) (eq vr vrr))
		new-id)
	       ((eq vl vlr)
		(acons vr vrr new-id))
	       ((eq vr vrr)
		(acons vl vlr new-id))
	       (t (acons vl vlr (acons vr vrr new-id))))))))



(defun cpa=initialise-q (equations)
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "Some A-equations.")
	   (effect  "Creates a new queue (cpa+queue) and transforms the "
		    "A-equations into word equations in the queue.")
	   (value   "The new queue."))
  (let ((queue (make-instance 'cpa+queue)))
    (cpa=init-q queue (unicp~eq-lhss equations) (unicp~eq-rhss equations))))


(defun cpa=init-q (queue lhss rhss)
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A queue (cpa*queue), left and right hand sides of "
		    "A-equations.")
	   (effect  "The left and right hand sides of the A-equations "
		    "are transformed into word equations in the queue.")
	   (value   "The altered queue."))
  (if (endp lhss)
      queue
    (progn
      (setf (cpa=l-todo queue)
	    (cons (a=term-to-word (first lhss)) (cpa=l-todo queue))
	    (cpa=r-todo queue)
	    (cons (a=term-to-word (first rhss)) (cpa=r-todo queue))
	    (cpa=qsize queue) (+ 1 (cpa=qsize queue)))
      (cpa=init-q queue (rest lhss) (rest rhss)))))


;;; +-----------------------------------------------------------------------
;;;  cpa+up, unicp~theo-init, unicp~new-constr
;;; +------------------------------------------------------------------------


(defclass cpa+up (unicp+elementary-up)
  ((word-equations :initarg :word-equations :initform NIL
		   :accessor cpa=word-eq)
   (stack :initform NIL :accessor cpa=stack))
  (:documentation "This is the class of A-unification problems."))


(defmethod unicp~theo-init ((theory th+a) general-up equations
			    identified discriminated constants const-restr)
  (declare (authors Kepser)
	   (edited  "8-OCT-1998")
	   (input   "An associative theory, a general unification problem, the "
		    "equations (constraints), and information on identified, "
		    "discriminated variables, constants and constant restrictions.")
	   (effect  "Creates a new A-theory unification problem. "
		    "Initially identifies all variable pairs from identified. "
		    "Transforms the equations into word problems stored in a queue. "
		    "Finally new information is calculated.")
           (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "  solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "  of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "  unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "  $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "  unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "  solution is lost if $y$ is in the constant"
		    "  restrictions of $x$).")
  	   (ignore discriminated constants const-restr))
  (let ((cpa*theory theory)
	(up (make-instance 'cpa+up
			    :general-up general-up
			    :equations equations
			    :theory theory))
	(words (cpa=apply-idents identified
				 (cpa=initialise-q equations))))
				  
    (declare (special cpa*theory))
    (setf (cpa=word-eq up) words)
    (multiple-value-bind (new-ident new-discr true-vars new-restr)
	(cpa=new-decs words)
      (values up new-ident new-discr true-vars nil new-restr nil))))



(defmethod unicp~new-constr
  ((up cpa+up) identified discriminated constants const-restr)
  (declare (authors Kepser)
	   (edited  "8-OCT-1998")
	   (input   "An A-theory unification problem and information on "
		    "identified and discriminated variables, constants and "
		    "constant restrictions.")
	   (effect  "Each variabel pair in identified will be identified.")
           (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "  solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "  of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "  unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "  $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "  unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "  solution is lost if $y$ is in the constant"
		    "  restrictions of $x$).")
	   (ignore discriminated const-restr))
  (if (and (endp identified) (endp constants))
      ;; nothing new for us
      (values up nil nil nil nil nil nil)
    (let ((cpa*theory (unicp~theory up))
	  (words (cpa=apply-idents identified (cpa=word-eq up))))
      (declare (special cpa*theory))
      (multiple-value-bind (new-ident new-discr true-vars new-restr)
	  (cpa=new-decs words)
	(values up new-ident new-discr true-vars nil new-restr nil)))))



;;; +-----------------------------------------------------------------------
;;;   unicp-solvable-p (calls the old a.lisp routine)
;;; +-----------------------------------------------------------------------

(defmethod unicp~solvable-p ((up cpa+up))
  ;; since it can only deduce some information, 
  ;; the deductive A component *must* call the standard
  ;; A-unification at the end, if all decisions are made
  ;; but we take the alredy reduced word equations, not the input ones
  (let* ((vars-as-consts (cpa=get-consts (unicp~variables up)
					 (unicp~theory up)))
	 (restr (cpa=get-restr vars-as-consts)))
    (a=decide-unify
     (cpa=get-repres (cpa=l-todo (cpa=word-eq up))
		     (cpa=get-repres (cpa=l-done (cpa=word-eq up))))
     (cpa=get-repres (cpa=r-todo (cpa=word-eq up))
		     (cpa=get-repres (cpa=r-done (cpa=word-eq up))))
     vars-as-consts restr NIL)))



(defun cpa=get-consts (vars theory &optional (consts NIL))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A list of variables and a theory.")
	   (effect  "None.")
	   (value   "A list of representatives of those variables "
		    "which are constants in the given theory."))
  (if (endp vars)
      (unicp~repres-list consts)
    (cpa=get-consts (rest vars) theory
		    (if (unicp~constant-p (first vars) theory)
			(cons (first vars) consts)
		      consts)))) 


(defun cpa=get-restr (consts &optional (restr NIL))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A list of constants.")
	   (effect  "None.")
	   (value   "A list of constant restrictions of the constants"
		    " using representants all over the place."))
  (if (endp consts)
      restr
    (let ((rest-list (unicp~repres-list
		      (unicp~var-label-less (first consts)))))
      (if (endp rest-list)
	  (cpa=get-restr (rest consts) restr)
	(cpa=get-restr (rest consts)
		       (acons (first consts) rest-list restr)))))) 



(defun cpa=get-repres (words &optional (rp-words NIL))
  (declare (edited  "04-NOV-1998")
	   (authors Kepser)
	   (input   "A list of words (lists of vars).")
	   (effect  "None.")
	   (value   "A list of words where each variable is replaced "
		    "by its representative."))
  (if (endp words)
      rp-words
    (cpa=get-repres
     (rest words)
     (cons (mapcar #'unicp~var-label-repres (first words)) rp-words)))) 



;;;+--------------------------------------------------------------------------
;;;  save/restore-choice-point                                                
;;;+--------------------------------------------------------------------------


#|
(defmethod unicp~up-copy ((up cpa+up))
  (declare (edited  "9-OCT-1998")
	   (authors Kepser)
	   (input   "A cp unification problem.")
	   (effect  "None.")
	   (value   "A copy of the cpa-UP that can be used when"
		    "backtracking is necessary.")) 
  (let ((copied-up (make-instance
		     'cpa+up
		     :general-up (unicp~general-up up)
		     :equations (unicp~equations up)
		     :theory (unicp~theory up)
		     :word-equations (cpa=copy-queue (cpa=word-eq up)))))
    copied-up)) 
|#


(defmethod unicp~up-save-choice-point ((up cpa+up))
  (push (cpa=copy-queue (cpa=word-eq up)) (cpa=stack up))
  up)


(defmethod unicp~up-restore-choice-point ((up cpa+up))
  (setf (cpa=word-eq up) (pop (cpa=stack up)))
  up)





;;;+--------------------------------------------------------------------------
;;;+ Incremental Version
;;;+--------------------------------------------------------------------------

(defmethod unicp~new-equations ((up cpa+up) equations identified
				discriminated constants const-restr)
  (declare (edited  "1998-10-9")
	   (authors  Kepser)
	   (input   "An elementary unification problem, a list of equations,"
		    "and some restrictions that are new for this problem:"
		    "a list of pairs of variables which have to be identified,"
		    "a list of pairs of variables which have to be discriminated,"
		    "a list of variables that should be treated as constants,"
		    "a list of constant restrictions (i.e. pairs of a constant"
		    "and a set of variables.")
	   (effect  "The ELEMENTARY-UP is updated with the new restrictions.")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP belongs"
		    "to is called.  It returns NIL if the unification problem is"
		    "not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains some new"
		    "constraints wich must be met if the unification problem should be solvable:"
		    "1. True, indicating that the unification problem is still solvable."
		    "2. A list of new identification of variables (i.e. sets of variables)."
		    "3. A list of pairs of variables which must not be unified."
		    "4. A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6. A list of pairs $(x . y)$ of variables where $\sigma(x)$"
		    "   is a subterm of $\sigma(y)$ for a unifier $\sigma$."
		    "7. A list of pairs $(x . y)$ of variables where no solution"
		    "   is lost if $y$ is in the constant restrictions of $x$.")
	   (ignore equations discriminated constants const-restr))
  ;; we do a new start
  (let ((cpa*theory (unicp~theory up))
	(words (cpa=apply-idents identified
				 (cpa=initialise-q (unicp~equations up)))))
    (declare (special cpa*theory))
    (setf (cpa=word-eq up) words)
    (multiple-value-bind (new-ident new-discr true-vars new-restr)
	(cpa=new-decs words)
      (values up new-ident new-discr true-vars nil new-restr nil))))

