;                    +--------------------------------+
;                    |********************************|
;                    |*                              *|
;                    |*      ACI - Unification       *|
;                    |*      Decision - Function     *|
;                    |*        by Felix Wolf         *|
;                    |*                              *|
;                    |********************************|
;                    +--------------------------------+
;
;
; Main-function: (aci~decide tl1 tl2
;                  &optional (reslist NIL) (var_as_const NIL)) 
;
; el :             two lists of terms (t1...tn) (s1...sn), which shall be
;                  aci-unified
; &optional
; reslist :        list of constant-restrictions
;                  ((v1 c11 c12 ..) (v2 c21 c22 ..) ..) 
; var_as_const :   list of variables which have to be considered as constants
;                  (v1 v2 ..)
;
; 
; This program contains four modules:
;
; 1. symset  : normalization
; 2. formset : creation of clauses for decision algorithm
; 3. res     : constant-restrictions
; 4. aci     : unification algorithm
;
;
; Please contact for questions felix@cantor.informatik.rwth-aachen.de 

;******************************************************************************
;* symset                                                                     *
;******************************************************************************

(in-package :keim)

(defstruct (symset+sel  (:conc-name symset~)
			(:constructor symset=make)
			(:print-function symset~print))
  symset_equation_list   ; "list of [symset+eq]" 
  var_list               ; "list of occuring variables"
  const_list)            ; "list of occuring constants"

(defstruct (symset+eq (:conc-name symset~eq-)
		      (:constructor symset=eq-make))
  var_eq                 ; (cons
                         ;    (list of var on left  side of equation)
                         ;    (list of var on right side of equation)
  const_eq               ; analogous
  const_intersect)       ; list of constants appearing on both sides
                         ; of equation

;            var-list:   (U Z Y X)
;            const-list: (A)
;
;            equation *** common constants
;
;            (Y) =  (U)                             NIL
;            (Z X Y) (A) =  (Y)                     NIL
;            (Y X) (A) =  (Z) (A)                   (A)


(defun symset~print (sel stream c)
  (declare (ignore c))
  (format stream " ~%")
  (format stream " var-list:   ~A~%" (reverse (symset~var_list sel)))
  (format stream " const-list: ~A~%~%" (reverse (symset~const_list sel)))
  (format stream " equation *** common constants~%~%")
  (mapc #'(lambda (eq) (symset+eq=print eq stream c))
	(reverse (symset~symset_equation_list sel)))
  (values))

(defun symset+eq=print (eq stream c)
  (declare (ignore c))
  (when (not (null (car (symset~eq-var_eq eq))))
	(format stream " ~A" (reverse (car (symset~eq-var_eq eq)))))
  (when (not (null (car (symset~eq-const_eq eq))))
	(format stream " ~A" (reverse (car (symset~eq-const_eq eq)))))
  (format stream " =")
  (when (not (null (cdr (symset~eq-var_eq eq))))
	(format stream " ~A" (reverse (cdr (symset~eq-var_eq eq)))))
  (when (not (null (cdr (symset~eq-const_eq eq))))
	(format stream " ~A" (reverse (cdr (symset~eq-const_eq eq)))))  
  (format stream "~40T~A~%" (symset~eq-const_intersect eq))) 

(defun symset=new ()
  (symset=make :symset_equation_list nil
	       :var_list nil
	       :const_list nil))

(defun symset=eq-new ()
  (symset=eq-make :var_eq (cons nil nil)
		  :const_eq (cons nil nil)
		  :const_intersect nil))

(defun symset~create (el var_as_const)
  (declare (edited  "25-JAN-1995")
	   (authors Felix)
	   (input   "Equation list ((s1.t1) ... (sn.tn)), si,ti keim-terms"
		    "var_as_const: list of varables which have"
		    "to be considered as constants")
	   (effect  "None.")
	   (value   "List of symset equations, list of variables,"
		    "list of constants [symset+sel]"))
      (if (null el)
	  (symset=new)
	(progn
	  (symset=mark-var-const var_as_const)
	  (symset=add-constant! (symset=create_aux! el (symset=new) 1)))))

(defun symset=add-constant! (sel)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "[symset+sel]")
	   (effect  "a further constant represented by a variable not"
		    "appearing in sel is added to its const_list")
	   (value   "sel modified by effect"))
  (push
   (fo~variable-create)
   (symset~const_list sel))
  sel)

(defun symset=mark-var-const (var_as_const)
  (mapc #'(lambda (var) (sym~set-symbol-binding! var T)) var_as_const))

(defun symset=variable-p (term)
  (and
   (sym~variable-p term)
   (not (sym~symbol-binding term))))

(defun symset=create_aux! (el sel counter)
  (declare (edited  "01-FEB-1995")
	   (authors Felix)
	   (input   "el equation-list ((s1.t1)...(sn.tn)), si,ti keim-terms"
		    "sel [symset+sel]"
		    "counter [number] is number of term to process next")
	   (effect  "sel will be modified destructivly")
	   (value   "None."))
  (if (null el)
      sel
    (progn
      (symset=equation! (car el) sel counter)
      (symset=create_aux! (cdr el) sel (+ 2 counter))
      sel)))
     
(defun symset=equation! (e sel counter)
  (declare (edited  "01-FEB-1995")
	   (authors Felix)
	   (input   "e equation (s.t), s,t keim-terms"
		    "sel [symset+sel]"
		    "counter [number] is number of term to process next")
	   (effect  "sel, e will be modified destructivly")
	   (value   "None."))
  (push
   (symset=eq-create e sel counter)
   (symset~symset_equation_list sel)))

(defun symset=eq-create (e sel counter)
  (let ((seq (symset=eq-new)))
    (progn
      (symset=term! (car e) seq sel counter)
      (symset=term! (cdr e) seq sel (1+ counter))
      seq)))

(defun symset=term! (term seq sel counter)
  (declare (edited  "01-FEB-1995")
	   (authors Felix)
	   (input   "keim-term"
		    "sel [symset+sel]"
		    "counter [number] is number of term to process next")
	   (effect  "sel, term will be modified destructivly")
	   (value   "None."))
  (symset=arglist! (list term) seq sel counter))

(defun symset=left-side (counter)    ; test if counter is on left side of  
  (not (zerop (mod counter 2))))     ; equation

(defun symset=arglist! (al seq sel counter)
  (declare (edited  "01-FEB-1995")
	   (authors Felix)
	   (input   "argument-list of a keim-term"
		    "symset"
		    "sel [symset+sel]"
		    "counter [number] is number of term to process next")
	   (effect  "sel will be modified destructivly")
	   (value   "symset corresponding to al"))
    
    (if (null al)
	seq
      
	(if (appl~p (car al))       ; if head of al is application 
	    (symset=arglist!        ; then continue with 
	     (cdr al)               ; tale of al
	     (symset=arglist! (appl~arguments (car al)) seq sel counter)
	                            ;; symset corresponding to head of al
	     sel                    ; modified sel
	     counter)               ; same counter

	  (let ((head (car al))
		(head-label (sym~label (car al))))
	    
	    (progn ; else
	      (when
	       (null head-label)          ; if term is not in const_list
					  ; or var_list of sel
	       (if                        ; then 
		   (symset=variable-p head)  ; if term is var
					; add to var_list of sel
		   (push head (symset~var_list sel))
					; else term is const and added to
		 (push head (symset~const_list sel))))
					; const_list of sel
	      (when
	       (or
		(null head-label)            ; if term (head of al) 
		(and                         ; is not in seq-side yet
		 (not (null head-label)) 
		 (> counter head-label)))
	     
	       (if                               ; then
		   (symset=left-side counter)    ; if term is from left side of
		   (if                           ; equation then
		       (symset=variable-p head)  ; if term is var
		       (push
			head
			(car (symset~eq-var_eq seq)))
					; add term to left-side-var-set
		     (push
		      head
		      (car (symset~eq-const_eq seq))))
					     ; else add term to
					     ; left-side-const-set
		 (if           ; if term is from right side of equation then
		     (symset=variable-p head)   ; if term is var
		     (push
		      head
		      (cdr (symset~eq-var_eq seq)))
					; add term to right-side-var-set
		   (progn               ; if term is const
		     (push
		    head
		    (cdr (symset~eq-const_eq seq)))
					; add term to right-side-const-set
		     (when              ; and if const is also on left side
		      (and
		       (not (null head-label))
		       (= (1- counter) head-label))
		      (push
		       head
		       (symset~eq-const_intersect seq))))))
                                        ; add to const-intersection
	       (sym~set-label! head counter))
					; set term-label to counter
	  
	      (symset=arglist! (cdr al) seq sel counter))))))
                                         ;; else continue with (cdr al)

;+----------------------------------------------------------------------------+
;| reset term-labels                                                          |
;+----------------------------------------------------------------------------+


(defun symset~slot-reset! (sel var_as_const)
  (declare (edited  "01-FEB-1995")
	   (authors Felix)
	   (input   "sel [symset+sel], var_as_const (see above)")
	   (effect  "slot-reset of terms from sel's var_list and const_list"
		    "and reset of marked var_as_const")
	   (value   "None."))
  (progn
    (mapc #'aci=slot-reset_term! var_as_const)
    (if (null sel)
	nil
      (progn
	(mapc #'(lambda (var) (sym~set-label! var nil))  
	      (symset~var_list sel))
	(mapc #'(lambda (const) (sym~set-label! const nil))
	      (symset~const_list sel))
	(mapc #'(lambda (const) (sym~set-symbol-binding! const nil))
	      (symset~const_list sel))))))

;******************************************************************************
;* formset                                                                    *
;******************************************************************************
;
;
;  outline of formset+formset
;  ==========================
;
;
;                     +------------------------+
;                     | +--------------------+ |
;                     | | +----------------+ | |                  |
;  + -  P(x1,c1)  (tv,| | |)               | | +-> (k, litlist)  -+
;  |    P(x2,c1)  (tv, cl)                 | |                    |
;  |      .                                | +---> (k, litlist)  -+
;  |      .                                |                      |
;  |    P(x1,c2)  (tv,cl) <---+            +-----> (k, | |)      -+
;  |      .                   |                        | |        |
;  |      .                   +------------------------+ |        |
;  + -  P(xn,cn)  (tv,cl) <------------------------------+        |
;  |    False     (nil,nil)                                       |
;  |                                                         (clauselist)     
;  |    * literals *
;  |
;  |
;  |         variables <-----------------+
;  +-------< +-+-+-+-+                   |     
;            | | | | |                   +---- (sel)      
;            +-+-+-+-+                   |
;            | | | | | constants <-------+    
;            +-+-+-+-+
;             (tab)
;
; tv : truthvalue (default true)
; cl : list of basic Horn propositions in which P occurs as negative literal
; clauselist : list of basic Horn propositions              
; k : number of negative literals in the basic Horn proposition
; litlist : list of positive literals in the conjunction of the conclusion
; tab : 2-dim array of literals P (var, const) which is true iff const
;       does not appear in (a solution for) var
; sel : [symset+sel]

(defstruct (formset+formset (:conc-name formset~)
		            (:constructor formset=make)
			    (:print-function formset=print))
  sel          ; "[symset+eq]"
  tab          ; tab (see above)
  clauselist)  ; clauselist (see above)

(defun formset=lit-create (value clauselist)
  (declare (edited  "03-FEB-1995")
	   (authors Felix)
	   (input   "value [boolean] : literal is evaluated to true"
		    "clauselist : list of basic Horn propositions" 
                    "in which literal occurs as negative (left side)")
	   (effect  "None.")
	   (value   "below"))
  (cons value clauselist))

(defun formset~lit-clist (lit)
  (cdr lit))

(defun formset=tab-create! (sel)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "[symset+sel]")
	   (effect  "terms of sel will be labeled")
	   (value   "tab (see above)"))
  (let*
      ((v (length (symset~var_list sel)))
       (c (length (symset~const_list sel)))
       (a (make-array (list v c))))
    (progn
      (loop for i from 0 to (1- v) do
	    (loop for j from 0 to (1- c) do
		    (setf (aref a i j) (formset=lit-create nil nil))))
      (formset=number-termlist! (symset~var_list sel) 0)
      (formset=number-termlist! (symset~const_list sel) 0)
      a)))

(defun formset~aref (tab var const)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "tab (see above), variable, constant")
	   (effect  "None.")
	   (value   "corresponding literal P(var,const)"))
	(aref tab (sym~label var) (sym~label const)))      

(defun formset=lit-entry! (var const clause tab)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "variable, constant, basic Horn clause, tab (see above)")
	   (effect  "clause will be added to clauselist"
		    "of literal P(var, const)")
	   (value   "None."))
  (push
   clause
   (cdr (formset~aref tab var const))))

(defun formset=number-termlist! (tl counter)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "list of terms")
	   (effect  "termlist will be numbered increasingly by using"
		    "label-slots beginning with counter")
	   (value   "None."))
  (if (null tl)
      nil
    (progn
      (sym~set-label! (car tl) counter)
      (formset=number-termlist! (cdr tl) (+ 1 counter)))))

(defun formset=const-left-p (c eq)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "constant, set-equation [symset+eq]")
	   (effect  "None.")
	   (value   "Constant appears on left side of equation."))
  (member c (car (symset~eq-const_eq eq)) :test #'term~equal))

(defun formset=const-in-equation-p (c eq)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "constant, set-equation [symset+eq]")
	   (effect  "None.")
	   (value   "Constant appears in equation."))
  (or
   (member c (car (symset~eq-const_eq eq)) :test #'term~equal)
   (member c (cdr (symset~eq-const_eq eq)) :test #'term~equal)))

(defun formset=var-in-equation-p (v eq)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "variable, set-equation [symset+eq]")
	   (effect  "None.")
	   (value   "Variable appears in equation."))
  (or
   (member v (car (symset~eq-var_eq eq)) :test #'term~equal)
   (member v (cdr (symset~eq-var_eq eq)) :test #'term~equal)))

;-----------------------------------------------------------------------------
; To check the solvability of the equation set, we form for every variable x
; and for every constant c in (formset=add-constant! sel) a propositional
; variable P(x,c) which is true iff a does not appear in x. Now corresponding
; to each equation we can form a set of Horn clauses relating these
; propositional variables. For instance, let uva = wxb be an equation, where
; a and b are constants and the rest are variables.
;
; 1. If c is a constant and does not appear in the equation, then we get
;    the clauses:
;
;    P(u,c) ^ P(v,c) => P(w,c) ^ P(x,c)
;    P(w,c) ^ P(x,c) => P(u,c) ^ P(v,c)
;
; 2. If b is a constant and does appear in the equation only on one side, then
;    we get the clauses:
;
;    P(u,b) ^ P(v,b) => False
;
; 3. We need to enforce each variable to have at least one constant, or
;    the clause 
;
;    P(x,a) ^ P(x,b) ^ P(x,c) => False
;
;    must be satisfied for each variable x in the equation.
;
;-----------------------------------------------------------------------------

(defun formset=const-in-eq-left! (c eq tab)
  ;
  ; P(v1,c) ^ .... ^ P(vn,c) => False for all vi on right side of equation
  ;
  (declare (edited  "21-FEB-1995")
	   (authors Felix)
	   (input   "constant, equation, tab (see above)")
	   (effect  "adding literals P(v,c) to tab for each var v"
		    "from right side of equation")
	   (value   "clause for constant appearing on left side and not"
		    "on right side")
	   (special formset*false))
  (let ((clause (cons
		 (length (cdr (symset~eq-var_eq eq)))
		 (list formset*false))))
    (mapc
     #'(lambda (v)
	 (formset=lit-entry! v c clause tab))
     (cdr (symset~eq-var_eq eq)))
    clause))

(defun formset=const-in-eq-right! (c eq tab)
  ;
  ; P(v1,c) ^ .... ^ P(vn,c) => False for all vi on left side of equation
  ;
  (declare (edited  "21-FEB-1995")
	   (authors Felix)
	   (input   "constant, equation, tab (see above)")
	   (effect  "adding literals P(v,c) to tab for each var v"
		    "from left side of equation")
	   (value   "clause for constant appearing on right side and not"
		    "on left side")
	   (special formset*false))
  (let ((clause (cons
		 (length (car (symset~eq-var_eq eq)))
		 (list formset*false))))
    (mapc
     #'(lambda (v)
	 (formset=lit-entry! v c clause tab))
     (car (symset~eq-var_eq eq)))
    clause))

(defun formset=const-not-in-eq-left! (c eq tab)
  ;
  ; P(v1,c) ^ .... ^ P(vn,c) => P(u1,c) ^ .... ^ P(um,c)
  ; for all var vi on left side and all var ui on right side
  ;  
  (declare (edited  "21-FEB-1995")
	   (authors Felix)
	   (input   "constant, equation, tab (see above)")
	   (effect  "adding literals P(v,c) to tab for each var v"
		    "from left side of equation")
	   (value   "first clause for constant c not appearing in equation"))
  (let ((clause
	 (cons
	  (length (car (symset~eq-var_eq eq)))
	  (mapcar #'(lambda (v)
		      (formset~aref tab v c))
		  (cdr (symset~eq-var_eq eq))))))
    (mapc #'(lambda (v)
	      (formset=lit-entry! v c clause tab))
	  (car (symset~eq-var_eq eq)))
    clause))

(defun formset=const-not-in-eq-right! (c eq tab)
  ;
  ; P(u1,c) ^ .... ^ P(un,c) => P(v1,c) ^ .... ^ P(vm,c)
  ; for all var vi on left side and all var ui on right side
  ;  
  (declare (edited  "21-FEB-1995")
	   (authors Felix)
	   (input   "constant, equation, tab (see above)")
	   (effect  "adding literals P(v,c) to tab for each var v"
		    "from right side of equation")
	   (value   "second clause for constant c not appearing in equation"))
  (let ((clause
	 (cons
	  (length (cdr (symset~eq-var_eq eq)))
	  (mapcar #'(lambda (v)
		      (formset~aref tab v c))
		  (car (symset~eq-var_eq eq))))))
    (mapc #'(lambda (v)
	      (formset=lit-entry! v c clause tab))
	  (cdr (symset~eq-var_eq eq)))
    clause))

(defun formset=clean-clauselist! (clauselist)
  ; removes all clauses P1 ^ ... ^ Pn => NIL from clauselist
  
  (mapcan #'(lambda (clause)
	      (when (not (null (cdr clause)))
		    (list clause)))
	  clauselist))  

(defun formset~create (sel)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "sel [symset+sel]")
	   (effect  "None.")
	   (value   "corresponding set of Horn formulae [formset+formset]"
		    "outlined above"))
  (let*
      ((tab (formset=tab-create! sel))
       (vl  (symset~var_list sel))
       (cl  (symset~const_list sel))
       (length-cl (length cl))
       (el  (symset~symset_equation_list sel))
       (formset*false (formset=lit-create nil nil))
       (clauselist nil))
    (declare (special formset*false))
    (labels
     ((const_formulae! (c eq)
		       (if (formset=const-in-equation-p c eq)
					     ; if const is in equation (2.)  
			   (when
			    (not
			     (member c (symset~eq-const_intersect eq)
				     :test #'term~equal))
			                     ; and not on both sides
			    (if (formset=const-left-p c eq)
				             ; if const is on left side
				(push
				 (formset=const-in-eq-left! c eq tab)
				 clauselist) ; add corresponding clause
			                     ; to clauselist
			      
			      (push          ; if const is on right  side
			       (formset=const-in-eq-right! c eq tab)
			       clauselist))) ; add corresponding clause
			                     ; to clauselist

			                   ; if const is not in equation (1.)  
			 (progn
			   (push   ; add for each side of equation
			    (formset=const-not-in-eq-right! c eq tab)
			    clauselist)
			   (push   ; corresponding clause to clauselist
			    (formset=const-not-in-eq-left! c eq tab)
			    clauselist))))
      (var_formulae! (v)
		     (let ((clause (cons
				    length-cl
				    (list formset*false))))
		       (push clause clauselist)
		       (mapc #'(lambda (c)
				 (formset=lit-entry! v c clause tab))
			     cl))))
                     ; add for v clause to clauselist (3.)
                     ; create for each constant c from cl a literal
                     ; P(v,c) and add clause to its cdr (see outline)
     ;;======================================================================
     (mapc #'(lambda (eq)
	       (mapc #'(lambda (c) (const_formulae! c eq)) cl)) el)
     ;; create formula for each constant in cl and each equation
     ;; in el (1. + 2.)
     
     (mapc #'(lambda (v) (var_formulae! v)) vl)
     ;; create formula for each variable in vl (3.)
     
     (setf clauselist (formset=clean-clauselist! clauselist))
     ;; remove redundant formulae from clauselist
     
     (formset=make :sel        sel
		   :tab        tab
		   :clauselist clauselist))))


(defun formset=print (formset stream c)
  (declare (ignore c))
  (let*
      ((tab (formset~tab formset))
       (sel (formset~sel formset))
       (clauselist (formset~clauselist formset))
       (vl (symset~var_list (formset~sel formset)))
       (cl (symset~const_list (formset~sel formset)))
       (formset*false (formset=lit-create nil nil)))
    (labels
     ((print-clause (clause)
		    (if (equal (cadr clause) formset*false) 
			(format stream " ~A ~A => False~%"
				(car clause)   ; number of not true evaluated
				               ; negative literals in clause
				(premise clause))
		      (format stream " ~A ~A => ~A~%"
			      (car clause)
			      (premise clause)
			      (conclusion clause))))
      (premise (clause)
		 (mapcan #'(lambda (c)
			   (mapcan
			    #'(lambda (v)
				(when
				 (member
				  clause
				  (formset~lit-clist (formset~aref tab v c)))
				 (list (cons c v))))
			    vl)) cl))

      ; (premise clause) is a list of literals pointing at clause
      
      (conclusion (clause)
		 (mapcan #'(lambda (c)
			   (mapcan
			    #'(lambda (v)
				(when
				 (member
				  (formset~aref tab v c)
				  (cdr clause))
				 (list (cons c v))))
			    vl)) cl)))

      ; (conclusion clause) is the list of literals from (cdr clause) 
      
     ;======================================================================
     (mapcar #'print-clause clauselist)
     (symset~print sel stream c))))
       
;******************************************************************************
;* Konstanten - Restriktion                                                   *
;******************************************************************************

; constant-restrictions are represented as follows:
;
; (list
;   (cons constant_1 (list var_1 ... var_n1))      (res) -+
;                   :                                     | (reslist)
;   (cons constant_m (list var_1 ... var_nm)))     (res) -+


(defun res~add-clauses! (formset reslist)

  ; adds for each restriction V(c) =  (x1 x2) clauses NIL => P(c x1),
  ; NIL => P(c x2) to (formset~clauselist formset)
  
  (if (null reslist)
      formset
    (progn
      (mapc #'
       (lambda (res)
	 (mapc #'
	  (lambda (var)
	    (when (and (sym~label (car res))
		       (sym~label var)
		       (member (car res) (symset~const_list (formset~sel formset))
			       :test #'term~equal)
		       (member var (symset~var_list (formset~sel formset))
			       :test #'term~equal))
		  (let ((clause
			 (cons
			  0
			  (list
			   (formset~aref
			    (formset~tab formset) var (car res))))))
		    (push clause (formset~clauselist formset)))))
	  (cdr res)))
       reslist)
      formset)))

(defun res~print (reslist)
  (format T " constant - restrictions~%~%")
  (if (null reslist)
      (format T "   ~A~%" nil)
    (mapc #'
     (lambda (res)
       (format T "   V(~A) =  ~A~%" (car res) (cdr res)))
     reslist)))

;******************************************************************************
;* ACI - Unifikation                                                          *
;******************************************************************************

(defvar aci**output)         ; output flag:  output if flag true
(defvar aci**unifier)        ; unifier flag: if problem unifiable and flag
                             ;               true, then the value of
                             ;               aci~unif is one possible
                             ;               unifier

(setf aci**output nil)         ; flag defaults
(setf aci**unifier nil)

(defmethod th~decide-unify ((th th+aci) tl1 tl2 restr const-vars)
  (aci~decide tl1 tl2 restr const-vars))

(defun aci~decide (tl1 tl2 &optional (reslist NIL) (var_as_const NIL))
  (declare (edited  "12-MAY-1995")
	   (authors Richts)
	   (input   "two list of Terms which shall be aci-unified"
		    "optional: a list of constant restrictions")
	   (effect  None)
	   (value   "see aci~unif"))
  (aci~unif (mapcar #'cons tl1 tl2) reslist var_as_const))

(defun aci~unif (el reslist var_as_const)
  (declare (edited  "30-MAR-1995")
	   (authors Felix)
	   (input   "el      ((s1.t1) ... (sn.tn))"
		    "reslist (above)")
	   (effect  "None. If aci**output then output of computation.")
	   (value   "(if (unifieable el)
                         (if aci**unifier
                             unifier
                           nil)
                       nil)"))
  (let*
      ((sel     nil)       ; declaration of variables  
       (formset nil)
       (q       nil)
       (consistent T))     ; flag for unifiability
    (unwind-protect
	(progn
	  (setf sel        (symset~create el var_as_const))
	  (symset~slot-reset! sel var_as_const)
	  (setf formset    (res~add-clauses! (formset~create sel) reslist))
	  (setf q          (aci=queue-create formset))
			   ; queue with all NIL => P1 ^ ... ^ Pn clauses 
	  (setf consistent (aci=pebbling! formset q consistent))
	  (when aci**output
		(progn
		  (format T "~A~%" formset)      ; output
		  (res~print reslist)
		  (format T "~%")))
	  (if aci**unifier                       
	      (if consistent
		  (aci=unifier formset el)       ; computation of unifier
;		  (progn
;		   (format T "~%~A~%" (aci=unifier formset el))
;		   (test~unifier-correct
;		     el var_as_const (aci=unifier formset el) sel))
		nil)
	    consistent))
      (aci=slot-reset! sel el var_as_const))))   ; reset of sym-slots


(defun aci=pebbling! (formset q consistent)
  ; 
  ; Algorithm:
  ;
  ; Initially the queue contains all clauses consistig only of positive
  ; literals. In the every-expression each clause on the queue is popped 
  ; and processed as follows: Let clause1 be the current head of the queue.
  ; The positive literals are evaluated to true and the counter k of clauses
  ; containing the positive literals of clause1 as negative literals is
  ; decremented by one for each true evaluated negative literal. If the
  ; counter reaches zero, the clause is entered into the queue. 
  ; 
  (let ((formset*false (cons nil nil)))
    (if (and
	 (not (aci=queue-empty-p q))   
	 consistent)
	(progn
	  (setf consistent
		(every #'     
		 (lambda (lit)     
		   (if (null (car lit))
		       (if (equal lit formset*false)
			   nil
			 (progn
			   (setf (car lit) T)
			   (mapc #'
			    (lambda (clause)
			      (decf (car clause))
			      (when (zerop (car clause))
				    (aci=queue-inq! q clause)))
			    (cdr lit))
			   T))
		     T))	     
		 (cdr (aci=queue-deq! q))))
	  (aci=pebbling! formset q consistent))
      consistent)))

;+----------------------------------------------------------------------------+
;| slot-reset-functions                                                       |
;+----------------------------------------------------------------------------+

(defun aci=slot-reset! (sel el var_as_const)
  (if (null sel)                               ; if sel not computed
      (progn     				 
	(aci=slot-reset_el! el)      ; reset on input-structure and
	(aci=slot-reset_var_as_const var_as_const))
				       ; reset marked var_as_const
    (symset~slot-reset! sel var_as_const)))  ; else reset on sel and 
                                       ; reset marked var_as_const
    
(defun aci=slot-reset_el! (el) ; reset on input-structure (equation-list)
  (if (null el)
      nil
    (progn
      (aci=slot-reset_term! (caar el))
      (aci=slot-reset_term! (cdar el))
      (aci=slot-reset_el! (cdr el)))))


(defun aci=slot-reset_term! (term) ; reset term
  (if (appl~p term)
      (mapc #'aci=slot-reset_term! (appl~arguments term)) 
    (progn
      (sym~set-label! term nil)                 ; reset sym~label-slot
      (sym~set-symbol-binding! term nil))))     ; reset sym~symbol-binding-slot
		
(defun aci=slot-reset_var_as_const (var_as_const) ; reset marked var_as_const
  (mapc #'aci=slot-reset_term! var_as_const))

;+----------------------------------------------------------------------------+
;| queue                                                                      |
;+----------------------------------------------------------------------------+

(defun aci=queue-new ()    ; value: empty queue
  (cons nil nil))

(defun aci=queue-inq! (queue element)     ; adds element to queue
  (let ((tale (list element)))
    (if (aci=queue-empty-p queue)
	(progn
	  (setf (car queue) tale)
	  (setf (cdr queue) tale))
      (progn
	(setf (cddr queue) tale)
	(setf (cdr  queue) tale)))))
	
(defun aci=queue-deq! (queue)             ; removes head from queue
  (let ((element (caar queue)))
    (setf (car queue) (cdar queue))
    (when (null (car queue))
	  (setf (cdr queue) nil))
    element))

(defun aci=queue-deq (queue)              ; value: head of queue
  (caar queue))

(defun aci=queue-create (formset)         ; creates queue with all
  (let                                    ; NIL => P1 ^ ... ^ Pn clauses
					  ; of formset
      ((q (aci=queue-new)))
    (mapc #'(lambda (clause)
	      (when (zerop (car clause)) (aci=queue-inq! q clause)))
    (formset~clauselist formset))
    q))
    
(defun aci=queue-empty-p (queue)          ; queue empty?
  (null (car queue)))

;+----------------------------------------------------------------------------+
;| computation of unifier                                                     |
;+----------------------------------------------------------------------------+

(defun aci=unifier (formset el)
  (declare (edited  "30-MAR-1995")
	   (authors Felix)
	   (input   "[formset+formset], equation list")
	   (effect  "None.")
	   (value   "unifier of el [subst+substitution]"))
  ;
  ; For each var v and all not to true evaluated literals
  ; P(v,c1), ..., P(v,cn) a substitution {v |-> (f cn (f cn-1 ...(f c2 c1)))}
  ; is created and added to unifier
  ;
  (let
      ((tab (formset~tab formset))
       (vl (symset~var_list (formset~sel formset)))
       (cl (symset~const_list (formset~sel formset)))
       (unifier (subst~create nil nil))         ; empty substitution
       (function (aci=take-function el)))       ; function appearing in el
    (mapc #'
     (lambda (var)
       (subst~insert-component!
	var
	(aci=term-create 
	 (mapcan #'         ; list of constants const appering in not true
	                    ; evaluated literals P(var, const)
	  (lambda (const) (aci=lit-true-check tab var const))
	  cl)
	 function)          ; function for term-creation
	unifier)) 
     vl)
    unifier))

(defun aci=lit-true-check (tab var const)
  (declare (edited  "30-MAR-1995")
	   (authors Felix)
	   (input   "tab var const")
	   (effect  "None.")
	   (value   "If P(var,const) then (list const)"
		    "                else nil"))
  (if (null (car (formset~aref tab var const)))
      (list const)
    nil))

(defun aci=term-create (constlist function)
  (declare (edited  "30-MAR-1995")
	   (authors Felix)
	   (input   "list of constants (c1 ... cn), function")
	   (effect  "None.")
	   (value   "term (f cn (f cn-1 .... (f c2 c1)))"))
  (if (null (cdr constlist))
      (car constlist)
    (appl~create function
		 (list
		  (car constlist)
		  (aci=term-create (cdr constlist) function)))))

(defun aci=take-function (el)
  (declare (edited  "30-MAR-1995")
	   (authors Felix)
	   (input   "equation list")
	   (effect  "None.")
	   (value   "function appearing in el"))
  (cond
   ((appl~p (caar el)) (appl~function (caar el)))
   ((appl~p (cdar el)) (appl~function (cdar el)))
   (T (aci=take-function (cdr el)))))
	
			




