(in-package "KEIM")



;;;+--------------------------------------------------------------------------+
;;;| label                                                                    |
;;;+--------------------------------------------------------------------------+


(defvar cpaci*debug-stream nil)

(defgeneric cpaci=label (term)
  (:method ((term sym+const))
    (unicp~term-label term))
  (:method ((var sym+var))
    (declare (special cpaci*theory))
    (unicp~var-label var cpaci*theory)))

(defgeneric cpaci=set-label! (term label)
  (:method ((term sym+const) label)
    (unicp~term-set-label! term label))
  (:method ((var sym+var) label)
    (declare (special cpaci*theory))
    (setf (unicp~var-label var cpaci*theory) label)))

(defun cpaci=constants (up)
  (declare (special cpaci*theory))
  (remove-if-not #'(lambda (var)
		     (unicp~constant-p var cpaci*theory))
		 (cpaci=sel-var_list (cpaci=formset-sel (cpaci~up-formset up)))))

;;;+--------------------------------------------------------------------------+
;;;| normalization                                                            |
;;;+--------------------------------------------------------------------------+

;;; Because ACI is the equational theory of sets, the equations of the
;;; unification problem will be transformed into equations between sets
;;; of variables, where each element occurs only once on each side of
;;; equation:
;;;
;;;            y = (f u u)
;;;            (f (f x z) (f x y)) = y
;;;            (f (f z z) (f x y)) = (f z (f u z))
;;;
;;;            will be transformed into
;;;
;;;            - var-list:   (U Z Y X)
;;;            - const-list: = var-list
;;;                            (for each var an  constant
;;;                            is created) 
;;;
;;;            - *** equations *** 
;;;
;;;              (Y) =  (U)                            
;;;              (Z X Y) (A) =  (Y)                    
;;;              (Y X) (A) =  (Z) (A)                  


(defstruct (cpaci+sel  (:conc-name cpaci=sel-)
		       (:constructor cpaci=sel-make)
 		       (:print-function cpaci=sel-print))
  equation_list          ; "list of [cpaci+eq]" 
  var_list)              ; "list of occuring variables"

(defstruct (cpaci+eq (:conc-name cpaci=eq-)
		     (:constructor cpaci=eq-make))
  var)                   ; (cons
                         ;    (list of var on left  side of equation)
                         ;    (list of var on right side of equation)

(defun cpaci=sel-print (sel stream c)
  (declare ; (ignore c)
	   (special cpaci*theory))
  (let ((variables (reverse (cpaci=sel-var_list sel))))
    (format stream "~%~% equations:")
    (mapc #'(lambda (eq)
	      (format stream "~12T")
	      (cpaci=eq-print eq stream c))
	  (reverse (cpaci=sel-equation_list sel)))
    (format stream "~% variables:  ~A"
	    (remove-if #'(lambda (var) (unicp~constant-p var cpaci*theory))
		       variables))
    (format stream "~% constants:  ~A"
	    (remove-if-not #'(lambda (var) (unicp~constant-p var cpaci*theory))
			   variables))
    (mapc #'(lambda (var)
	      (format stream "~%   ~A ~A ~18,2T< ~A ~30,5T=/= ~A"
		      (if (unicp~constant-p var cpaci*theory) "Const:" "Var:  ")
		      var
		      (intersection variables (unicp~var-label-less var))
		      (intersection variables (unicp~vars-discriminated var))))
	  variables)
    (values)))

(defun cpaci=eq-print (eq stream c)
  (declare (ignore c))
  (when (not (null (car (cpaci=eq-var eq))))
    (format stream " ~A" (reverse (car (cpaci=eq-var eq)))))
  (format stream " =")
  (when (not (null (cdr (cpaci=eq-var eq))))
    (format stream " ~A~%" (reverse (cdr (cpaci=eq-var eq)))))) 

(defun cpaci=sel-new ()
  (cpaci=sel-make :equation_list nil
		  :var_list nil))

(defun cpaci=eq-new ()
  (cpaci=eq-make :var (cons nil nil)))

(defun cpaci~sel-create (tl1 tl2)
  (declare (edited  "22-NOV-1995")
	   (authors Felix)
	   (input   "two lists of terms (s1..sn) (t1...tn)"
		    "representing a list of equations (s1=t1...sn=tn)")
	   (effect  "")
	   (value   "sel-structure:"
		    "equations transformed into equations between sets"
		    "of the occuring symbols, i.e. variables"))
  (declare (special cpaci*theory))
  (let
      ((sel (cpaci=sel-new))
       (counter 1))
    (mapc #'
     (lambda (t1 t2)
       (let
	   ((seq (cpaci=eq-new))) 
	 (cpaci=term! (list t1) sel seq counter)
	 (cpaci=term! (list t2) sel seq (1+ counter))
	 (push seq (cpaci=sel-equation_list sel))
	 (setf counter (+ 2 counter))))
     tl1 tl2)
    sel))

(defun cpaci=left-side-eq (counter)  ; test if counter is on left side of
  (not (zerop (mod counter 2))))     ; equation

(defun cpaci=term! (tl sel seq counter)
  (declare (special cpaci*theory))
  (if (null tl)
      seq
    
      (if (appl~p (car tl))		; if head of tl is application
	  (let
	      ((seq
		(cpaci=term! (appl~arguments (car tl)) sel seq counter)))
	    (cpaci=term!		; then continue with 
	     (cdr tl)			; tale of tl
	     sel			; modified sel
	     seq			; symset corresponding to head of tl
	     counter))			; same counter
      
	  (let* ((head (unicp~var-label-repres (car tl)))
		 (head-label (cpaci=label head)))
	
	    (progn			; else
	      (when
		  (null head-label)	; if term is not in var_list of sel
		(push head (cpaci=sel-var_list sel)))
					; then add to var_list of sel
	      (when
		  (or
		   (null head-label)	; if term (head of tl) 
		   (and			; is not in seq-side yet
		    (not (null head-label)) 
		    (> counter head-label)))
	   
		(if			; then
		 (cpaci=left-side-eq counter) ; if term is from left side of
					; equation then
		 (push head (car (cpaci=eq-var seq)))
					; add term to left-side-var-set
                 (push head (cdr (cpaci=eq-var seq))))
					; if term is from right side
					; of equation then
					; add term to right-side-var-set
		(cpaci=set-label! head counter))
					; set term-label to counter
	  
	      (cpaci=term! (cdr tl) sel seq counter))))))


(defun cpaci=variables-label-reset! (variables)
  (mapc
   #'(lambda (var) (cpaci=set-label! var nil))
   variables))

;;;+--------------------------------------------------------------------------+
;;;| create clauses                                                           |
;;;+--------------------------------------------------------------------------+
;;;
;;;
;;;  outline of cpaci+formset
;;;  ========================
;;;
;;;  +------------------------------------------------------+
;;;  | +--------------------------------------------------+ |
;;;  | |                                                  | |
;;;  | |                                                  | |
;;;  | |                      +-------------------+       | |
;;;  | |                      | +---------------+ |       | |
;;;  | |                      | |               | |       | |            |
;;;  | |                      | |               | |       | |            |
;;;  | |   +--[ P(x1,c1)  (tv | |      lit-pos) | +-> (c  | |    cl-pos)-+
;;;  | |   |    P(x2,c1)  (tv lit-neg  lit-pos) |                        |
;;;  | + --|--> P(x3,c1)  (tv lit-neg    | |  ) |                        |
;;;  |     |      .                      | |    +---> (c cl-neg  cl-pos)-+
;;;  |     |      .                      | |                             |
;;;  |     |      .                      | +------>   (c cl-neg   | |  ) | 
;;;  |     |      .                      |                        | |    |
;;;  |     |      .                      +-------->   (c cl-neg   | | |)-+
;;;  |     |                                                      | | |  |
;;;  +-----|--> P(x1,c2)  (tv lit-neg  lit-pos) <-----------------+ | |  |
;;;        |      .                                                 | |  |
;;;        |      .                                                 | |  |
;;;        +--[ P(xn,cn)  (tv lit-neg  lit-pos) <-------------------+ |  |
;;;        |    False     (:false :false lit-pos) <-------------------+  |
;;;        |                                                             |
;;;        |     * literals *                                     (clauselist)
;;;        |
;;;        |
;;;        |        variables <-----------------+
;;;        +------< +-+-+-+-+                   |     
;;;                 | | | | |                   +---- (sel)      
;;;                 +-+-+-+-+                   |
;;;                 | | | | | 'constants' <-----+    
;;;                 +-+-+-+-+
;;;                   (tab)
;;;
;;; tv : truthvalue (:true, :false ,default nil)
;;; lit-neg : list of basic Horn propositions in which P occurs as negative
;;;           literal
;;; lit-pos : list of basic Horn propositions in which P occurs as positive
;;;           literal
;;; clauselist : list of basic Horn propositions              
;;; c : number of negative literals in the basic Horn proposition
;;; premise : list of literals in the premise of the clause
;;; conclusion : list of literals in the conclusion of the clause
;;; tab : 2-dim array of literals P (var, const) which is true iff const
;;;       does not appear in (a solution for) var
;;; sel : [symset+sel]

(defstruct (cpaci+formset (:conc-name cpaci=formset-)
			  (:constructor cpaci=formset-make)
			  (:print-function cpaci=formset-print))
  sel					; "[cpaci+sel]"
  tab					; tab (see above)
  false					; literal False
  clauselist)                           

;;;+--------------------------------------------------------------------------+
;;;| literals, clauses                                                        |
;;;+--------------------------------------------------------------------------+
;;;
;;; For solving the unification problem this algorithm uses Horn clauses of
;;; the following form:
;;;
;;;           not(P1) v ... v not(Pn) v (Q1 ^ ... ^ Qm)
;;;
;;;           negative literals         positive literals
;;;
;;;           means (P1 ^ ... ^ Pn) => (Q1 ^ ... ^ Qm)      


(defmacro cpaci=cl-count (clause)   ; counter of not to True evaluated
  `(first ,clause))	            ; negative literals

(defmacro cpaci=cl-neg (clause)     ; list of negative literals in clause
  `(second ,clause))

(defmacro cpaci=cl-pos (clause)     ; list of positive literals in clause
  `(third ,clause))

(defmacro cpaci=lit-tv (lit)        ; truthvalue of literal
  `(first ,lit))	         

(defmacro cpaci=lit-neg (lit)       ; list of clauses in which lit occurs as
  `(second ,lit))                   ; negative literal

(defmacro cpaci=lit-pos (lit)       ; list of clauses in which lit occurs as
  `(third ,lit))                    ; positive literal

(defun cpaci=lit-create (value lit-neg lit-pos)
  (declare (edited  "03-FEB-1995")
	   (authors Felix)
	   (input   "value [boolean] : literal is evaluated to true"
		    "lit-neg: list of basic Horn propositions" 
                    "in which literal occurs as negative (left side)"
		    "lit-pos: list of basic Horn propositions" 
                    "in which literal occurs as positive (right side)")
	   (effect  "None.")
	   (value   "below"))
  (list value lit-neg lit-pos))

(defun cpaci=lit-pos-entry! (var const clause tab)
  (declare (edited "07-MAR-1995")
	   (authors Felix)
	   (input   "variable, constant, basic Horn clause, tab (see above)")
	   (effect  "clause will be added to lit-pos"
		    "of literal P(var, const)")
	   (value   "None.")
	   (special cpaci*theory))
  (push
   clause
   (cpaci=lit-pos (cpaci=aref tab var const))))

(defun cpaci=lit-neg-entry! (var const clause tab)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "variable, constant, basic Horn clause, tab (see above)")
	   (effect  "clause will be added to lit-neg"
		    "of literal P(var, const)")
	   (value   "None.")
	   (special cpaci*theory))
  (push
   clause
   (cpaci=lit-neg (cpaci=aref tab var const))))

;;;+--------------------------------------------------------------------------+
;;;| tab                                                                      |
;;;+--------------------------------------------------------------------------+

(defun cpaci=tab-create! (sel)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "[cpaci+sel]")
	   (effect  "terms of sel will be labeled")
	   (value   "tab (see above)")
	   (special cpaci*theory))
  (let*
      ((v (length (cpaci=sel-var_list sel)))
       (a (make-array (list v v))))
    (progn
      (loop for i from 0 to (1- v) do
	    (loop for j from 0 to (1- v) do
		  (setf (aref a i j) (cpaci=lit-create nil nil nil))))
      (cpaci=number-termlist! (cpaci=sel-var_list sel) 0)
      a)))

(defun cpaci=number-termlist! (tl counter)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "list of terms")
	   (effect  "termlist will be numbered increasingly by using"
		    "label-slots beginning with counter")
	   (value   "None.")
	   (special cpaci*theory))
  (if (null tl)
      nil
      (progn
	(cpaci=set-label! (car tl) counter)
	(cpaci=number-termlist! (cdr tl) (+ 1 counter)))))

(defun cpaci=tab-print (formset &optional (stream cpaci*debug-stream))
  (let ((tab (cpaci=formset-tab formset))
	(vl (cpaci=sel-var_list (cpaci=formset-sel formset))))
    (format stream "~%~%")
    (mapc #'(lambda (const)
	      (format stream "~9,6T~A" const))
	  vl)
    (format stream " ~%")
    (mapc #'(lambda (var)
	      (format stream "   ~A" var)
	      (mapc #'(lambda (const)
			(format stream "~9,6T~A" (cpaci=lit-tv (cpaci=aref tab var const))))
		    vl)
	      (format stream " ~%"))
	  vl)
    ))

(defun cpaci=aref (tab var const)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "tab (see above), variable, constant")
	   (effect  "None.")
	   (value   "corresponding literal P(var,const)")
	   (special cpaci*theory))
  (aref tab (cpaci=label var) (cpaci=label const)))

;;;+--------------------------------------------------------------------------+
;;;| formset-create                                                           |
;;;+--------------------------------------------------------------------------+


;;;----------------------------------------------------------------------------
;;; To check the solvability of the equation set, we form for every variable x
;;; and for every constant c in sel a propositional variable P(x,c)
;;; which is true iff a does not appear in x. Now corresponding
;;; to each equation we can form a set of Horn clauses relating these
;;; propositional variables. For instance, let uva = wxb be an equation, where
;;; a and b are constants and the rest are variables.
;;;
;;; 1. If c is a constant and does not appear in the equation, then we get
;;;    the clauses:
;;;
;;;    P(u,c) ^ P(v,c) => P(w,c) ^ P(x,c)
;;;    P(w,c) ^ P(x,c) => P(u,c) ^ P(v,c)
;;;
;;; 2. We need to enforce each variable to have at least one constant, or
;;;    the clause 
;;;
;;;    P(x,a) ^ P(x,b) ^ P(x,c) => False
;;;
;;;    must be satisfied for each variable x in the equation.
;;;
;;;----------------------------------------------------------------------------


(defun cpaci=const-not-in-eq-left! (c eq tab)
  ;;
  ;; P(v1,c) ^ .... ^ P(vn,c) => P(u1,c) ^ .... ^ P(um,c)
  ;; for all var vi on left side and all var ui on right side
  ;;  
  (declare (edited  "21-FEB-1995")
	   (authors Felix)
	   (input   "constant, equation, tab (see above)")
	   (effect  "adding literals P(v,c) to tab for each var v"
		    "from left side of equation")
	   (value   "first clause for constant c not appearing in equation")
	   (special  cpaci*theory))
  (let ((clause
	 (list
	  (length (car (cpaci=eq-var eq)))
	  nil
	  nil)))
    (mapc #'
     (lambda (v)
       (cpaci=lit-neg-entry! v c clause tab)
       (push (cpaci=aref tab v c) (cpaci=cl-neg clause)))
     (car (cpaci=eq-var eq)))
    (mapc #'
     (lambda (v)
       (cpaci=lit-pos-entry! v c clause tab)
       (push (cpaci=aref tab v c) (cpaci=cl-pos clause)))
     (cdr (cpaci=eq-var eq)))
    clause))

(defun cpaci=const-not-in-eq-right! (c eq tab)
  ;;
  ;; P(u1,c) ^ .... ^ P(un,c) => P(v1,c) ^ .... ^ P(vm,c)
  ;; for all var vi on left side and all var ui on right side
  ;;  
  (declare (edited  "21-FEB-1995")
	   (authors Felix)
	   (input   "constant, equation, tab (see above)")
	   (effect  "adding literals P(v,c) to tab for each var v"
		    "from right side of equation")
	   (value   "second clause for constant c not appearing in equation")
	   (special  cpaci*theory))
  (let ((clause
	 (list
	  (length (cdr (cpaci=eq-var eq)))
	  nil
	  nil)))
    (mapc #'
     (lambda (v)
       (cpaci=lit-neg-entry! v c clause tab)
       (push (cpaci=aref tab v c) (cpaci=cl-neg clause)))
     (cdr (cpaci=eq-var eq)))
    (mapc #'
     (lambda (v)
       (cpaci=lit-pos-entry! v c clause tab)
       (push (cpaci=aref tab v c) (cpaci=cl-pos clause)))
     (car (cpaci=eq-var eq)))
    clause))

(defun cpaci=clean-clauselist! (clauselist)
  ;;
  ;; removes all clauses P1 ^ ... ^ Pn => NIL from clauselist
  ;;
  (mapcan #'
   (lambda (clause)
     (when (not (null (cpaci=cl-pos clause)))
       (list clause)))
   clauselist))

;;;+--------------------------------------------------------------------------+

(defun cpaci~formset-create (sel)
  (declare (edited  "07-MAR-1995")
	   (authors Felix)
	   (input   "sel [cpaci+sel]")
	   (effect  "")
	   (value   "corresponding set of Horn formulae [cpaci+formset]"
		    "outlined above")
	   (special  cpaci*theory))
  (let*
      ((tab (cpaci=tab-create! sel))
       (vl (cpaci=sel-var_list sel))
       (length-vl (length vl))
       (el  (cpaci=sel-equation_list sel))
       (cpaci*false (cpaci=lit-create :false :false nil))
       (clauselist nil))
    (declare (special cpaci*false))
    (labels
	((const_formulae! (c eq)         ; if const is not in equation (1.)  
	   (progn
	     (push                       ; add for each side of equation
	      (cpaci=const-not-in-eq-right! c eq tab)
	      clauselist)
	     (push                       ; corresponding clause to clauselist
	      (cpaci=const-not-in-eq-left! c eq tab)
	      clauselist)))
	 (var_formulae! (v)
	   (let ((clause (list
			  length-vl
			  nil
			  (list cpaci*false))))
	     (push clause (cpaci=lit-pos cpaci*false))
	     (push clause clauselist)
	     (mapc #'(lambda (c)
		       (cpaci=lit-neg-entry! v c clause tab)
		       (push
			(cpaci=aref tab v c)
			(cpaci=cl-neg clause)))
		   vl))))
      ;; add for v clause to clauselist (2.)
      ;; create for each constant c a literal
      ;; P(v,c) and add clause to its lit-pos (see outline),
      ;; also add each literal to premise-list of the clause
      ;;======================================================================
      (mapc #'(lambda (eq)
		(mapc #'(lambda (c) (const_formulae! c eq)) vl)) el)
      ;; create formula for each constant in vl and each equation
      ;; in el (1.)
     
      (mapc #'(lambda (v) (var_formulae! v)) vl)
      ;; create formula for each variable in vl (2.)
     
      (setf clauselist (cpaci=clean-clauselist! clauselist))
      ;; remove redundant formulae from clauselist
     
      (cpaci=formset-make :sel        sel
			  :tab        tab
			  :false      cpaci*false
			  :clauselist clauselist))))

(defun cpaci=formset-print (formset stream c)
  (declare (special  cpaci*theory))
  (let* ((tab (cpaci=formset-tab formset))
	 (sel (cpaci=formset-sel formset))
	 (clauselist (cpaci=formset-clauselist formset))
	 (vl (cpaci=sel-var_list (cpaci=formset-sel formset))))

    
    (cpaci=sel-print sel stream c)
    (cpaci=tab-print formset stream)
    
    (labels ((print-clause (clause)
	       (if (equal (cpaci=lit-neg (first (cpaci=cl-pos clause)))
			  :false)
		   (format stream "~% ~A ~A => FALSE"
			   (cpaci=cl-count clause)
			   ;; number of not true evaluated negative literals in clause
			   (premise clause))
		   (format stream "~% ~A ~A => ~A"
			   (cpaci=cl-count clause)
			   (premise clause)
			   (conclusion clause))))
	     
	     (premise (clause)
	       ;; (premise clause) is the list of literals from (cpaci=cl-neg clause)
	       (mapcan #'(lambda (c)
			   (mapcan #'(lambda (v)
				       (when (member (cpaci=aref tab v c)
						     (cpaci=cl-neg clause))
					 (list (cons v c))))
				   vl))
		       vl))
	     
	     (conclusion (clause)
	       ;; (conclusion clause) is the list of literals from (cpaci=cl-pos clause) 
  	       (mapcan #'(lambda (c)
			   (mapcan #'(lambda (v)
				       (when (member (cpaci=aref tab v c)
						     (cpaci=cl-pos clause))
					 (list (cons v c))))
				   vl))
		       vl)))

      (mapcar #'print-clause clauselist))))


;;;+--------------------------------------------------------------------------+
;;;| constant restrictions                                                    |
;;;+--------------------------------------------------------------------------+

;;; constant-restrictions are represented as follows:
;;;
;;; (list
;;;   (list constant_1   var_1 ... var_n1)      (restr) -+
;;;                     :                                | (const-restr)
;;;   (list constant_m   var_1 ... var_nm))     (restr) -+

(defun cpaci~restr-add-clauses! (formset const-restr)
  (declare (edited  "22-NOV-1995")
	   (authors Felix)
	   (input   "formset [cpaci+formset], list of constant-restrictions")
	   (effect  "adds for each restriction V(c) = (x1 x2)"
		    "clauses NIL => P(c x1), NIL => P(c x2)"
		    "to (cpaci=formset-clauselist formset)")
	   (value   "")
	   (special  cpaci*theory))
  ;(format cpaci*debug-stream "~% cpaci-restr-add-clauses!: ~A" const-restr)
  (let ((tab (cpaci=formset-tab formset)))
    (mapc #'(lambda (restr)
	      (let ((const (unicp~cr-const restr)))
		(mapc #'(lambda (var)
			  #+weg(format cpaci*debug-stream "~%    add-lit: ~A:~A,  ~A:~A"
				       var (cpaci=label var)
				       const (cpaci=label const))
			  (when (and (cpaci=label const)
				     (cpaci=label var)
				     (unicp~vars-discriminated-p var const))
			    (let ((clause
				   (list 0 nil (list (cpaci=aref tab var const)))))
			      (push clause (cpaci=formset-clauselist formset))
			      (cpaci=lit-pos-entry! var const clause tab))))
		    (unicp~cr-varlist restr))))
	  const-restr)
    formset))

(defun cpaci=restr-print (const-restr &optional (stream cpaci*debug-stream))
  (format stream "~% constant - restrictions")
  (if (null const-restr)
      (format stream "~%   ~A" nil)
      (mapc #'
       (lambda (restr)
	 (format stream "~%   V(~A) =  ~A" (car restr) (cdr restr)))
       const-restr)))

;;;+--------------------------------------------------------------------------+
;;;| constants (ind <> ACI)                                                   |
;;;+--------------------------------------------------------------------------+

(defun cpaci~const-add-clauses! (formset constants)
  (declare (edited  "22-NOV-1995")
	   (authors Felix)
	   (input   "formset [cpaci+formset], list of variables x with"
		    "ind(x) <> ACI")
	   (effect  "adds for each 'variable' x from constants"
		    "a clause P(x x) => False and a clause"
		    "NIL => P(x y) for each y<>x (dicriminated)")
	   (value   ""))
  (declare (special  cpaci*theory))
  (let ((tab (cpaci=formset-tab formset))
	(vl (cpaci=sel-var_list (cpaci=formset-sel formset)))
	(cpaci*false (cpaci=formset-false formset)))
    (mapc #'(lambda (var)
	      (when (cpaci=label var)
		(let ((clause (list
			       1
			       (list (cpaci=aref tab var var))
			       (list cpaci*false))))
		  (push clause (cpaci=lit-pos cpaci*false))
		  (push clause (cpaci=formset-clauselist formset))
		  (cpaci=lit-neg-entry! var var clause tab))
		(mapc #'(lambda (const)
			  (when (unicp~vars-discriminated-p const var)
			    (let ((clause (list
					   0
					   nil
					   (list (cpaci=aref tab var const)))))
			      (push clause (cpaci=formset-clauselist formset))
			      (cpaci=lit-pos-entry! var const clause tab))))
		      vl)))
	  constants)
    formset))

;;;+--------------------------------------------------------------------------+
;;;| dicrimination                                                            |
;;;+--------------------------------------------------------------------------+

(defun cpaci~discrim-add-clauses! (formset discrim-list)
  (declare (edited  "26-NOV-1996")
	   (authors Richts)
	   (input   )
	   (effect  )
	   (value   )
	   (special cpaci*theory))
  (let ((tab (cpaci=formset-tab formset)))
    (mapc #'(lambda (discrim)
	      (let ((var1 (unicp~id-left discrim))
		    (var2 (unicp~id-right discrim)))
		(when (and (cpaci=label var1)
			   (cpaci=label var2))
		  (when (or (unicp~constant-p var1 cpaci*theory)
			    (unicp~vars-lesseq-p var1 var2))
		    (let ((clause (list
				   0
				   nil
				   (list (cpaci=aref tab var1 var2)))))
		      (push clause (cpaci=formset-clauselist formset))
		      (cpaci=lit-pos-entry! var1 var2 clause tab)))
		  (when (or (unicp~constant-p var2 cpaci*theory)
			    (unicp~vars-lesseq-p var2 var1))
		    (let ((clause (list
				   0
				   nil
				   (list (cpaci=aref tab var2 var1)))))
		      (push clause (cpaci=formset-clauselist formset))
		      (cpaci=lit-pos-entry! var2 var1 clause tab))))))
	  discrim-list)
    formset))

;;;+--------------------------------------------------------------------------+
;;;| pebbling                                                                 |
;;;+--------------------------------------------------------------------------+

(defun cpaci~false-propagate! (formset qf consistent)
  (declare (edited  "27-NOV-1995")
	   (authors Felix)
	   (input   "formset [cpaci+formset]"
		    "queue with all P => False clauses"
		    "predicate consistent")
	   (effect  "For each clause from the queue, i.e. each clause where"
		    "the right-hand side is False and there is only one"
		    "literal P on the left-hand side that has not been set"
		    "to True, P and all right-hand sides containig P are"
		    "set to False. Furthermore all literals P being set to"
		    "False this way are stored in cpaci*falselit.")
	   (value   "")
	   (special cpaci*falselit))
  (let ((cpaci*false (cpaci=formset-false formset)))
    (if (and
	 (not (cpaci=queue-empty-p qf))   
	 consistent)
	(let ((lit (some #'(lambda (lit)
			     (if (equal (cpaci=lit-tv lit) :true)
				 nil
				 lit))
			 (cpaci=cl-neg (cpaci=queue-deq! qf)))))
	  (if (or
	       (null lit)
	       (eq (cpaci=lit-tv lit) :false))
	      (cpaci~false-propagate! formset qf consistent)
	      (progn
		(setf (cpaci=lit-tv lit) :false)
		(push lit cpaci*falselit)
		(mapc #'(lambda (clause)
			  (when (not (eq (first (cpaci=cl-pos clause)) cpaci*false))
			      (setf (cpaci=cl-pos clause) (list cpaci*false))
			      (push clause (cpaci=lit-pos cpaci*false)))
			  (when (equal (cpaci=cl-count clause) 1)
			    (cpaci=queue-inq! qf clause)))
		      (cpaci=lit-pos lit))
		(cpaci~false-propagate! formset qf consistent))))
	consistent)))


(defun cpaci~true-propagate! (formset qt consistent)
  (declare (edited  "27-NOV-1995")
	   (authors Felix)
	   (input   "- formset [cpaci+formset]"
		    "- queue with all clauses consisting only of positive"
		    "  literals and all clauses, where the counter of not"
		    "  to true evaluated negative literals is zero"
		    "- predicate consistent")
	   (effect  "Each clause on the queue is popped and processed as"
		    "follows: Let clause1 be the current head of the queue."
		    "The positive literals are evaluated to true and the"
		    "counter k of clauses containing the positive literals"
		    "of clause1 as negative literals is decremented by one"
		    "for each true evaluated negative literal. If the counter"
		    "reaches zero, the clause is entered into the queue.")
	   (value   ""))
  (let ((cpaci*false (cpaci=formset-false formset)))
    (if (and
	 (not (cpaci=queue-empty-p qt))   
	 consistent)
	(progn
	  (setf consistent
		(every #'     
		 (lambda (lit)
		   (cond
		    ((eq lit cpaci*false) nil)
		    ((equal (cpaci=lit-tv lit) :false) nil)
		    ((null (cpaci=lit-tv lit))
		     (setf (cpaci=lit-tv lit) :true)
		     (mapc #'
		      (lambda (clause)
			(decf (cpaci=cl-count clause))
			(when (zerop
			       (cpaci=cl-count clause))
			  (cpaci=queue-inq!
			   qt clause)))
		      (cpaci=lit-neg lit))
		     T)
		    (T T)))
		 (cpaci=cl-pos (cpaci=queue-deq! qt))))
	  (cpaci~true-propagate! formset qt consistent))
      consistent)))
  
;;;+--------------------------------------------------------------------------+
;;;| queue                                                                    |
;;;+--------------------------------------------------------------------------+

(defun cpaci=queue-new ()		; value: empty queue
  (cons nil nil))

(defun cpaci=queue-inq! (queue element)	; adds element to queue
  (let ((tale (list element)))
    (if (cpaci=queue-empty-p queue)
	(progn
	  (setf (car queue) tale)
	  (setf (cdr queue) tale))
	(progn
	  (setf (cddr queue) tale)
	  (setf (cdr  queue) tale)))))

(defun cpaci=queue-deq! (queue)		; removes head from queue
  (let ((element (caar queue)))
    (setf (car queue) (cdar queue))
    (when (null (car queue))
      (setf (cdr queue) nil))
    element))

(defun cpaci=queue-deq (queue)		; value: head of queue
  (caar queue))

(defun cpaci~queue-true-create (formset)
  (declare (edited  "01-DEC-1995")
	   (authors Felix)
	   (input   "formset [cpaci+formset]")
	   (effect  "")
	   (value   "queue with all NIL => P1 ^ ... ^ Pn clauses"
		    "of formset, i.e all clauses with counter of not to"
		    "true evaluated negative literals = 0"))                   
  (let					
      ((q (cpaci=queue-new)))            
    (mapc #'(lambda (clause)             
	      (when (zerop (cpaci=cl-count clause))
		(cpaci=queue-inq! q clause)))
	  (cpaci=formset-clauselist formset))
    q))

(defun cpaci~queue-false-create (formset)
  (declare (edited  "01-DEC-1995")
	   (authors Felix)
	   (input   "formset [cpaci+formset]")
	   (effect  "")
	   (value   "creates queue with all P => False clauses of formset"))
  (let					
      ((q (cpaci=queue-new))
       (cpaci*false (cpaci=formset-false formset)))
    (mapc #'(lambda (clause)
	      (when
		  (and
		   (equal (cpaci=cl-count clause) 1)
		   (eq (first (cpaci=cl-pos clause)) cpaci*false))
		(cpaci=queue-inq! q clause)))
	  (cpaci=lit-pos (cpaci=formset-false formset)))
    q))

(defun cpaci=queue-empty-p (queue)	; queue empty?
  (null (car queue)))

;;;+--------------------------------------------------------------------------+
;;;| aci - unification                                                        |
;;;+--------------------------------------------------------------------------+

(defun cpaci~values! (up &optional (constants nil c-supplied-p) discriminated)
  ;; This function computes the incremental output by inspecting the
  ;; actually modified truthvalues of the literals. Besides the normal 
  ;; rules there are computations to detect further possible identifications
  ;; in cpaci=values1 and cpaci=values2.
  ;;
  (declare (special  cpaci*theory cpaci*falselit))
  (let* ((formset (cpaci~up-formset up))
	 (tab (cpaci=formset-tab formset))
	 (vl (cpaci=sel-var_list (cpaci=formset-sel formset)))
	 (el (cpaci=sel-equation_list (cpaci=formset-sel formset)))
	 (new-const-restr nil)
	 (new-ind-aci nil)
	 (new-identified (append (cpaci=values1 vl tab)
				 (cpaci=values2
				  (if c-supplied-p
				      constants
				      (cpaci=constants up))
				  el))))
    (labels ((cpaci=new-cr (var const lit)
	       (when (and (not (eq var const))
			  (eq (cpaci=lit-tv lit) :false)
			  (unicp~constant-p const cpaci*theory))
		 (push (unicp~cr-create var (list const)) new-const-restr)))
	     (cpaci=new-ind (var const lit)
	       (when (and (not (eq var const))
			  (eq (cpaci=lit-tv lit) :false)
			  (unicp~constant-p const cpaci*theory)
			  (unicp~vars-discriminated-p var const))
		 (push var new-ind-aci))))
      (mapc #'(lambda (var)
		(mapc #'(lambda (const)
			  (let ((lit (cpaci=aref tab var const)))
			    (when (member lit cpaci*falselit)
			      (cpaci=new-cr var const lit)
			      (cpaci=new-ind var const lit))))
		      vl)
		(when (and (eq :true (cpaci=lit-tv (cpaci=aref tab var var)))
			   (not (unicp~var-label-theory var)))
		  (push var new-ind-aci)))
	    vl)
      (mapc #'(lambda (const)
		(when (cpaci=label const)
		  (mapc #'(lambda (var)
			    (let ((lit (cpaci=aref tab var const)))
			      (cpaci=new-cr var const lit)
			      (cpaci=new-ind var const lit)))
			vl)))
		constants)
      (mapc #'(lambda (discrim)
		(let ((var1 (unicp~id-left discrim))
		      (var2 (unicp~id-right discrim)))
		  (when (and (cpaci=label var1)
			     (cpaci=label var2))
		    (let ((lit12 (cpaci=aref tab var1 var2))
			  (lit21 (cpaci=aref tab var2 var1)))
		      (cpaci=new-ind var1 var2 lit12)
		      (cpaci=new-ind var2 var1 lit21)))))
	    discriminated))
    (values up
	    new-identified
	    nil
	    (delete-duplicates new-ind-aci)
	    nil
	    new-const-restr
	    nil)))


(defun cpaci=values1 (varlist tab)
  ;; (1) If there is a row in (cpaci=formset-tab formset) with all literals
  ;;     set to true except one literal P(v,c) set to false, then v and c
  ;;     are identified.
  (mapcan #'
   (lambda (var)
     (let ((false-lits 0)
	   temp-identified)
       (when (and (every #'(lambda (const)
			     ;; searching for a row (true ... true false true ... true)
			     ;; in (cpaci=formset-tab formset)
			     (let ((lit (cpaci=aref tab var const)))
			       (cond ((null (cpaci=lit-tv lit)) nil)
				     ((and (eql false-lits 0)
					   (equal  (cpaci=lit-tv lit) :false))
				      (setf false-lits 1)
				      (setf temp-identified (unicp~id-create var const)))
				     ((and (eql false-lits 1)
					   (equal  (cpaci=lit-tv lit) :false))
				      nil))))
			 varlist)
		  (eql false-lits 1)
		  (not (unicp~vars-identified-p
			(unicp~id-left temp-identified)
			(unicp~id-right temp-identified))))
	 (list temp-identified))))
     varlist))

(defun cpaci=values2 (constants el)
  ;; (2) If there is an equation in the problem with only one single constant 
  ;;     on one side, then this constant is identified with all symbols
  ;;     on the other side of the equation.
  (mapcan #'(lambda (const)
	      (mapcan #'(lambda (eq)
			  (let ((var-left (car (cpaci=eq-var eq)))
				(var-right (cdr (cpaci=eq-var eq))))
			    (cond ((and (cpaci=length-one-p var-left)
					(term~equal const (first var-left)))
				   (mapcar #'(lambda (var)
					       (unicp~id-create const var))
					   var-right))
				  ((and (cpaci=length-one-p var-right)
					(term~equal const (first var-right)))
				   (mapcar #'(lambda (var)
					       (unicp~id-create const var))
					 var-left))
				  (t nil))))
		      el))
	  constants))


(defun cpaci=length-one-p (list)
  (and (not (null list))
       (null (rest list))))

    
(defclass cpaci+up (unicp+elementary-up)
  ((formset  :initform nil :accessor cpaci~up-formset))
  (:documentation "This is the class of aci-up."))

(defun cpaci=create-up (theory general-up equations)
  (declare (special cpaci*theory))
  ;(print (unicp~variables general-up)) (print (mapcar #'cpaci=label (unicp~variables general-up)))
  (cpaci=variables-label-reset! (unicp~variables general-up))
  (let*
      ((up (make-instance 'cpaci+up
			  :general-up general-up
			  :equations equations
			  :theory theory))
       (tl1 (unicp~eq-lhss (unicp~equations up)))
       (tl2 (unicp~eq-rhss (unicp~equations up)))
       ;; (f (format T "~A ~A~%" tl1 tl2))
       (sel (cpaci~sel-create tl1 tl2)))           ; creation of symbol sets
       ;; (f (cpaci=sel-print sel T T))
    (cpaci=variables-label-reset! (unicp~variables general-up))
    (let ((formset (cpaci~formset-create sel)))      ; creration of Horn formulae
      (setf (cpaci~up-formset up) formset)
      up)))
  

(defun cpaci=create-new-up (up)
  (declare (special cpaci*theory))
  (let* ((general-up (unicp~general-up up))
	 (equations (unicp~equations up)))
    (cpaci=create-up cpaci*theory general-up equations))) 

(defun cpaci=actual-state (up)
  (let ((cpaci*falselit nil))
    (declare (special cpaci*theory cpaci*falselit))
    (let* ((formset (cpaci~up-formset up))
	   (sel (cpaci=formset-sel formset))
	   (variables (cpaci=sel-var_list sel))
	   (actual-constants (cpaci=constants up))
	   (actual-const-restr
	    (mapcan #'(lambda (var)
			(let ((less (unicp~var-label-less var)))
			  (when less
			    (list (unicp~cr-create var less)))))
		    variables)))
      (cpaci=actual-constraints up actual-constants actual-const-restr))))
      

(defun cpaci=actual-constraints (up constants const-restr)
  (declare (special cpaci*theory))
  (let* ((formset (cpaci~up-formset up))
	 (formset (cpaci~restr-add-clauses! formset const-restr))
	 (formset (cpaci~const-add-clauses! formset constants)))
    (when (and (cpaci~true-propagate!
		formset (cpaci~queue-true-create formset) T)
	       (cpaci~false-propagate!
		formset (cpaci~queue-false-create formset) T))
      (setf (cpaci~up-formset up) formset)
      up)))
       

(defmethod unicp~theo-init ((theory th+aci) general-up equations
			    identified discriminated constants const-restr)
  (declare (edited  "23-OCT-1995")
	   (authors Felix)
	   (input   "theory [], general-up, equations")
	   (effect  "The initial set of Horn-formulae is created and"
		    "processed.")
	   (value   "see unicp~new-constr")
	   (ignore identified))

  (let ((cpaci*theory theory)
	(cpaci*falselit nil))
    (declare (special cpaci*theory cpaci*falselit))
    (let* ((up (cpaci=create-up theory general-up equations)))
      (when cpaci*debug-stream
	(cpaci=formset-print (cpaci~up-formset up) cpaci*debug-stream nil))
      (unicp~new-constr up nil discriminated constants const-restr))))
      
(defmethod unicp~new-constr ((up cpaci+up) identified discriminated constants const-restr)
  (declare (edited  "18-JUL-1995")
	   (authors Felix)
	   (input   "unification-problem [cpaci+up]"
		    "const-restr: list of constant-restrictions")
	   (effect  "")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "  solvable."
		    "2. A list of new identification of variables (i.e. sets"
		    "  of variables)."
		    "3. A list of pairs of variables which must not be"
		    "  unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "  $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "  unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "  solution is lost if $y$ is in the constant"
		    "  restrictions of $x$)."))
  (setf cpaci*debug-stream nil)
  (let ((cpaci*theory (unicp~theory up))
	(cpaci*falselit nil))
	; here all literals which are evaluated to False in this call
        ; of unicp~new-constr are stored for computation of incremental output

    (declare (special cpaci*theory cpaci*falselit))
    (when (and nil (equal "M1" (sym~name (first (th~symbols (unicp~theory up))))))
      (setf cpaci*debug-stream t)
      (format t "~%~% UP:   ~A, ~A~% ident: ~A ~% discr: ~A ~% const: ~A ~% c-res: ~A "
	      (unicp~theory up) (sym~name (first (th~symbols (unicp~theory up))))
	      identified discriminated constants const-restr)
      (cpaci==format-labels1 up)
      (cpaci==format-labels identified))
    (cond ((cpaci=important-identified-p! identified)
	   (when (and nil (equal "M1" (sym~name (first (th~symbols (unicp~theory up))))))
	     (format t "~% Neuer Graph!"))
	   (let ((new-up (cpaci=actual-state (cpaci=create-new-up up))))
	     (when cpaci*debug-stream
	       (cpaci=formset-print (cpaci~up-formset new-up) cpaci*debug-stream nil))
	     (when new-up
	       (cpaci~values! new-up))))
	
	  (t
	   (let* ((formset (cpaci~up-formset up))
		  (formset (cpaci~restr-add-clauses! formset const-restr))
		  (formset (cpaci~const-add-clauses! formset constants))
		  (formset (cpaci~discrim-add-clauses! formset discriminated))
		  (result (and (cpaci~true-propagate!
				formset (cpaci~queue-true-create formset) T)
			       (cpaci~false-propagate!
				formset (cpaci~queue-false-create formset) T))))
	     (when result
	       (setf (cpaci~up-formset up) formset))
	     (when cpaci*debug-stream
	       (cpaci=formset-print formset cpaci*debug-stream nil))
	     (if result
		 (cpaci~values! up constants discriminated)
		 nil))))))
	     

(defun cpaci=update-label (var1 var2)
  (cpaci=set-label! var2 (cpaci=label var1)))

(defun cpaci=important-identified-p! (identified)
  (some #'(lambda (id)
	    (let ((var1 (unicp~id-left id))
		  (var2 (unicp~id-right id)))
	      (cond ((null (cpaci=label var1))
		     (when (cpaci=label var2)
		       (cpaci=update-label var2 var1))
		     nil)
		    ((null (cpaci=label var2))
		     (cpaci=update-label var1 var2)
		     nil)
		    ((not (= (cpaci=label var1) (cpaci=label var2)))
		     t))))
	identified))


(defun cpaci==format-labels (identified)
  (format t "~% ")
  (mapc #'(lambda (id)
	    (format t " (~A:~A, ~A:~A)"
		    (unicp~id-left id) (cpaci=label (unicp~id-left id))
		    (unicp~id-right id) (cpaci=label (unicp~id-right id))))
	identified))

(defun cpaci==format-labels1 (up)
  (let ((vars (top~free-variables (unicp~equations up))))
    (format t "~% ")
    (mapc #'(lambda (var)
	      (format t " ~A<-~A:~A  " (unicp~var-label-repres var) var (cpaci=label var)))
	  vars)))

;;;+--------------------------------------------------------------------------+
;;;+ save/restore-choice-point                                                +
;;;+--------------------------------------------------------------------------+


(defmethod unicp~up-save-choice-point ((up cpaci+up))
  ;;
  ;; up will be restored by computing a new one completely, so nothing has to
  ;; be saved
  ;;
  nil)



(defmethod unicp~up-restore-choice-point ((up cpaci+up))
  ;;
  ;; computes by reading the actual identifications an restrictions
  ;; from the restored variable-slots a complete new up which
  ;; is equal to the old one
  ;;
  (let ((cpaci*theory (unicp~theory up)))
    (declare (special cpaci*theory))
    (let ((new-up (cpaci=create-new-up up)))
      (cpaci=actual-state new-up))))
	 


  
;;;+--------------------------------------------------------------------------+
;;;+ unicp~solvable-p                                                         +
;;;+--------------------------------------------------------------------------+

(defmethod unicp~solvable-p ((up cpaci+up))
  t)


;-----------------------------------------------------------------------------







