(in-package "KEIM")

     ;;; this file contains three modules
     ;;;  i)    the original algorithm (all parts named loewenheim=alg... )
     ;;;  ii)   the constant elimination algorithm (named loewenheim=ce... )
     ;;;  iii)  the normalization algorithm (named loewenheim=norm... )
     ;;; 
     ;;;
     ;;; The main function is loewenheim~unify
     ;;; This function needs three arguments:
     ;;;   1) a term or a list of terms (where the terms are unified pairwise
     ;;;                                  (even number of terms!!!))
     ;;;   2) a list of variables, which are used like constants
     ;;;   3) a list of restrictions
     ;;;  




(defmethod th~unify ((th th+br) tl1 tl2 restr const-vars)
  (let ((unifier (loewenheim~unify (mapcan #'list tl1 tl2) const-vars restr)))
    (if unifier
	(list unifier)
	nil)))


(defun loewenheim~unify (term-list var-as-const restrictions)
  (declare (edited  "18-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list of terms.")
	   (effect  "none.")
	   (value   "a substitution."))
  (let* ((theory (if (eq (class-of (th~term-theory (car term-list))) (th~theory-class 'br))
		     (th~term-theory (car term-list))
		     (th~term-theory (cadr term-list))))
	 (loewenheim*add (th~symbol theory :plus))
	 (loewenheim*mul (th~symbol theory :mal))
	 (loewenheim*one (th~symbol theory 1))
	 (loewenheim*zero (th~symbol theory 0))
	 (loewenheim*vac var-as-const))
    (declare (special loewenheim*add))
    (declare (special loewenheim*mul))
    (declare (special loewenheim*one))
    (declare (special loewenheim*zero))
    (declare (special loewenheim*vac))
    (let ((subst (cond ((null (cdr term-list)) (loewenheim=alg (loewenheim=alg-simplify (loewenheim=norm (car term-list)))))
		       ((null (cddr term-list)) (loewenheim=alg
				    (loewenheim=norm
				     (appl~create loewenheim*add (cons (first term-list)
								   (list (second term-list)))))))
		       (t (let* ((norm-terms (loewenheim=order-norm-and-conc-terms term-list))
				 (var-lists (loewenheim=order-create-var-list norm-terms))
				 (new-term-list (loewenheim=order norm-terms var-lists)))
			    (loewenheim=unify-prepare new-term-list))))))
      (if (null restrictions) subst
	  (loewenheim=const-elimination subst restrictions)))))


(defun loewenheim=unify-prepare (term-list-list)
  (if (null term-list-list) (subst~create nil nil)
      (if (null (car (car  term-list-list))) (loewenheim=unify-prepare (cdr term-list-list))
	  (let ((subst1 (loewenheim=unify-prep1 (car term-list-list))))
	    (if (null subst1) nil
		(let ((subst2 (loewenheim=unify-prepare (cdr term-list-list))))
		  (if (null subst2) nil
		      (subst~compose-substitution subst2 subst1))))))))


   (defun loewenheim=unify-prep1 (term-list)
     (if (null (cdr term-list)) (loewenheim=alg (car term-list))
	   (loewenheim=unify-prep1 (cons (loewenheim=alg-simplify (append
		   			        	  (append (first term-list)
			   			        	  (second term-list))
				   		          (loewenheim=alg-new-mult (first term-list)
							          (second term-list))))
			         	  (cddr term-list)))))



     ;;; this part of the module is written for reuse
     ;;; it contains a function (loewenheim=order), which is called with a list of variable-lists
     ;;; and a list of objects (no specific type) (both lists with the same length)
     ;;; the function creates a list of groups of objects, such that the variable-lists
     ;;; in each group are disjunct to the lists in other groups


(defun loewenheim=order (term-list var-lists)
  (declare (edited  "31-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list of normalized terms, represented by lists.")
	   (effect  "none.")
	   (value   "a list of disjunct groups of terms in respect to their variables."))
  (mapcar #'cdr (loewenheim=order1 term-list var-lists)))


   (defun loewenheim=order1 (term-list var-lists)
     (if (null term-list) nil
         (let ((new-list (loewenheim=order1 (cdr term-list) (cdr var-lists))))
	   (loewenheim=order-insert-and-merge new-list (car term-list) (car var-lists)))))


(defun loewenheim=order-insert-and-merge (work-list term var-list)
  (if (null work-list) (list (cons var-list (list term)))
      (let ((new-list (loewenheim=order-insert-and-merge (cdr work-list) term var-list)))
	(if (loewenheim=order-member-p (car (car work-list)) (car (car new-list)))
	    (cons (cons (loewenheim=order-merge-var-list (car (car work-list)) (car (car new-list)))
		 	(append (cdr (car work-list))
				(cdr (car new-list))))
	         (cdr new-list))
	    (append new-list (list (car work-list)))))))


(defun loewenheim=order-member-p (var-list-1 var-list-2)
  (if (null var-list-1) nil
      (if (member (car var-list-1) var-list-2 :test #'term~equal) T
	  (loewenheim=order-member-p (cdr var-list-1) var-list-2))))


(defun loewenheim=order-merge-var-list (var-list-1 var-list-2)
  (if (null var-list-1) nil
      (let ((new-list (or (loewenheim=order-merge-var-list (cdr var-list-1) var-list-2)
			  var-list-2)))
	  (if (member (car var-list-1) new-list :test #'term~equal) new-list
	    (cons (car var-list-1) new-list)))))


(defun loewenheim=order-norm-and-conc-terms (term-list)
  (if (null term-list) nil
      (cons (append
	       (loewenheim=norm (car term-list))
	       (loewenheim=norm (cadr term-list)))
       (loewenheim=order-norm-and-conc-terms (cddr term-list)))))


(defun loewenheim=order-create-var-list (term-list)
  (if (null term-list) nil
      (cons (loewenheim=order-find-all-vars (car term-list))
	    (loewenheim=order-create-var-list (cdr term-list)))))


(defun loewenheim=order-find-all-vars (norm-term)
  (if (null norm-term) nil
      (let ((var-list (loewenheim=order-find-all-vars (cdr norm-term))))
	(loewenheim=order-find-new-vars var-list (car norm-term)))))


(defun loewenheim=order-find-new-vars (var-list list)
  (declare (special loewenheim*vac))
  (if (null list) var-list
      (if (and
	   (sym~variable-p (car list))
	   (and (not (member (car list) var-list :test #'term~equal))
		(not (member (car list) loewenheim*vac :test #'term~equal))))
	  (loewenheim=order-find-new-vars (cons (car list) var-list) (cdr list))
	  (loewenheim=order-find-new-vars var-list (cdr list)))))

     ;;; end of loewenheim=order
		

(defun loewenheim=term-equal (list-1 list-2)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input  "Two lists representing a term." )
	   (effect  "none.")
	   (value   "T if lists are equal in respekt to "
		    "commutativity and associativity."))
  (if (null list-1)
      (if (null list-2) T
	  Nil)
      (if (member (car list-1) list-2 :test #'term~equal)
	  (loewenheim=term-equal (cdr list-1) (loewenheim=delete-term (car list-1) list-2))
	  Nil)))


(defun loewenheim=delete-term (element list)
  (if (term~equal element (car list)) (cdr list)
      (cons (car list) (loewenheim=delete-term element (cdr list)))))





     ;;; Part 1: The unification algorithm based on Loewenheim's formula:
     ;;; 
     ;;; t1, t2 : two terms of free Boolean ring
     ;;; {x1, ... , xn} all variables in t1 and t2
     ;;; s={xi<-bi | 1<=i<=n} a substitution with: s(t1)=s(t2)
     ;;; then the substitution
     ;;; d:={xi <- yi+g(t1+t2)*(yi+bi) | 1<=i<=n} with g={xi<-yi | 1<=i<=n}
     ;;; is a most general unifier of t1 and t2.

     ;;; the particular solution (the substitution s) is computed by a method called
     ;;; "Basis of Constants" method
     ;;; input : term t
     ;;; output: BR-unifier s of term t and 0
     ;;;
     ;;; to do this the term is first translated in its polynomial form and then splitted into
     ;;; a homogeneous and a inhomogeneous part
     ;;; then the part of the orthogonal basis of the initial Boolean ring is computed, which is
     ;;; needed to represent the inhomogeneous part
     ;;; with this information the unifier can be computed




(defun loewenheim=alg (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a substitution if possible, nil else."))
  (declare (special loewenheim*zero))
  (declare (special loewenheim*one))
  (loewenheim=alg-2 list (subst~create nil nil)))





(defun loewenheim=alg-2 (list subst)
  (let ((new-subst-right  (loewenheim=alg-compute-subst
                            (loewenheim=alg-compute-part-of-basis
                              (loewenheim=alg-get-all-constants list nil)
                              (loewenheim=alg-get-c0 list nil)
                               nil)
                            (loewenheim=alg-get-all-variables list nil)
                            (loewenheim=alg-polynomial-form list))))
    (loewenheim=alg-3 subst new-subst-right list (loewenheim=alg-compute-new-var-list
					    (loewenheim=alg-get-all-variables list nil)
					    nil))))

(defun loewenheim=alg-3 (subst subst-right term var-names)
  (if (null subst-right) subst
      (let ((new-var (loewenheim=alg-get-new-name (car (car subst-right)) var-names)))
	
	(let ((new-subst-right (loewenheim=alg-simplify (append (list (list new-var))
				      (loewenheim=alg-new-mult (loewenheim=alg-subst-term term var-names)
							 (append (list (list new-var))
								 (car (cdr (car subst-right)))))))))
      (loewenheim=alg-3 (subst~add-component (car (car subst-right))
					   (loewenheim=alg-create-new-term
					    (loewenheim=alg-simplify new-subst-right))
					   subst)
		      (cdr subst-right)
		      term var-names)))))

(defun loewenheim=alg-get-new-name (var var-names)
  (if (null var-names) var
      (if (term~equal var (car (car var-names)))
	  (second (car var-names))
	  (loewenheim=alg-get-new-name var (cdr var-names)))))

(defun loewenheim=alg-subst-term (term var-names)
  (if (null term) nil
      (append (list (loewenheim=alg-st-2 (car term) var-names))
	      (loewenheim=alg-subst-term (cdr term) var-names))))

(defun loewenheim=alg-st-2 (term var-names)
  (if (null term) nil
      (append (list (loewenheim=alg-get-new-name (car term) var-names))
	      (loewenheim=alg-st-2 (cdr term) var-names))))

(defun loewenheim=alg-compute-new-var-list (vars new-var-list)
  (if (null vars) new-var-list
      (loewenheim=alg-compute-new-var-list (cdr vars) (append (list (append (list (car vars))
							      (list (sym~rename-var (car vars)))))
							      new-var-list))))
						      


(defun loewenheim=alg-simplify (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a list, representing the simplified term."))
  (loewenheim=alg-simpl-add (loewenheim=alg-simpl-mul list)))


(defun loewenheim=alg-simpl-mul (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a list, representing the term after simplifying the multiplications"))
  (if (null list) list
      (cons (loewenheim=alg-del-ones (remove-duplicates (loewenheim=alg-get-zero (car list))
						      :test #'term~equal))
	    (loewenheim=alg-simpl-mul (cdr list)))))


(defun loewenheim=alg-get-zero (list)
  (declare (special loewenheim*zero))
  (if (member loewenheim*zero list :test #'term~equal) (list loewenheim*zero)
      list))


(defun loewenheim=alg-del-ones (list)
  (declare (special loewenheim*one))
  (if (and (null (cdr list)) (term~equal loewenheim*one (car list))) list
      (remove loewenheim*one list :test #'term~equal)))


(defun loewenheim=alg-simpl-add (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a list, representing the term after simplifying the additions."))
  (loewenheim=alg-get-non-zero (loewenheim=alg-del-inverse list)))

(defun loewenheim=alg-del-inverse (list)
  (declare (special loewenheim*zero))
  (if (null (cdr list)) list
      (if (member (car list) (cdr list) :test #'loewenheim=term-equal)
          (cons (list loewenheim*zero)
	        (loewenheim=alg-del-inverse (loewenheim=alg-delete-member (car list) (cdr list))))
          (cons (car list) (loewenheim=alg-del-inverse (cdr list))))))


(defun loewenheim=alg-delete-member (element list)
  (if (loewenheim=term-equal element (car list)) (cdr list)
      (cons (car list) (loewenheim=alg-delete-member element (cdr list)))))


(defun loewenheim=alg-get-non-zero (list)
  (declare (special loewenheim*zero))
  (if (null (cdr list)) list
      (cond ((loewenheim=term-equal (list loewenheim*zero) (car list))
	     (loewenheim=alg-get-non-zero (cdr list)))
	    ((loewenheim=term-equal (list loewenheim*zero) (cadr list))
	     (loewenheim=alg-get-non-zero (cons (car list) (cddr list))))
	  (t (cons (car list) (loewenheim=alg-get-non-zero (cdr list)))))))


(defun loewenheim=alg-new-mult (list-1 list-2)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "two lists, representing normalized terms")
	   (effect  "none.")
	   (value   "a list, representing a normalized term, created by multiplying"
		    "the two original terms."))
  (declare (special loewenheim*zero))
  (if (null list-2) (list (list loewenheim*zero))
  (if (or (null list-1) (null list-2)) list-1
      (append (loewenheim=alg-new-mult1 (car list-1) list-2)
	      (loewenheim=alg-new-mult (cdr list-1) list-2)))))

   (defun loewenheim=alg-new-mult1 (element list-2)
     (if (null list-2) list-2
         (cons (append element (car list-2))
	         (loewenheim=alg-new-mult1 element (cdr list-2)))))


(defun loewenheim=alg-create-new-term (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "the term, represented by a term+term"))
  (loewenheim=alg-create-add list))


(defun loewenheim=alg-create-add (list)
  (declare (special loewenheim*add))
  (if (null (cddr list))
      (if (null (cdr list)) (loewenheim=alg-create-mult (car list))
	  (appl~create loewenheim*add (cons (loewenheim=alg-create-mult (car list))
				      (list (loewenheim=alg-create-mult (cadr list))))))
      (appl~create loewenheim*add (cons (loewenheim=alg-create-mult (car list))
				  (list (loewenheim=alg-create-add (cdr list)))))))


(defun loewenheim=alg-create-mult (list)
  (declare (special loewenheim*mul))
  (if (null (cddr list))
      (if (null (cdr list)) (car list)
	  (appl~create loewenheim*mul (cons (car list)
				      (list (cadr list)))))
      (appl~create loewenheim*mul (cons (car list)
			          (list (loewenheim=alg-create-mult (cdr list)))))))








(defun loewenheim=alg-polynomial-form (term)
  (declare (edited  "16-SEP-1998")
	   (authors Hendrik Loehn)
	   (input   "a term in normalized form")
	   (effect  "none.")
	   (value   "a term in polynomial form sorted by criterion 'smallest sets of variables'"))
  (loewenheim=alg-reverse-p-form
   (loewenheim=alg-create-polynomial-form term (loewenheim=alg-create-new-term-form
					 (loewenheim=alg-get-all-variables term nil)))
   nil))

(defun loewenheim=alg-reverse-p-form (p-form rev-p-form)
  (if (null p-form) rev-p-form
      (loewenheim=alg-reverse-p-form (cdr p-form) (append (list (car p-form)) rev-p-form))))
  



(defun loewenheim=alg-create-new-term-form (var-list)
  (loewenheim=alg-cntf2 (loewenheim=alg-create-new-var-list var-list)))


(defun loewenheim=alg-cntf2 (new-var-list)
  (if (null new-var-list) nil
      (append (list (list (car new-var-list)))
	      (loewenheim=alg-cntf2 (cdr new-var-list)))))

(defun loewenheim=alg-create-new-var-list (var-list)
  (let ((n (list-length var-list)))
    (loewenheim=alg-potenz var-list n n)))

(defun loewenheim=alg-potenz (vars n m)
  (if (equal m 1) (loewenheim=alg-pot1 vars n 1)
      (append (loewenheim=alg-pot1 vars n m)
	      (loewenheim=alg-potenz vars n (- m 1)))))

(defun loewenheim=alg-pot1 (vars n m)
  (if (equal m 1) (mapcar #'list vars)
      (loewenheim=alg-pot2 vars n (- m 1))))

(defun loewenheim=alg-pot2 (vars n m)
  (if (equal 1 (- n m))
      (list vars)
      (append (loewenheim=alg-pot-mult (car vars) (loewenheim=ce-pot1 (cdr vars) (- n 1) m))
	      (loewenheim=alg-pot2 (cdr vars) (- n 1) m))))

(defun loewenheim=alg-pot-mult (elem list)
  (if (null list) nil
      (cons (loewenheim=alg-pot-add elem (car list))
	    (loewenheim=alg-pot-mult elem (cdr list)))))

(defun loewenheim=alg-pot-add (elem1 elem2)
  (if (atom elem2) (cons elem1 (list elem2))
      (cons elem1 elem2)))
	      


(defun loewenheim=alg-create-polynomial-form (term new-term-form)
  (if (null term) new-term-form
      (loewenheim=alg-create-polynomial-form
       (cdr term) (loewenheim=alg-add-new-element (car term) new-term-form))))


(defun loewenheim=alg-add-new-element (element new-term-form)
  (if (null new-term-form) nil
  (let ((cu (loewenheim=alg-right-vu-p element (car (car new-term-form)))))
	(if cu
      (append (list (append (list (car (car new-term-form)))
		          (list (append (car (cdr (car new-term-form))) (list cu)))))
	      (cdr new-term-form))
      (append (list (car new-term-form))
	      (loewenheim=alg-add-new-element element (cdr new-term-form)))))))

(defun loewenheim=alg-right-vu-p (element var-list)
  (if (null var-list) element
      (if (member (car var-list) element :test #'term~equal)
	  (loewenheim=alg-right-vu-p (remove (car var-list) element :test #'term~equal) (cdr var-list))
	  nil)))

  
(defun loewenheim=alg-get-all-variables (term vars)
  (declare (edited  "16-SEP-1998")
	   (authors Hendrik Loehn)
	   (input   "a term an a list of variables [normally 'NIL']")
	   (effect  "none.")
	   (value   "a list of all variables in the term"))
  (declare (special loewenheim*vac))
  (if (null term) vars
      (loewenheim=alg-get-all-variables
       (cdr term)
       (append vars
	       (remove-duplicates
		(mapcan #' (lambda (atom)
			     (if (and (and (sym~variable-p atom)
					   (not (member atom loewenheim*vac :test #'term~equal)))
				      (not (member atom vars :test #'term~equal)))
				 (list atom)
				 nil))
			   (car term))
		:test #'term~equal)))))

(defun loewenheim=alg-get-all-constants (term consts)
   (declare (edited  "16-SEP-1998")
	   (authors Hendrik Loehn)
	   (input   "a term and a list of constants [normally 'NIL']")
	   (effect  "none.")
	   (value   "a list of all constants in the term"))
  (declare (special loewenheim*vac))
  (declare (special loewenheim*zero))
  (if (null term) consts
      (loewenheim=alg-get-all-constants
       (cdr term)
       (append consts
	       (remove-duplicates
		(mapcan #' (lambda (atom)
			     (if (and (or (sym~constant-p atom)
					   (member atom loewenheim*vac :test #'term~equal))
				      (and (not (member atom consts :test #'term~equal))
					   (not (term~equal atom loewenheim*zero))))
				 (list atom)
				 nil))
			   (car term))
		:test #'term~equal)))))


  
(defun loewenheim=alg-get-c0 (term c0)
   (declare (edited  "16-SEP-1998")
	   (authors Hendrik Loehn)
	   (input   "a term")
	   (effect  "none.")
	   (value   "the inhomogeneous part of the polynomial form"))
  (declare (special loewenheim*zero))
  (if (null term) nil
	    (if (and (loewenheim=alg-get-c0-2-p (car term))
		     (not (term~equal (car (car term))loewenheim*zero)))
		(append (list (append c0 (car term)))
			(loewenheim=alg-get-c0 (cdr term) c0))
                (loewenheim=alg-get-c0 (cdr term) c0))))

(defun loewenheim=alg-get-c0-2-p (list)
  (declare (special loewenheim*vac))
  (if (null list) T
      (if (and (sym~variable-p (car list))
	       (not (member (car list) loewenheim*vac :test #'term~equal)))
	  Nil
	  (loewenheim=alg-get-c0-2-p (cdr list)))))


(defun loewenheim=alg-compute-part-of-basis (consts c0 c0-basis)
   (declare (edited  "16-SEP-1998")
	   (authors Hendrik Loehn)
	   (input   "a list of constants, the inhomogenous part of the term and a dummy")
	   (effect  "none.")
	   (value   "the part of the orthogonal basis needed to represent c0"))
  (mapcar #'loewenheim=alg-simplify (mapcar #'loewenheim=norm (loewenheim=alg-cpob-1 consts c0 c0-basis))))



(defun loewenheim=alg-cpob-1 (consts c0 c0-basis)
  (if (null c0) nil
      (append (loewenheim=alg-cpob-2 consts (car c0) c0-basis)
	      (loewenheim=alg-cpob-1 consts (cdr c0) nil))))

(defun loewenheim=alg-cpob-2 (consts c c-basis)
  (if (null consts) c-basis
      (loewenheim=alg-cpob-2 (cdr consts) c (loewenheim=alg-cpob-3 (car consts) c c-basis))))

(defun loewenheim=alg-cpob-3 (const c c-basis)
  (declare (special loewenheim*add))
  (declare (special loewenheim*mult))
  (declare (special loewenheim*one))
  (if (member const c :test #'term~equal)
      (if (null c-basis)
	  (list const)
	  (loewenheim=alg-cpob-4 const c c-basis))
      (if (null c-basis)
	  (append
	   (list const)
	   (list (appl~create loewenheim*add (cons const (list loewenheim*one)))))
	  (loewenheim=alg-cpob-4 const c c-basis))))

(defun loewenheim=alg-cpob-4 (const c c-basis)
  (declare (special loewenheim*add))
  (declare (special loewenheim*mul))
  (declare (special loewenheim*one))
  (if (null c-basis) nil
      (append
       (append
	(list (appl~create loewenheim*mul (cons const (list (car c-basis)))))
	(if (member const c :test #'term~equal) nil
	    (list (appl~create loewenheim*mul
			       (cons (appl~create loewenheim*add (cons const (list loewenheim*one)))
				     (list (car c-basis)))))))
       (loewenheim=alg-cpob-4 const c (cdr c-basis)))))
     
(defun loewenheim=alg-compute-subst (D vars p-form)
   (declare (edited  "16-SEP-1998")
	   (authors Hendrik Loehn)
	   (input   "a part of the orthogonal basis, a list of varsand the term in polynomial form")
	   (effect  "none.")
	   (value   "a substitution, which is a unifier for the term and 0"))
  (loewenheim=alg-cs-2 D p-form (mapcar #'list vars)))

(defun loewenheim=alg-cs-2 (D p-form new-subst)
  (if (null D) new-subst
      (loewenheim=alg-cs-2 (cdr D) p-form (loewenheim=alg-cs-3 (car D) p-form new-subst))))

(defun loewenheim=alg-cs-3 (elem p-form new-subst)
  (declare (special loewenheim*zero))
  (if (null p-form) nil
      (if (not (term~equal
		(car (car (loewenheim=alg-simplify (loewenheim=alg-new-mult elem (car (cdr (car p-form)))))))
		loewenheim*zero))
	  (loewenheim=alg-cs-add-elem elem (car (car p-form)) new-subst)
	  (loewenheim=alg-cs-3 elem (cdr p-form) new-subst))))

(defun loewenheim=alg-cs-add-elem (elem vars new-subst)
  (if (null new-subst) nil
      (if (member (car (car new-subst)) vars :test #'term~equal)
	  (append (list (append (list (car (car new-subst)))
			  (list (append elem (car (cdr (car new-subst)))))))
		  (loewenheim=alg-cs-add-elem elem vars (cdr new-subst)))
	  (append (list (car new-subst))
		  (loewenheim=alg-cs-add-elem elem vars (cdr new-subst))))))

















     ;;;  Part II : The constant elimination algorithm
     ;;;
     ;;;  c0 = set of all constants to be eliminated
     ;;;  v0 = set of all variables occuring in the terms of the problems
     ;;;  D  = set of all possible products of elements in c0 including "1"


(defun loewenheim=const-elimination (subst restrictions)
  (declare (edited  "17-MAR-1998")
	   (authors Hendrik)
	   (input   "A substitution and a list of constant restrictions.")
	   (effect  "none.")
	   (value   "a new substitution, in which all restrictions are ?...?"
		    "if one can be found, nil else."))
  (if (null subst) nil
      (let* ((old-ti (loewenheim=ce-get-ti restrictions))
	     (ci (loewenheim=ce-create-ci old-ti restrictions))
	     (ti (mapcar #'(lambda (ti) (subst~apply subst ti)) old-ti))
	     (ti-norm (mapcar #'loewenheim=norm ti))
	     (v0 (loewenheim=ce-create-v0 ti-norm))
	     (c0 (loewenheim=ce-create-c0 restrictions))
	     (d (loewenheim=ce-create-d c0))
	     (new-subst (loewenheim=ce-gen-subst v0 d))) 
	(let* ((norm-terms (mapcan #'(lambda (list c)
				       (loewenheim=ce-get-only-ci (loewenheim=ce-create-tijd list d) c d))
				   (mapcar #'loewenheim=norm
					   (mapcar #'(lambda (list)
						       (subst~apply new-subst list)) ti))
				   ci))
		    (var-lists (loewenheim=order-create-var-list norm-terms))
		    (new-term-list (loewenheim=order norm-terms var-lists)))
	       (let ((elimination (loewenheim=unify-prepare new-term-list)))
		 (if (null elimination) nil
		     (let ((final-subst (subst~compose-substitution
					 (subst~apply elimination new-subst)
					 subst)))
		       (loewenheim=ce-make-new-subst (subst~domain final-subst)
					       (subst~codomain final-subst)))))))))


(defun loewenheim=ce-create-ci(ti restrictions)
  (declare (edited  "17-MAR-1998")
	   (authors Hendrik)
	   (input   "a list of terms in which some constants are restricted (mostly variables only).")
	   (effect  "none.")
	   (value   "a list of constants, which are restricted."))
  (mapcar #'(lambda (var1)
	      (mapcan #'(lambda (var2)
			  (if (member var1 var2 :test #'term~equal)
			      (list (car var2))
			      nil))
		      restrictions))
	  ti))


(defun loewenheim=ce-get-only-ci (list c d)
  (if (null d) nil
      (if (loewenheim=ce-get-only-ci2 c (car d))
	  (cons (car list) (loewenheim=ce-get-only-ci (cdr list) c (cdr d)))
	  (loewenheim=ce-get-only-ci (cdr list) c (cdr d)))))


   (defun loewenheim=ce-get-only-ci2 (c prefix)
     (if (null c) nil
         (if (member (car c) prefix :test #'term~equal) t
	     (loewenheim=ce-get-only-ci2 (cdr c) prefix))))


(defun loewenheim=ce-make-new-subst (domain co-domain)
  (if (null co-domain) (subst~create nil nil)
      (subst~insert-component (car domain)
			      (loewenheim=alg-create-new-term
			       (loewenheim=alg-simplify
				(loewenheim=norm (car co-domain))))
			      (loewenheim=ce-make-new-subst (cdr domain) (cdr co-domain)))))


(defun loewenheim=ce-gen-subst (v0 d)
  (if (null v0) (subst~create nil nil)
      (subst~insert-component (car v0)
			      (loewenheim=alg-create-new-term
			          (mapcan #'(lambda (var)
					     (list (append var
							   (list (sym~rename-var (car v0))))))
					  d))
			      (loewenheim=ce-gen-subst (cdr v0) d))))


(defun loewenheim=ce-create-tijd (list d)
  (loewenheim=ce-check-for-nil (loewenheim=ce-create-tijd1 (loewenheim=alg-simplify list) d (loewenheim=ce-prepare-list d))))


(defun loewenheim=ce-check-for-nil (list)
  (declare (special loewenheim*zero))
  (if (null list) nil
      (if (null (car list)) (cons (list (list loewenheim*zero))
				    (loewenheim=ce-check-for-nil (cdr list)))
	  (cons (car list) (loewenheim=ce-check-for-nil (cdr list))))))


   (defun loewenheim=ce-create-tijd1 (list d new-list)
     (if (null list) new-list
         (loewenheim=ce-create-tijd1 (cdr list) d (loewenheim=ce-create-tijd2 (car list) d new-list))))


   (defun loewenheim=ce-create-tijd2 (list d new-list)
     (if (null (cdr d)) (list (cons list (car new-list)))
         (let ((term-part (loewenheim=ce-is-ci-in-tijd-p list (car d))))
	   (if term-part (cons (append (list term-part) (car new-list))
			       (cdr new-list))
	       (cons (car new-list)
		     (loewenheim=ce-create-tijd2 list (cdr d) (cdr new-list)))))))


(defun loewenheim=ce-prepare-list (d)
  (if (null d) nil
      (cons nil (loewenheim=ce-prepare-list (cdr d)))))


(defun loewenheim=ce-is-ci-in-tijd-p (list const-list)
  (if (null const-list) list
      (let ((list-part (loewenheim=ce-is-ci-in-tijd2-p list (car const-list))))
	(if list-part (loewenheim=ce-is-ci-in-tijd-p list-part (cdr const-list))
	    nil))))


   (defun loewenheim=ce-is-ci-in-tijd2-p (list const)
     (declare (special loewenheim*one))
     (if (null list) nil
         (if (term~equal (car list) const)
	     (if (null (cdr list)) (list loewenheim*one)
	         (cdr list))
             (let ((test (loewenheim=ce-is-ci-in-tijd2-p (cdr list) const)))
	       (if test (cons (car list) test)
		   nil)))))


(defun loewenheim=ce-create-v0 (term-list)
  (let ((new-list (if (atom (car (car term-list))) term-list
		      (mapcan #'(lambda (elem) elem) term-list))))
    (loewenheim=ce-c-v0-get-vars new-list nil)))


(defun loewenheim=ce-c-v0-get-vars (term vars)
  (declare (special loewenheim*vac))
  (if (null term) vars
      (loewenheim=ce-c-v0-get-vars
       (cdr term)
       (append vars
	       (remove-duplicates
		       (mapcan #' (lambda (atom)
			         (if (and (and (sym~variable-p atom)
					  (not (member atom loewenheim*vac :test #'term~equal)))
				          (not (member atom vars :test #'term~equal)))
				     (list atom)
				     nil))
				  (car term))
		       :test #'term~equal)))))
				  

(defun loewenheim=ce-get-ti (restrictions)
  (if (null restrictions) nil
      (loewenheim=ce-new-ti (cdr (car restrictions)) (loewenheim=ce-get-ti (cdr restrictions)))))


(defun loewenheim=ce-new-ti (ti old-ti)
  (if (null ti) old-ti
      (if (member (car ti) old-ti :test #'term~equal)
	  (loewenheim=ce-new-ti (cdr ti) old-ti)
	  (loewenheim=ce-new-ti (cdr ti) (cons (car ti) old-ti)))))


(defun loewenheim=ce-create-c0 (restrictions)
  (if (null restrictions) nil
      (cons (car (car restrictions))
	    (loewenheim=ce-create-c0 (cdr restrictions)))))


(defun loewenheim=ce-create-d (c0)
  (declare (special loewenheim*one))
  (let ((n (list-length c0)))
    (append (loewenheim=ce-potenz c0 n n) (list (list loewenheim*one)))))


(defun loewenheim=ce-potenz (c0 n m)
  (if (equal m 1) (loewenheim=ce-pot1 c0 n 1)
      (append (loewenheim=ce-pot1 c0 n m)
	      (loewenheim=ce-potenz c0 n (- m 1)))))


   (defun loewenheim=ce-pot1 (C0 n m)
     (if (equal m 1) (mapcar #'list c0)
         (loewenheim=ce-pot2 C0 n (- m 1))))


   (defun loewenheim=ce-pot2 (C0 n m)
      (if (equal 1 (- n m))
         (list c0)
         (append (loewenheim=ce-pot-mult (car c0) (loewenheim=ce-pot1 (cdr c0) (- n 1) m))
	         (loewenheim=ce-pot2 (cdr c0) (- n 1) m))))


(defun loewenheim=ce-pot-mult (elem list)
  (if (null list) nil
      (cons (loewenheim=ce-pot-add elem (car list))
	    (loewenheim=ce-pot-mult elem (cdr list)))))


(defun loewenheim=ce-pot-add (elem1 elem2)
  (if (atom elem2) (cons elem1 (list elem2))
      (cons elem1 elem2)))




     ;;;  Part III : The normalization algorithm
     ;;;
     ;;;  the original term is normalized. The new term has the form of lists.


(defun loewenheim=norm (term)
  (declare (edited  "17-MAR-1998")
	   (authors Hendrik)
	   (input   "a term*term.")
	   (effect  "none.")
	   (value   "a list of lists of terms, representing the term in its normalized form"
		    "i.e. a disjunction of conjunctios."))
  (declare (special loewenheim*add))
  (declare (special loewenheim*mul))
  (cond ((sym~constant-p Term) (list (list Term)))
	((sym~variable-p Term) (list (list Term)))
	((term~equal (term~top Term) loewenheim*add)
	  (append (loewenheim=norm (car (appl~arguments Term)))
		  (loewenheim=norm (cadr (appl~arguments Term)))))
	((term~equal (term~top Term) loewenheim*mul)
	  (loewenheim=norm-mult-new (loewenheim=norm (car (appl~arguments Term)))
			 (loewenheim=norm (cadr (appl~arguments Term)))))
	)
  )

(defun loewenheim=norm-mult-new (list-1 list-2)
  (if (null list-1) list-1
      (append (loewenheim=norm-mult-each-new (car list-1) list-2)
	      (loewenheim=norm-mult-new (cdr list-1) list-2))))

(defun loewenheim=norm-mult-each-new (elem list)
  (if (null list) list
      (cons (append elem (car list))
	    (loewenheim=norm-mult-each-new elem (cdr list)))))
