
(in-package "KEIM")

(export '(misc~lqueue-create misc~lqueue-lists misc~lqueue-next-lists
	  misc~lqueue-insert! misc~lqueue-delete!
	  misc~queue-create misc~queue-front misc~queue-list
	  misc~queue-insert! misc~queue-delete!
	  misc~lqueue-tuples))

#|
The following functions define a generalization of a queue.  The concept of
queues is generalized in two ways:  First, instead of elements we want to insert lists
of elements into the queue.  The delete operation should delete these lists in
the first-in-first-out manner known from queues.  Additionally we do not want to
access the first element of the queue or the first list, but the list of all
elements, i.e. a list built by appending (not listing) all lists that are in the
queue.

Example:

(setf queue (queue-create))
(insert queue '(1 2 3))
(queue-list queue)       -->  (1 2 3)
(insert queue '())
(queue-list queue)       -->  (1 2 3)
(insert queue '(a b c))
(queue-list queue)       -->  (1 2 3 a b c)
(delete queue)
(queue-list queue)       -->  (a b c)
(delete queue)
(queue-list queue)       -->  (a b c)
(delete queue)
(queue-list queue)       -->  ()
(delete queue)
(queue-list queue)       -->  Error

Second, we want this mechanism to be performed on a set of parallel queues.  So
instead of `(progn (insert queue1 list1) (insert queue2 list2) ...)' and
`(progn (delete queue1) (delete queue2) ...)' we can write `(insert queue list1
list2 ...)' and `(delete queue)'
|#

(defvar misc*queue-length 5
  "This variable defines the number of parallel queues.")

;;; The following figure displays the structure of multiple queue for lists with
;;; two queues.  Only the first queue is displayed; the structure of the second
;;; is analogous; note that it must contain n lists, too.  LIST_1 is the last
;;; list that has beEn inserted into the first queue; LIST_N is the first list
;;; that has been inserted into the the first queue and that has not been
;;; deleted from the queue yet.
;;; 
;;;
;;;            list_n                 list_2              list_1
;;;       ________________       ________________   ________________
;;;      /                \     /                \ /                \
;;; (old l_n[1] .. l_n[m_n] ... l_2[1] .. l_2[m_2] l_1[1] .. l_1[m_1])    <- queue_1
;;; ^             ^                                         ^
;;; |             |                                         |    ;; These pointers
;;; |             |                                         |    ;; point  to  the
;;; +---+         +-----+                       +-----------+    ;; last cons-cell
;;;     |               |                       |                ;; of the lists. 
;;;     |               |                       |
;;;  ((q_1[n] q_2[n]) (q_1[n-1] q_2[n-1]) ... (q_1[1] q_2[1]))  ;; (*) 
;;;  ^                                       ^^
;;;  |                                       ||<-;; This list-rears pointer is actually
;;;  +----------+    +-----------------------+|  ;; a copy of (q_1[1] q_2[1]).
;;;             |    |       +----------------+
;;;             |    |       |
;;;          (front rear list-rears)  <-- This is the object representing
;;;                                       the multiple queue.
;;;
;;;  (*) This list is treated as normal queue: Elements are popped from the
;;;  front and added to the end, where the elements are tuples pointing to the
;;;  corresponging lists in the queues.
;;;

(defmacro misc=lqueue-front (queue)
  `(first ,queue))

(defmacro misc=lqueue-rear (queue)
  `(second ,queue))

(defmacro misc=lqueue-list-fronts (queue)
  `(first (first ,queue)))

(defmacro misc=lqueue-list-next-fronts (queue)
  `(second (first ,queue)))

(defmacro misc=lqueue-list-rears (queue)
  `(third ,queue))


(defun misc~lqueue-create ()
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input   "None.")
	   (effect  "None.")
	   (value   "A new multiple queue for lists"
		    "(i.e. a queue generalized as descirbed above)."))
  (let ((heads (make-sequence 'list misc*queue-length)))
    (mapl #'(lambda (tail) (setf (first tail) (list nil)))
	  heads)
    (list nil (mapcar #'list heads) heads)))

(defun misc~lqueue-insert! (queue &rest new-lists)
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input   "A multiple queue and some lists.")
	   (effect  "The lists are inserted into QUEUE.")
	   (value   "Undefined."))
  (when (not (= misc*queue-length (length new-lists)))
    (error "cbcomb: The number of new queue entries (~A) is not ~A."
	   (length new-lists) misc*queue-length))
  (mapc #'(lambda (rear new-list)
	    (setf (rest rear) new-list))
	(misc=lqueue-list-rears queue) new-lists)
  (let ((new-element (cons (copy-list (misc=lqueue-list-rears queue)) nil)))
    (mapl #'(lambda (rear-tail)
	      (setf (car rear-tail)
		    (last (car rear-tail))))
	  (misc=lqueue-list-rears queue))
    (setf (cdr (misc=lqueue-rear queue))
	  new-element)
    (setf (misc=lqueue-rear queue)
	  (cdr (misc=lqueue-rear queue)))
    (if (null (misc=lqueue-front queue))
	(setf (misc=lqueue-front queue) new-element))
    queue))

(defun misc~lqueue-delete! (queue)
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input   "A multiple queue.")
	   (effect  "In each single queue in QUEUE, the list that has been"
		    "inserted first and has not been deleted yet is deleted.")
	   (value   "Undefined."))
  (when (null (misc=lqueue-front queue))
    (error "misc:  Queue is empty!"))
  (pop (misc=lqueue-front queue))
  queue)

(defun misc~lqueue-lists (queue)
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input   "A multiple queue.")
	   (effect  "None.")
	   (value   "A list containing for each single queue in QUEUE a list with"
		    "its elements.  This is the list that is built by appending"
		    "all lists that have been inserted in this single queue and"
		    "have not been deleted yet."))
  (mapcar #'rest (misc=lqueue-list-fronts queue)))

(defun misc~lqueue-next-lists (queue)
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input   "A multiple queue.")
	   (effect  "None.")
	   (value   "The same as in misc~lqueue-lists but without the first"
		    "list, i.e. the list that will be deleted next."))
  (mapcar #'rest (misc=lqueue-list-next-fronts queue)))


;;; some stuff that can be usefull for monitoring multiple queues

(defun misc~lqueue-tuples (queue)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A multiple queue.")
	   (effect  "None.")
	   (value   "A list containing the content of each queue.  The lists"
		    "contained in each queue are not appended like in"
		    "misc~lqueue-lists but listed.  So the structure of the"
		    "queues and the lists they are containing becomes visible."))
  (maplist #'(lambda (queue-tail)
	       (if (second queue-tail)
		   (mapcar #'(lambda (list next-list)
			       (misc=sublist (rest list) (rest next-list)))
			   (first queue-tail) (second queue-tail))
		   (mapcar #'rest (first queue-tail))))
	   (misc=lqueue-front queue)))

(defun misc=sublist (list end-pointer)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A list and a pointer pointing to some tail-list of LIST (or"
		    "NIL).")
	   (effect  "None.")
	   (value   "The subsequence of LIST from the beginning of LIST up to"
		    "END-POINTER (i.e. up to the end of LIST if END-POINTER is"
		    "NIL)."))
  (cond ((eq list end-pointer)
	 nil)
	((null list) (error "MISC(unimok): LIST is empty, but END-POINTER is not NIL."))
	(t (cons (first list) (misc=sublist (rest list) end-pointer)))))


;;; normal queues

(defun misc~queue-create ()
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "None.")
	   (effect  "None.")
	   (value   "An empty queue."))
  (cons nil nil))

(defun misc~queue-front (queue)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A queue.")
	   (effect  "An error is signalled if QUEUE is empty.")
	   (value   "The first element of queue."))
  (if (car queue)
      (caar queue)
      (error "misc: Queue is empty!")))

(defun misc~queue-list (queue)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A queue.")
	   (effect  "None.")
	   (value   "A list of all elements in QUEUE."))
  (car queue))

(defun misc~queue-insert! (queue element)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A queue and an element.")
	   (effect  "ELEMENT is added to the end of QUEUE.")
	   (value   "The changed QUEUE."))
  (if (car queue)
      (setf (cddr queue) (list element)
	    (cdr queue) (cddr queue))
      (setf (car queue) (list element)
	    (cdr queue) (car queue)))
  queue)

(defun misc~queue-delete! (queue)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A queue.")
	   (effect  "The first element is deleted from QUEUE.  An error is"
		    "signalled if QUEUE is empty.")
	   (value   "The changed QUEUE."))
  (if (car queue)
      (pop (car queue))
      (error "misc: Queue is empty!"))
  queue)



