
(in-package "KEIM")

;#| This modules contains the datastructures and basic algorithms for unification
;   with constraints. |#

(export '(unicp+unification-problem
	  unicp~general-up-create unicp~variables
	  unicp~gup-choice-points unicp~gup-label-stack
	  unicp~gup-up-queue unicp~gup-constraint-lqueue
	  unicp+elementary-up
	  unicp~general-up unicp~equations unicp~theory
	  unicp~theo-init unicp~new-constr unicp~solvable-p
	  unicp~up-copy unicp~up-save-choice-point unicp~up-restore-choice-point
	  unicp+var-label unicp~orig-var-label unicp~set-orig-var-label!
	  unicp~var-label unicp~var-label-theories
	  unicp~var-label-identified unicp~var-label-discrim-lists
	  unicp~var-label-theory unicp~var-label-not-theory
	  unicp~var-label-less unicp~var-label-not-less
	  unicp~var-label-greater unicp~var-label-not-greater
	  unicp~var-label-mark
	  unicp~var-label-repres unicp~vars-label-repres unicp~repres-list
	  unicp~vars-identified-p unicp~vars-discriminated-p unicp~vars-discriminated
	  unicp~vars-label-initialize! unicp~vars-identify! unicp~vars-discriminate!
	  unicp~vars-lesseq-p unicp~vars-less-p unicp~constant-p unicp~shared-variable-p
	  unicp~term-label
	  unicp~eq-create unicp~eq-lhss unicp~eq-rhss
	  unicp~id-create unicp~id-left unicp~id-right
	  unicp~vi-create unicp~vi-var unicp~vi-theory
	  unicp~cr-create unicp~cr-const unicp~cr-varlist))

(eval-when (load compile eval)
 (defclass unicp+unification-problem (uni+unification-problem)
   ((variables         :initarg :variables        :reader unicp~variables)
    (shared-variables  :initarg :shared-variables :reader unicp~shared-variables)
    (choice-points     :initform nil            :accessor unicp~gup-choice-points)
    (label-stack       :initform nil            :accessor unicp~gup-label-stack)
    (up-queue          :initform nil            :accessor unicp~gup-up-queue)
    (up-list           :initform nil            :accessor unicp~gup-up-list)
    (constraint-lqueue :initform nil            :accessor unicp~gup-constraint-lqueue))
   (:documentation
    "The class of unification problems for the combination algorithm with constraints.")))

(defun unicp~general-up-create (initial-problem &key
						(variables nil s-variables)
						(shared-variables nil s-shared-variables))
  (let* ((vars (if s-variables
		   variables
		   (top~free-variables initial-problem)))
	 (shared-vars (if s-shared-variables
			  shared-variables
			  vars)))
    (unless variables
      (unicp~vars-label-initialize! vars))
    (make-instance 'unicp+unification-problem
		   :initial-problem initial-problem
		   :previous-solutions nil
		   :variables vars
		   :shared-variables shared-vars)))

(eval-when (load compile eval)
 (defclass unicp+elementary-up (keim+object)
   ((general-up :initarg :general-up :reader unicp~general-up)
    (equations  :initarg :equations  :reader unicp~equations)
    (theory     :initarg :theory     :reader unicp~theory)
    (comb-variables :initform :undefined :accessor unicp=comb-variables))
   (:documentation
    "This is the abstract superclass of all elementary unification problems used in
    the combination algorithm with constraints.")))

(defun unicp~comb-variables (up)
  (declare (edited  "07-OCT-1996")
	   (authors Richts)
	   (input   "An elementary unification problem.")
	   (effect  "The comb-variables slot is initialized if this has not been"
		    "done before.")
	   (value   "The list of variables in UP that also occur in other up's."))
  (if (eq (unicp=comb-variables up) :undefined)
      (setf (unicp=comb-variables up)
	    (remove-if-not #'unicp~shared-variable-p
			   (top~free-variables (unicp~equations up))))
      (unicp=comb-variables up)))
      

(defmethod unicp~variables ((up unicp+elementary-up))
  (unicp~variables (unicp~general-up up)))

(defmethod unicp~shared-variables ((up unicp+elementary-up))
  (unicp~shared-variables (unicp~general-up up)))


(defgeneric unicp~theo-init (theory general-up equations
				    identified discriminated constants const-restr)
  (declare (edited  "06-APR-1995")
	   (authors Richts)
	   (input   "A theory, a list of pure equations (i.e. pairs of terms),"
		    "a list of pairs of variables which have to be identified,"
		    "a list of pairs of variables which have to be discriminated,"
		    "a list of variables that should be treated as constants,"
		    "a list of constant restrictions (i.e. pairs of a constant"
		    "and a set of variables.")
	   (effect  "None.")
	   (value   "The elementary algorithm for THEORY is called.  It returns"
		    "the same values as unicp~new-constraints with the exception"
		    "that the first value is a new elementary unification problem"
		    "if the problem is solvable.")))

(defgeneric unicp~new-constr (elementary-up identified discriminated constants const-restr)
  (declare (edited  "06-APR-1995")
	   (authors Richts)
	   (input   "An elementary unification problem and some restrictions"
		    "that are new for this problem:"
		    "a list of pairs of variables which have to be identified,"
		    "a list of pairs of variables which have to be discriminated,"
		    "a list of variables that should be treated as constants,"
		    "a list of constant restrictions (i.e. pairs of a constant"
		    "and a set of variables.")
	   (effect  "The ELEMENTARY-UP is updated with the new restrictions.")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP belongs"
		    "to is called.  It returns NIL if the unification problem is"
		    "not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains some new"
		    "constraints wich must be met if the unification problem should be solvable:"
		    "1. True, indicating that the unification problem is still solvable."
		    "2. A list of new identification of variables (i.e. sets of variables)."
		    "3. A list of pairs of variables which must not be unified."
		    "4. A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6. A list of pairs $(x . y)$ of variables where $\sigma(x)$"
		    "   is a subterm of $\sigma(y)$ for a unifier $\sigma$."
		    "7. A list of pairs $(x . y)$ of variables where no solution"
		    "   is lost if $y$ is in the constant restrictions of $x$.")))

(defgeneric unicp~solvable-p (elementary-up)
  (declare (edited  "19-DEC-1995")
	   (authors Richts)
	   (input   "An elementary unification problem.")
	   (effect  "The ELEMENTARY-UP might be destroyed.")
	   (value   "True iff the unification problem consisting of the original"
		    "equations of ELEMENTARY-UP and the constraints made on its variables"
		    "is solvable.")))

(defgeneric unicp~more-information-p (elementary-up)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "An elementary unification problem.")
	   (effect  "None.")
	   (value   "True, iff there might still be constraints that can be"
		    "computed for ELEMENTARY-UP, because this computation has"
		    "been postponed due to high complexity." ))
  (:method ((up unicp+elementary-up))
     nil))


#|
\subsection{Backtracking}

When backtracking is necessary, the information in the following kinds of
structures must be set back:
\begin{description}
\item[Variables] Most information about the current status of the decomposition
algorithm is stored in the variables.  But also some elementary algorithms store
information in the variables.  All this information is united in the
unicp+var-label object in the variables' label slots.  For backtracking this
information is saved and set back by the decompostion algorithm.
\item[Elementary unification problems]  Some information of the elementary
unification algorithms is stored in the elementary unification problems.  The
simplest way for backtracking is to copy the unification problems at choice
points and use these copies when tracking back.
\item[Terms] Some algorithms store information in non-variable terms.  This
cannot result in conflicts with other theories since only the algorithm for the
theory of a term's function symbols can use this term.  For backtracking this
information must be copied and set back by the elementary unfication algorithm.
\end{description}

Unification algorithms that use non-variable terms for their information must
supply methods for UNICP~UP-SAVE-CHOICE-POINT and
UNICP~UP-RESTORE-CHOICE-POINT.  For algorithms that do not use non-variable
terms, the default is to copy and store the unification problem in
UNICP~UP-SAVE-CHOICE-POINT and to return it in UNICP~UP-RESTORE-CHOICE-POINT.
Note that this default is overwritten when methods for
UNICP~UP-SAVE-CHOICE-POINT and UNICP~UP-RESTORE-CHOICE-POINT are defined,
i.e. algorithms doing this must alse handle the information in their unification
problem.

Unification algorithms that define new slots for their unification problems must
supply a method for UNICP~UP-COPY.  The default method for algorithms that do
not define new slots is to return the (uncopied) object.  This causes no
problems since the slots defined in UNICP+ELEMENTARY-UP are not changed and must
therefore not be set back.

Unification algorithms that only use variables to store their information need
not to bother about backtracking since the default methods for their unification
problems work correct and the variables are set back by the decomposition
algorithm.
|#

(defgeneric unicp~up-copy (up)
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input   "A cp unification problem.")
	   (effect  "None.")
	   (value   "A copy of UP that can be used when backtracking is necessary."))
  (:method ((up unicp+elementary-up))
    up))

(defgeneric unicp~up-save-choice-point (up)
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input  "A cp unification problem.")
	   (effect "The current status of UP is saved on a stack, so it can be"
		   "restored with UNICP~UP-RESTORE-CHOICE-POINT.")
	   (value  "Undefined."))
  (:method ((up unicp+elementary-up))
    (keim~put up :unicp=stack (cons (unicp~up-copy up)
				    (keim~get up :unicp=stack)))))

(defgeneric unicp~up-restore-choice-point (up)
  (declare (edited  "11-AUG-1995")
	   (authors Richts)
	   (input   "A cp unification problem.")
	   (effect  "UP may be changed.")
	   (value   "A unification problem (not necessarily eq to UP) with the"
		    "last status of UP that was stored with UNICP~UP-SAVE-CHOICE-POINT."))
  (:method ((up unicp+elementary-up))
    (prog1 (car (keim~get up :unicp=stack))
      (keim~put up :unicp=stack (cdr (keim~get up :unicp=stack))))))


;#|
;\subsection{Data for Variables}

;Most information used in the combination algorithm is stored in the variables.
;For this purpose a class is defined whose instances are stored in the sym~label
;slot and that contains several slots in order to store this information.  This
;class is a subclass of keim+object, i.e. it has a property list and the
;elementary theories should use this property list if they want to store
;additional data in the variables. |#

(eval-when (load compile eval)
 (defclass unicp+var-label (keim+object)
   ((theories       :initform nil
		    :accessor unicp~var-label-theories)
    (identified     :initarg  :identified
		    :accessor unicp~var-label-identified)
    (discrim-lists  :initform nil
		    :accessor unicp~var-label-discrim-lists)
    (theory         :initform nil
		    :accessor unicp~var-label-theory)
    (not-theory     :initform nil  
		    :accessor unicp~var-label-not-theory)
    (less           :initform nil
		    :accessor unicp~var-label-less)
    (not-less       :initform nil
		    :accessor unicp~var-label-not-less)
    (greater        :initform nil
		    :accessor unicp~var-label-greater)
    (not-greater    :initform nil
		    :accessor unicp~var-label-not-greater)
    (mark           :initform nil
		    :accessor unicp~var-label-mark))
   (:documentation
    "This is the abstract superclass of all elementary unification problems used in
    the combination algorithm with constraints.")))

(defun unicp=var-label-create (var)
  (make-instance 'unicp+var-label :identified (list var)))


; Since we want to call some algorithms that use the old label slot we redefine
; the the sym~label and sym~set-label! methods: if the label of a variable
; contains a unicp+var-label object, they now access its mark slot.  (This works
; only if the sym~label and sym~set-label! access the mark slot via the
; variable; via the shared symbol the original label slot is returned.  This is
; neccessary in order to access the unicp+var-label object itself.  Therefore
; only algorithms which do not access the label slot via the shared symbol
; themselves can be used with the unicp+var-labels in effect.)

#+small-keim
(defun unicp~orig-label (var)
  (skeim~unicp-label var))

#+small-keim
(defun unicp~set-orig-label! (var value)
  (skeim~set-unicp-label! value var))

#-small-keim
(defun unicp~orig-label (var)
  (shsym~label (sym~shared-symbol var)))

#-small-keim
(defun unicp~set-orig-label! (var value)
  (shsym~set-label! (sym~shared-symbol var) value))

#-small-keim
(defmethod sym~label ((var sym+var))
  (let ((label (unicp~orig-label var)))
    (if (typep label 'unicp+var-label)
	(unicp~var-label-mark label)
	label)))

#-small-keim
(defmethod sym~set-label! ((var sym+var) value)
  (let ((label (unicp~orig-label var)))
    (if (typep label 'unicp+var-label)
	(setf (unicp~var-label-mark label) value)
	(unicp~set-orig-label! var value))))


(defun unicp~vars-label-initialize! (variables)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A list of variables.")
	   (effect  "The label slot of each variable is set to a new var-label"
		    "object.")
	   (value   "Undefined."))
  (mapc #'(lambda (var)
	    (unicp~set-orig-label! var (unicp=var-label-create var)))
	variables))


;;; Selectors for var-labels

(defmethod unicp~var-label-theories ((sym sym+sym))
  (unicp~var-label-theories (unicp~orig-label sym)))

(defmethod unicp~var-label-identified ((sym sym+sym))
  (unicp~var-label-identified (unicp~orig-label sym)))

(defmethod unicp~var-label-discrim-lists ((sym sym+sym))
  (unicp~var-label-discrim-lists (unicp~orig-label sym)))

(defmethod unicp~var-label-mark ((sym sym+sym))
  (unicp~var-label-mark (unicp~orig-label sym)))

(defmethod unicp~var-label-less ((sym sym+sym))
  (unicp~var-label-less (unicp~orig-label sym)))

(defmethod unicp~var-label-greater ((sym sym+sym))
  (unicp~var-label-greater (unicp~orig-label sym)))

(defmethod unicp~var-label-theory ((sym sym+sym))
  (unicp~var-label-theory (unicp~orig-label sym)))

(defmethod unicp~var-label-not-theory ((sym sym+sym))
  (unicp~var-label-not-theory (unicp~orig-label sym)))

;;; Modifiers for var-labels

(defmethod (setf unicp~var-label-theories) (value (sym sym+sym))
  (setf (unicp~var-label-theories (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-identified) (value (sym sym+sym))
  (setf (unicp~var-label-identified (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-discrim-lists) (value (sym sym+sym))
  (setf (unicp~var-label-discrim-lists (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-theory) (value (sym sym+sym))
  (setf (unicp~var-label-theory (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-not-theory) (value (sym sym+sym))
  (setf (unicp~var-label-not-theory (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-less) (value (sym sym+sym))
  (setf (unicp~var-label-less (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-not-less) (value (sym sym+sym))
  (setf (unicp~var-label-not-less (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-greater) (value (sym sym+sym))
  (setf (unicp~var-label-greater (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-not-greater) (value (sym sym+sym))
  (setf (unicp~var-label-not-greater (unicp~orig-label sym)) value))

(defmethod (setf unicp~var-label-mark) (value (sym sym+sym))
  (setf (unicp~var-label-mark (unicp~orig-label sym)) value))



;;; Identification


(defun unicp~var-label-repres (var)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A variable.")
	   (effect  "None.")
	   (value   "The representative of VAR (a variable, too).  The"
		    "representative changes if two variables are identified."
		    "The representative might be VAR itself."))
  (first (unicp~var-label-identified var)))

(defun unicp~vars-label-repres (var-list)
  (declare (ignore var-list))
  (break "unicp~~vars-label-repres nicht benutzen!"))

(defun unicp~repres-list (var-list)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A list of variables.")
	   (effect  "None.")
	   (value   "The representatives of VAR-LIST (a list of varibles, too)"
		    "without duplicates."))
  (if (listp var-list)
      (delete-duplicates (mapcar #'unicp~var-label-repres var-list))
      (error "UNICP~~REPRES-LIST: ~A mu eine Liste von Variablen sein." var-list)))

(defun unicp~vars-identified-p (var1 var2)
  (eq (unicp~var-label-repres var1) (unicp~var-label-repres var2)))

(defun unicp~vars-identify! (var1 var2)
  (declare (edited  "20-DEC-1995")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "The two variables are idententified, i.e. their"
		    "identity-classes are unioned.  The representative of VAR1"
		    "becomes the representative of the new identity-class.  Each"
		    "variable in the identity-classes of VAR1 or VAR2 gets this"
		    "new identity-class and for each variable discriminated from"
		    "VAR1 or VAR2 the old identity-classes in the discrim-lists"
		    "slot are replaced by the new one.  Furthermore, in the"
		    "greater- and less-list of the new representative, each"
		    "variable identified with the representative is removed.")
	   (value   "A list of identifications which are consequences of the"
		    "identification of VAR1 and VAR2, i.e. all pairs of"
		    "variables identified with VAR1 and variables identified"
		    "with VAR2 (including VAR1 and VAR2, respectively),  but"
		    "excluding the pair of VAR1 and VAR2."))
  (when (unicp~vars-discriminated-p var1 var2)
    (error "UNICP: Programmierfehler, ~A und ~A sollen identifiziert werden,~
                   sind aber schon diskiminiert." var1 var2))
  (unless (unicp~vars-identified-p var1 var2)
    (let* ((eq-class1 (unicp~var-label-identified var1))
	   (eq-class2 (unicp~var-label-identified var2))
	   (new-eq-class (append eq-class1 eq-class2))
	   (new-discrim-classes (remove-duplicates #+destr delete-duplicates
						   (append (unicp~var-label-discrim-lists var1)
							   (unicp~var-label-discrim-lists var2)))))
      (mapc #'(lambda (var)
		(setf (unicp~var-label-identified var) new-eq-class))
	    new-eq-class)
      (let ((new-repres (unicp~var-label-repres var1)))
	(setf (unicp~var-label-less new-repres)
	      (remove-if #'(lambda (var) (unicp~vars-identified-p var new-repres))
			 (unicp~var-label-less new-repres))
	      (unicp~var-label-greater new-repres)
	      (remove-if #'(lambda (var) (unicp~vars-identified-p var new-repres))
			 (unicp~var-label-greater new-repres))))
      (mapc #'(lambda (eq-class)
		(let ((new-value (cons new-eq-class
				       (remove-if #'(lambda (eq-cl)
						      (or (eq eq-cl eq-class1)
							  (eq eq-cl eq-class2)))
						  (unicp~var-label-discrim-lists
						   (first eq-class))
						  :count 2))))
		  (mapc #'(lambda (var)
			    (setf (unicp~var-label-discrim-lists var)
				  new-value))
			eq-class)))
	    new-discrim-classes)
      (mapcan #'(lambda (v1)
		  (mapcan #'(lambda (v2)
			      (if (and (eq v1 var1) (eq v2 var2))
				  nil
				  (list (unicp~id-create v1 v2))))
			  eq-class2))
	      eq-class1))))



;;; Discrimination


(defun unicp~vars-discriminated (var)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A variable.")
	   (effect  "None.")
	   (value   "All variables that are discriminated from VAR."))
  (apply #'append (unicp~var-label-discrim-lists var)))

(defun unicp~vars-discriminated-p (var1 var2)
  (member (unicp~var-label-identified var1) (unicp~var-label-discrim-lists var2)
	  :test #'equal))

(defun unicp~vars-discriminate! (var1 var2)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "The two variables are discriminated.  The variables"
		    "identified with VAR1 are added to the discim-lists of the"
		    "variables identified with VAR2 (including VAR1 and VAR2,"
		    "respectively), and vice versa.")
	   (value   "A list of discriminitions which are consequences of the"
		    "discrimination of VAR1 and VAR2, i.e. all pairs of"
		    "variables identified with VAR1 and variables identified"
		    "with VAR2 (including VAR1 and VAR2, respectively),  but"
		    "excluding the pair of VAR1 and VAR2."))
  (cond ((unicp~vars-identified-p var1 var2)
	 (error "UNICP: Programmierfehler, ~A und ~A sollen diskiminiert werden,~
                        sind aber schon identifiziert." var1 var2))
	((unicp~vars-discriminated-p var1 var2)
	 nil)
	(t
	 (let* ((eq-class1 (unicp~var-label-identified var1))
		(eq-class2 (unicp~var-label-identified var2))
		(new-discrim-classes1 (cons eq-class2
					    (unicp~var-label-discrim-lists var1)))
		(new-discrim-classes2 (cons eq-class1
					    (unicp~var-label-discrim-lists var2))))
	   (mapc #'(lambda (var)
		     (setf (unicp~var-label-discrim-lists var) new-discrim-classes1))
		 eq-class1)
	   (mapc #'(lambda (var)
		     (setf (unicp~var-label-discrim-lists var) new-discrim-classes2))
		 eq-class2)
	   (mapcan #'(lambda (v1)
		       (mapcan #'(lambda (v2)
				   (if (and (eq v1 var1) (eq v2 var2))
				       nil
				       (list (unicp~id-create v1 v2))))
			       eq-class2))
		   eq-class1)))))


;;; Ordering

(defun unicp~vars-lesseq-p (var1 var2)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "None.")
	   (value   "True, iff VAR1 is less than VAR2."))
  (member var1 (unicp~var-label-less var2) :test #'term~equal))

(defun unicp~vars-less-p (var1 var2)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "None.")
	   (value   "True, iff VAR1 is less than VAR2 and VAR1 is discriminated"
		    "from VAR2."))
  (and (unicp~vars-lesseq-p var1 var2)
       (some #'(lambda (eq-class)
		 (member var1 eq-class :test #'term~equal))
	     (unicp~var-label-discrim-lists var2))))


;;; Labels for theories and copying

(defun unicp~constant-p (var theory)
  (declare (edited  "18-OCT-1996")
	   (authors Felix)
	   (input   "A variable and a theory.")
	   (effect  "None.")
	   (value   "Is the variable a constant in the theory?"
		    "I.e. belongs the variable to another theory or is the variable"
		    "not allowed to belong to this theory?"))
  (let
      ((var-label-theory (unicp~var-label-theory var))
       (var-label-not-theory (unicp~var-label-not-theory var)))
    (or (and
	 (not (null var-label-theory))
	 (not (eq theory var-label-theory)))
	(member theory var-label-not-theory))))
      

(defun unicp~shared-variable-p (var)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A variable.")
	   (effect  "None.")
	   (value   "True, iff VAR is a shared variable, i.e. it occurs in"
		    "equations of more than one theory."))
  (rest (unicp~var-label-theories var)))

(defun unicp~var-label (var theory)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A variable and a theory.")
	   (effect  "None.")
	   (value   "Each variable has a label for each theory (in the property"
		    "list of its label-object).  This function returns VAR's"
		    "label for THEORY and it can also be used with setf to set"
		    "the label to a new value."))
  (keim~get (unicp~orig-label var) theory))

(defsetf unicp~var-label (var theory) (new-label)
  `(progn (keim~put (unicp~orig-label ,var) ,theory ,new-label)
          ,new-label))

(defgeneric unicp~var-label-copy (object)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A label object.")
	   (effect  "None.")
	   (value   "A copy of OBJECT.  All slots except IDENTIFIED and"
		    "DISCRIM-LISTS are copied with COPY-TREE.  The theory labels"
		    "in the property list are copied with UNICP~TH-VALUE-COPY,"
		    "i.e. with the method specified by the corresponding theory."))
  (:method ((var-label unicp+var-label))
     (let ((new-label (unicp=var-label-create nil)))
       (setf (unicp~var-label-theories new-label)
	     (copy-tree (unicp~var-label-theories var-label)))
       (setf (unicp~var-label-identified new-label)
	     #+destr(copy-tree (unicp~var-label-identified var-label))
	     (copy-tree (unicp~var-label-identified var-label)))
       (setf (unicp~var-label-discrim-lists new-label)
	     #+destr(copy-tree (unicp~var-label-discrim-lists var-label))
	     (copy-tree (unicp~var-label-discrim-lists var-label)))
       (setf (unicp~var-label-theory new-label)
	     (copy-tree (unicp~var-label-theory var-label)))
       (setf (unicp~var-label-not-theory new-label)
	     (copy-tree (unicp~var-label-not-theory var-label)))
       (setf (unicp~var-label-less new-label)
	     (copy-tree (unicp~var-label-less var-label)))
       (setf (unicp~var-label-not-less new-label)
	     (copy-tree (unicp~var-label-not-less var-label)))
       (setf (unicp~var-label-greater new-label)
	     (copy-tree (unicp~var-label-greater var-label)))
       (setf (unicp~var-label-not-greater new-label)
	     (copy-tree (unicp~var-label-not-greater var-label)))
       (do* ((ptail (keim~plist var-label) (rest (rest ptail)))
	     (key (car ptail) (car ptail))
	     (value (second ptail) (second ptail)))
	    ((null ptail) nil)
	 (keim~put new-label key
		   (if (th~theory-p key)
		       (unicp~th-value-copy key value)
		       value)))
       new-label)))

(defgeneric unicp~th-value-copy (theory value)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A theory (which specifies the method that will be executed)"
		    "and a value.")
	   (effect  "None.")
	   (value   "A copy of VALUE.  THEORY determines how to copy VALUE; the"
		    "default is to use COPY-TREE."))
  (:method ((theory th+general) value)
     (copy-tree value)))

(defun unicp~term-label (term)
  (term~label term))

(defun unicp~term-set-label! (term label)
  (term~set-label! term label))


;;; Abstract Datatypes for equations, identifications, constant restrictions, ...


(defun unicp~eq-create (lhss rhss)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "Two list of terms of equal length.")
	   (effect  "None.")
	   (value   "An equations-object consisting of the left-hand sides LHSS"
		    "and the right-hand sides RHSS."))
  (list lhss rhss))

(defun unicp~eq-lhss (equations)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "An equations-object.")
	   (effect  "None.")
	   (value   "The left-hand sides of the equations; a list of terms."))
  (first equations))

(defun unicp~eq-rhss (equations)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "An equations-object.")
	   (effect  "None.")
	   (value   "The right-hand sides of the equations; a list of terms."))
  (second equations))

(defun unicp~eq-append (equations1 equations2)
  (unicp~eq-create (append (unicp~eq-lhss equations1)
			   (unicp~eq-lhss equations2))
		   (append (unicp~eq-rhss equations1)
			   (unicp~eq-rhss equations2))))



(defun unicp~id-create (var1 var2)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "None.")
	   (value   "The identification-object consisting of the two variables"
		    "VAR1 and VAR2."))
  (list var1 var2))

(defun unicp~id-left (id)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "An identification-object.")
	   (effect  "None.")
	   (value   "The left variable of the identification."))
  (first id))

(defun unicp~id-right (id)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "An identification-object.")
	   (effect  "None.")
	   (value   "The right variable of the identification."))
  (second id))


(defun unicp~vi-create (var theory)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "A variable and a theory.")
	   (effect  "None.")
	   (value   "The variable-index, indexing VAR with THEORY."))
  (list var theory))

(defun unicp~vi-var (var-index)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "A variable-index")
	   (effect  "None.")
	   (value   "The variable which is indexed."))
  (first var-index))

(defun unicp~vi-theory (var-index)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "A variable-index")
	   (effect  "None.")
	   (value   "The theory of VAR-INDEX."))
  (second var-index))


(defun unicp~cr-create (const var-list)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "A variable and a list of variables (or another variable).")
	   (effect  "None.")
	   (value   "The constant restriction consisting of CONST and VAR-LIST."
		    "This means CONST must not occur in any variable in VAR-LIST,"
		    "or in other words, CONST is greater than each variables in VAR-LIST."))
  (if (sym~variable-p var-list)
      (list const var-list)
      (cons const var-list)))

(defun unicp~cr-const (const-restr)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "A constant-restriction.")
	   (effect  "None.")
	   (value   "The variable which is the restricted constant in CONST-RESTR."))
  (first const-restr))

(defun unicp~cr-varlist (const-restr)
  (declare (edited  "31-OCT-1995")
	   (authors Richts)
	   (input   "A constant-restriction.")
	   (effect  "None.")
	   (value   "The list of variables in CONST-RESTR."))
  (rest const-restr))


;;; Algorithms for regular and collaps-free theories

(defclass unicp+standard-up (unicp+elementary-up)
  ()
  (:documentation "This is the class of all unification problem,
                   for which no special algorithm is defined."))

(defun unicp=standard-up-create (general-up equations theory)
  (make-instance 'unicp+standard-up
		 :general-up general-up :equations equations :theory theory))

(defmethod unicp~new-constr ((up unicp+standard-up) ident discrim constants const-restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A standard unification problem and four lists of constraints.")
	   (effect  "None.")
	   (value   "Multiple value: UP and six NIL.  Since there is no special"
		    "algorithm for this theory, no constraints can be computed."))
  (declare (ignore ident discrim constants const-restr))
  (values up nil nil nil nil nil nil))

(defmethod unicp~solvable-p ((up unicp+standard-up))
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A standard unification problem.")
	   (effect  "None.")
	   (value   "True, iff UP is solvable with the restrictions made so far."
		    "For standard unification problems (i.e. if no special"
		    "algorithm for unification with constraints exists) the"
		    "decision procedures for unification problems with constant"
		    "restrictions must be called."))
  (let* ((theory (unicp~theory up))
	 (variables (unicp~variables up))
	 (reps (unicp~repres-list variables))
	 (ident (subst~create variables (mapcar #'unicp~var-label-repres variables)))
	 (constants (remove-if #'(lambda (th) (or (null th) (eq th theory)))
			       reps :key #'unicp~var-label-theory))
	 (const-restr (mapcar #'(lambda (const)
				  (unicp~cr-create const (unicp~var-label-less const)))
			      constants)))
    (th~decide-unify theory
		     (subst~apply ident (unicp~eq-lhss (unicp~equations up)))
		     (subst~apply ident (unicp~eq-rhss (unicp~equations up)))
		     const-restr
		     constants)))

(defmethod unicp~theo-init ((th th+general) gup equations ident discrim consts const-restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A theory, a general unification problem, some equations,"
		    "and four lists of constraints.")
	   (effect  "None.")
	   (value   "A new standard unification problem is created.  Since"
		    "nothing is known about THEORY, no constraints can be set"
		    "up.  The result is a multiple value: the new unification"
		    "problem and six NIL."))
  (declare (ignore ident discrim consts const-restr))
  (let ((up (unicp=standard-up-create gup equations th)))
    (values up nil nil nil nil nil nil)))

(defmethod unicp~theo-init ((th th+regular) gup equations ident discrim consts const-restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A regular theory, a general unification problem, some"
		    "equations, and four lists of constraints.")
	   (effect  "None.")
	   (value   "A new standard unification problem is created.  Since"
		    "THEORY is only known to be regular, nothing but the"
		    "following constraints can be set up:  For each equation"
		    "x=t, the variables in t must be less (or equal) than x. The"
		    "result is a multiple value: the new unification problem,"
		    "four NIL, the computed constraints, and one more NIL."))
  (declare (ignore ident discrim consts const-restr))
  (let ((up (unicp=standard-up-create gup equations th)))
    (values up nil nil nil nil
	    (unicp=standard-const-restr equations)
	    nil)))

(defmethod unicp~theo-init ((th th+collapse-free) gup equations ident discrim consts const-restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A collaps-free theory, a general unification problem, some"
		    "equations, and four lists of constraints.")
	   (effect  "None.")
	   (value   "A new standard unification problem is created.  Since"
		    "THEORY is only known to be collaps-free, nothing but the"
		    "following constraints can be set up:  For each equation"
		    "x=t, x must have the theory index TH. The result is a"
		    "multiple value: the new unification problem, two NIL, the"
		    "computed constraints, and three more NIL."))
  (declare (ignore ident discrim consts const-restr))
  (let ((up (unicp=standard-up-create gup equations th)))
    (values up nil nil
	    (unicp=standard-theory-indices equations)
	    nil nil nil)))
		    
(defmethod unicp~theo-init ((th th+fine) gup equations ident discrim consts const-restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A fine theory (i.e. a theory that is regular and"
		    "collaps-free), a general unification problem, some"
		    "equations, and four lists of constraints.")
	   (effect  "None.")
	   (value   "A new standard unification problem is created.  Since"
		    "THEORY is only known to be collaps-free and regular,"
		    "nothing but the constraints computed in these cases (see"
		    "the corresponding methods) can be set up. The result is a"
		    "multiple value: the new unification problem, two NIL, the"
		    "computed theory indices, another NIL, the computed constant"
		    "restrictions, and one more NIL."))
  (declare (ignore ident discrim consts const-restr))
  (let ((up (unicp=standard-up-create gup equations th)))
    (values up nil nil
	    (unicp=standard-theory-indices equations)
	    nil
	    (unicp=standard-const-restr equations)
	    nil)))

(defmethod unicp~theo-init ((th th+simple) gup equations ident discrim consts const-restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A simple theory, a general unification problem, some"
		    "equations, and four lists of constraints.")
	   (effect  "None.")
	   (value   "A new standard unification problem is created.  Since"
		    "THEORY is only known to be simple, nothing but the"
		    "constraints computed for collaps-free and regular theories"
		    "(see the corresponding methods) and the following"
		    "constraints can be set up: For each equation x=t, the"
		    "variables in t must be discriminated from x. The result is"
		    "a multiple value: the new unification problem, NIL, the"
		    "computed discriminations, the computed theory indices, NIL,"
		    "the computed constant restrictions, and one more NIL."))
  (declare (ignore ident discrim consts const-restr))
  (let ((up (unicp=standard-up-create gup equations th)))
    (values up nil
	    (unicp=standard-discrim equations)
	    (unicp=standard-theory-indices equations)
	    nil
	    (unicp=standard-const-restr equations)
	    nil)))

(defun unicp=standard-const-restr (equations)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Some equations.")
	   (effect  "None.")
	   (value   "A list of constant restrictions.  For each equation x=t it"
		    "contains x >= V(t) where V(t) are the variables in t."))
  (mapcan #'(lambda (term1 term2)
	      (cond ((and (sym~variable-p term1) (sym~variable-p term2))
		     nil)
		    ((sym~variable-p term1)
		     (let ((vars (top~free-variables term2)))
		       (if vars
			   (list (unicp~cr-create term1 vars)))))
		    ((sym~variable-p term2)
		     (let ((vars (top~free-variables term1)))
		       (if vars
			   (list (unicp~cr-create term2 vars)))))
		    (t nil)))
	  (unicp~eq-lhss equations)
	  (unicp~eq-rhss equations)))

(defun unicp=standard-theory-indices (equations)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Some equations.")
	   (effect  "None.")
	   (value   "A list of variables.  For each equation x=t it contains x."))
  (mapcan #'(lambda (term1 term2)
	      (cond ((and (sym~variable-p term1) (sym~variable-p term2))
		     nil)
		    ((sym~variable-p term1)
		     (list term1))
		    ((sym~variable-p term2)
		     (list term2))
		    (t nil)))
	  (unicp~eq-lhss equations)
	  (unicp~eq-rhss equations)))

(defun unicp=standard-discrim (equations)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Some equations.")
	   (effect  "None.")
	   (value   "A list of discriminations.  For each equation x=t it"
		    "contains a discrimination of x and all variables in t."))
  (mapcan #'(lambda (term1 term2)
	      (cond ((and (sym~variable-p term1) (sym~variable-p term2))
		     nil)
		    ((sym~variable-p term1)
		     (mapcar #'(lambda (var)
				 (unicp~id-create term1 var))
			     (top~free-variables term2)))
		    ((sym~variable-p term2)
		     (mapcar #'(lambda (var)
				 (unicp~id-create term2 var))
			     (top~free-variables term1)))
		    (t nil)))
	  (unicp~eq-lhss equations)
	  (unicp~eq-rhss equations)))

#|

(defmethod unicp~theo-init ((theory th+ac) general-up equations
			    identified discriminated constants const-restr)
  (call-next-method))


(defmethod unicp~theo-init ((theory th+aci) general-up equations
			    identified discriminated constants const-restr)
  (call-next-method))


(defmethod unicp~theo-init ((theory th+free) general-up equations
			    identified discriminated constants const-restr)
  (call-next-method))



(defmethod unicp~theo-init ((theory th+ac) general-up equations
			    identified discriminated constants const-restr)
  (unicp=standard-up-create general-up equations theory))


(defmethod unicp~theo-init ((theory th+aci) general-up equations
			    identified discriminated constants const-restr)
  (unicp=standard-up-create general-up equations theory))


(defmethod unicp~theo-init ((theory th+free) general-up equations
			    identified discriminated constants const-restr)
  (unicp=standard-up-create general-up equations theory))

(defmethod th~decide-unify ((th th+ac) tl1 tl2 restr const-vars)
  (ac~unify tl1 tl2 restr const-vars t))
|#


(defun unicp=var-label-identified (v)
  (declare (ignore v))
  (break "unicp=var-label-identified nicht mehr benutzen!"))

(defun unicp=var-label-not-identified (v)
  (declare (ignore v))
  (break "unicp=var-label-not-identified nicht mehr benutzen!"))

