;;; -*- Package: HADES; Syntax: Common-lisp; Mode: LISP -*-

(in-package "HADES")

;;; solving DIOphantine equations with the method of Eric Domenjoud

(defvar dioed*trace nil)

(defun dioed-solve.equations (matrix &optional (constrained 0))
  (declare (edited  "16-NOV-1991 11:37")
	   (authors RICHTS)
	   (input   "A m*n-matrix (a list of m lists of n numbers).")
	   (effect  "None.")
	   (value   "The minimal solutions of the diophantine equations (a list of vectors).")
	   (example "'((1 1 -2 0) (2 0 -1 -1)) has the solutions '([1 1 1 1] [2 0 1 3] [1 3 2 0])."))
  (let* ((*m* (length matrix))
	 (*n* (length (car matrix))))
    (declare (special *m* *n*))
    (let ((result (dioed=minimal.solutions (dioed=basis (dioed=convert.matrix matrix)))))
      (when dioed*trace (format dioed*trace "~%Result: ~A" (my=print result)))
      ;result
      (delete-if #'(lambda (solution)
		     (do ((i 0 (1+ i))
			  (sum 0 (+ sum (elt solution i))))
			 ((= i constrained) (> sum 1))))
		 result))))

#|
(defun dioed-solve.equations (matrix)
  (let* ((*m* (length matrix))
	 (*n* (length (car matrix))))
    (declare (special *m* *n*))
    (let* ((new-matrix (dioed=convert.matrix matrix))
	   (basis (time (dioed=basis new-matrix)))
	   (result nil)) ;(time (dioed=minimal.solutions basis))))
      (when dioed*trace (format dioed*trace "~%Result: ~A" (my=print result)))
      result)))
|#

(defmacro dioed=matrix.elt (matrix i j)
  `(elt (elt ,matrix ,i) ,j))


(defmacro 2nd-value (form)
  `(multiple-value-bind (first second)
       ,form
     (ignore first)
     second))

;;; Basis solutions

(defun dioed=basis (matrix)
  (declare (edited  "31-JUL-1991 19:01")
	   (authors RICHTS)
	   (input   "A m*n-matrix (an array).")
	   (effect  "None.")
	   (value   "A set of solutions S0(MATRIX) (a list of vectors)."
		    " They are the minimal solution with minimal carrier or equivalent"
		    " the solution of the rows, whose convex hull contains (0 ... 0), or equivalent"
		    " the determinants |ei0 ... eim| of each subset {Ai0 ... A1m} of rows of MATRIX."
		    "                  |Ai0 ... Aim|    ")
	   (special *n* *m*))
  (let* ((determinants (dioed=determinants *m* *n* matrix))
	 (basis (mapcan #'(lambda (subset)
			    (dioed=basis-solution subset matrix determinants))
			(dioed=subsets (1+ *m*) *n*)))
	 (minimal-basis (dioed=minimal.set basis)))
    (when dioed*trace
      (if (= (length basis) (length minimal-basis))
	  (format dioed*trace "~%Basis Solutions:  ~A"
		  (my=print minimal-basis))
	  (format dioed*trace "~%Basis Solutions(minimal):  ~A"
		  (my=print minimal-basis))))
    minimal-basis))
#|      (do ((i 0 (1+ i)))
	  ((= i (length determinants)))
	(format t "~% ~3D:  ~3D , ~3D" i (car (aref determinants i)) (cdr (aref determinants i)))))))|#

(defun dioed=basis-solution (subset matrix determinants)
  (declare (edited  "31-JUL-1991 19:50")
	   (authors RICHTS)
	   (input   "A subset {i0,...,im} of {0,...,n} (a list of numbers),"
		    "an m*n-Matrix (an Array) and the allready computed subdeterminants of MATRIX.")
	   (effect  "Subdeterminants of MATRIX, computed for the first time, are stored into DETERMINANTS.")
	   (value   "If there exists a solution for [Ai1 ... Aim+1] then a list of that solution:"
		    "  |ei0 ... eim| "
		    "  |Ai0 ... Aim| divided by the gcd of its elements (a list of a vector of length n),"
		    " else nil.")
	   (special *n*))
  (let* ((bits (mapcar #'(lambda (n) (expt 2 n)) subset))
	 (bit-all (apply #'+ bits))
	 (solution (make-array *n* :initial-element 0)))
    (do* ((i 0 (1+ i))
	  (subset-tail subset (cdr subset-tail))
	  (bits-tail bits (cdr bits-tail))
	  (sign 1 (- sign))
	  (coeff (setf (elt solution (car subset-tail))
		       (* sign (dioed=basis-coeff subset i (- bit-all (car bits-tail)) matrix determinants)))
		 (setf (elt solution (car subset-tail))
		       (* sign (dioed=basis-coeff subset i (- bit-all (car bits-tail)) matrix determinants))))
	  (gcdiv coeff (gcd gcdiv coeff))
	  (gcd-sign (if (zerop (signum coeff)) 0 (signum coeff))
		    (cond ((zerop gcd-sign) (if (zerop (signum coeff)) 0 (signum coeff)))
			  ((zerop (signum coeff)) gcd-sign)
			  ((= gcd-sign (signum coeff)) gcd-sign)
			  (t nil))))
	 ((or (null (cdr subset-tail)) (null gcd-sign))
	  (when (and gcd-sign (plusp gcdiv))
	    (if (= 1 gcdiv gcd-sign)
		(list solution)
		(do ((factor (/ gcd-sign gcdiv))
		     (i 0 (1+ i)))
		    ((= i *n*) (list solution))
		  (setf (elt solution i) (* (elt solution i) factor)))))))))

(defun dioed=basis-coeff (columns without bit matrix determinants)
  (declare (edited  "31-JUL-1991 20:55")
	   (authors RICHTS)
	   (input   "A set of m+1 numbers (i0..ik..im) and a number k between 0 and m"
		    "which represent n columns (i0..ik-1 ik+1 im) of the matrix,"
		    "a number which represent the same columns or nil,"
		    "a m*n-matrix (an array) and the allready computed subdeterminants of matrix.")
	   (effect  "If a new subdeterminant is computed its value is stored in determinants.")
	   (value   "The value of the determinant of the submatrix which consist of the last n rows of the n columns.")
	   (special *m*))
  (if (zerop without)
      (dioed=determinant (cdr columns) bit matrix determinants)
      (let* ((without-1-pair (nthcdr (1- without) columns))
	     (without-pair (cdr without-1-pair)))
	(setf (cdr without-1-pair) (cdr without-pair))
	(prog1
	  (dioed=determinant columns bit matrix determinants)
	  (setf (cdr without-1-pair) without-pair)))))

#|alte Version:
(defun dioed=basis-solution (subset matrix determinants)
  (declare (edited  "31-JUL-1991 19:50")
	   (authors RICHTS)
	   (input   )
	   (effect  )
	   (value   )
	   (special *n*))
  (let* ((bits (mapcar #'(lambda (n) (expt 2 n)) subset))
	 (bit (apply #'+ bits))
	 (solution (make-array *n* :initial-element 0))
	 (header (list '***)))
    (dioed=compute.basis-solution header header subset bit bits 1 matrix determinants solution)
    (if (dioed=test.basis-solution solution)
	(list solution))))

(defun dioed=compute.basis-solution (set1 last-set1 set2 bit-all bit2 sign matrix determinants result)
  (when set2
    (let ((rest-set2 (cdr set2)))
      (setf (cdr last-set1) (cdr set2))
      (setf (aref result (car set2)) (* sign (dioed=determinant (cdr set1) (- bit-all (car bit2)) matrix determinants)))
      (setf (cdr last-set1) set2
	    (cdr set2) nil)
      (dioed=compute.basis-solution set1 (cdr last-set1) rest-set2 bit-all (cdr bit2) (- sign) matrix determinants result))))

(defun dioed=test.basis-solution (solution)
  (declare (edited  "04-AUG-1991 14:38")
	   (authors RICHTS)
	   (input   )
	   (effect  )
	   (value   )
	   (special *n*))
  (let ((sign 0)
	(gcdiv 0))
    (when (every #'(lambda (coeff)
		     (cond ((zerop coeff) t)
			   ((zerop sign)
			    (setf sign (signum coeff)
				  gcdiv coeff))
			   ((= sign (signum coeff))
			    (setf gcdiv (gcd gcdiv coeff)))
			   (t nil)))
		 solution)
      (setf gcdiv (* sign gcdiv))
      (when (/= gcdiv 1)
	(do ((i 0 (1+ i)))
	    ((= i *n*))
	  (setf (aref solution i) (/ (aref solution i) gcdiv))))
      t)))
|#

;;; minimal solutions

(defun dioed=minimal.solutions (basis-solutions)
  (declare (edited  "16-NOV-1991 12:10")
	   (authors RICHTS)
	   (input   "The set of the basis-solutions S0(A) of a m*n-matrix A (a list of vectors of length n).")
	   (effect  "None.")
	   (value   "The set of all minimal solutions of the diophantine equations Ax=0 (a list of vectors)."))
  (when basis-solutions
    (let ((n (length basis-solutions)))
      (dioed=minimal.set (dioed=linear.independent.basis basis-solutions (1- (expt 2 n)) n nil nil)
			 basis-solutions))))

(defun dioed=linear.independent.basis (basis-solutions subset pos subsets solutions)
  (declare (edited  "16-NOV-1991 12:13")
	   (authors RICHTS)
	   (input   "The basis-solutions S0(A) (a list of vectors of length n),"
		    "a subset of these solutions (a number representing this subset with its bits),"
		    "the number of one solution saying which subsets need not to be regarded,"
		    "a set of subsets of BASIS-SOLUTIONS for which a solution has allready been computed (a list of numbers)"
		    "and a list of the solutions that have allready been computed (a list of vectors).")
	   (effect  "None.")
	   (values  "If the SUBSET of BASIS-SOLUTIONS P contains linear independent vectors, all linear combinations"
		    "   with rational coefficients 0<=a<1 of the vectors of P which are vectors with natural numbers"
		    "   are computed;"
		    "else this is done for for the subsets of P."
		    "The linear combination of P is not computed if it has been computed for a superset of P"
		    " (P is a subset of a member of SUBSETS)."
		    "1. A list containing the new solutions and SOLUTIONS (a list of vectors)."
		    "2. A list containing the SUBSETS and the subsets of BASIS-SOLUTIONS for which new solutions"
		    "   have been computed (a list of numbers)."))
  (if (member subset subsets :test #'dioed=subset.p)
      (values solutions subsets)
      (let ((subset-basis-solutions (dioed=basis.subset basis-solutions subset)))
	(multiple-value-bind (new-solutions independent)
	    (dioed=solutions.of.independent.basis
	      subset-basis-solutions
	      (dioed=transpose.matrix (apply #'vector subset-basis-solutions)))
	  (if independent
	      (values (nconc new-solutions solutions) (cons subset subsets))
	      (do ((i 0 (1+ i))
		   (set-i 1 (ash set-i 1)))
		  ((= i pos) (values solutions subsets))
		(multiple-value-setq (solutions subsets)
		  (dioed=linear.independent.basis basis-solutions (logxor subset set-i) i subsets solutions))))))))

(defun dioed=solutions.of.independent.basis (basis-solutions matrix-m)
  (declare (edited  "16-NOV-1991 12:38")
	   (authors RICHTS)
	   (input   "A subset of basis-solutions S0(A) (a list of vectors of length n) and"
		    "an n*k-matrix which contains these vectors as columns"
		    "(a list of n vectors of length k, each vector representing a row).")
	   (effect  "MATRIX-M is destroyed.")
	   (values  "1. If the vectors in BASIS-solutions are linear independent, all linear combinations"
		    "   with rational coefficients 0<=a<1 of the vectors of BASIS-SOLUTIONS which are"
		    "   vectors with natural numbers, and NIL else."
		    "2. True, iff the vectors in BASIS-solutions are linear independent."))
  (let ((matrix-t (dioed=transform.matrix matrix-m 0)))
    (if matrix-t
	(let ((diagonal (dioed=diagonal matrix-t))
	      (ddd (when dioed*trace (format dioed*trace "~%M = ~A~%T:  ~A ^ -1"
					  (my-print basis-solutions nil) (my-print matrix-t nil))))
	      (matrix-t-invers (dioed=invert.matrix matrix-t)))
	  (when dioed*trace (format dioed*trace "~%  = ~A" (my-print matrix-t-invers nil) ddd))
	  (values (dioed=linear.combinations basis-solutions diagonal matrix-t-invers)
		  t))
	(values nil nil))))

(defun dioed=transform.matrix (matrix row)
  (declare (edited  "16-NOV-1991 14:46")
	   (authors RICHTS)
	   (input   "A n*k-matrix (a list with n vektors with k elements, each vector representing a row)"
		    "   where the rows from 0 to ROW-1 are allready transformed and"
		    "a number.")
	   (effect  "The matrix is transformed destructivly into a matrix where the first k rows are"
		    "an upper triangle matrix and the other rows are 0.")
	   (value   "The changed MATRIX if it could be transformed into an upper triangle matrix,"
		    "NIL else."))
  (let ((pivot (dioed=matrix.min matrix row)))
    (when pivot
      #+kklisp(until (dioed=matrix.add.rows matrix row pivot)
	(setf pivot (dioed=matrix.min matrix row)))
      (do ((piv (dioaed=matrix.min matrix row) (dioaed=matrix.min matrix row)))
	  ((dioaed=matrix.add.rows matrix row piv) (setf pivot piv)))
      (let ((pivot-row (elt matrix pivot)))
	(when (minusp (elt pivot-row row))
	  (dioed=vector.* -1 pivot-row row))
	(setf (elt matrix pivot) (elt matrix row)
	      (elt matrix row) pivot-row))
      (if (= (1+ row) (length (elt matrix 0)))
	  matrix
	  (dioed=transform.matrix matrix (1+ row))))))
  
(defun dioed=matrix.min (matrix column)
  (declare (edited  "16-NOV-1991 14:53")
	   (authors RICHTS)
	   (input   "A n*k-matrix (a list with n vektors with k elements, each vector representing a row) and"
		    "a number representing a column in MATRIX")
	   (effect  "None.")
	   (value   "The number of the row (from row COLUMN row n) with the smallest number not equal to 0"
		    "  in column COLUMN (i.e. the j with Mjc=min({Mcc,...,Mnc}\{0}), if there is one,"
		    "NIL, if all these numbers are 0."))
  (let ((matrix-length (length matrix)))
    (do* ((coeff nil (elt (elt matrix i) column))
	  (minimum nil
		   (if minimum
		       (if (< 0 (abs coeff) (abs (elt (elt matrix minimum) column))) i minimum)
		       (unless (zerop coeff) i)))
	  (i column (1+ i)))	 
	((= i matrix-length) minimum))))

(defun dioed=matrix.add.rows (matrix column pivot)
  (declare (edited  "16-NOV-1991 15:03")
	   (authors RICHTS)
	   (input   "A n*k-matrix (a list with n vektors with k elements, each vector representing a row) and"
		    "two numbers, the fist representing the actual column and row of the matrix transformation"
		    "and the second representing the number of the row with smallest =/= 0 in column COLUMN.")
	   (effect  "All rows in MATRIX from COLUMN to n are subtracted by an integer multiple of row PIVOT"
		    "such that the new element in column COLUMN of the row is as small as possible.")
	   (value   "True, iff all elements in column COLUMN (without that in row PIVOT) has been set to 0."))
  (do ((all.zero t)
       (i (1- (length matrix)) (1- i)))
      ((< i column) all.zero)
    (when (/= i pivot)
      (let* ((row (elt matrix i))
	     (pivot-row (elt matrix pivot))
	     (row-length (length row)))
	(multiple-value-bind (factor rest)
	    (truncate (elt row column) (elt pivot-row column))
	  (unless (zerop rest) (setf all.zero nil))
	  (do ((j column (1+ j)))
	      ((= j row-length))
	    (setf (elt row j) (- (elt row j) (* factor (elt pivot-row j))))))))))

(defun dioed=diagonal (matrix)
  (declare (edited  "16-NOV-1991 15:25")
	   (authors RICHTS)
	   (input   "A n*k-matrix (a list with n vektors with k elements, each vector representing a row).")
	   (effect  "None.")
	   (value   "A list of the elements on the diagonal of matrix: (M00,...,Mkk)."))
  (do* ((i (length (elt matrix 0)) (1- i))
	(diagonal nil (cons (elt (elt matrix i) i) diagonal)))
       ((zerop i) diagonal)))

(defun dioed=invert.matrix (matrix)
  (declare (edited  "16-NOV-1991 15:27")
	   (authors RICHTS)
	   (input   "A n*k-matrix (a list with n vektors with k elements, each vector representing a row)")
	   (effect  "MATRIX is changed destructively.")
	   (value   "The inverted MATRIX."))
  (do* ((i (1- (length (elt matrix 0))) (1- i))
	(pivot (dioed=matrix.elt matrix i i) (dioed=matrix.elt matrix i i)))
       ((zerop i) (progn (setf (dioed=matrix.elt matrix 0 0) (/ pivot))
			 (dioed=vector.* (/ pivot) (elt matrix 0) 1)
			 matrix))
    (setf (dioed=matrix.elt matrix i i) (/ pivot))
    (dioed=vector.* (/ pivot) (elt matrix i) (1+ i))
    (do* ((j (1- i) (1- j)))
	 ((minusp j))
      (let ((elt-ji (dioed=matrix.elt matrix j i)))
	(setf (dioed=matrix.elt matrix j i) (- (/ elt-ji pivot)))
	(dioed=row.- (elt matrix j) elt-ji (elt matrix i) (1+ i))))))
      
(defun dioed=linear.combinations (basis-solutions diagonal t-invers)
  (declare (edited  "16-NOV-1991 15:28")
	   (authors RICHTS)
	   (input   "A subset of the basis solutions S0(A) (a list of k vectors with n elements),"
		    "the diagonal of the corresponding matrix T (a list of k numbers) and"
		    "T^-1 ,a k*k-matrix (a list of k vectors).")
	   (effect  "None.")
	   (value   "The linear combinations of vectors in BASIS-SOLUTIONS (a list of vectors):"
		    "  BASIS-SOLUTIONS * [T^-1 * x] "
		    "where [z] is the floor of z and x is a vector with x<<DIAGONAL ,i.e. x(i)<D(i) f.a. i=0,...,k."))
  (let ((x-vector (map 'vector #'identity diagonal)))
    (dioed=compute.solutions basis-solutions x-vector t-invers (1- (length x-vector)))))

(defun dioed=compute.solutions (basis-solutions x-vector t-invers pos)
  (declare (edited  "16-NOV-1991 15:39")
	   (authors RICHTS)
	   (input   "A set of solutions (a list of k vectors with n elements),"
		    "a vector with k elements, a k*k-matrix (a list of k vectors) and"
		    "a number representing a position in x")
	   (effect  "None, the numbers in X-VECTOR are drecremented but restored in the end.")
	   (value   "The linear combinations of vectors in BASIS-SOLUTIONS (a list of vectors):"
		    "  BASIS-SOLUTIONS * [T^-1 * x]  "
		    "where [z] is the floor of z and x is a vector with x(i)<X-VECTOR(i) f.a. i=0,...,POS."))
  (let ((start-value (elt x-vector pos)))
    (do* ((value start-value (decf (elt x-vector pos)))
	  (solutions nil
		     (nconc (if (zerop pos)
				(dioed=compute.solution basis-solutions x-vector t-invers)
				(dioed=compute.solutions basis-solutions x-vector t-invers (1- pos)))
			    solutions)))
	((zerop value)
	 (progn (setf (elt x-vector pos) start-value)
		solutions)))))

(defun dioed=compute.solution (basis-solutions x-vector t-invers)
  (declare (edited  "16-NOV-1991 15:47")
	   (authors RICHTS)
	   (input   "A set of solutions (a list of k vectors with n elements),"
		    "a vector with k elements and a k*k-matrix (a list of k vectors).")
	   (effect  "None.")
	   (value   "A list with the linear combination of the vectors in BASIS-SOLUTIONS (a list with one vector):"
		    "  BASIS-SOLUTIONS * [T^-1 * X-VECTOR]  where [z] is the floor of z."))
  (unless (every #'zerop x-vector)
    (let* ((alpha (dioed=compute.alpha x-vector t-invers))
	   (solution (dioed=vector.add basis-solutions alpha)))
      (when dioed*trace (format dioed*trace "~%X: ~A  ->  alpha: ~A  ->  ~55Tsolution: ~A"
			     (my-print x-vector nil) (my-print alpha nil) (my-print solution nil)))
      (list solution))))

(defun dioed=compute.alpha (x-vector t-invers)
  (declare (edited  "16-NOV-1991 15:56")
	   (authors RICHTS)
	   (input   "A vector with k elements and a k*k-matrix (a list of k vectors).")
	   (effect  "None.")
	   (value   "The floor of the product of T-INVERS and X-VECTOR: [T^-1 * X-VECTOR]"))
  (do* ((i (length x-vector) (1- i))
	(alpha nil
	       (cons (2nd-value (floor (dioed=vector.scalar.product x-vector (elt t-invers i) i)))
		     alpha)))
       ((zerop i) alpha)))


;;; Matrices and determinants


(defun dioed=convert.matrix (matrix)
  (declare (edited  "31-JUL-1991 18:57")
	   (authors RICHTS)
	   (input   "A m*n-matrix (a list of m lists with n integers).")
	   (effect  "None.")
	   (value   "An array containing the numbers of MATRIX.")
	   (special *m* *n*))
  (let ((array (make-array (list *m* *n*) :element-type 'integer)))
    (do ((i 0 (1+ i))
	 (rows matrix (cdr rows)))
	((= i *m*))
      (do ((j 0 (1+ j))
	   (coeffs (car rows) (cdr coeffs)))
	  ((= j *n*))
	(setf (aref array i j) (car coeffs))))
    array))

(defun dioed=transpose.matrix (matrix)
  (declare (edited  "04-AUG-1991 15:05")
	   (authors RICHTS)
	   (input   "A n*k-matrix (a vector of k vectors with n elements).")
	   (effect  "None.")
	   (value   "The transposed k*n-matrix (a vector of n vectors with k elements)."))
  (let* ((m (length matrix))
	 (n (length (elt matrix 0)))
	 (new-matrix (make-array n)))
    (do ((i 0 (1+ i)))
	((= i n))
      (do ((new-row (make-array m))
	   (j 0 (1+ j)))
	  ((= j m)
	   (setf (elt new-matrix i) new-row))
	(setf (elt new-row j) (elt (elt matrix j) i))))
    new-matrix))

(defun dioed=determinants (m n matrix)
  (declare (edited  "20-AUG-1991 15:43")
	   (authors RICHTS)
	   (input   "Two numbers and a m*n-matrix (an array).")
	   (effect  "None.")
	   (value   "A vector where the value of the subdeterminants of matrix can be stored."))
  (ignore m matrix)
  (make-array (1- (expt 2 n)) :initial-element nil))

(defun dioed=determinant (columns bit-or-nil matrix determinants)
  (declare (edited  "31-JUL-1991 20:55")
	   (authors RICHTS)
	   (input   "A list of n numbers which represent the clomuns the matrix,"
		    "nil or a number which represent the same columns,"
		    "a matrix (an array) and the allready computed subdeterminants of matrix (a vector).")
	   (effect  "If a new subdeterminant is computed its value is stored in DETERMINANTS.")
	   (value   "The value of the determinant of the submatrix which consist of the last n rows of the n COLUMNS.")
	   (special *m*))
  (let* ((n (length columns))
	 (row (- *m* n))
	 (bit (if bit-or-nil bit-or-nil (apply #'+ (mapcar #'(lambda (n) (expt 2 n)) columns))))
	 (result (aref determinants bit)))
    (cond (result
	   (incf (car result))
	   (cdr result))
	  ((> n 2)
	   (do ((column-tail columns (cdr column-tail))
		(sign 1 (- sign))
		(sum 0 (+ sum (* sign (aref matrix row (car column-tail))
				 (dioed=determinant (remove (car column-tail) columns) nil matrix determinants)))))
	       ((null column-tail)
		(progn (setf (aref determinants bit) (cons 0 sum)) sum))))
	  ((= n 2) (let ((res (- (* (aref matrix row (car columns)) (aref matrix (1+ row) (cadr columns)))
				 (* (aref matrix (1+ row) (car columns)) (aref matrix row (cadr columns))))))
		     (setf (aref determinants bit) (cons 0 res))
		     res))
	  ((= n 1) (aref matrix row (car columns)))
	  (t (break "n < 1")))))

(defun dioed=row.- (row1 factor row2 &optional (start 0))
  (declare (edited  "16-NOV-1991 16:26")
	   (authors RICHTS)
	   (input   "A vector of length n, a number, another vector length n and "
		    "a number representing a position in the vectors.")
	   (effect  "A multiple of ROW2 is subtracted from ROW1:  for i := START to n  ROW1(i) := ROW1(i) - FACTOR * ROW2(i).")
	   (value   "The changed ROW1."))
  (do ((i (1- (length row1)) (1- i)))
      ((< i start) row1)
    (setf (elt row1 i) (- (elt row1 i) (* factor (elt row2 i))))))

(defun dioed=vector.add (vectors factors)
  (declare (edited  "16-NOV-1991 16:31")
	   (authors RICHTS)
	   (input   "A list of n vectors and n numbers.")
	   (effect  "None.")	
	   (value   "The sum of the VECTORS multiplied with the coresponding FACTOR."))
  (let ((result (make-array (length (car vectors)) :initial-element 0)))
    (mapc #'(lambda (vector factor)
	      (dioed=vector.+ result factor vector))
	  vectors factors)
    result))

(defun dioed=vector.scalar.product (vector1 vector2 &optional (start 0))
  (declare (edited  "16-NOV-1991 16:35")
	   (authors RICHTS)
	   (input   "Two vectors of length n and number representing a position in the vectors.")
	   (effect  "None.")
	   (value   "The scalar product of the vectors: V1(START)*V2[START) + ... + V1(n) * V2(n)."))
  (do* ((i (length vector1) (1- i))
	(sum 0 (+ sum (* (elt vector1 i) (elt vector2 i)))))
       ((= i start) sum)))

(defun dioed=vector.* (factor vector &optional (start 0))
  (declare (edited  "16-NOV-1991 16:37")
	   (authors RICHTS)
	   (input   "A number, a vector of length n and a number representing a position in VECTOR.")
	   (effect  "Each element in VECTOR is multiplied with FACTOR.")
	   (value   "The changed VECTOR."))
  (do ((i (1- (length vector)) (1- i)))
      ((< i start) vector)
    (setf (elt vector i) (* factor (elt vector i)))))

(defun dioed=vector.+ (vector1 factor vector2)
  (declare (edited  "16-NOV-1991 16:39")
	   (authors RICHTS)
	   (input   "A vector of length n, a number and another vector of length n.")
	   (effect  "A multiple of VECTOR2 is added to VECTOR1:   VECTOR1 := FACTOR * VECTOR2.")
	   (value   "The changed VECTOR1."))
  (do ((i (1- (length vector1)) (1- i)))
      ((< i 0) vector1)
    (setf (elt vector1 i) (+ (elt vector1 i) (* factor (elt vector2 i))))))

(defun dioed=subsets (m end)
  (declare (edited  "20-AUG-1991 15:45")
	   (authors RICHTS)
	   (input   "Two numbers.")
	   (effect  "None.")
	   (value   "All subsets with m elements which are numbers in [0 end[."))
  (cond ((= m 0) (list nil))
	((> m end) nil)
	(t (let ((subsets1 (dioed=subsets (1- m) (1- end))))
	     (mapl #'(lambda (subsets-tail)
		       (push (1- end) (car subsets-tail)))
		   subsets1)
	     (nconc (dioed=subsets m (1- end)) subsets1)))))



;;; searching minimal solutions


(defun dioed=minimal.set (solutions &optional minimal-set)
  (declare (edited  "16-NOV-1991 16:42")
	   (authors RICHTS)
	   (input   "A set of solutions (a list of vectors) and"
		    "a set of solutions which is minimal (another list of vectors).")
	   (effect  "MINIMAL-SET is destroyed.")
	   (value   "The set with the minimal solutions of SOLUTIONS and MINIMAL-SET (a list of vectors)."))
  (delete nil (dioed=insert.minimal.solutions solutions minimal-set)))

(defun dioed=insert.minimal.solutions (solutions minimal-set)
  (declare (edited  "16-NOV-1991 16:45")
	   (authors RICHTS)
	   (input   "A set of solutions (a list of vectors) and"
		    "a set of solutions which is minimal (another list of vectors and NILs).")
	   (effect  "MINIMAL-SET is destroyed.")
	   (value   "The set with the minimal solutions of SOLUTIONS and MINIMAL-SET (a list of vectors)."))
  (if solutions
      (let ((solution (car solutions)))
	(when nil dioed*trace (format t ;dioed*trace
				      "~%Insert ~A in minimal set ~A." (my-print solution nil) (my-print minimal-set nil)))
	(if #+kklisp(somel #'(lambda (solution-tail)
		      (when (car solution-tail)
			(multiple-value-bind (sol1.<=.sol2 sol1.>=.sol2)
			    (dioed=solution.<= solution (car solution-tail))
			  (cond (sol1.>=.sol2 t)
				(sol1.<=.sol2 (setf (car solution-tail) nil))
				(t nil)))))
		  minimal-set)
	    (do ((solution-tail minimal-set (rest solution-tail))
		 (flag nil))
		((or flag (null solution-tail)) flag)
		(when (car solution-tail)
		      (multiple-value-bind (sol1.<=.sol2 sol1.>=.sol2)
			  (dioaed=solution.<= solution (car solution-tail))
			(cond (sol1.>=.sol2 (setf flag t))
			      (sol1.<=.sol2 (setf (car solution-tail) nil))))))
	    (dioed=insert.minimal.solutions (cdr solutions) minimal-set)
	    (dioed=insert.minimal.solutions (cdr solutions) (cons solution minimal-set))))
      minimal-set))

(defun dioed=solution.<= (solution1 solution2)
  (declare (edited  "16-NOV-1991 16:48")
	   (authors RICHTS)
	   (input   "Two vectors of equal length.")
	   (effect  "None.")
	   (values  "1. True, iff SOLUTION1 <= SOLUTION2"
		    "2. True, iff SOLUTION2 <= SOLUTION1"
		    "Note that <= is only a partial order."))
  ;(format t "  ~A  ~A" (my-print solution1 nil) (my-print solution2 nil))
  (do* ((i (length solution1) (1- i))
	(sol1.<=.sol2 t (and sol1.<=.sol2 (<= (elt solution1 i) (elt solution2 i))))
	(sol1.>=.sol2 t (and sol1.>=.sol2 (>= (elt solution1 i) (elt solution2 i)))))
       ((or (zerop i) (and (not sol1.<=.sol2) (not sol1.>=.sol2)))
	(values sol1.<=.sol2 sol1.>=.sol2))))


;;; sets as numbers


(defun dioed=subset.p (set1 set2)
  (declare (edited  "16-NOV-1991 16:51")
	   (authors RICHTS)
	   (input   "Two sets (two numbers which represent sets].")
	   (effect  "None.")
	   (value   "True, iff SET1 is a subset of SET2."))
  (= (logand set1 set2) set1))

(defun dioed=basis.subset (basis-solutions subset)
  (declare (edited  "16-NOV-1991 16:52")
	   (authors RICHTS)
	   (input   "A set (a list) and a number representing a subset of BASIS-SOLUTIONS.")
	   (effect  "None.")
	   (value   "A list of the elements of BASIS-SOLUTIONS in SUBSET."))
  (do ((position (expt 2 (1- (length basis-solutions))) (ash position -1))
       (solutions basis-solutions (cdr solutions))
       (solutions-subset nil (if (plusp (logand position subset))
				 (cons (car solutions) solutions-subset)
				 solutions-subset)))
      ((null solutions) (nreverse solutions-subset))))


;;; Bounds (not needed)

(defun dioed=bound (rows)
  (declare (edited  "20-AUG-1991 15:30")
	   (authors RICHTS)
	   (input   "A list of lists of number representing the rows of matrix.")
	   (effect  "None.")
	   (value   "A number which is a bound for the length (degree)"
		    "of all minimal positive integer solutions of matrix.")
	   (special *m* *n*))
  (let* ((*m* (length rows))
	 (*n* (length (car rows)))
	 (matrix (dioed=convert.matrix rows))
	 (determinants (dioed=determinants *m* *n* matrix)))
    (declare (special *m* *n*))
    (do* ((column-lists (dioed=subsets (1+ *m*) *n*) (cdr column-lists))
	  (determinant (dioed=bound.determinant (car column-lists) matrix determinants)
		       (dioed=bound.determinant (car column-lists) matrix determinants))
	  (maximum determinant (max maximum (abs determinant))))
	 ((null (cdr column-lists))
	  (let ((bound (* (- *n* *m*) maximum)))
	    (format t "~%Bound: ~A" bound)
	    bound)))))

(defun dioed=bound.determinant (columns matrix determinants)
  (do ((column-tail columns (cdr column-tail))
       (sign 1 (- sign))
       (sum 0 (+ sum (* sign (dioed=determinant (remove (car column-tail) columns) nil matrix determinants)))))
      ((null column-tail)
       (progn (format t "~%~A:  ~A" columns sum) sum))))


;;; printing

(defun my-print (object &optional (stream t))
  (if (equal stream t)
      (format t "~%~A" (my=print object))
      (format stream "~A" (my=print object))))

(defun my=print (object)
  (cond ((vectorp object) (format nil "[~{~A ~}]" (map 'list #'(lambda (object) (my=print object)) object)))
	((listp object) (mapcar #'(lambda (object) (my=print object)) object))
	(t object)))

#|


; Domenjoud-Beispiele:

(my-print (setf res (dioed-solve.equations '((1 2 -3 -2 -4) (2 -1 -3 2 5)))))    ;Bsp. 1:  10 loesungen
(my-print (setf res (dioed-solve.equations '((10 -7 -8 3 -11) (12 -9 -7 3 13)))))      ;Bsp. 2: 240 loesungen
(my-print (setf res (dioed-solve.equations '((1 2 -1 0 -2 -1) (0 -1 -2 2 0 1) (2 0 1 -1 -2 0)))))      ;Bsp. 3: 13 loesungen
(my-print (setf res (dioed-solve.equations '((-10 0 20 -1 -21) (9 1 -17 2 19)))))       ;Bsp. 3: 0 loesungen
(my-print (setf res (dioed-solve.equations '((6 4 -5 -4 -1) (3 5 2 -2 -4)))))          	;Bsp. ED: 8 Loesungen
(my-print (setf res (dioed-solve.equations '((1 2 -2 -1) (1 0 -2 0)))))         	;Bsp. CD: 2 Loesungen

(my-print (setf res (dioed-solve.equations '((1 1 -1 -1)))))

(my-print (setf res (dioed-solve.equations '((1 2 -1 -1 -2)))))
(my-print (setf res (dioed-solve.equations '((2 1 -1 -1 -1)))))

(my-print (setf res (dioed-solve.equations '((1 1 -2 0) (2 0 -1 -1)))))
(my-print (setf res (dioed-solve.equations '((1 1 -1 -1) (2 0 -2 0)))))
(my-print (setf res (dioed-solve.equations '((1 1 -1 -1) (2 0 -1 -1) (1 0 -1 0)))))


;ac-test 9:
(my-print (setf res (dioed-solve.equations '((-1 -1 1 1 1 -1 1)) 5)))
(my-print (setf res (dioed-solve.equations '((1 1 1 -1 -1 -1)))))


(my-print (setf res (dioed=invert.matrix (vector (vector 1 -5 -2) (vector 0 9 0) (vector 0 0 3)))))
(my-print (setf res (dioed=invert.matrix (vector (vector 1 0 -4) (vector 0 1 -3) (vector 0 0 6)))))

(my-print (setf res (dioed-solve.equations '((2 2 -3)))))
(my-print (setf res (dioed-solve.equations '((-3 2 -1)))))
(my-print (setf res (dioed-solve.equations '((2 2 -3) (-3 2 -1)))))

(with-open-file (out "F:>richts>trace>dioed" :direction :output)
  (setf dioed*trace out)
  (dioed-solve.equations '((1 2 -3 -2 -4) (2 -1 -3 2 5)))
  (terpri dioed*trace)(terpri dioed*trace)(terpri dioed*trace)
  (dioed-solve.equations '((1 2 -1 0 -2 -1) (0 -1 -2 2 0 1) (2 0 1 -1 -2 0)))
  (terpri dioed*trace)(terpri dioed*trace)(terpri dioed*trace)
  (my-print (setf res (dioed-solve.equations '((6 4 -5 -4 -1) (3 5 2 -2 -4)))))
  (setf dioed*trace nil))
|#

#+weg(defmethod print-object :around ((object vector) stream)
  (if (eq (class-of object) (find-class 'vector))
      (format stream "[~{~A ~}]" nil (map 'list #'(lambda (object) object) object))
    (call-next-method)))

(defun pr (object stream)
  (if (eq (class-of object) (find-class 'vector))
      (format stream "[~{~A ~}]" (map 'list #'(lambda (object) object) object))
    (print object stream)))

(defmethod print-object :around ((object vector) stream)
  (if (eq (class-of object) (find-class 'vector))
      (format stream "[~{~A ~}]" (coerce object 'list))
    (call-next-method)))
