(in-package :KEIM)

#|
Theory:
Smolka and Treinen's FT
see: Gert Smolka & Ralf Treinen: Records for Logic Programming
Journal of Logic Programming, 18(3), April 1994, pp. 229-258
Restriction: We only allow conjunctions of atomic constraints.
|#



;+-----------------------------------------------------------------------------------+
;+ Slots for variables:
;+ binding, sort, features, minors
;+
;+ The value of the binding slot is another variable.
;+ The sort slot contains a sort predicate.
;+ The features slot is a list of (f . v) pairs where f is a feature predicate
;+ and v is a variable.
;+ The minors slot is a list of variables. It contains the transitive closure
;+ of the variables reachable via the features. All these variables must be
;+ smaller than the variable in the ordering, hence "minors".
;+
;+ If the binding slot is non-empty, the information in the other three slots
;+ is meaningless and ignored. The variable was identified with the one in the
;+ binding slot. The real information is to be found at the end of the binding chain.
;+ If the binding slot is empty, the variable is a represetant.
;+ 
;+-----------------------------------------------------------------------------------+


(defmethod cpfeat=binding ((var sym+var))
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable.")
	   (effect  "None.")
	   (value   "The value of the variable's binding slot."))
  (declare (special cpfeat*theory))
  (first (unicp~var-label var cpfeat*theory)))


(defmethod cpfeat=sort ((var sym+var))
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable.")
	   (effect  "None.")
	   (value   "The value of the variable's sort slot."))
  (declare (special cpfeat*theory))
  (second (unicp~var-label var cpfeat*theory)))


(defmethod cpfeat=features ((var sym+var))
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable.")
	   (effect  "None.")
	   (value   "The value of the variable's features slot."))
  (declare (special cpfeat*theory))
  (third (unicp~var-label var cpfeat*theory)))


(defmethod cpfeat=minors ((var sym+var))
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable.")
	   (effect  "None.")
	   (value   "The value of the variable's minors slot."))
  (declare (special cpfeat*theory))
  (fourth (unicp~var-label var cpfeat*theory)))


(defmethod cpfeat=set-binding! ((var sym+var) binding)
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable and a binding (also a variable).")
	   (effect  "The variable's binding slot is set to binding.")
	   (value   "The binding."))
  (declare (special cpfeat*theory))
  (cond ((term~equal var binding)
	 (error "cpfeat: trying to bind variable ~A to itself!" var))
	((unicp~var-label var cpfeat*theory)
	 (setf (first (unicp~var-label var cpfeat*theory)) binding))
	(T (setf (unicp~var-label var cpfeat*theory)
		 (list binding nil nil nil)))))


(defmethod cpfeat=set-sort! ((var sym+var) sort)
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable and a sort.")
	   (effect  "The variable's sort slot is set to sort.")
	   (value   "The sort."))
  (declare (special cpfeat*theory))
  (if (unicp~var-label var cpfeat*theory)
      (setf (second (unicp~var-label var cpfeat*theory)) sort)
    (setf (unicp~var-label var cpfeat*theory) (list nil sort nil nil))))


(defmethod cpfeat=set-features! ((var sym+var) features)
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable and a list of features.")
	   (effect  "The variable's features slot is set to features.")
	   (value   "The features."))
  (declare (special cpfeat*theory))
  (if (unicp~var-label var cpfeat*theory)
      (setf (third (unicp~var-label var cpfeat*theory)) features)
    (setf (unicp~var-label var cpfeat*theory) (list nil nil features nil))))


(defmethod cpfeat=add-feature! ((var sym+var) feature)
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable and a feature.")
	   (effect  "The feature is added to the variable's features slot.")
	   (value   "The features."))
  (declare (special cpfeat*theory))
  (if (unicp~var-label var cpfeat*theory)
      (setf (third (unicp~var-label var cpfeat*theory))
	    (cons feature (cpfeat=features var)))
    (setf (unicp~var-label var cpfeat*theory) (list nil nil (list feature) nil))))


(defmethod cpfeat=set-minors! ((var sym+var) minors)
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable and a list of minor variables.")
	   (effect  "The variable's minors slot is set to minors.")
	   (value   "The minors."))
  (declare (special cpfeat*theory))
  (if (unicp~var-label var cpfeat*theory)
      (setf (fourth (unicp~var-label var cpfeat*theory)) minors)
    (setf (unicp~var-label var cpfeat*theory) (list nil nil nil minors))))


;+------------------------------------------------------------------------+
; sorts
;+------------------------------------------------------------------------+


(defun cpfeat=sort-unify (sort1 sort2)
  (declare (authors Kepser)
	   (input   "Two sorts.")
	   (effect  "None.")
	   (value   "The largest common subsort, if the sorts are "
		    "unifiable; nil else."))
  (cond ((and (null sort1) (null sort2))
	 :top)
	((null sort1)
	 sort2)
	((null sort2)
	 sort1)
	((term~equal sort1 sort2)
	 sort1)
	(T ; nontrivial case
	 (rest (assoc sort2 (fth~sort-unifiers sort1) :test #'term~equal))
	 )))


(defun cpfeat=sort-constraint-p (constraint)
  (declare (authors Kepser)
	   (input   "A constraint.")
	   (effect  "None.")
	   (value   "Non-nil, iff constraint is a sort constraint. "
		    "We take sort constraints to be unary predicates."))
  (not (rest (appl~arguments constraint))))



;+------------------------------------------------------------------------+
;+ unification of two feature structres
;+ a merge/find algorithm
;+ see e.g. "Integrating efficient records into concurrent constraint
;+  programming" by Peter Van Roy, Michael Mehl and Ralf Scheidhauer
;+ in: 8th International Symposium on Programming Languages, Implementations,
;+ Logics, and Programs (PLILP 96), Aachen, Germany, Sept. 1996.
;+------------------------------------------------------------------------+


(defun cpfeat=find! (var)
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A variable.")
	   (effect  "All binding slots in the binding chain are set to the end "
		    "of the binding chain.")
	   (value   "The end of the binding chain."))
  (declare (special cpfeat*theory))
  (let ((binding (cpfeat=binding var)))
    (if binding
	(cpfeat=set-binding! var (cpfeat=find! binding))
      var))) 


(defun cpfeat=unify-pairlist (var-pairs)
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A list of variable pair to be unified.")
	   (effect  "Destructively unifies all variable pairs.")
	   (value   "Non-nil, iff all variable pairs are unifiable."))
  (declare (special cpfeat*theory))
  (if (endp var-pairs)
      T ; unification succeeded!
      (if (cpfeat=unify (caar var-pairs) (cdar var-pairs))
	  (cpfeat=unify-pairlist (rest var-pairs))
	NIL ; unification failed!
	)))


(defun cpfeat=merge (f-list var &optional (var-pairs nil))
  (declare (edited  "01-SEP-1997")
	   (authors Stephan)
	   (input   "A list of features and a variable.")
	   (effect  "Merges (destructively) the f-list into the variable's features "
		    "slot. If a feature is not present in the features slot, it is "
		    "simply added. If a feature is present, the two variables that the "
		    "feature points to are added to the list of variable pairs to be "
		    "unified.")
	   (value   "Non-nil, iff the f-list can be merged into the features slot."))
  (declare (special cpfeat*theory))
  (if (endp f-list)
      var-pairs
    (let ((x (rest (assoc (caar f-list) (cpfeat=features var) :test #'term~equal))))
      (if x
	  (cpfeat=merge (rest f-list) var
			(acons x (cdar f-list) var-pairs))
	(progn
	  (cpfeat=add-feature! var (first f-list))
	  (cpfeat=merge (rest f-list) var var-pairs)))))) 


(defun cpfeat=unify-in (x y sort)
  (declare (authors Kepser)
	   (edited  "26-AUG-1997")
	   (input   "Two variables x and y, and a sort.")
	   (effect  "FT-unifies y (the smaller one) into x, making x "
		    "be the result and setting its sort to sort.")
	   (value   "Non-nil, iff that succeeds.")
	   (special cpfeat*theory))
  (if (term~p sort)
      (cpfeat=set-sort! x sort))
  (cpfeat=set-binding! y x)
  (cpfeat=unify-pairlist
   (cpfeat=merge (cpfeat=features y) x)))


(defun cpfeat=unify (x y)
  (declare (authors Kepser)
	   (edited  "26-AUG-1997")
	   (input   "Two variables.")
	   (effect  "Performs (destructive) FT-unification on them.")
	   (value   "Non-nil, iff FT-unification succeeds.")
	   (special cpfeat*theory))
  (if (not (unicp~vars-discriminated-p x y))
      (let* ((x1 (cpfeat=find! x))
	     (y1 (cpfeat=find! y))
	     (newsort (cpfeat=sort-unify (cpfeat=sort x1)
					 (cpfeat=sort y1))))
	(if (term~equal x1 y1)
	    T ; x and y are already unified!
	  (if newsort
	      ; if not newsort, then the unification of x and y failed!
	      (if (> (length (cpfeat=features x1))
		     (length (cpfeat=features y1)))
		  (cpfeat=unify-in x1 y1 newsort)
		(cpfeat=unify-in y1 x1 newsort))))))) 



;+----------------------------------------------------------------------------+
;+ theo-init, new-constr , values                                             +
;+                                                                            +
;+----------------------------------------------------------------------------+


(defclass cpfeat+up (unicp+elementary-up)
  ()
  (:documentation
   "This is the class of feature-constraints problems."))


(defmethod unicp~theo-init ((theory th+feat) general-up constraints
                            identified discriminated constants const-restr)
  (declare (authors Kepser)
	   (edited  "26-AUG-1997")
	   (input   "A feature theory, a general unification problem, the "
		    "equations (constraints), and information on identified, "
		    "discriminated variables, constants and constant restrictions.")
	   (effect  "Creates a new feature theory unification problem. "
		    "Initially identifies all variable pairs from identified. "
		    "Processes the constraints creating feature trees. "
		    "Finally new information is calculated.")
           (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "  solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "  of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "  unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "  $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "  unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "  solution is lost if $y$ is in the constant"
		    "  restrictions of $x$).")
  	   (ignore discriminated constants const-restr))
  (let* ((up (make-instance 'cpfeat+up
			    :general-up general-up
			    :equations constraints
			    :theory theory))
	 (cpfeat*theory theory)
	 (variables (top~free-variables (unicp~equations up))))
    (declare (special cpfeat*theory))
    (if identified
	(cpfeat=initial-identification identified))
    (if (cpfeat=process-constraints constraints)
	(progn
	  (cpfeat=all-sorted variables)
	  (multiple-value-bind (new-ident new-discr new-indexed)
	      (cpfeat=new-info variables)
	    (values
	     up  new-ident  new-discr  new-indexed  nil
	     (cpfeat=new-order (cpfeat=calc-minors variables))
	     nil)))
	nil)))



(defun cpfeat=initial-identification (identified)
  (declare (authors Kepser)
	   (edited  "26-AUG-1997")
	   (effect  "Identifies all variable pairs of identified by "
		    "setting the binding slot of one to point to the other. "
		    "On inital identification, no feature structures "
		    "exist yet, thus no FT-unification is needed.")
	   (value   "None.")
	   (special cpfeat*theory))
  (when (not (endp identified))
    (if (not (term~equal (cpfeat=find! (unicp~id-left  (first identified)))
			 (cpfeat=find! (unicp~id-right (first identified)))))
	(cpfeat=set-binding! (cpfeat=find! (unicp~id-left  (first identified)))
			     (cpfeat=find! (unicp~id-right (first identified)))))
    (cpfeat=initial-identification (rest identified)))) 


(defun cpfeat=identify (identified)
  (declare (authors Kepser)
	   (edited  "26-AUG-1997")
	   (effect  "Tries to FT-unify all variable pairs of identified.")
	   (value   "Non-nil, iff all pairs were identified.")
	   (special cpfeat*theory))
  (or (endp identified)
      (and (cpfeat=unify (unicp~id-left  (first identified))
			 (unicp~id-right (first identified)))
	   (cpfeat=identify (rest identified)))))


(defun cpfeat=process-sort-constraint (sc)
  (declare (authors Kepser)
	   (input   "A sort constraint.")
	   (effect  "If the sort and the old sort of the variable "
		    "are unifiable, the sort slot of the variable "
		    "is set to the result of the sort unification.")
	   (value   "The new sort, if sort-unification worked, "
		    "else nil, indicating unsatisfiability.")
	   (special cpfeat*theory))
  (let* ((sort (appl~function sc))
	 (var (cpfeat=find! (first (appl~arguments sc))))
	 (new-sort (cpfeat=sort-unify (cpfeat=sort var) sort)))
    (if new-sort
	(cpfeat=set-sort! var new-sort))))


(defun cpfeat=process-feature-constraint (fc)
  (declare (authors Kepser)
	   (input   "A feature constraint.")
	   (effect  "If the first argument variable does not yet contain "
		    "this feature, it is just added to the features "
		    "list; else it tries to unify the second argument "
		    "variable with whatever this feature already points "
		    "to in the first argument variable's features list.")
	   (value   "Non-nil, if the feature constraint "
		    "is satisfiable, else nil.")
	   (special cpfeat*theory))
  (cpfeat=unify-pairlist
   (cpfeat=merge
    (list (cons (appl~function fc)
		(cpfeat=find! (second (appl~arguments fc)))))
    (cpfeat=find! (first (appl~arguments fc))))))


(defun cpfeat=process-constraints (constraints)
  (declare (authors Kepser)
	   (edited  "01-SEP-1997")
	   (input   "A list of constraints.")
	   (effect  "The feature trees corresponding to the constraints are "
		    "created (in the variables' sort and features slots.")
	   (value   "Non-nil, iff all constraints can be processed without "
		    "causing inconsistencies.")
	   (special cpfeat*theory))
  (cond ((endp constraints)
	 T) ; all constraints are satisfiable
	((cpfeat=sort-constraint-p (first constraints))
	 (and (cpfeat=process-sort-constraint (first constraints))
	      (cpfeat=process-constraints (rest constraints))))
	(T ; it's a feature constraint
	 (and (cpfeat=process-feature-constraint (first constraints))
	      (cpfeat=process-constraints (rest constraints)))))) 


(defun cpfeat=all-sorted (variables)
  (declare (edited  "08-SEP-1997")
	   (authors Kepser)
	   (input   "The variables of the problem.")
	   (effect  "Checks, if each variable's representant has a sort."
		    "If not, issues an error message.")
	   (value   "Not defined. (nil)")
	   (special cpfeat*theory))
  (when (not (endp variables))
    (if (not (cpfeat=sort (cpfeat=find! (first variables))))
	(error "Keim(cpfeat): Underspecified feature theory!~%Variable ~A has no sort!"
	       (first variables))
	(cpfeat=all-sorted (rest variables)))))


(defmethod unicp~new-constr
  ((up cpfeat+up) identified discriminated constants const-restr)
  (declare (authors Kepser)
	   (edited  "26-AUG-1997")
	   (input   "A feature theory unification problem and information on "
		    "identified and discriminated variables, constants and "
		    "constant restrictions.")
	   (effect  "Each variabel pair in identified will be FT-unified.")
           (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "  solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "  of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "  unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "  $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "  unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "  solution is lost if $y$ is in the constant"
		    "  restrictions of $x$).")
  (ignore discriminated const-restr))
  (let ((cpfeat*theory (unicp~theory up))
	(variables (unicp~variables up)))
    (declare (special cpfeat*theory))
    (if identified
	(if (cpfeat=identify identified)
	    (multiple-value-bind (new-ident new-discr new-indexed)
		(cpfeat=new-info variables)
	      (values
	       up  new-ident  new-discr  new-indexed  nil
	       (cpfeat=new-order (cpfeat=calc-minors variables))
	       nil))
	  nil)
      (if constants
	  (values up nil nil nil nil nil
		  (cpfeat=new-order
		   (remove-if #'cpfeat=binding variables))
		  nil)
	(values up nil nil nil nil nil nil nil)))))


(defmethod unicp~solvable-p ((up cpfeat+up))
  t)


;+----------------------------------------------------------------------------+
; new informations: identification, discrimination, theory index
;+----------------------------------------------------------------------------+


(defun cpfeat=new-info (variables &optional  (new-ident NIL)
				  (new-discr NIL) (my-vars NIL))
  (declare (authors Kepser)
	   (edited "01-SEP-1997")
	   (input  "The variables of the feature theory unification problem.")
	   (effect "None.")
	   (value  "New information on identified and discriminated variables "
		   "and theory indices.")
	   (special cpfeat*theory))
  (if (endp variables)
      (values new-ident new-discr my-vars)
    (let ((var (first variables)))
      (cpfeat=new-info
       (rest variables)
       (let ((v (cpfeat=find! var)))
	 (if (or (term~equal v var)
		 (unicp~vars-identified-p v var))
	     new-ident
	   (cons (unicp~id-create var v) new-ident)))
       (cpfeat=new-discr (cpfeat=find! var) (rest variables) new-discr)
       (if (and (not (keim~equal (unicp~var-label-theory var)
				 cpfeat*theory))
		(cpfeat=features (cpfeat=find! var)))
	   (cons var my-vars)
	 my-vars)))))


(defun cpfeat=new-discr (var vars discr)
  (declare (authors Kepser)
	   (edited "01-SEP-1997")
	   (input  "A variable, some other variables, and a list of newly "
		   "discriminated variable pairs.")
	   (effect "None.")
	   (value  "For each v in vars: if v and var are not yet discriminated "
		   "and their sorts don't unify, add the pair (var v) to the "
		   "newly discriminated variable pairs.")
	   (special cpfeat*theory))
  (if (endp vars)
      discr
    (cpfeat=new-discr
     var (rest vars)
     (let ((v (cpfeat=find! (first vars))))
       (if (or (unicp~vars-discriminated-p var v)
	       (cpfeat=sort-unify (cpfeat=sort var)
				  (cpfeat=sort v)))
	   discr
	 (cons (unicp~id-create var v) discr)))))) 


;+----------------------------------------------------------------------+
;+ for ordering information
;+----------------------------------------------------------------------+

(defun cpfeat=initial-minors (variables &optional (reps nil))
  (declare (authors Kepser)
	   (edited  "01-SEP-1997")
	   (input   "The variables of the feature theory unification problem.")
	   (effect  "Initialise the minors slot with the representatives "
		    "of the variables reachable via the features.")
	   (value   "The representatives.")
	   (special cpfeat*theory))
  (if (endp variables)
      reps
    (let ((v (first variables)))
      (if (null (cpfeat=binding v))
	  (cpfeat=set-minors!
	   v (cpfeat=initial-minors-a v (cpfeat=features v))))
      (cpfeat=initial-minors
       (rest variables)
       (if (null (cpfeat=binding v))
	   (cons v reps)
	 reps)))))


(defun cpfeat=initial-minors-a (var features &optional (minors nil))
  (declare (authors Kepser)
	   (edited "01-SEP-1997")
	   (input  "A variable and its features.")
	   (effect "None.")
	   (value  "The variables that the features point to are collected.")
	   (special cpfeat*theory))  
  (if (endp features)
      (remove-duplicates minors :test #'term~equal)
      (cpfeat=initial-minors-a var (rest features)
			       (cons (cpfeat=find! (cdar features)) minors)))) 


(defun cpfeat=ws (variables &optional (vars variables))
  (declare (edited  "13-AUG-1997")
           (authors Kepser)
           (input   "A list of variables")
           (effect  "Transitive closure on the minors slot via the warshall "
                    "algorithm.")
           (value   "None.")
	   (special cpfeat*theory))
  (when (not (endp variables))
    (cpfeat=ws-a (first variables) (cpfeat=minors (first variables)) vars)
    (cpfeat=ws (rest variables) vars)))


(defun cpfeat=ws-a (var succ variables)
  (declare (edited  "13-AUG-1997")
           (authors Kepser)
           (input   "A variable, its succesors and a list of variables.")
           (effect  "For each v of variables: if var is in the minors slot, "
                    "then succ is added to the minors slot.")
           (value   "None.")
	   (special cpfeat*theory))
  (when (not (endp variables))
    (if (member var (cpfeat=minors (first variables)) :test #'term~equal)
        (cpfeat=set-minors! (first variables)
                            (union succ (cpfeat=minors (first variables))
				   :test #'term~equal)))
    (cpfeat=ws-a var succ (rest variables))))


(defun cpfeat=calc-minors (variables)
  (declare (authors Kepser)
	   (edited "01-SEP-1997")
	   (input  "The feature theory unification problem's variables.")
	   (effect "The representants' minors slots are filled appropriately. ")
	   (value  "The representants.")
	   (special cpfeat*theory))
  (let ((reps (cpfeat=initial-minors variables)))
    (cpfeat=ws reps)
    reps))


(defun cpfeat=new-order (variables &optional (new-order nil))
  (declare (authors Kepser)
	   (edited "01-SEP-1997")
	   (input  "A list of variables.")
	   (effect "None.")
	   (value  "New ordering information.")
	   (special cpfeat*theory))
  (if (endp variables)
      new-order
    (let* ((v (first variables))
	   (nmin (cpfeat=find-min (cpfeat=minors v) v)))
      (cpfeat=new-order
       (rest variables)
       (if (null nmin)
	   new-order
	 (cons (unicp~cr-create v nmin) new-order))))))


(defun cpfeat=find-min (variables var &optional (nmin nil))
  (declare (authors Kepser)
	   (edited "01-SEP-1997")
	   (input  "A list of variables and a variable.")
	   (effect "None.")
	   (value  "All those from variables that have a theory index different from "
		   "the current theory and aren't yet specified as being less than "
		   "variable.")
	   (special cpfeat*theory))
  (if (endp variables)
      nmin
    (cpfeat=find-min
     (rest variables) var
     (let* ((v (first variables))
	    (th (unicp~var-label-theory v)))
       (if (or  (null th)  ; has no theory index
		(keim~equal th cpfeat*theory) ; index is equal to current
		(unicp~vars-lesseq-p v var)) ; ordering already known
	   nmin
	 (cons v nmin)))))) 







;+-----------------------------------------------------------------------------
;+ Incremental Version
;+
;+-----------------------------------------------------------------------------

(defmethod unicp~new-equations ((up cpfeat+up) constraints identified
				discriminated constants const-restr) 
  (declare (edited  "1998-09-17")
	   (authors  Kepser)
	   (input   "An elementary unification problem, a list of constraints,"
		    "and some restrictions that are new for this problem:"
		    "a list of pairs of variables which have to be identified,"
		    "a list of pairs of variables which have to be discriminated,"
		    "a list of variables that should be treated as constants,"
		    "a list of constant restrictions (i.e. pairs of a constant"
		    "and a set of variables.")
	   (effect  "The ELEMENTARY-UP is updated with the new restrictions.")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP belongs"
		    "to is called.  It returns NIL if the unification problem is"
		    "not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains some new"
		    "constraints wich must be met if the unification problem should be solvable:"
		    "1. True, indicating that the unification problem is still solvable."
		    "2. A list of new identification of variables (i.e. sets of variables)."
		    "3. A list of pairs of variables which must not be unified."
		    "4. A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6. A list of pairs $(x . y)$ of variables where $\sigma(x)$"
		    "   is a subterm of $\sigma(y)$ for a unifier $\sigma$."
		    "7. A list of pairs $(x . y)$ of variables where no solution"
		    "   is lost if $y$ is in the constant restrictions of $x$.")
	   (ignore discriminated constants const-restr))
    (let ((cpfeat*theory (unicp~theory up))
	  (variables (unicp~variables up)))
      (declare (special cpfeat*theory))
      (if (and (cpfeat=process-constraints constraints)
	       (cpfeat=identify identified))
	  (progn
	    (cpfeat=all-sorted variables)
	    (multiple-value-bind (new-ident new-discr new-indexed)
		(cpfeat=new-info variables)
	      (values
	       up  new-ident  new-discr  new-indexed  nil
	       (cpfeat=new-order (cpfeat=calc-minors variables))
	       nil)))
	  nil)))

