(in-package "KEIM")

;;; Syntactic Unification of Terms
;;; Solutions can be either finite trees or rational trees.
;;; The switch is cpfree*rational, its default is NIL.

(defvar cpfree*rational NIL
  "Switch for solution domain of syntactic unification: finite
vs. rational trees.")


;;; The unification algorithm is a variant of the union-find algorithm.  If
;;; the solution domain is that of finite trees, the standard, the
;;; occurs-check is done at the end by means of a cyclicity test. If rational
;;; trees are the solution domain, this occurs check is simply left out.


;+----------------------------------------------------------------------------+
;+ label, binding, weight                                                     +
;+----------------------------------------------------------------------------+

;;;  ** terms, constants **                      label     binding
;;;
;;;  (unicp~term-label term)                     car       cdr
;;;
;;;
;;;  ** variables **                             label     binding    weight
;;;
;;;  (unicp~var-label var cpfree*theory)    first     second     third

(defgeneric cpfree=label (term)
  (declare (special cpfree*theory))
  (:method ((term term+term))
      (car (unicp~term-label term)))
  (:method ((term sym+const))
      (car (unicp~term-label term)))
  (:method ((var sym+var))
      (first (unicp~var-label var cpfree*theory))))

(defgeneric cpfree=binding (term)
  (declare (special cpfree*theory))
  (:method ((term term+term))
      (cdr (unicp~term-label term)))
  (:method ((term sym+const))
      (cdr (unicp~term-label term)))
  (:method ((var sym+var))
      (second (unicp~var-label var cpfree*theory))))

(defgeneric cpfree=set-label! (term label)
  (declare (special cpfree*theory))
  (:method ((term term+term) label)
      (if (unicp~term-label term)
	  (setf (car (unicp~term-label term)) label)
	(unicp~term-set-label! term (cons label nil))))
  (:method ((term sym+const) label)
      (if (unicp~term-label term)
	  (setf (car (unicp~term-label term)) label)
	(unicp~term-set-label! term (cons label nil))))
  (:method ((var sym+var) label)
	   (if (unicp~var-label var cpfree*theory)
	       (setf (first (unicp~var-label var cpfree*theory)) label)
	     (setf (unicp~var-label var cpfree*theory) (list label nil 0)))))

(defgeneric cpfree=set-binding! (term binding)
  (declare (special cpfree*theory))
  (:method ((term term+term) binding)
      (if (unicp~term-label term)
	  (setf (cdr (unicp~term-label term)) binding)
	(term~set-label! term (cons nil binding))))
  (:method ((term sym+const) binding)
      (if (unicp~term-label term)
	  (setf (cdr (unicp~term-label term)) binding)
	(unicp~term-set-label! term (cons nil binding))))
  (:method ((var sym+var) binding)
	   (if (unicp~var-label var cpfree*theory)
	       (setf (second (unicp~var-label var cpfree*theory)) binding)
	     (setf (unicp~var-label var cpfree*theory) (list nil binding 0)))))

;; For a term u without pointer is the weight of u the number of variables v
;; with (cpfree=collapsing-find! v) = u. 

(defun cpfree=weight (var)
  (declare (special cpfree*theory))
  (let ((weight (third (unicp~var-label var cpfree*theory))))
    (if weight
	weight
      0)))

(defun cpfree=inc-weight! (var amount)
  (declare (special cpfree*theory))
  (if (unicp~var-label var cpfree*theory)
      (setf (third (unicp~var-label var cpfree*theory))
	    (+ (third (unicp~var-label var cpfree*theory)) amount))
    (setf (unicp~var-label var cpfree*theory) (list nil nil amount))))
  
;+----------------------------------------------------------------------------+
;+ variable-p, union, find, variables                                         +
;+----------------------------------------------------------------------------+

#|
(defun cpfree=variable-p (term)
  (declare (special cpfree*theory))
  (and
   (sym~variable-p term)
   (null (unicp~var-label-theory term))
   (eq (unicp~var-label-theory term) cpfree*theory)))

(defun cpfree=variables (up)     ; variables of the problem 
  (mapcan #'                     ; 'cpfree=variable-p'
   (lambda (var)
     (if (cpfree=variable-p var)
	 (list var)
       nil))
   (unicp~variables up)))
|#

(defun cpfree=variable-p (term)
  (sym~variable-p term))

(defun cpfree=variables (up)     ; variables of the problem 
  (unicp~variables up))

(defun cpfree=collapsing-find! (term)  ; value:  end of the binding chain      
  (declare (special cpfree*theory))    ; effect: sets binding of term to value
  (let       
      ((binding (cpfree=binding term)))
    (if (null binding)
	term
      (cpfree=set-binding! term (cpfree=collapsing-find! binding)))))

(defun cpfree=find (term)        ; value: end of the binding chain
  (declare (special cpfree*theory))
  (let
      ((binding (cpfree=binding term)))
    (if (null (cpfree=binding term))
	term
      (cpfree=find binding))))

(defun cpfree=union_with_weight (term1 term2)
  (declare (special cpfree*theory))
  (if
       ; sets pointer from the term with less weight to the pointer with
       ; greater weight
      
      (> (cpfree=weight term1) (cpfree=weight term2))
      (progn
	(cpfree=set-binding! term1 term2)
	(cpfree=inc-weight! term2 (1+ (cpfree=weight term1))))

    (progn
      (cpfree=set-binding! term2 term1)
      (cpfree=inc-weight! term1 (1+ (cpfree=weight term2))))))

(defun cpfree=union (term1 term2)     ; sets pointer from term1 to term2
  (declare (special cpfree*theory))
  (cpfree=set-binding! term1 term2))

(defun cpfree=function (k) ; function of k, if k is application, else k
  (if (appl~p k)
      (appl~function k)
    k))

;+----------------------------------------------------------------------------+
;+ dag, cycle-test                                                            +
;+----------------------------------------------------------------------------+

(defun cpfree=dag-create! (k1 k2)
  (declare (edited  "12-JUL-1995")
	   (authors Felix)
	   (input   "k1, k2: two terms (or nodes)")
	   (effect  "DAG is created")
	   (value   "T if there is no clash-failure, else nil")
	   (special cpfree*theory))

  ; see also unify3-algorithm from Unification Theory by
  ; Franz Baader and Joerg H. Siekmann
  
  (let
      ((temp nil))
    (if (eq k1 k2)
	T
      (progn
	(when (cpfree=variable-p k2)
	      (progn
		(setf temp k1)    ; if there is a var-node, k1 is the var-node
		(setf k1 k2)
		(setf k2 temp)))
	(cond ((cpfree=variable-p k1)  
	       (if (cpfree=variable-p k2)
		   (if (not (term~equal k1 k2))
		       (cpfree=union_with_weight k1 k2))
		 (cpfree=union k1 k2)) 
	       T)
	      ((not (term~equal (cpfree=function k1) (cpfree=function k2)))
	       nil)
	      ((sym~p k1)   ; two equal symbols
	                    ; (constants or constant variables)
	       T)           ; should better not be unioned.
	      (T
	       (cpfree=union k1 k2)
	       (every #'
		(lambda (k1 k2)
		  (cpfree=dag-create!
		   (cpfree=collapsing-find! k1)
		   (cpfree=collapsing-find! k2)))
		(fo~arguments k1) (fo~arguments k2))
	       ))))))


(defun cpfree=cycle-test-a (term)
  (declare (edited  "12-AUG-1997")
	   (authors Kepser)
	   (input   "A term.")
	   (effect  "See below: cpfree=cycle-test.")
	   (value   "See below: cpfree=cycle-test.")
   ;; This is just an interface function. If cpfree*rational, no cycle test
   ;; is needed, so none is done. Else the cycle-test cpfree=cycle-test is
   ;; called.
   )
  (if cpfree*rational
      :cpfree=true
    (cpfree=cycle-test term)))


(defun cpfree=cycle-test (term)
  (declare (edited  "11-AUG-1997")
	   (authors Richts)
	   (input   "A term.")
	   (effect  "The DAG starting at TERM is searched for a cycle and each"
		    "node is labelled with the list of variables occorring in the"
		    "DAG starting at that node."
		    "Before exploring the DAG, TERM is labelled with :cpfree=false,"
		    "so a cycle can be detected.  If no cycle was found (i.e. no"
		    ":cpfree=false was found), TERM is labelled with the variables"
		    "that were found in the DAG (including TERM if it is a variable);"
		    "if there are no variables it is labelled with :cpfree=true.")
	   (value   "NIL if a cycle is detected; otherwise, a list of"
		    "variables occurring in the DAG starting at TERM; or"
		    ":cpfree=true if this list is empty."))
  (declare (special cpfree*theory))
  (let
      ((next (cpfree=binding term))
       (label (cpfree=label term))
       (term-is-var (cpfree=variable-p term)))
           
    (cond
     ((and (null next) term-is-var) (list term)) ; term is var and leaf
                                                 ; in the DAG

     ((eq label :cpfree=false) nil)        ; there is a cycle in the DAG
                                           ; found out before
     
     ((not (null label)) label)            ; term is already checked and
             				   ; labeled with the occuring vars
                                           ; or :cpfree=true
     (T 
      (cpfree=set-label! term :cpfree=false)  ; else a cycle is assumed
	  
      (cond
	   
       ;; if term is var and there is no cycle reachable from the
       ;; succsesor of term, then term is labeled with its occuring vars 
       ;; including itself as o.k., else nil is returned, and term is left as
       ;; :cpfree=false
       
       (term-is-var                       
	(let ((result (cpfree=cycle-test next)))
	  (cond
	    ((eq result :cpfree=true)
	     (cpfree=set-label! term (list term)))
	    (result
	     (cpfree=set-label! term (cons term result)))
	    (T nil))))
	    
       ;; if term is application and there is no cycle reachable from
       ;; the arguments of the application, then the term will be
       ;; labeled with its occuring vars including its own if existing or  
       ;; :cpfree=true as o.k., else nil is returned and term is left
       ;; as :cpfree=false
       
       ((appl~p term)                     
	(let
	    ((result_list
	      (cpfree=cycle-test_list (appl~arguments term) nil))
	     (var-arguments (top~free-variables (appl~arguments term))))
	  (cond
	    ((eq result_list :cpfree=true)
	     (if var-arguments
		 (cpfree=set-label! term var-arguments)
		 (cpfree=set-label! term :cpfree=true)))
	    (result_list
	     (cpfree=set-label! term (append var-arguments result_list)))
	    (T nil))))

       ;; if term is const, :cpfree=true is returned

       (T (cpfree=set-label! term :cpfree=true) :cpfree=true))))))
       

(defun cpfree=cycle-test_list (termlist buffer)
  ;;
  ;;  input by calling from outside: termlist nil 
  ;;  effect: applies cpfree=cycle-test to all terms in the list
  ;;  value:  - if cpfree=cycle-test of all terms is :cpfree=true
  ;;            then :cpfree=true
  ;;          - if cpfree=cycle-test of all terms is their occuring vars or
  ;;            :cpfree=true then the lists of the occuring vars are appended
  ;;            and returned
  ;;          - if cpfree=cycle-test of one term in the list is nil then nil
  ;;
  (declare (special cpfree*theory))
  (if (null termlist)
      (if (null buffer)
	  :cpfree=true
	(remove-duplicates buffer))
    (let
	((result (cpfree=cycle-test (car termlist))))
      (cond
       ((null result) nil)
       ((eq result :cpfree=true)
	(cpfree=cycle-test_list (cdr termlist) buffer))
       (T (cpfree=cycle-test_list (cdr termlist) (append result buffer))))))) 

;+----------------------------------------------------------------------------+
;+ theo-init, new-constr , values                                             +
;+                                                                            +
;+----------------------------------------------------------------------------+

(defclass cpfree+up (unicp+elementary-up)
  ()
  (:documentation
   "This is the class of free-unification problems."))


(defmethod unicp~theo-init ((theory th+free) general-up equations
			    identified discriminated constants const-restr)
  (declare (edited  "17-NOV-1995")
	   (authors Felix)
	   (input   "free theory"
		    "general-up"
		    "a pair of termlists with same length"
		    "%"
		    "%"
		    "%")
	   (effect  "")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "  solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "  of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "  unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "  $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "  unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "  solution is lost if $y$ is in the constant"
		    "  restrictions of $x$)."))
  (let*
      ((up (make-instance 'cpfree+up
			  :general-up general-up
			  :equations equations
			  :theory theory))
       (cpfree*theory theory)
       (variables (cpfree=variables up))
       (tl1 (unicp~eq-lhss (unicp~equations up)))
       (tl2 (unicp~eq-rhss (unicp~equations up))))
    (declare (special cpfree*theory))
    #+weg(format t "   ~{~5A~}~%   ~{~5A~} Labels" variables
		 (mapcar #'(lambda (var) (unicp~var-label var theory)) variables))
    (if (and
	 (every #'
	  (lambda (t1 t2)
	    (cpfree=dag-create!
	     (cpfree=collapsing-find! t1)
	     (cpfree=collapsing-find! t2)))
	  tl1 tl2)
	 (every #'cpfree=cycle-test-a variables))
	(if identified
	    (unicp~new-constr
	     up identified discriminated constants const-restr)
	    (let*
		((identification (cpfree=identification variables nil))
		 (no-constants (cpfree=no-constants variables))
		 (subterm (cpfree=subterm variables))
		 (not-unified (cpfree=not-unified subterm variables)))
	      #+weg(format t "   ~{~5A~}~%   ~{~5A~} Labels" variables
			   (mapcar #'(lambda (var) (unicp~var-label var theory)) variables))
	      (values
	       up
	       identification
	       not-unified 
	       no-constants
	       nil
	       subterm
	       nil)))
	nil)))


(defmethod unicp~new-constr
    ((up cpfree+up) identified discriminated constants const-restr)
  (declare (edited  "18-JUL-1995")
	   (authors Felix)
	   (input   "unification-problem [cpfree+up]"
		    ""
		    ""
		    "const-restr: list of constant-restrictions")
	   (effect  "")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "  solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "  of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "  unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "  $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "  unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "  solution is lost if $y$ is in the constant"
		    "  restrictions of $x$).")
	   (ignore discriminated constants const-restr))
  (let
      ((cpfree*theory (unicp~theory up))
       (variables (cpfree=variables up)))   ;
    (declare (special cpfree*theory))
    (mapc #'cpfree=label-reset! variables)             ;
    (if (every #'
	 (lambda (pair)
	   (cpfree=dag-create!
	    (cpfree=collapsing-find! (unicp~id-left pair))
	    (cpfree=collapsing-find! (unicp~id-right pair))))
	 identified)
	(if identified
	    (cpfree=values up variables)
	    (values up nil nil nil nil nil))
	nil)))

(defmethod unicp~solvable-p ((up cpfree+up))
  t)

(defun cpfree=values (up variables)
  (declare (special cpfree*theory))
  (if (every #'cpfree=cycle-test-a variables)
      (let* ((identification (cpfree=identification variables nil))
	     (no-constants (cpfree=no-constants variables))
	     (subterm (cpfree=subterm variables))
	     (not-unified (cpfree=not-unified subterm variables)))
	(values up
		identification
		not-unified 
		no-constants
		nil
		subterm
		nil))
      nil))

;+----------------------------------------------------------------------------+
;+ identification                                                             +
;+----------------------------------------------------------------------------+

(defun cpfree=identification (variables buffer)
  ;;
  ;; value: the shortest set of new identifications 
  ;;        =  a list of all pairs (x y) with find(x) = find(y) except
  ;;           - (y x) if (x y) is in the list
  ;;           - (x x)
  ;;           - (x z) if (x y) and (y z) is in the list
  ;;           - (x y) if (x y) is already identified
  ;;
  (declare (special cpfree*theory))
  (if (null variables)
      buffer
    (let ((new-id (cpfree=compare (car variables) (cdr variables))))
      (if new-id
	  (cpfree=identification (cdr variables) (cons new-id buffer))
	(cpfree=identification (cdr variables) buffer)))))

(defun cpfree=compare (var varlist)
  ;;
  ;; value: if existing a pair (var y) with:
  ;;                     y is the first occurence in varlist with
  ;;                     - find(var) = find(y)
  ;;                     - var and y are not identified already
  ;;        else nil
  ;;
  (declare (special cpfree*theory))
  (if (null varlist)
      nil
    (if (and
	 (cpfree=ident-equal var (car varlist))
	 (not (unicp~vars-identified-p var (car varlist))))
	(unicp~id-create var (car varlist))
      (cpfree=compare var (cdr varlist)))))

(defun cpfree=ident-equal (t1 t2)
  (let
      ((next1 (cpfree=collapsing-find! t1))
       (next2 (cpfree=collapsing-find! t2)))
    (or ;(eq next1 next2)
	(if (and
	     (appl~p next1)
	     (appl~p next2)
	     (term~equal (cpfree=function next1) (cpfree=function next2)))
	    (every #'
	     (lambda (t1 t2)  (cpfree=ident-equal t1 t2))
	     (fo~arguments next1) (fo~arguments next2))
	    (term~equal next1 next2)))))
	
;+----------------------------------------------------------------------------+
;+ not-unified                                                                +
;+----------------------------------------------------------------------------+

(defun cpfree=not-unified (subterm variables)
  ;;
  ;; value: a list of all pairs (x < y), which are given by the new
  ;;        constant-restrictions (subterm)
  ;;
  (append
   (mapcan #'
    (lambda (const-restr)
      (let ((x (first const-restr)))
	(mapcan #'
	 (lambda (y) (list (unicp~id-create x y))) (rest const-restr))))
    subterm)
   (mapcon #'
    (lambda (var-tail)
      (mapcan #'
       (lambda (var)
	 (let
	     ((find1 (cpfree=find (first var-tail)))
	      (find2 (cpfree=find var)))
	   (when
	       (and
		(not (sym~variable-p find1))
		(not (sym~variable-p find2))
		(not (term~equal
		      (term~top find1) (term~top find2)))) 
	   (list (unicp~id-create (first var-tail) var)))))
       (rest var-tail)))
    variables)))
   

;+----------------------------------------------------------------------------+
;+ no-constants                                                               +
;+----------------------------------------------------------------------------+

(defun cpfree=no-constants (variables)
  ;;
  ;; value: a list of variables var with
  ;;        - ind(var) not yet free
  ;;        - bound to a constant or application   
  ;;
  (declare (special cpfree*theory))
  (mapcan #'
   (lambda (var)
     (and
      (cpfree=label var)
      (not (cpfree=variable-p (cpfree=find var)))
      (not (eq (unicp~var-label-theory var) cpfree*theory))
      (list var)))
   variables))

;+----------------------------------------------------------------------------+
;+ subterm                                                                    +
;+----------------------------------------------------------------------------+

(defun cpfree=subterm (variables)
  ;;
  ;; value: a list of new restrictions (x y1 y2 ... yn) with
  ;;        sigma(y_i) is a subterm of sigma(x) for a
  ;;        unifier sigma 
  ;;
  (declare (special cpfree*theory))
  (if (not cpfree*rational)
      (mapcan #'
       (lambda (var)
	 (when (and
		(not (null (cpfree=label var)))
		(listp (cpfree=label var)))
	   (let* ((find-var (cpfree=find var))
		  (var-list (remove-duplicates (set-difference
						(rest (cpfree=label var))
						(unicp~var-label-less var))))
		  (var-list (mapcan #'
			     (lambda (y) (if (eq find-var (cpfree=find y))
					     nil
					   (list y)))
			     var-list)))
	     (if var-list
		 (list (unicp~cr-create var var-list))
	       nil))))
       variables)))

;+----------------------------------------------------------------------------+
;+ id-variables, label-reset!                                                 +
;+----------------------------------------------------------------------------+
  
(defun cpfree=id-variables (identified)
  ;;
  ;; value: a list of all variables occuring in identified
  ;;
  (remove-duplicates
   (mapcan #'(lambda (pair) (list (unicp~id-left pair) (unicp~id-right pair))) identified)))

(defun cpfree=label-reset! (term)
  (declare (special cpfree*theory))
  (let
      ((next (cpfree=binding term)))
    (when (not (null (cpfree=label term)))
	  (cpfree=set-label! term nil)
	  (cond
	   ((cpfree=variable-p term) (cpfree=label-reset! next))
	   ((appl~p term)
	    (mapc #'cpfree=label-reset! (appl~arguments term)))))))
     
	      

#|
(defmethod unicp~var-label-less ((sym sym+symbol))
  (unicp~var-label-less (sym~label sym)))
 value: list von variablen, die <,>

(defmethod unicp~var-label-greater ((sym sym+symbol))
  (unicp~var-label-greater (sym~label sym)))
    (setf (unicp~var-label-less (first variables)) constants)
    (push new-var (unicp~var-label-less (first variables)))

(variables (unicp~variables up))
(theory (unicp~theory up))


(defmethod unicp~var-label-theory ((sym sym+sym))
  (unicp~var-label-theory (sym~label sym)))

(defmethod (setf unicp~var-label-theory) (value (sym sym+sym))
  (setf (unicp~var-label-theory (sym~label sym)) value))

|#

;+----------------------------------------------------------------------------+
;+ save/restore-choice-point                                                  +
;+----------------------------------------------------------------------------+


(defmethod unicp~up-save-choice-point ((up cpfree+up))
  ;;
  ;; up will be restored by computing a new one completely, so nothing has to
  ;; be saved
  ;;
  nil)

(defmethod unicp~up-restore-choice-point ((up cpfree+up))
  ;;
  ;; computes by reading  the actual identifications from the restored 
  ;; variable-slots a complete new up which is equal to the old one
  ;;
  (let*
      ((variables (unicp~variables up))
       (general-up (unicp~general-up up))
       (equations (unicp~equations up))
       (cpfree*theory (unicp~theory up))
       (actual-identification
	(mapcon #'
	 (lambda (var-tail)
	   (mapcan #'
	    (lambda (var)
	      (when (unicp~vars-identified-p (first var-tail) var)
		(list (unicp~id-create (first var-tail) var))))
	    (rest var-tail)))
	 variables)))
    (declare (special cpfree*theory))
    (cpfree=slot-reset! variables)
    (unicp~theo-init
     cpfree*theory general-up equations actual-identification nil nil nil)))

(defun cpfree=slot-reset! (termlist)
  ;;
  ;; resets recursively all slots of the terms in termlist to nil 
  ;;
  (declare (special cpfree*theory))
  (mapc #'
   (lambda (term)
     (let
	 ((next (cpfree=binding term)))
       (cpfree=set-label! term nil)
       (when next
	 (cpfree=set-binding! term nil)
	 (cpfree=slot-reset! (list next)))
       (when (appl~p term) (cpfree=slot-reset! (appl~arguments term)))))
   termlist))
	 
	     
  

;+--------------------------------------------------------------------------
;+ incremental version
;+--------------------------------------------------------------------------

(defmethod inc~new-equations ((up cpfree+up) equations
                                 identified discriminated constants const-restr)
  (declare (edited  "1998-09-17")
	   (authors Kepser)
	   (input   "An elementary unification problem, a list of equations,"
		    "and some restrictions that are new for this problem:"
		    "a list of pairs of variables which have to be identified,"
		    "a list of pairs of variables which have to be discriminated,"
		    "a list of variables that should be treated as constants,"
		    "a list of constant restrictions (i.e. pairs of a constant"
		    "and a set of variables.")
	   (effect  "The ELEMENTARY-UP is updated with the new restrictions.")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP belongs"
		    "to is called.  It returns NIL if the unification problem is"
		    "not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains some new"
		    "constraints wich must be met if the unification problem should be solvable:"
		    "1. True, indicating that the unification problem is still solvable."
		    "2. A list of new identification of variables (i.e. sets of variables)."
		    "3. A list of pairs of variables which must not be unified."
		    "4. A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6. A list of pairs $(x . y)$ of variables where $\sigma(x)$"
		    "   is a subterm of $\sigma(y)$ for a unifier $\sigma$."
		    "7. A list of pairs $(x . y)$ of variables where no solution"
		    "   is lost if $y$ is in the constant restrictions of $x$.")
	   (ignore discriminated constants const-restr))
  (let
      ((cpfree*theory (unicp~theory up))
       (variables (cpfree=variables up)))
    (declare (special cpfree*theory))
    (mapc #'cpfree=label-reset! variables)
    (if (and (every #'(lambda (pair)
			(cpfree=dag-create!
			 (cpfree=collapsing-find! (unicp~id-left pair))
			 (cpfree=collapsing-find! (unicp~id-right pair))))
		    identified)
	     (every #'(lambda (term1 term2)
			(cpfree=dag-create!
			 (cpfree=collapsing-find! term1)
			 (cpfree=collapsing-find! term2)))
		    (unicp~eq-lhss equations)
		    (unicp~eq-rhss equations)))
	(cpfree=values up variables)
	nil)))


