;;;                    +--------------------------------+
;;;                    |********************************|
;;;                    |*                              *|
;;;                    |*      Free - Unification      *|
;;;                    |*      Decision - Function     *|
;;;                    |*        by Felix Wolf         *|
;;;                    |*                              *|
;;;                    |********************************|
;;;                    +--------------------------------+
;;;
;;;
;;; Main-functions: (free~decide tl1 tl2  &optional reslist var_as_const)
;;;                 (free~unif tl1 tl2  &optional reslist var_as_const)
;;;                 (uni~unify-list tl1 tl2)
;;;
;;; tl1, tl2 :       two lists of terms which have to be unified
;;;                     (t1...tn) (s1...sn)
;;; reslist :        list of constant-restrictions
;;;                     ((c1 v11 v12 ..) (c2 v21 v22 ..) ..) 
;;; var_as_const :   list of variables which have to be considered as constants
;;;                  (v1 v2 ..)
;;; value:           ~decide: T        if unifier exists
;;;                           nil      else
;;;                  ~unif    unifier  if unifier exists
;;;                           nil      else
;;;                 
;;;                  uni~unify-list  (see below)    
;;;
;;;
;;; For questions please contact felix@cantor.informatik.rwth-aachen.de

(in-package :keim)

;;;+--------------------------------------------------------------------------+
;;;| main-functions                                                           |
;;;+--------------------------------------------------------------------------+

(defun uni~unify-list (termlist1 termlist2)
  (declare (edited  "29-JAN-1996")
	   (authors Felix)
	   (input   "Two lists of terms which have to be unified"
		    "(t1...tn) (s1...sn).")
	   (effect  "If uni*permanent-bindings is TRUE the variable bindings"
		    "established during the algorithm are kept. The bindings"
		    "between applications are deleted.")
	   (value   "If uni*permanent-bindings is TRUE the unification problem"
		    "{t1=s1, ..., tn=sn} is only decided and TRUE or nil"
		    "is returned. Else it is decided and a most general"
		    "unifier or nil is returned."))
  (if uni*permanent-bindings
      (free=decide-permanent-bindings termlist1 termlist2)
      (free~unif termlist1 termlist2)))

(defun free=decide-permanent-bindings (tl1 tl2)
  (unwind-protect                     ; restore label-slots
       
       (progn
	 (let
	    ((variables (free=variables (append tl1 tl2))))
	  (if (every #'
	       (lambda (t1 t2)
		 (free=dag-create!
		  (free=collapsing-find! t1)
		  (free=collapsing-find! t2)))
	       tl1 tl2)
	      (free=subst-create nil variables nil)
	      nil)))
    
    (progn
      (mapc #'(lambda (term) (free=label-reset! term)) (append tl1 tl2))
      (mapc #'(lambda (term) (free=term-binding-reset! term)) (append tl1 tl2))
      )))


(defmethod th~decide-unify ((th th+free) tl1 tl2 restr const-vars)
  (free~decide tl1 tl2 restr const-vars))


(defun free~decide (tl1 tl2 &optional reslist var_as_const)
  (declare (edited  "28-JUN-1995")
	   (authors Felix)
	   (input   "see above")
	   (effect  "None.")
	   (value   "Has the problem {t1=s1,...,tn=sn} an unifier?"))
  (free=main tl1 tl2 reslist var_as_const nil))
  
(defmethod th~unify ((th th+free) tl1 tl2 restr const-vars)
  (let ((unifier (free~unif tl1 tl2 restr const-vars)))
    (if unifier
	(list unifier)
      nil)))

(defun free~unif (tl1 tl2 &optional reslist var_as_const)
  (declare (edited  "28-JUN-1995")
	   (authors Felix)
	   (input   "see above")
	   (effect  "None.")
	   (value   "if existing the unifier of the problem {t1=s1,...,tn=sn}"
		    "else nil"))
  (free=main tl1 tl2 reslist var_as_const T))

;;;+--------------------------------------------------------------------------+
;;;| dag-create, subst-create, cycle-test                                     |
;;;+--------------------------------------------------------------------------+

(defun free=main (tl1 tl2 reslist var_as_const unif-flag)
  (unwind-protect                     ; restore slots of terms and symbols

      (progn
	(free=mark-var_as_const! var_as_const)
	(let
	    ((variables (free=variables (append tl1 tl2))))
	  (if (every #'
	       (lambda (t1 t2)
		 (free=dag-create!
		  (free=collapsing-find! t1)
		  (free=collapsing-find! t2)))
	       tl1 tl2)
	      (free=subst-create reslist variables unif-flag)
	    nil)))
    
    (progn
      (mapc #'(lambda (var) (keim~remprop var :const)) var_as_const)
      (mapc #'(lambda (term) (free=dag-reset! term)) (append tl1 tl2))
      (mapc #'(lambda (res) (keim~remprop (car res) :free=res)) reslist))))


(defun free=dag-create! (k1 k2)
  (declare (edited  "12-JUL-1995")
	   (authors Felix)
	   (input   "k1, k2: two terms (or nodes)")
	   (effect  "slots will be modified")
	   (value   "DAG if there is no clash-failure, else nil"))

  ; see also unify3-algorithm from Unification Theory by
  ; Franz Baader and Joerg H. Siekmann
  
  (let
      ((temp nil))
    (if (eq k1 k2)
	T
      (progn
	(when (free=variable-p k2)
	      (progn
		(setf temp k1)    ; if there is a var-node, k1 is the var-node
		(setf k1 k2)
		(setf k2 temp)))
	(cond ((free=variable-p k1)  
	       (if (free=variable-p k2)
		   (if (not (term~equal k1 k2))
		       (free=union_with_weight k1 k2))
		 (free=union k1 k2)) 
	       T)
	      ((not (term~equal (free=function k1) (free=function k2)))
	       nil)
	      ((sym~p k1)   ; two equal symbols
	                    ; (constants or constant variables)
	       T)           ; should better not be unioned.
	      (T
	       (free=union k1 k2)
	       (every #'
		(lambda (k1 k2)
		  (free=dag-create!
		   (free=collapsing-find! k1)
		   (free=collapsing-find! k2)))
		(fo~arguments k1) (fo~arguments k2))
	       ))))))


(defun free=subst-create (reslist variables unif-flag)
  (declare (edited  "12-JUL-1995")
	   (authors Felix)
	   (input   "variables: list of variables of the problem, see above")
	   (value   "if unif-flag if DAG is acyclic unifier"
		    "             else nil"
		    "else if DAG is acyclic T"
		    "     else nil"))
  (let
      ((subst (subst~create nil nil)))
    (mapc #'
     (lambda (res) (keim~put (car res) :free=res (cdr res))) reslist)
    (if unif-flag
	(when
	 (every #'
	  (lambda (var)
	    (free=cycle-test var unif-flag))
	  variables)
	 (progn
	   (mapc #'
	    (lambda (var)
	      (when (not (null (free=label var)))
		    (subst~insert-component! var (free=label var) subst)))
	    variables)
	   subst))
      (every #'
	  (lambda (var)
	    (free=cycle-test var unif-flag))
	  variables))))
     

(defun free=cycle-test (term unif-flag)
  (declare (edited  "12-JUL-1995")
	   (authors Felix)
	   (input   "term, node in a DAG"
		    "unif-flag [BOOLEAN] T   if unifier shall be computed"
		    "                    nil else")
	   (value   "unifier or T if from term no cycle"
		    "             is reachable in the DAG"
		    "nil          else"))
  (if unif-flag
      (free=cycle-test-unif term unif-flag)
    (free=cycle-test-decide term unif-flag)))

(defun free=cycle-test-unif (term unif-flag)
  (let
      ((next (free=binding term))
       (label (free=label term))
       (term-is-var (free=variable-p term)))
    
    (cond
     ((and (null next) term-is-var) term) ; term is var and leaf in the DAG
     
     ((eq label :free=false) nil)         ; there is a cycle in the DAG
                                          ; found out before
     
     ((eq label :free=true) term)         ; term is const and already
				 	  ; checked o.k.
     
     ((not (null label)) label)           ; term is already checked and
             				  ; labeled with the substitute 
     (T 
      (free=set-label! term :free=false)  ; else a cycle is assumed
	  
      (cond
	   
       ;; if term is var and there is no cycle reachable from the
       ;; succsesor of term, then term is labeled with its substitute
       ;; as o.k., else nil is returned, and term is left as
       ;; :free=false
       
       (term-is-var                       
	(let ((result (free=cycle-test-unif next unif-flag)))
	  (when result                    
		(free=set-label! term result))))
       
       ;; if term is application and there is no cycle reachable from
       ;; the arguments of the application, then the term will be
       ;; labeled with its substitute as o.k., else nil is returned
       ;; and term is left as :free=false
       
       ((appl~p term)                     
	(let
	    ((result_list
	      (free=cycle-test_list (appl~arguments term) nil unif-flag)))
	  (if (null result_list)          
	      nil
	    (free=set-label!              
	     term                         
	     (appl~create (appl~function term) result_list)))))
       
       ;; if term is const and there is no cycle from the variables of 
       ;; the const-restriction to term, then term is labeled with
       ;; :free=true as o.k., else nil is returned and term is left as
       ;; :free=false
       
       (T                                 
	(if (every #'
	     (lambda (resvar)             
	       (free=cycle-test-unif resvar unif-flag))
	     (keim~get term :free=res))   
	    (progn                        
	      (free=set-label! term :free=true)
	      term)                       
	  nil)))))))                        
                                              
(defun free=cycle-test-decide (term unif-flag) 
  (let
      ((next (free=binding term))
       (label (free=label term))
       (term-is-var (free=variable-p term)))

    (cond
     ((eq label :free=true) T)             ; term is already labeled as o.k.
     ((and (null next) term-is-var) T)     ; term is leaf in the DAG
     ((eq label :free=false) nil)          ; there is a cycle in the DAG
                                           ; found out before

     (T
	(free=set-label! term :free=false) ; else a cycle is assumed
	(cond

       ;; if term is var and there is no cycle reachable from the
       ;; succsesor of term, then term is labeled with :free=true
       ;; as o.k., else nil is returned, and term is left as
       ;; :free=false

	 ((free=variable-p term)
	  (when
	   (free=cycle-test-decide next unif-flag)
	   (free=set-label! term :free=true)))
	 
       ;; if term is application and there is no cycle reachable from
       ;; the arguments of the application, then the term will be
       ;; labeled with :free=true as o.k., else nil is returned
       ;; and term is left as :free=false

	 ((appl~p term)
	  (let
	      ((result_list
		(free=cycle-test_list (appl~arguments term) nil  unif-flag)))
	    (if (null result_list)
		nil
	      (free=set-label! term :free=true))))
	        
       ;; if term is const and there is no cycle from the variables of 
       ;; the const-restriction to term, then term is labeled with
       ;; :free=true as o.k., else nil is returned and term is left as
       ;; :free=false
       
	 (T
	  (when
	   (every #'
	    (lambda (resvar) (free=cycle-test-decide resvar unif-flag))
	    (keim~get term :free=res))
	   (progn
	     (free=set-label! term :free=true)
	     T))))))))
  
(defun free=cycle-test_list (termlist buffer unif-flag)
  (if (null termlist)
      (reverse buffer)
    (let
	((result (if unif-flag
		     (free=cycle-test-unif (car termlist) unif-flag)
		   (free=cycle-test-decide (car termlist) unif-flag))))
      (if (null result)
	  nil
	(free=cycle-test_list
	 (cdr termlist) (cons result buffer) unif-flag)))))
		   
	
;;;+--------------------------------------------------------------------------+
;;;| union-with-weight, union, collapsing-find, find                          |
;;;+--------------------------------------------------------------------------+


;;; For a term u without pointer is the weight of u the number of 
;;; variables v with (free=collapsing-find! v) = u. 


(defun free=inc-weight (term amount) 
  (if (zerop (keim~get term :weight 0))
      (keim~put term :weight 1)
    (keim~put term :weight (+ (keim~get term :weight) amount))))

(defun free=union_with_weight (term1 term2)
  (if
       ; sets pointer from the term with less weight to the pointer with
       ; greater weight
      
      (> (keim~get term1 :weight 0) (keim~get term2 :weight 0))
      (progn
	(free=set-binding! term1 term2)
	(free=inc-weight term2 (1+ (keim~get term1 :weight 0))))

    (progn
      (free=set-binding! term2 term1)
      (free=inc-weight term1 (1+ (keim~get term2 :weight 0))))))

(defun free=union (term1 term2)     ; sets pointer from term1 to term2
  (free=set-binding! term1 term2))

(defun free=collapsing-find! (term)

  ;; Is (free=collapsing-find u) = v then the pointers of all terms occuring
  ;; during the run of free=collapsing-find! will be set direct on v.

  (let
      ((binding (free=binding term)))
    (if (null binding)
	term
      (free=set-binding! term (free=collapsing-find! binding)))))
      

(defun free=mark-var_as_const! (var_as_const) 
    (mapc #'(lambda (var) (keim~put var :const T)) var_as_const))

(defun free=variable-p (term)
  (and
   (sym~variable-p term)
   (not (keim~get term :const nil))))

(defun free=variables (termlist) ; variables of the problem 'free=variable-p'
  (mapcan #'
   (lambda (var)
     (if (free=variable-p var)
	 (list var)
       nil))
   (top~free-variables termlist)))

(defun free=function (k) ; function of an application
  (if (appl~p k)
      (appl~function k)
    k))

;;;+--------------------------------------------------------------------------+
;;;| label, binding                                                           |
;;;+--------------------------------------------------------------------------+

(defgeneric free=label (term))

(defmethod free=label ((sym sym+sym))
  (sym~label sym))

(defmethod free=label ((term term+term))
  (term~label term))


(defgeneric free=set-label! (term label))

(defmethod free=set-label! ((sym sym+sym) label)
  (sym~set-label! sym label))

(defmethod free=set-label! ((term term+term) label)
  (term~set-label! term label))


(defgeneric free=binding (term))

(defmethod free=binding ((sym sym+sym))
  (sym~symbol-binding sym))

(defmethod free=binding ((term term+term))
  (term~binding term))


(defgeneric free=set-binding! (term label))

(defmethod free=set-binding! ((sym sym+sym) label)
  (sym~set-symbol-binding! sym label))

(defmethod free=set-binding! ((term term+term) label)
  (term~set-binding! term label))

;;;+--------------------------------------------------------------------------+
;;;| reset of slots                                                           |
;;;+--------------------------------------------------------------------------+

(defun free=dag-reset! (term)
  (free=set-binding! term nil)
  (free=set-label! term nil)
  (keim~remprop term :weight)
  (when (appl~p term)
	(mapc #'free=dag-reset! (appl~arguments term))))
    
(defun free=label-reset! (term)
  (free=set-label! term nil)
  (when (appl~p term)
	(mapc #'free=label-reset! (appl~arguments term))))

(defun free=term-binding-reset! (term)
  (when (appl~p term)
    (free=set-binding! term nil)
    (mapc #'free=term-binding-reset! (appl~arguments term))))





