;;; -*- Package: KEIM; Syntax: Common-lisp; Mode: LISP -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;

(in-package  "KEIM")

(mod~defmod type :uses (env keim mod post )
	    :documentation "Datastructures and basic functionality of simple types."
	    :exports (
		      type+type
		      type+complex
		      type+primitive
		      type+constant
		      type+variable
		      type~constant-create
		      type~variable-create
		      type~apply
		      type~abstract
		      type~p
		      type~primitive-p
		      type~constant-p
		      type~variable-p
		      type~complex-p
		      type~ground-p
		      type~n-range
		      type~n-domain
		      type~c-range
		      type~c-domain
		      type~i
		      type~o
		      type~i-p
		      type~o-p
		      type~predicate-p
		      type~function-create
		      type~predicate-create
		      type+substitution
		      type~substitution-create
		      type~subst-domain
		      type~subst-codomain
		      type~subst-get-component
		      type~subst-empty-p
		      type~subst-apply
		      type~unify
		      type~read-new-type-variables
		      type~read-new-type-variable
		      type~enter-type-variable
		      type~read-new-type-constants
		      type~read-new-type-constant
		      type~enter-type-constant
		      type~read-existing-type
		      type~env-lookup
		      )
	    )

#{
\section{Types}\label{mod:type}

Types are a well-known mechanism for syntactically ensuring the
consistency of higher-order logic. \keim\ types are inductively built
up from type variables and type constants by making function types: if
$\beta$, $\alpha^1$, \ldots, $\alpha^n$ are types then there is a
function type, written as ($\alpha^1, \ldots,\alpha^n\rightarrow
\beta$), such that $\beta$ is the type of the range and the cartesian
product of the types $\alpha^1$, \ldots, $\alpha^n$ is the domain. 

In the presence of higher-order types functions do not have a fixed
arity: The function $+$ can be viewed to take 2 arguments and yield
an in dividual, the sum. Therefore it is of the type 
$(\iota,\iota)\to\iota$. On the other hand it can be viewed as a function taking
one argument, e.g. 1 and yielding the 
unary function {\tt inc} which increases each number by 1. Generally we see that the types 
$(\alpha^1,\ldots,\alpha^n\rightarrow ((\beta^1,\ldots,\beta^n\rightarrow \gamma)))$ and   
$(\alpha^1,\ldots,\alpha^n,\beta^1,\ldots,\beta^n\rightarrow \gamma)$ have to be equal. There are two normal forms
for types: 
\begin{itemize}
\item {\em n-normal form}, where the number of occurrences of the function constructor $\rightarrow$ is
minimal (in this case the range of the type is primitive and the number of arguments of each
function is maximal), and 
\item {\em c-normal form} where occurrences of $\rightarrow$ is unary (in this case the number
occurrances of $\rightarrow$ is maximal). This form of types corresponds to the {\em Curried
form of functions}
\end{itemize}
Corresponding to these normal forms many {\bf TYPE} functions have {\bf N} and {\bf C} forms.

 This file contains the primitive types i and o to represent higher order types
 but everything can be a primitive type (for example terms to represent sorts of functions).
 A new primitive type must be a subclass of type+primitive or understand all the methods defined for this class.

 A type is in n-normal-form if the range of all subtypes is primitive (t0 = i or o)
 and in curry-normal-form if the domain of all complex subtypes contains only one element (n = 1).
#}
;; 
;;
;;  Implementation: The types are represented in n-normal-form. If the range of the c-normal-form must be constructed
;;  it is saved into a slot.




(eval-when (load compile eval) 
  (defclass type+fo-type (type+type)
      ()
    (:documentation "The class of all first order types in KEIM.")))

(eval-when (load compile eval) 
  (defclass type+symbol ()
      ((arity :initarg :arity :reader type~arity))
    (:documentation "This is an abstract mixin class; it contains an arity slot for symbols.")))


(eval-when (load compile eval) 
  (defclass type+individual (type+fo-type)
      ()
    (:documentation "The class of all types which range over the individuals.")))

(eval-when (load compile eval) 
  (defclass type+truth-value (type+fo-type)
      ()
    (:documentation "The class of all types which range over the truth values.")))


(eval-when (load compile eval) 
  (defclass type+function (type+symbol type+individual)
      ()
    (:documentation "The class of all types for functions, constants, and variables.")))

(eval-when (load compile eval) 
  (defclass type+predicate (type+symbol type+truth-value)
      ()
    (:documentation "The class of all types for predicates.")))

(eval-when (load compile eval) 
  (defclass type+junctor (type+symbol type+truth-value)
      ()
    (:documentation "The class of all types for junctors.")))


(defvar type*individual (make-instance 'type+individual))

(defvar type*truth-value (make-instance 'type+truth-value))

(defun type~individual ()
  type*individual)

(defun type~truth-value ()
  type*truth-value)


;;;
;;;;; Constructors


(defun type~apply (applied-type argument-types)
  (declare (edited  "08-AUG-1991 11:27")
	   (authors RICHTS)
	   (input   "A type (t1,...,tn) -> t0 and a list of types (s1 ... sm) or a type s1.")
	   (effect  "An error if not (m<=n and t1=s1,...,tm=sm).")
	   (value   "The type (tm+1,...,tn) -> t0  if m<=n and t1=s1,...,tm=sm.")
	   (example "(i,i) -> o   i --> (i) -> o)"))
  (cond ((not (= (type~arity applied-type) (length argument-types)))
	 (error "The type ~A with can not be applied to ~A arguments (~A)."
		applied-type (length argument-types) argument-types))
	((some #'(lambda (type) (< 0 (type~arity type))) argument-types)
	 (error "A type can only be applied to types with arity 0 (~A)." argument-types))
	((type~function-p applied-type)
	 (if (every #'(lambda (type) (type~individual-p type)) argument-types)
	     type*individual
	   (error "The function type ~A can not be applied to an argument which is not an individual type."
		  applied-type)))
	((type~predicate-p applied-type)
	 (if (every #'(lambda (type) (type~individual-p type)) argument-types)
	     type*truth-value
	   (error "The predicate type ~A can not be applied to an argument which is not an individual type."
		  applied-type)))
	((type~junctor-p applied-type)
	 (if (every #'(lambda (type) (type~truth-value-p type)) argument-types)
	     type*truth-value
	   (error "The junctor type ~A can not be applied to an argument which is not an truth value type."
		  applied-type)))
	(t
	 (error "The type ~A is not known to be applicable to any argument types."
		applied-type))))



;;;;; Predicates

(defmethod keim~equal ((type1 type+fo-type) (type2 type+fo-type))
  (declare (edited  "08-AUG-1991 13:08")
	   (authors RICHTS)
	   (input   "Two types.")
	   (effect  "None.")
	   (value   "True iff the two types are equal."))
  (and (eq (class-of type1) (class-of type2))
       (= (type~arity type1) (type~arity type2))))


(defun type~p (object)
  (declare (edited  "21-OCT-1991 11:09")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True iff OBJECT is a type."))
  (typep object 'type+fo-type))

(defun type~individual-p (object)
  (typep object 'type+individual))

(defun type~truth-value-p (object)
  (typep object 'type+truth-value))

(defun type~function-p (object)
  (typep object 'type+function))

(defun type~predicate-p (object)
  (typep object 'type+predicate))

(defun type~junctor-p (object)
  (typep object 'type+junctor))

(defun type~symbol-p (object)
  (typep object 'type+symbol))


;;;
;;;;; Selectors

(defgeneric type~arity (type)
  (declare (edited  "08-AUG-1991 13:00")
	   (authors RICHTS)
	   (input   "A type.")
	   (effect  "None.")
	   (value   "The arity of the type.")
	   (example "For type+symbols a method is generated by the defclass; for all other Types 0"
		    "is returned."))
  (:method ((type type+fo-type))
     0))


; Printing

(defmethod print-object ((type type+fo-type) stream)
  (when (< 0 (type~arity type))
	(format stream "~A(~A) -> " (if (type~junctor-p type) "o" "i")
		(type~arity type)))
  (format stream (if (type~individual-p type) "i" "o")))



(defun type~function-create (arity)
  (declare (edited  "08-AUG-1991 11:13")
	   (authors RICHTS)
	   (input   "A number n.")
	   (effect  "None.")
	   (value   "The structure representing the type (i1,...,in) -> i.")
	   (example "3 --> (i,i,i) -> i"))
  (make-instance 'type+function :arity arity))

(defun type~predicate-create (arity)
  (declare (edited  "08-AUG-1991 11:13")
	   (authors RICHTS)
	   (input   "A number n.")
	   (effect  "None.")
	   (value   "The structure representing the type (i1,...,in) -> o.")
	   (example "3 --> (i,i,i) -> o"))
  (make-instance 'type+predicate :arity arity))

(defun type~junctor-create (arity)
  (declare (edited  "08-AUG-1991 11:13")
	   (authors RICHTS)
	   (input   "A number n.")
	   (effect  "None.")
	   (value   "The structure representing the type (i1,...,in) -> o.")
	   (example "3 --> (i,i,i) -> o"))
  (make-instance 'type+junctor :arity arity))




    

(defmethod type~ground-p ((type type+type))
 t)
