;;; -*- Package: KEIM; Syntax: Common-lisp; Mode: LISP -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;

(in-package  "KEIM")

(mod~defmod type :uses (env keim mod post )
	    :documentation "Datastructures and basic functionality of simple types."
	    :exports (
		      type+type
		      type+complex
		      type+primitive
		      type+constant
		      type+variable
		      type~constant-create
		      type~variable-create
		      type~apply
		      type~abstract
		      type~p
		      type~primitive-p
		      type~constant-p
		      type~variable-p
		      type~complex-p
		      type~ground-p
		      type~n-range
		      type~n-domain
		      type~c-range
		      type~c-domain
		      type~i
		      type~o
		      type~i-p
		      type~o-p
		      type~predicate-p
		      type~function-create
		      type~predicate-create
		      type+substitution
		      type~substitution-create
		      type~subst-domain
		      type~subst-codomain
		      type~subst-get-component
		      type~subst-empty-p
		      type~subst-apply
		      type~unify
		      type~read-new-type-variables
		      type~read-new-type-variable
		      type~enter-type-variable
		      type~read-new-type-constants
		      type~read-new-type-constant
		      type~enter-type-constant
		      type~read-existing-type
		      type~env-lookup
		      )
	    )

#{
\section{Types}\label{mod:type}

Types are a well-known mechanism for syntactically ensuring the
consistency of higher-order logic. \keim\ types are inductively built
up from type variables and type constants by making function types: if
$\beta$, $\alpha^1$, \ldots, $\alpha^n$ are types then there is a
function type, written as ($\alpha^1, \ldots,\alpha^n\rightarrow
\beta$), such that $\beta$ is the type of the range and the cartesian
product of the types $\alpha^1$, \ldots, $\alpha^n$ is the domain. 

In the presence of higher-order types functions do not have a fixed
arity: The function $+$ can be viewed to take 2 arguments and yield
an in dividual, the sum. Therefore it is of the type 
$(\iota,\iota)\to\iota$. On the other hand it can be viewed as a function taking
one argument, e.g. 1 and yielding the 
unary function {\tt inc} which increases each number by 1. Generally we see that the types 
$(\alpha^1,\ldots,\alpha^n\rightarrow ((\beta^1,\ldots,\beta^n\rightarrow \gamma)))$ and   
$(\alpha^1,\ldots,\alpha^n,\beta^1,\ldots,\beta^n\rightarrow \gamma)$ have to be equal. There are two normal forms
for types: 
\begin{itemize}
\item {\em n-normal form}, where the number of occurrences of the function constructor $\rightarrow$ is
minimal (in this case the range of the type is primitive and the number of arguments of each
function is maximal), and 
\item {\em c-normal form} where occurrences of $\rightarrow$ is unary (in this case the number
occurrances of $\rightarrow$ is maximal). This form of types corresponds to the {\em Curried
form of functions}
\end{itemize}
Corresponding to these normal forms many {\bf TYPE} functions have {\bf N} and {\bf C} forms.

 This file contains the primitive types i and o to represent higher order types
 but everything can be a primitive type (for example terms to represent sorts of functions).
 A new primitive type must be a subclass of type+primitive or understand all the methods defined for this class.

 A type is in n-normal-form if the range of all subtypes is primitive (t0 = i or o)
 and in curry-normal-form if the domain of all complex subtypes contains only one element (n = 1).
#}
;; 
;;
;;  Implementation: The types are represented in n-normal-form. If the range of the c-normal-form must be constructed
;;  it is saved into a slot.




(defun type=create (n-range n-domain &key c-range)
  (declare (edited  "08-AUG-1991 11:08")
	   (authors RICHTS)
	   (input   "A type t0 and a list of types (t1 ... tn).")
	   (effect  "None.")
	   (value   "The structure representing the type (t1,...,tn) -> t0."))
  (error "Typen haben sich geaendert."))


;;;
;;;;; Constructors

(defun type~constant-create (name)
  (declare (edited  "27-Jan-1992 19:30")
	   (authors Kohlhase)
	   (input   "A print name for the type")
	   (effect  "None.")
	   (value   "A type constant with name NAME.")
	   (example "(type~constant-create 'a) --> A"))
  (error "Typen haben sich geaendert."))


(defun type~variable-create (name)
  (declare (edited  "27-Jan-1992 19:30")
	   (authors Kohlhase)
	   (input   "A print name for the type")
	   (effect  "None.")
	   (value   "A type variable with name NAME.")
	   (example "(type~variable-create 'x) --> X"))
  (error "Typen haben sich geaendert."))




(defun type~abstract (abstracted-type argument-types)
  (declare (edited  "08-AUG-1991 11:29")
	   (authors RICHTS)
	   (input   "A type (t1,...,tn) -> t0 and a list of types (s1 ... sm) or a type s1.")
	   (effect  "None.")
	   (value   "The type (s1,...,sm,t1,...,tn) -> t0.")
	   (example "(i,i) -> o   ((i) -> o i) --> ((i) -> o,i,i,i) -> o"))
  (error "Typen haben sich geaendert."))




;;;;; Predicates

(defmethod keim~equal ((type1 type+type) (type2 type+type))
  (error "Typen haben sich geaendert."))


(defmethod keim~equal ((type1 type+complex) (type2 type+complex))
  (error "Typen haben sich geaendert."))

(defun type~primitive-p (type)
  (error "Typen haben sich geaendert."))

(defun type~constant-p (type)
  (error "Typen haben sich geaendert."))

(defun type~variable-p (type)
  (error "Typen haben sich geaendert."))

(defun type~complex-p (type)
  (error "Typen haben sich geaendert."))

(defgeneric type~ground-p (type)
  (:method ((type type+complex))
  (error "Typen haben sich geaendert."))
  (:method ((type type+constant))
  (error "Typen haben sich geaendert."))
  (:method ((type type+variable))
  (error "Typen haben sich geaendert.")))


;;;
;;;;; Selectors



(defmethod keim~set-name! ((type type+primitive) name)
  (declare (ignore name))
  (error "It is not legal to alter the name of a type."))

(defgeneric type~n-range (type)
  (:method ((type type+primitive))
  (error "Typen haben sich geaendert."))
  (:method ((type type+complex))
  (error "Typen haben sich geaendert.")))
     
(defgeneric type~n-domain (type)
  (:method ((type type+primitive))
  (error "Typen haben sich geaendert."))
  (:method ((type type+complex))
  (error "Typen haben sich geaendert.")))

(defgeneric type~c-range (type)
  (:method ((type type+complex))
  (error "Typen haben sich geaendert."))
  (:method ((type type+primitive))
  (error "Typen haben sich geaendert.")))

(defgeneric type~c-domain (type)
  (:method ((type type+complex))
  (error "Typen haben sich geaendert."))
  (:method ((type type+primitive))
  (error "Typen haben sich geaendert.")))


; Printing

(defmethod print-object ((type type+primitive) stream)
  (let ((format-string 
	 ;; if type o or i, and *print-pretty*, then lowercase it.
	 ;; if type is a keyword, print the colon
	 ;; otherwise just print the symbol's name
	 (cond ((or (type~o-p type) (type~i-p type))
		(if *print-pretty* "~(~A~)" "~A"))
	       ((keywordp (keim~name type)) "~S")
	       (t "~A"))))
  (format stream format-string (keim~name type))))

(defmethod print-object ((type type+complex) stream)
  (declare (edited  "23-AUG-1991 19:22")
	   (authors RICHTS)
	   (input   )
	   (effect  )
	   (value   ))
  (declare (special *print-level* *print-length*))
  (cond ((or (null *print-length*) (< (length (type~n-domain type)) *print-length*))
	 (format stream "(~A~{,~A~}) -> ~A"
		 (car (type~n-domain type))
		 (cdr (type~n-domain type))
		 (type~n-range type)))
	(t (format stream "(~{~A,~}...) -> ~A"
		   (subseq (type~n-domain type) 0 (1- *print-length*))
		   (type~n-range type)))))


#{
\subsection{Primitive types}

The type constants I for individuals and O for truth values are predefined.
They should occur in every environment.
#}

;; Changing definitions to *not* use lower case I and O, which cause 
;; confusion when reading in types.  Instead use capital I and O, but
;; change print-object to print these guys as lower case by default. DAN

(defvar type*i (type~constant-create 'i))
(defvar type*o (type~constant-create 'o))

(defun type~i ()
  (declare (edited  "21-AUG-1991 19:30")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "None.")
	   (value   "A structure representing the type i."))
  (error "Typen haben sich geaendert."))

(defun type~o ()
  (declare (edited  "21-AUG-1991 19:30")
	   (authors RICHTS)
	   (input   "None.")
	   (effect  "None.")
	   (value   "A structure representing the type o."))
  (error "Typen haben sich geaendert."))


(defun type~i-p (type)
  (declare (edited  "23-AUG-1991 12:24")
	   (authors RICHTS)
	   (input   "A type.")
	   (effect  "None.")
	   (value   "True iff the type is the primitive type i."))
  (eq (type~i) type))

(defun type~o-p (type)
  (declare (edited  "27-JAN-1992 12:24")
	   (authors RICHTS)
	   (input   "A type.")
	   (effect  "None.")
	   (value   "True iff the type is the primitive type o."))
  (eq (type~o) type))


#+weg(defgeneric type~predicate-p (type)
  (declare (edited  "27-JAN-1992 12:24")
	   (authors RICHTS)
	   (input   "A type.")
	   (effect  "None.")
	   (value   "True iff the type is the form type (A) -> o")
	   (example "(i,i) -> o --> T"
		     "(i,i) -> i --> NIL"))
  (:method ((type type+primitive)) nil)
  (:method ((type type+complex)) (keim~equal (type~o) (type~n-range type))))

#{
\subsection{Constructors for first order function and predicate types}
#}



;;Bindings for Type Variables


(defun type=create-binding-substitution ()
  (declare (edited  "07-NOV-1991 12:01")
	   (authors KOHLHASE)
	   (input   "None.")
	   (effect  "None.")
	   (value   "A substitution that contains all variables in the actual binding-list as its domain"
		    "and the terms in the binding slots of these variables as codomain."
		    "The codomain-terms are built by inserting the type=bindings, i. e. the binding slots are regarded."))
  (error "Typen haben sich geaendert."))

(defgeneric type=insert-bindings (type)
  (declare (edited  "05-NOV-1991 10:55")
	   (authors KOHLHASE)
	   (input   "A type.")
	   (effect  "None.")
	   (value   "A new type where the bindings of all subterms of TYPE are inserted into the term recursively."))
  (:method ((var type+variable))
  (error "Typen haben sich geaendert."))
  (:method ((const type+constant))
  (error "Typen haben sich geaendert."))
  (:method ((type type+complex))
  (error "Typen haben sich geaendert.")))




(defmethod post~read-object (typevars (env env+environment) (indicator (eql :type-variables)))
  (dolist (typevar typevars nil)
    (type~enter-type-variable typevar env)))
	    
(defmethod post~read-object (typevar (env env+environment) 
			    (indicator (eql :type-variable)))
  (type~enter-type-variable typevar env))

(defun type~enter-type-variable (symbol env)
  (declare
   (authors nesmith)
   (input "a symbol representing a type variable and an environment")
   (effect "Reads the type variable declaration into the environment")
   (value "undefined")
   (example "(type~enter-type-variable 'type3 env)"))
  (let ((thing (env~lookup-object symbol env)))
    (cond ((and thing (not (type~variable-p thing)))
	   (error "Can't declare ~S as a type variable, because it already exists in the environment as a ~A. " symbol (class-name (class-of thing))))
	  (thing nil)
	  ((not (symbolp symbol))
	   (error "Can't declare ~S as a type variable, because it is not a symbol." symbol))
	  (t (env~enter symbol (type~variable-create symbol) env)
	     nil))))

(defun type~read-new-type-constants (typeconsts env)
  (declare
   (authors nesmith)
   (input "a POST representation of a list of type constants
and an environment")
   (effect "Reads the type constant declarations into the environment")
   (value "undefined"))
  (post~read-object typeconsts env :type-constants))

(defun type~read-new-type-constant (typeconst env)
  (declare
   (authors nesmith)
   (input "a POST representation of a type constant and an environment")
   (effect "Reads the type constant declaration into the environment")
   (value "undefined"))
  (post~read-object typeconst env :type-constant))


(defmethod post~read-object (typevars (env env+environment) 
			     (indicator (eql :type-constants)))
  (dolist (typevar typevars nil)
    (type~enter-type-constant typevar env)))
	    
(defmethod post~read-object (typevar (env env+environment) 
			    (indicator (eql :type-constant)))
  (type~enter-type-constant typevar env))


(defun type~enter-type-constant (symbol env)
  (declare
   (authors nesmith)
   (input "a Lisp symbol and an environment")
   (effect "Declares the symbol as a type constant in the environment")
   (value "undefined"))
  (let ((thing (env~lookup-object symbol env)))
    (cond ((and thing (not (type~constant-p thing)))
	   (error "Can't declare ~S as a type constant, because it already exists in the environment as a ~A. " symbol (class-name (class-of thing))))
	  (thing nil)
	  ((not (symbolp symbol))
	   (error "Can't declare ~S as a type constant, because it is not a symbol." symbol))
	  (t (env~enter symbol (type~constant-create symbol) env)
	     nil))))

; existing-types

(defun type~read-existing-type (type env)
  (declare
   (authors huang nesmith)
   (input "A POST representation of an existing TYPE and an environment ENV.")
   (effect "None")
   (value "The type is looked up in the environment and a type is returned.
If the type is not in the environment an error will be signaled.")
   (example "(type~read-existing-type '(i i) env) --> (i) -> i"))
  (type~env-lookup type env))

(defmethod post~read-object ((type symbol) (env env+environment) 
			    (indicator (eql :existing-type)))
  (type~env-lookup type env))

(defmethod post~read-object ((type cons) (env env+environment) 
			    (indicator (eql :existing-type)))
  (type~env-lookup type env))

(defgeneric type~env-lookup (type env)
  (declare 
   (authors NESMITH)
   (input   "A POST representation of an existing TYPE and an environment ENV.")
   (effect  "None.")
   (value   "The type is looked up in the environment and a type is returned.
If the type is not in the environment an error will be signaled.")
   (example "(type~read-existing-type '(i i) env) --> (i) -> i"))
  (:method ((type list) (env env+environment))
	   (if (= (length type) 1)
	       (type~env-lookup (car type) env)
	     (type=env-lookup-complex type env)))
  (:method ((type symbol) (env env+environment))
	   (type=env-lookup-symbol type env)))

(defun type=env-lookup-symbol (type env)
  (declare
   (authors nesmith)
   (input "an object and an environment")
   (effect "Looks up the object in environment, returns it if it is associated with a type, otherwise error")
   (value "undefined"))
  (let ((obj (env~lookup-object type env)))
    (cond ((not obj) 
	   (error "~A is not an existing type." type))
	  ((not (type~p obj))
	   (error "~A is declared in the environment as a ~A, not as a type."
		  type (class-of obj)))
	  (t obj))))


(defun type=env-lookup-complex (type env)
  (declare
   (authors nesmith)
   (input "an object and an environment")
   (effect "Looks up the object in environment, returns it if it is associated with a type, otherwise error")
   (value "undefined"))
  (if (consp type)
      (type~abstract (type~env-lookup (car type) env)
		     (mapcar #'(lambda (x) (type~env-lookup x env))
			     (reverse (cdr type))))
    (error "complex type  ~S should be a cons" type)))

    

