;;;****************************************************************************
;;;*                                                                          *
;;;*                                Smallkeim                                 *
;;;*                                                                          *
;;;*       A short version of keim for the combination of unification         *
;;;*                              algorithms                                  *
;;;*                                                                          *
;;;****************************************************************************

;;; The following functions and classes are exported and are able to simulate
;;; the keim-system for the combination algorithm
 

;;; appl~arguments             ;
;;; appl~create                ;
;;; appl~function              ;
;;; appl~p                     ;

;;; env~create                 ;

;;; fo~arguments               ;
;;; fo~variable-create         ;

;;; keim~equal                 ;
;;; keim~get                   ;
;;; keim+object                ;
;;; keim~plist                 ;
;;; keim~put                   ;
;;; keim~remprop               ;
;;; keim~set-plist!            ;

;;; post~read-object-list      ;

;;; shsym~label                ;
;;; shsym~set-label!           ;

;;; sym+const                  ;
;;; sym+sym                    ;
;;; sym+var                    ;
;;; sym~constant-p             ;
;;; sym~label                  ;
;;; sym~name                   ;
;;; sym~p                      ;
;;; sym~set-label!             ;
;;; sym~set-symbol-binding!    ;
;;; sym~shared-symbol          ; 
;;; sym~symbol-binding         ;
;;; sym~variable-p             ;

;;; term+term                  ; 
;;; term~binding               ;
;;; term~equal                 ;
;;; term~label                 ;
;;; term~read                  ;
;;; term~set-binding!          ;
;;; term~set-label!            ;
;;; term~top                   ;

;;; subst~apply                ;
;;; subst~create               ;
;;; subst~insert-component!    ;

;;; top~free-variables         ;

;;; uni+unification-problem    ;

(in-package "COMMON-LISP-USER")
(eval-when (load compile eval)
  (let
      ((package (find-package "COMMON-LISP-USER")))
    (rename-package package (package-name package)
		    (cons "KEIM" (package-nicknames package)))))
(in-package "KEIM")

;;; Simulation of a loaded KEIM-package 

;;;+--------------------------------------------------------------------------+
;;;| class-definitions:                                                       |
;;;+--------------------------------------------------------------------------+
;;;
;;;                        
;;;                             keim+object
;;;                             /      \  \
;;;                            /        \  subst+substitution
;;;                        term+term     \
;;;                        /       \      uni+unification-problem
;;;                       /         \
;;;               appl+appl          sym+sym
;;;                                  /     \
;;;                                 /       \
;;;                          sym+var         sym+const
;;;
;;;
;;;   keim+object:  superclass of all classes
;;;   term+term:    superclass of all concrete terms
;;;   appl+appl:    class of all composed terms, e.g. (f x a)
;;;   sym+sym:      superclass of variables, constants and symbols of functions
;;;   sym+var:      class of variables
;;;   sym+const:    class of constants and symbols of functions
;;;
;;;
;;;   subst+substitution:       class of substitutions
;;;   uni+unification-problem:  class of unification problems

(eval-when (load compile eval)

  (defclass keim+object ()
    ((plist :initform nil
	    :reader keim=plist
	    :writer keim=write-plist!))
    (:documentation "This is the superclass to all other classes."))
  
  (defclass term+term (keim+object)
    ((label :initform nil
	    :reader term=term-label
	    :writer term=write-term-label!)
     (binding :initform nil
	      :reader term=term-binding
	      :writer term=write-term-binding!))
    (:documentation "This is the superclass to all concrete terms"))
  
  (defclass appl+appl (term+term)
    ((function :initarg :function
	       :reader appl~function 
	       :writer appl=write-function!) 
     (arguments :initarg :arguments
		:reader appl~arguments 
		:writer appl=write-arguments!))
    (:documentation "This is the class of all composed terms, e.g. (f x a)"))

;;; Variables, constants and symbols of functions (constants with arity
;;; =|= 0) are shared objects and can occur in of more than one
;;; application.

  (defclass sym+sym (term+term)
    ((symbol :initarg :symbol
	     :reader sym=symbol
	     :writer sym=write-symbol!))
    (:documentation
     "The superclass of variables, constants and symbols of functions"))
  
  (defclass sym+var (sym+sym)
    ((unicp-label :initform nil
		  :reader skeim~unicp-label
		  :writer skeim~set-unicp-label!))
    (:documentation "The class of variables"))

  (defclass sym+const (sym+sym)
    ((arity :initarg :arity
	    :reader const=arity
	    :writer const=write-arity!))
    (:documentation "The class of constants and symbols of functions"))
  

  (defclass subst+substitution (keim+object)
    ((domain :initarg :domain
	     :reader  subst~domain
	     :writer  subst=write-domain!)
     (codomain :initarg :codomain
	       :reader subst~codomain
	       :writer subst=write-codomain!))
    (:documentation "The class of substitutions."))


  (defclass uni+unification-problem (keim+object)
    ((initial-problem :initarg :initial-problem
		      :reader uni=initial-problem)
     (previous-solutions :initarg :previous-solutions
			 :accessor uni=previous-solutions))
    (:documentation "The class of unification problems."))
  )

;;;+--------------------------------------------------------------------------+
;;;| appl                                                                     |
;;;+--------------------------------------------------------------------------+

(defun appl~create (function arguments)
  (make-instance 
   'appl+appl
   :function function
   :arguments arguments))


(defun appl~p (object)
  (declare (edited  "12-SEP-1991 14:25")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True iff OBJECT is an application."))
  (typep object 'appl+appl))

;;;+--------------------------------------------------------------------------+
;;;| env                                                                      |
;;;+--------------------------------------------------------------------------+

(defun env~create ()
  (list :environment))

(defun env=get (name env) 
  (let
      ((object (some
		#'(lambda (pair)
		    (if (eql name (car pair))
			(cdr pair)
			nil))
		(rest env))))
    (if object
	object
	(error "There is no object in environment with this name: ~A" name))))

;;;+--------------------------------------------------------------------------+
;;;| fo                                                                       |
;;;+--------------------------------------------------------------------------+

(defgeneric fo~arguments (term)
  (declare (edited  "02-AUG-1991 20:46")
	   (authors RICHTS)
	   (input   "A term.")
	   (effect  "None.")
	   (value   "Nil if term is a constant or variable,"
		    "a list containing the arguments of the application"
		    "otherwise.")
	   (example "Let c be an object constant, x an object variable"
		    "c --> nil"
		    "x --> nil"))
  (:method ((variable sym+var))
	   nil)
  (:method ((constant sym+const))
	   nil)
  (:method ((application appl+appl))
	   (appl~arguments application)))

		     
(defvar fo*variable-counter 0)

;;; A counter for uniquely numerating new variables. Increased by 1 each
;;; time, when fo~variable-create is called.

(defun fo~variable-create (&optional name &rest term)
  (declare (edited  "25-JUL-1991 13:03")
	   (authors huang RICHTS)
	   (input   "name: a symbol."
		    " A reference term")
	   (effect  "None.")
	   (value   "The new variable (if NAME is nil its name will"
		    "be \"x-<number>\").")
	   (example "(fo~variable-create) --> x-2"
		    "(fo~variable-create) --> x-3"
		    "(term~type (fo~variable-create)) --> i")
	   (ignore term))
  (make-instance
   'sym+var
   :symbol (if name
	       name
	       (format nil "x-~A" (incf fo*variable-counter)))))


;;;+--------------------------------------------------------------------------+
;;;| keim                                                                     |
;;;+--------------------------------------------------------------------------+


(defgeneric keim~equal (keim-object1 keim-object2)
  (declare (edited  "06-DEC-3707 14:65")
	   (authors RICHTS)
	   (input   "Two objects or object lists.")
	   (effect  "None.")
	   (value   "True, iff the two objects are equal."))
  (:method (object1 object2)
   (eq object1 object2))
  (:method ((keim-object1 keim+object) (keim-object2 keim+object))
   (eq keim-object1 keim-object2))
  (:method ((termlist1 list) (termlist2 list))
   (and (= (length termlist1) (length termlist2))
	(every #'keim~equal termlist1 termlist2))))

(defun keim~get (term indicator &optional default)
  (declare (edited  "13-MAR-1997 12:57")
	   (authors FELIX)
	   (input   "A TERM")
	   (effect  "None.")
	   (value   "The value to INDICATOR in the property-list of TERM"
		    "or DEFAULT if INDICATOR isn't found."))
  (do ((plist (keim=plist term) (cddr plist)))
      ((or (eql (car plist) indicator) (null plist))
       (if plist (cadr plist) default))))

(defgeneric keim~plist (keim-object)
  (declare (edited  "11-SEP-1992 14:06")
	   (authors RICHTS)
	   (input   "A KEIM-object.")
	   (effect  "None.")
	   (value   "The property list of this object."))
  (:method ((anything keim+object))
	   (keim=plist anything)))

(defun keim~put (term indicator value)
  (declare (edited  "13-MAR-1997 13:19")
	   (authors FELIX)
	   (input   "A TERM, a symbol and the new value.")
	   (effect  "The property-list of TERM is changed by adding"
		    "the new property-value pair of INDICATOR and VALUE"
		    "or by replacing the old value for INDICATOR.")
	   (value   "VALUE."))
  (do ((plist (keim=plist term) (cddr plist)))
      ((or (eql (car plist) indicator) (null plist))
       (if plist
	   (setf (cadr plist) value)
	   (keim=write-plist!
	    (cons indicator (cons value (keim=plist term)))
	    term))))
  value)

(defun keim~remprop (term indicator)
  (declare (edited  "13-MAR-1997 13:20")
	   (authors FELIX)
	   (input   "A TERM and a symbol.")
	   (effect  "The property-value pair for INDICATOR is removed"
		    "from the property-list of KEIM-OBJECT.")
	   (value   "True, iff INDICATOR was an indicator in the"
		    "property-list of TERM."))
  (keim~remprops term (list indicator)))

(defun keim~remprops (term indicators)
  (declare (edited  "13-MAR-1997 13:28")
	   (authors FELIX)
	   (input   "A TERM and a list of symbols.")
	   (effect  "The property-value pairs for the indicators in"
		    "INDICATORS are removed from the property-list of"
		    "TERM.")
	   (value   "True, iff every member in INDICATORS was an indicator"
		    "in the property-list of TERM."))
  (let* ((removed-indicators nil)
	 (plist (do ((plist-tail (keim=plist term) (cddr plist-tail)))
		    ((or
		      (null plist-tail)
		      (not (member (car plist-tail) indicators)))
		     plist-tail)
		  (push (car plist-tail) removed-indicators))))
    (keim=write-plist! plist term)
    (do ((plist-tail plist))
	((null (cddr plist-tail)))
      (if (member (caddr plist-tail) indicators)
	  (progn (push (caddr plist-tail) removed-indicators)
		 (pop (cddr plist-tail))
		 (pop (cddr plist-tail)))
	  (setf plist-tail (cddr plist-tail))))
    (every
     #'(lambda (indicator) (member indicator removed-indicators))
     indicators))) 

(defgeneric keim~set-plist! (keim-object plist)
  (declare (edited  "11-SEP-1992 14:11")
	   (authors RICHTS)
	   (input   "A KEIM-object and list of associations.")
	   (effect  "The property list of KEIM-OBJECT is set to PLIST.")
	   (value   "The new property-list of KEIM-OBJECT, i.e. PLIST."))
  (:method ((keim-object keim+object) plist)
   (keim=write-plist! plist keim-object)))

;;;+--------------------------------------------------------------------------+
;;;| post                                                                     |
;;;+--------------------------------------------------------------------------+

(defun post~read-object-list (object-list env)
  (mapc
   #'(lambda (element)
       (let ((key (first element))
	     (objects (rest element)))
	 (case key
	   (constants (post=read-constants objects env))  
	   (variables (post=read-variables objects env))
	   (signatures (post=read-signatures objects env)))))
   object-list))

(defun post=read-constants (objects env)
  (mapc
   #'(lambda (object)
       (let
	   ((symbol (first object))
	    (kind (rest object))
	    (arity nil))
	 (if (listp kind)
	     (setf arity (1- (length kind)))
	     (setf arity 0))
	 (push
	  (cons symbol
		(make-instance
		 'sym+const
		 :symbol (symbol-name symbol) 
		 :arity arity))
	  (rest env))))
   objects))

(defun post=read-variables (objects env)
  (mapc
   #'(lambda (object)
       (let
	   ((symbol (first object)))
	 (push
	  (cons symbol
		(make-instance
		 'sym+var
		 :symbol (symbol-name symbol)))
	  (rest env))))
   objects))

(defun post=read-signatures (objects env)
  (mapc #'(lambda (signature) (post=read-signature signature env)) objects))

(defun post=read-signature (signature env)
  (let* ((constants
	  (mapcar
	   #'(lambda (object)
	       (let
		   ((symbol (first object))
		    (kind (rest object))
		    (arity nil)
		    (const nil))
		 (if (listp kind)
		     (setf arity (1- (length kind)))
		     (setf arity 0))
		 (setf const (make-instance
			      'sym+const
			      :symbol (symbol-name symbol) 
			      :arity arity))
		 (push (cons symbol const) (rest env))
		 const))
	   (rest signature)))
	 (theory (th~theory-create (first signature) constants)))
    (mapc
     #'(lambda (const) (keim~put const :th=theory theory))
     constants)))

;;;+--------------------------------------------------------------------------+
;;;| subst                                                                    |
;;;+--------------------------------------------------------------------------+

(defun subst~apply (substitution object)
  (declare (edited  "04-NOV-1991 11:14")
       	   (authors RICHTS)
	   (input   "A substitution and an object.")
	   (effect  "None.")
       	   (value   "The new object with its terms instantiated by"
		    "SUBSTITUTION.")
	   (example "{(X --> A) (Y --> B)} (Q X (G X Y)) --> (Q A (G A B))"))
  (unwind-protect
       (subst=apply-aux substitution object)
    (subst=label-reset! object)))

(defun subst=label-reset! (object)
  (if (appl~p object)
      (progn
	(term~set-label! object nil)
	(mapc #'subst=label-reset! (appl~arguments object)))
      (term~set-label! object nil)))

(defun subst=apply-aux (substitution object)
  (cond
    ((term~label object) (term~label object))
    ((appl~p object)
     (let
	 ((new 
	   (make-instance
	    'appl+appl
	    :function (appl~function object) 
	    :arguments (mapcar
			#'(lambda (object) (subst~apply
					    substitution object))
			(appl~arguments object)))))
       (term~set-label! object new)
       new))
    ((sym~constant-p object) object)
    ((sym~variable-p object)
     (let
	 ((new (if (subst~get-component object substitution)
		   (subst~get-component object substitution)
		   object)))
       (term~set-label! object new)
       new))))
 
       
(defun subst~create (domain codomain &optional copy)
  (declare (edited  "04-NOV-1991 11:00")
	   (authors RICHTS)
	   (input   "A list of variables and a list of terms."
		    "Optional a flag.")
	   (effect  "None.")
       	   (value   "The new substitution with the domain DOMAIN and the"
		    "codomain CODOMAIN. If COPY = T the input-lists are"
		    "copied.")
	   (example "(x y) (a (f a)) --> {(X --> A) (Y --> (F A))}"))
  (if copy
      (make-instance 'subst+substitution
		     :domain (copy-list domain)
		     :codomain (copy-list codomain))
      (make-instance 'subst+substitution
		     :domain domain
		     :codomain codomain)))

(defun subst~empty-p (substitution)
  (declare (edited  "04-NOV-1991 11:02")
	   (authors RICHTS)
	   (input   "A substitution.")
	   (effect  "None.")
       	   (value   "True iff SUBSTITUTION is empty.")
	   (example "{(X --> A) (Y --> (F A))}-->NIL"
		     "{}--> NIL"))
  (null (subst~domain substitution)))

(defun subst~get-component (variable substitution)
  (declare (edited  "04-NOV-1991 11:13")
	   (authors RICHTS prckln)
	   (input   "A variable and a substitution.")
	   (effect  "None.")
	   (value   "The codomain term corresponding to VARIABLE or NIL"
		    "if there is none.")
	   (example "X {(X --> F) (Y --> (F A))} --> F"))
  (some #'(lambda (var term)
	    (if (keim~equal var variable) term nil))
	(subst~domain substitution)
	(subst~codomain substitution)))

(defun subst~insert-component! (variable term substitution)
  (declare (edited  "04-NOV-1991 11:09")
	   (authors RICHTS)
	   (input   "A variable, a term and a substitution.")
	   (effect  "The pair [VARIABLE --> TERM] is added at the front of the"
		    "domain and codomain lists"
		    "of SUBSTITUTION.")
       	   (value   "The changed SUBSTITUTION.")
	   (remark  "There is no check if the added component"
		    "[VARIABLE --> TERM] is in the domain and codomain."
		    "Also occurences of VARIABLE in the codomain terms are not"
		    "substituted.")
	   (example "X B {(X --> F) (Y --> (F A))} -->"
		     "{(X --> B) (X --> F) (Y --> (F A))}"))
  (let ((new-domain (cons variable (subst~domain substitution)))
	(new-codomain (cons term (subst~codomain substitution))))
    (subst=write-domain! new-domain substitution)
    (subst=write-codomain! new-codomain substitution)
    substitution))


;;;+--------------------------------------------------------------------------+
;;;| shsym                                                                    |
;;;+--------------------------------------------------------------------------+

#|
(defun shsym~label (sym)
  (term~label sym))

(defun shsym~set-label! (sym value)
  (term~set-label! sym value))
|#

;;;+--------------------------------------------------------------------------+
;;;| sym                                                                      |
;;;+--------------------------------------------------------------------------+

(defun sym~p (term)
  (declare (edited  "12-SEP-1991 13:48")
	   (authors RICHTS)
	   (input   "A term.")
	   (effect  "None.")
	   (value   "True iff {\vb TERM} is a logical symbol")
	   (example "X --> T"
		    "C --> T"
		    "(P x) --> NIL"))
  (typep term 'sym+sym))

(defun sym~variable-p (object)
  (declare (edited  "12-SEP-1991 13:42")
	   (authors RICHTS)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True iff {\vb OBJECT} is a variable.")
	   (example "X --> T"
		    "C --> NIL"))
  (typep object 'sym+var))

(defun sym~constant-p (object)
  (declare (edited  "12-SEP-1991 13:42")
	   (authors RICHTS)
	   (input   "A term.")
	   (effect  "None.")
	   (value   "True iff TERM is a constant.")
	   (example "C --> T"
		    "X --> NIL"))
  (typep object 'sym+const))

(defun sym~label (sym)
  (term~label sym))

(defun sym~set-label! (sym value)
  (term~set-label! sym value))

(defun sym=print (symbol stream)
  (declare (edited  "08-AUG-1991 13:30")
	   (authors RICHTS)
	   (input   "A term symbol SYMBOL and an output stream STREAM.")
	   (effect  "SYMBOL is printed to STREAM")
	   (value   "Undefined."))
  (princ (sym~name symbol) stream))

(defun sym~symbol-binding (term)
  (term~binding term))

(defun sym~set-symbol-binding! (term value)
  (term~set-binding! value term))

(defun sym~name (sym)
  (sym=symbol sym))

(defun sym~shared-symbol (sym)
  sym)


;;;+--------------------------------------------------------------------------+
;;;| term                                                                     |
;;;+--------------------------------------------------------------------------+

;;; Here a pseudo-term~copy to prevent errors in example-files - a real
;;; copy-function is not neccessary for our purposes. 

(defgeneric term~p (object)
  (declare (edited  "06-DEC-1991 14:04")
	   (authors RICHTS)
	   (input   "A LISP object.")
	   (effect  "None.")
	   (value   "True, iff OBJECT is a concrete term, namely, of class term+term"))
  (:method ((thing t)) nil)
  (:method ((term term+term))
	   t))

(defgeneric term~binding (term)
  (declare (edited  "18-OCT-1991 17:18")
	   (authors RICHTS)
	   (input   "A term.")
	   (effect  "An error if TERM is a constant or variable and has"
		    "a term-binding and a symbol-binding.")
	   (value   "The term-binding or the symbol-binding if TERM has"
		    "these slots and one of them contains something."))
  (:method ((term term+term))
    (term~term-binding term)))

(defun term~copy (term)
  term)

(defun term~equal (term1 term2)
  (cond
    ((and (sym~p term1) (sym~p term2)) 
     (eq (sym=symbol term1) (sym=symbol term2)))
    ((and (appl~p term1) (appl~p term2))
     (and
      (eq (appl~function term1) (appl~function term2))
      (every #'term~equal (appl~arguments term1) (appl~arguments term2))))
    (T nil)))

(defgeneric term~label (term) 
  (declare (edited  "16-JUN-1992 16:40" )
	   (authors KOHLHASE )
	   (input   "A term." )
	   (effect  "None." )
	   (value   "The label of TERM." )
	   (example "The following evaluations are carried out succesively:"
		    "(term~label term1) --> NIL"
		    "(term~set-label! term1 't1)"
		    "(term~label term1) --> T1"))
  (:method ((term term+term))
	   (term=term-label term)))


(defun term~read (object env)
  (if (listp object)
      (make-instance
       'appl+appl
       :function (env=get (car object) env) 
       :arguments (mapcar
		   #'(lambda (argument) (term~read argument env))
		   (cdr object)))
      (env=get object env)))

(defgeneric term~set-binding! (term new-binding)
  (declare (edited  "26-AUG-1992 13:32")
	   (authors RICHTS)
	   (input   "Terms TERM and NEW-BINDING.")
	   (effect  "The binding cell of TERM is bound to NEW-BINDING.")
	   (value   "Undefined."))
  (:method ((term term+term) new-binding)
	   (term~set-term-binding! term new-binding)))

(defgeneric term~set-label! (term label) 
  (declare (edited  "16-JUN-1992 16:40" )
	   (authors huang KOHLHASE )
	   (input   "A term and a label." )
	   (effect  "The label of TERM is set to LABEL." )
	   (value   "Undefined" ))
  (:method ((term term+term) label)
	   (term=write-term-label! label term)))

(defgeneric term~top (term)
  (declare (edited  "02-AUG-1991 20:40")
	   (authors RICHTS)
	   (input   "A term.")
	   (effect  "None.")
	   (value   "If TERM is a constant or variable TERM itself,"
		    "if it's an application the applied function of the"
		    "application in n-normal form,"
                    " that is, if ((f a b) b) is the term then f,"
		    "if it's an abstraction the top of the scope of the"
		    "abstraction."))
  (:method ((application appl+appl))
	   (term~top (appl~function application)))
  (:method ((sym sym+sym))
	   sym)
  )

;;;+--------------------------------------------------------------------------+
;;;| top                                                                      |
;;;+--------------------------------------------------------------------------+

(defun top~free-variables (term)
  (reverse (delete-duplicates (top=free-variables term))))

(defgeneric top=free-variables (term)
  ;;the recursive part of top~free-variables.
  (:method ((termlist list))
	   (mapcan #'top=free-variables termlist))
  (:method ((variable sym+var))
	   (list variable))
  (:method ((constant sym+const))
	   nil)
  (:method ((application appl+appl))
	   (mapcan #'top=free-variables
		   (appl~arguments application))))

;;;+--------------------------------------------------------------------------+
;;;| pretty-print                                                             |
;;;+--------------------------------------------------------------------------+


(defmethod print-object ((appl appl+appl) stream)
  (declare (edited  "02-AUG-1991 21:44")
	   (authors RICHTS)
	   (input   )
	   (effect  )
	   (value   )
	   (special *print-length*))
  (cond	((or (null *print-length*) 
	     (> *print-length* (length (appl~arguments appl))))
	 (format stream "(~A~{ ~A~})" (appl~function appl) 
		 (appl~arguments appl)))
	(t (format stream "(~A~{ ~A~} ...)"
		   (appl~function appl)
		   (subseq (appl~arguments appl) 0 (1- *print-length*))))))

(defmethod print-object ((sym sym+sym) stream)
  (declare (edited  "02-AUG-1991 21:44")
	   (authors RICHTS)
	   (input   )
	   (effect  )
	   (value   ))
  (sym=print sym stream))


(defmethod print-object ((substitution subst+substitution) stream)
  (declare (edited  "05-NOV-1991 12:22")
	   (authors RICHTS)
	   (input   )
	   (effect  )
	   (value   )
	   (special *print-length*))
  (cond ((subst~empty-p substitution)
	 (format stream "{}"))
	(t
	 (format stream "{(~A --> ~A)"
		 (car (subst~domain substitution))
		 (car (subst~codomain substitution)))
	 (do ((i (if  *print-length* (1- *print-length*) -1) (1- i))
	      (domain-tail (cdr (subst~domain substitution))
			   (cdr domain-tail))
	      (codomain-tail (cdr (subst~codomain substitution))
			     (cdr codomain-tail)))
	     ((or (null domain-tail) (zerop i))
	      (if domain-tail
		  (format stream " ...}")
		  (format stream "}")))
	   (format stream " (~A --> ~A)"
		   (car domain-tail)
		   (car codomain-tail))))))

;;;+--------------------------------------------------------------------------+
;;;| features                                                                 |
;;;+--------------------------------------------------------------------------+

(pushnew :small-keim *features*)
