
#|
Special functions and methods for handling feature theories.
They should later replace the general ones in th.lisp

Main part is a sort hierarchy extractor.
A sort hierarchy is a special kind of finite partial order: It is an
upper semi lattice with the additional restriction that if two sorts s1 and s2
have common subsorts then the supremum of the set of subsorts is a lower bound
of s1 and s2 and thus their infimum. This restriction simplifies the calculation
of unifiers (infima) of sorts: One only needs to know the subsorts of a sort and
can simply calculate the supremum of the common subsorts.
|#

(in-package :KEIM)


(export '(fth~subsorts fth~set-subsorts! fth~add-subsort!
	  fth~sort-unifiers fth~set-sort-unifiers! fth~add-sort-unifier!))



;+------------------------------------------------------------------------------
;+ Slot for each sort: its subsorts and unifiers
;+ subsorts is just a list of sorts, always containing the sort itself
;+ (reflexive transitive closure).
;+
;+ unifiers is an a-list: Each pair consists of a sort that the sort unifiers
;+ with as the key and the unifier as the datum.
;+ Note that unification information is stored doubly: If sorts s1 and s2 unify
;+ with unifier u, then s1's unifiers' slot contains (s2 . u) and s2's unifiers'
;+ slot contains (s1 . u).
;+
;+ Accessors
;+------------------------------------------------------------------------------


(defun fth~subsorts (sort)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A sort.")
	   (effect  "None.")
	   (value   "The subsorts of the sort (including itself)."))
  (keim~get sort :fth=subsorts nil))


(defun fth~set-subsorts! (sort subsorts)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A sort and its subsorts.")
	   (effect  "The subsorts slot of the sort is set to subsorts.")
	   (value   "The subsorts."))
  (keim~put sort :fth=subsorts subsorts))


(defun fth~add-subsort! (sort subsort)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A sort and a subsort.")
	   (effect  "The subsort is added to the subsorts slot of sort.")
	   (value   "The subsorts."))
  (keim~put sort :fth=subsorts
	    (cons subsort (keim~get sort :fth=subsorts nil))))


(defun fth~sort-unifiers (sort)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A sort.")
	   (effect  "None.")
	   (value   "The list of sort unifiers. This is an a-list. Each element is a "
		    "pair of another sort that sort unifies with and the result sort "
		    "of the unification."))
  (keim~get sort :fth=unifiers nil))


(defun fth~set-sort-unifiers! (sort unifiers)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A sort and its unifiers.")
	   (effect  "The unifiers slot is set to unifiers.")
	   (value   "The unifiers."))
  (keim~put sort :fth=unifiers unifiers))


(defun fth~add-sort-unifier! (sort unifier)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A sort and a unifier.")
	   (effect  "The unifier is added to the unifiers slot of sort.")
	   (value   "The unifiers."))
  (keim~put sort :fth=unifiers
	    (cons unifier (fth~sort-unifiers sort))))


;+------------------------------------------------------------------------------
;+ processing the sort hierarchy
;+
;+------------------------------------------------------------------------------


(defun fth=initial-subsorts (sorts)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A list of sorts.")
	   (effect  "Each subsorts slot is initialised with a list of just the "
		    "sort itself and each unifiers slot is initialised with the "
		    "trivial unifier that sort unifies with itself returning sort.")
	   (value   "Nil."))
  (when (not (endp sorts))
    (let ((s (first sorts)))
      (fth~set-subsorts! s (list s))
      (fth~set-sort-unifiers! s (list (cons s s)))
      (fth=initial-subsorts (rest sorts)))))


(defun fth=check-sorts (candidates sorts env)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A list of candidates (potential sorts) or a single candidate, "
		    "a list of known sorts, a KEIM environment.")
	   (effect  "Creates an error, if one of the candidates is not a known sort.")
	   (value   "The true sorts from candidates. Actually, the candidates are not "
		    "KEIM terms, but Lisp terms. They are converted into KEIM "
		    "constants using term~read, hence the need of the environment."))
  (if (listp candidates)
      (if (endp candidates)
	  nil
	  (let ((c (term~read (first candidates) env)))
	    (if (member c sorts :test #'term~equal)
		(cons c (fth=check-sorts (rest candidates) sorts env))
		(error "KEIM(th): Sort hierarchy: ~A is not a sort.~%" (first candidates)))))
      (if (member (term~read candidates env) sorts :test #'term~equal)
	  (term~read candidates env)
	  (error "KEIM(th): Sort hierarchy: ~A is not a sort.~%" candidates))))


(defun fth=process-subsorts (sortlist sorts env)
  (declare (edited  "29-AUG-1997")
	   (authors Kepser)
	   (input   "A sort hierachy as a list of lists of sorts and their immediate "
		    "subsorts, a list of known sorts and a KEIM environment.")
	   (effect  "Each sort-subsorts information is checked, whether it is OK. "
		    "If so, each sort's subsort slot is filled the reflexive "
		    "transitive closure of its subsort.")
	   (value   "Nil."))
  (if (endp sortlist)
      (fth=tc sorts)
      (let ((rs (fth=check-sorts (first sortlist) sorts env)))
	(if (listp rs)
	    (fth~set-subsorts! (first rs) rs))
	(fth=process-subsorts (rest sortlist) sorts env))))


(defun fth=tc (sorts &optional (srts sorts))
    (declare (edited  "13-AUG-1997")
	     (authors Kepser)
	     (input   "A list of sorts")
	     (effect  "transitive closure on the subsorts slot via the warshall"
		      "algorithm.")
	     (value   "None."))
    (when (not (endp sorts))
      (fth=tc-a (first sorts) (fth~subsorts (first sorts)) srts)
      (fth=tc (rest sorts) srts)))


(defun fth=tc-a (sort subsorts sorts)
  (declare (edited  "13-AUG-1997")
	   (authors Kepser)
	   (input   "A sort, its subsorts and a list of sorts.")
	   (effect  "For each s of sorts: if sort is in the subsorts slot, "
		    "then subsorts is added to the subsorts slot.")
	   (value   "None."))
  (when (not (endp sorts))
    (if (member sort (fth~subsorts (first sorts)) :test #'term~equal)
	(fth~set-subsorts! (first sorts)
			   (union subsorts (fth~subsorts (first sorts))
				  :test #'term~equal)))
    (fth=tc-a sort subsorts (rest sorts))))


(defun fth=supremum (sorts)
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "A list of sorts (where the subsorts slots are filled).")
	   (effect  "Causes an error, if the set has no supremum.")
	   (value   "The supremum of the sorts. This is NO general function to calculate "
		    "a supremum. It requires that the supremum of sorts is one of the "
		    "sorts. See also comments at the start of the file."))
  (if (not (endp sorts))
      (fth=sup-a (first sorts) (rest sorts) sorts)))


(defun fth=sup-a (sup sorts all-s)
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "A sort (the candidate supremum), a list of sorts still to be "
		    "processed and the set of all sorts for which to calculate the supremum.")
	   (effect  "Causes an error, if no supremum is found.")
	   (value   "The supremum of sup and sorts, if it is amongst sup and sorts."))
  (if (endp sorts)
      (if (set-exclusive-or all-s (fth~subsorts sup) :test #'term~equal)
	  (error "KEIM(th): Your sort hierarchy is incorrect!~%The set ~A has no supremum within itself.~%" all-s)
	  sup)
      (let* ((sub-sup (fth~subsorts sup))
	     (next (first sorts))
	     (sub-next (fth~subsorts next)))
	(cond ((and (member next sub-sup :test #'term~equal)
		    (not (member sup sub-next :test #'term~equal)))
	       (fth=sup-a sup (rest sorts) all-s))
	      ((and (member sup sub-next :test #'term~equal)
		    (not (member next sub-sup :test #'term~equal)))
	       (fth=sup-a next (rest sorts) all-s))
	      ((and (member sup sub-next :test #'term~equal)
		    (member next sub-sup :test #'term~equal))
	       ;; we found a cycle in the order
	       (error "KEIM(th): Your sort hierarchy is NOT a partial order!~%~A and ~A form a cycle.~%"
		      sup next))
	      (T ; sup and next can't be compared. Could be a fault, could be not.
	         ; we go on with the old sup, but test at the end.
	       (fth=sup-a sup (rest sorts) all-s))))))


(defun fth=unifiables (sorts)
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "The set of all sorts.")
	   (effect  "Each sort's unifiers slot is bound to the sort's unifiers. ")
	   (value   "Not defined (nil)."))
  (when (not (endp sorts))
    (fth=unifiables-a (first sorts) (rest sorts))
    (fth=unifiables (rest sorts))))


(defun fth=unifiables-a (sort sorts)
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "A sort and a list of sorts.")
	   (effect  "The sort is checked against each element of sorts to see, if the "
		    "two unify. If so, the sort and the element's unifiers slots are "
		    "updated accordingly.")
	   (value   "Not defined (nil)."))
  (if (not (endp sorts))
      (let* ((s (first sorts))
	     (uni (fth=supremum (intersection (fth~subsorts sort)
					      (fth~subsorts s)
					      :test #'term~equal))))
	(when uni
	    (fth~add-sort-unifier! sort (cons s uni))
	    (fth~add-sort-unifier! s (cons sort uni)))
	(fth=unifiables-a sort (rest sorts)))))




;+-----------------------------------------------------------------------------------
;+  POST methods
;+
;+-----------------------------------------------------------------------------------




(defmethod post~read-object ((signatures list) env
			     (indicator (eql :signatures)))
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "A list of signatures, a KEIM environment, and the "
		    "indicator :signatures.")
	   (effect  "All signatures are read in, created (as KEIM objects) and "
		    "added to the environment.")
	   (value   "The extended environment."))
  (if signatures
      (let* ((signature (first signatures))
	     (theory (first signature))
	     (constants (mapcar #'(lambda (const)
				    (post~read-object const env :constant))
				(if (eql theory 'feat)
				    (butlast (rest signature))
				    (rest signature)))))
	(if (eql theory 'feat)
	    (fth=create-ft constants (first (last signature)) env)
	    (th=create-theory theory constants))
	(post~read-object (rest signatures) env :signatures))))


(defun th=create-theory (theory-name constants)
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "A theory name and a list of KEIM constants, the symbols "
		    "of that theory.")
	   (effect  "The thory object is created and all constants are marked "
		    "as belonging to that theory.")
	   (value   "Not defined."))
  (let ((theory (th~theory-create theory-name constants)))
    (mapc #'(lambda (const)
	      (keim~put const :th=theory theory))
	  constants)))


(defun fth=create-ft (constants hierarchy env)
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "A list of KEIM constants (the features and sorts), "
		    "a sort hierarchy and a KEIM environment.")
	   (effect  "A feature theory object is created, and the features and sorts are "
		    "marked with this theory. The sort hierarchy is processed and all "
		    "subsorts are stored in each sort's subsort slot. "
		    "Each sort's unifiers are calculated and stored in the unifiers "
		    "slot.")
	   (value   "Not defined (nil)."))
  (th=create-theory 'feat constants)
  (let ((sorts (fth=get-sorts constants)))
    (fth=initial-subsorts sorts)
    (if (not (eql (first hierarchy) 'sort-hierarchy))
	(error "KEIM(th): key word 'sort-hierarchy' missing!")
	(progn
	  (fth=process-subsorts (rest hierarchy) sorts env)
	  (fth=unifiables sorts)))))


(defun fth=get-sorts (constants &optional (sorts nil))
  (declare (edited  "01-SEP-1997")
	   (authors Kepser)
	   (input   "A list of feature and sort predicates, which are technically "
		    "KEIM constants.")
	   (effect  "None.")
	   (value   "The list of all sort predicates (unary predicates) in "
		    "constants."))
  (if (endp constants)
      sorts
      (let ((c (first constants)))
	(fth=get-sorts (rest constants)
		       (if (keim~equal (term~type c) (type~predicate-create 1))
			   (cons c sorts)
			   sorts)))))

