
(in-package "KEIM")

(export '(th+general th+collapse-free th+regular th+fine th+simple
	  th~symbols th~axioms
	  th+free th+a th+ac th+aci
	  th~theory-class th~theory-create th~theory-p th~term-theory
	  th~unify th~decide-unify))

#|
Each symbol has a theory slot.  The theories are ordered hierarchically.
|#

(defgeneric th~symbols (theory)
  (declare (edited  "12-DEC-1994")
	   (authors Richts)
	   (input   "A theory or a symbol.")
	   (effect  "None.")
	   (value   "A list with the symbols of the theory.")))

(defgeneric th~axioms (theory)
  (declare (edited  "12-DEC-1994")
	   (authors Richts)
	   (input   "A theory or a symbol.")
	   (effect  "None.")
	   (value   "The list of axioms of the theory.")))


;;; Abstract superclasses of theories

(defclass th+general (keim+object) ;(keim+unif-class)
  ((symbolnames :allocation :class :initform nil :accessor th~symbol-names)
   (unification-type :allocation :class :initform nil :accessor th~unification-type)
   (symbols :initarg :symbols :accessor th~symbols)
   (axioms :initarg :axioms :accessor th~axioms))
  (:documentation "This is the superclass of all equational theories."))

(defclass th+collapse-free (th+general)
  ()
  (:documentation "This is the superclass of all collapse-free equational theories."))

(defclass th+regular (th+general)
  ()
  (:documentation "This is the superclass of all regular equational theories."))

(defclass th+fine (th+collapse-free th+regular)
  ()
  (:documentation "This is the superclass of all collapse-free and regular equational theories."))

(defclass th+simple (th+fine)
  ()
  (:documentation "This is the superclass of all simple equational theories."))


;;; Theory classes

(defclass th+free (th+simple)
  ((symbolnames :allocation :class :initform '())
   (unification-type :allocation :class :initform :unitary))
  (:documentation "This is the class of the empty theory."))

(defmethod th~theory-class ((name (eql 'free)))
  (find-class 'th+free))

(defclass th+a (th+simple)
  ((symbolnames :allocation :class :initform '(:conc))
   (unification-type :allocation :class :initform :infinitary))
  (:documentation "This is the class of all theories of semigroups."))

(defmethod th~theory-class ((name (eql 'a)))
  (find-class 'th+a))

(defclass th+ac (th+simple)
  ((symbolnames :allocation :class :initform '(:plus))
   (unification-type :allocation :class :initform :finitary))
  (:documentation "This is the class of all theories of Abelean semigroups."))

(defmethod th~theory-class ((name (eql 'ac)))
  (find-class 'th+ac))

(defclass th+aci (th+regular)
  ((symbolnames :allocation :class :initform '(:union))
   (unification-type :allocation :class :initform :finitary))
  (:documentation "This is the class of all theories of idempotent, Abelean semigroups."))

(defmethod th~theory-class ((name (eql 'aci)))
  (find-class 'th+aci))


(defclass th+br (th+general)       
  ((symbolnames :allocation :class :initform '(:plus :mal 0 1))
   (unification-type :allocation :class :initform :finitary))
  (:documentation "This is the class of all theories of Boolean rings."))

(defmethod th~theory-class ((name (eql 'br)))
  (find-class 'th+br))


(defclass th+rat-tree (th+general)       
  ((symbolnames :allocation :class :initform nil)
   (unification-type :allocation :class :initform :unitary))
  (:documentation "This is the class of all theories of rational trees."))

(defmethod th~theory-class ((name (eql 'rat-tree)))
  (find-class 'th+rat-tree))


(defclass th+feat (th+general)       
  ((symbolnames :allocation :class :initform nil)
   (unification-type :allocation :class :initform :unitary))
  (:documentation "This is the class of all theories of features."))

(defmethod th~theory-class ((name (eql 'feat)))
  (find-class 'th+feat))



#|
Instances of theories are defined in the \post declarations:

\begin{postsyntax}
\syntax{\nt{theory-inst-decl} ::= (signatures (\nt{theory-name} (\nt{function-name} \nt{type})+)+).}
\end{postsyntax}

|#

(defun th~theory-create (name symbols)
  (declare (edited  "14-MAR-1995")
	   (authors Richts)
	   (input   "A name of a theory or a name of a theory class, and a list of symbols.")
	   (effect  "An error occurs if there is no theory with name NAME.")
	   (value   "An instance of the theory class."))
  (let ((theory-class (or (find-class name nil)
			  (th~theory-class name))))
    (if (null theory-class)
	(error "KEIM (th): There is no theory with name ~A" name)
      (let ((theory (make-instance theory-class :symbols symbols)))
	(if (and (th~symbol-names theory)
		 (not (= (length (th~symbols theory))
			 (length (th~symbol-names theory)))))
	    (error "KEIM (th): The theory ~A has a signature of the form ~S~
                    but was instantiated with the symbols ~A."
		   name (th~symbol-names theory) symbols)
	  theory)))))

(defun th~theory-p (object)
  (typep object 'th+general))

(defmethod top=replace-free-variables-and-rename ((theory th+general) variables)
  theory)

(defvar th*free (th~theory-create 'free nil))


(defgeneric th~term-theory (term)
  (declare (edited  "16-MAR-1995")
	   (authors Kepser Richts)
	   (input   "A term.")
	   (effect  "None.")
	   (value   "The theory instance TERM belongs to. (th*free if"
		    "it is free)."))
  (:method ((term appl+appl))
     (th~term-theory (appl~function term)))
  (:method ((sym sym+var))
     nil)
  (:method ((sym sym+sym))
     (keim~get sym :th=theory th*free)))

(defun th~symbol-theory (symbol)
  (declare (edited  "14-MAR-1995")
	   (authors Richts)
	   (input   "A term symbol.")
	   (effect  "None.")
	   (value   "The theory instance SYMBOL belongs to. (NIL if it is free)."))
  ;(error "th~~symbol-theory should be replaced by th~~term-theory")
  (th~term-theory symbol))

(defun th~symbol (theory key)
  (declare (edited  "16-JAN-1997")
	   (authors Felix)
	   (input   "A theory and the key for a special symbol of this theory")
	   (effect  "None.")
	   (value   "The special symbol characterized by KEY."))
  (or
   (some #'(lambda (name symbol) (if (eql name key)
				     symbol
				     nil))
	(th~symbol-names theory)
	(th~symbols theory))
   (error "Keim(TH): ~A is not a symbol in ~A" key theory)))

(defmethod post~read-object ((signatures list) env
			    (indicator (eql :signatures)))
  (if signatures
      (let* ((signature (first signatures))
	     (constants (mapcar #'(lambda (const)
				   (post~read-object const env :constant))
				(rest signature)))
	     (theory (th~theory-create (first signature) constants)))
	(mapc #'(lambda (const)
		  (keim~put const :th=theory theory))
	      constants)
	(post~read-object (rest signatures) env :signatures))))

#|
(defun aci~decide (tl1 tl2 &optional (restr NIL) (var_as_const NIL))
  (declare (edited  "12-MAY-1995")
	   (authors Richts)
	   (input   "two list of Terms which shall be ac-unified"
		    "optional: a list of constant restrictions")
	   (effect  None)
	   (value   "see aci~unif"))
  (aci~unif (mapcar #'cons tl1 tl2) restr var_as_const))
|#

(defgeneric th~unify (theory termlist1 termlist2  constant-restrictions constant-variables)
  (declare (edited  "23-MAY-1995")
	   (authors Richts)
	   (input   "A theory (i.e. an instance of a subclass of th+general);"
		    "two lists of equal length with terms built from variables,"
		    "free constants, and function symbols of THEORY;"
		    "a list of linear constant restrictions of the form"
		    "((c1 . x1 x2 ...) (c2 . y1 y2 ...) ...);"
		    "and a list of variables that are treated as constants.")
	   (effect  "None.")
	   (value   "A list of complete unifiers for pairwise unification of"
		    "TERMLIST1 and TERMLIST2 with CONSTANT-RESTRICTIONS"
		    "and CONSTANT-VARIABLES.")))


(defgeneric th~decide-unify (theory termlist1 termlist2  constant-restrictions constant-variables)
  (declare (edited  "23-MAY-1995")
	   (authors Richts)
	   (input   "A theory (i.e. an instance of a subclass of th+general);"
		    "two lists of equal length with terms built from variables,"
		    "free constants, and function symbols of THEORY;"
		    "a list of constant restrictions of the form"
		    "((c1 . x1 x2 ...) (c2 . y1 y2 ...) ...);"
		    "and a list of variables that are treated as constants.")
	   (effect  "None.")
	   (value   "True, iff there is a solution for pairwise unification of"
		    "TERMLIST1 and TERMLIST2 with CONSTANT-RESTRICTIONS"
		    "and CONSTANT-VARIABLES.")))

#|
;Examples:

(setf env (env~create))

(post~read-object-list '((type-constants i)
			 (constants (a i) (b i)(f (i i)))
			 (signatures (a (conc (i i i))) (a (cons (i i i)))
				     (ac (* (i i i))) (ac (+ (i i i)))
				     (aci (cup (i i i))) (aci (union (i i i)))))
		       env)

(setf t1 (term~read '(conc (* a b) (cons (+ a b) (conc (cup a b) (cons (union a b) b)))) env))
(setf t2 (term~read '(cup a b) env))
(setf t3 (term~read '(* a b) env))

(mapcar #'(lambda (pos)
	    (th~symbol-theory (term~at-position t1 pos)))
	(term~positions t1 #'sym~constant-p))
|#
